// addNodes - to load named nodes into the correct named nodes of Diversifly
//
// Authors:
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  addNodes is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

// streams
ifstream data_file;

// example addNodes dtkAugment DSO
class addNodes : public dtkAugment {
public:
  addNodes(dpf *app) ;
  ~addNodes(void) ;
  int preFrame(void);
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;

  // counter for the nodes
  int cnt;
  pfGroup* rootGroup;
  pfNode** objectPtrs;
  char** file_names;
  char* inName;
  unsigned char lnum;
  void createNewNode(void);
  pfNode *newNode;
  int pargc ;
  char **pargv ;
} ;

#define NUMBER_ALLOWED 5

// create object and set description.
// be sure to validate it!
addNodes::addNodes(dpf *app_in) :
  dtkAugment("addNodes")
{
  setDescription(" loads files specified by DPF_NODE_FILES"
		 " environment variable and "
		 " assigns each named node defined in these"
                 " files to their appropriate node in Diversifly."
		 " This version was compiled on 8/19/03.") ;
  app = app_in;
  validate();
  cnt = 0 ;
  objectPtrs = NULL;
  file_names = NULL;
  lnum = 1;
  pargv = NULL;
} ;

// destructor
addNodes::~addNodes(void) {
}

#define MIN_BUTTONS 3

int addNodes::preFrame(void) {
  display = app->display();

  if(!display) return DTK_CALLBACK_ERROR;

  // just call buttonNavControl::preFrame() once at the start.
  return DTK_REMOVE_CALLBACK;
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig() 
int addNodes::postConfig(void) {
  // first get the display and make and set the group node
  display = app->display();
  rootGroup = new pfGroup;
  display->world()->addChild(rootGroup);
  
  char *env = getenv("DPF_NODE_FILES") ;
  char *original_list ;
  char *line = new char[2047];
  
  if (env) {
    // read in the list of file names
    original_list = strdup(env) ;
    // copy the original list
    char *end = original_list ;
    for(;*end;) {
      for(; *end != ':' && *end; end++);
      if(*end) { *end = '\0'; end++; }
      // add to the list of files whose names are in original list
      if (strlen(original_list)>0) {
	// read each file and fill the filea and fileb arrays with names
	//printf("opening file: %s\n",original_list);
	FILE *fp = fopen(original_list, "r");        
        if (!fp) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "ABswitchC::preConfig: error opening input file %s\n",
		     original_list) ;
	  return DTK_CALLBACK_ERROR;
        }

	while ((fgets(line,250,fp)!=NULL)) {
	  dpf_charsToArgcArgv(line, &pargc, &pargv) ;
	  if (line != NULL && pargc > 0) {
	    inName = pargv[0];
	    // create a new node by loading the data
	    createNewNode();
	  }
	} //end while
	original_list = end;
      }
    }
  }

  char *old_path = (char *) pfGetFilePath();
  char *path = NULL;
  if(old_path)
    {
      // copy old_path from static memory to the heap.
      old_path = dtk_strdup(old_path);
      path = (char *) dtk_malloc(strlen(old_path) +
				 strlen("%s:.:"DPF_DATA_DIR));
      sprintf(path,"%s:.:"DPF_DATA_DIR, old_path);
      pfFilePath(path);
    }
  else // old_path == NULL
    pfFilePath(".:"DPF_DATA_DIR);

  return DTK_REMOVE_CALLBACK ;
}

void addNodes::createNewNode(void) {
  //read in new pfNode (inName) and load

  // get the parent node name to hang the loaded nodes
  pfGroup* group = (pfGroup*)pfdLoadFile(inName);

  // get the child
  int num = group->getNumChildren();
  pfNode *lNode = NULL;
  pfGroup *group2;
  if (num > 0 )
  for ( int j = 0; j < num; ++j ) {
    group2 = (pfGroup*)group->getChild(j); 
    const char* nName = group2->getName();

    // see how many child nodes came in
    int numChild = group2->getNumChildren();

    for ( int jj = 0; jj < numChild; ++jj ) {
      lNode = (pfNode*)group2->getChild(jj);
      if (nName != NULL) {
	if (nName[0] == 'n') {
	  // cout << "adding node to nav" << endl;
	  display->nav()->addChild(lNode);
	}
	if (nName[0] == 's') {
	  display->scene()->addChild(lNode);
	}
	if (nName[0] == 'e') 
	  display->ether()->addChild(lNode);
	if (nName[0] == 'w') 
	  display->world()->addChild(lNode);
      }
      else {
	// default to the world
	display->world()->addChild(lNode);
      }
    }
  }
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to be called once, and the object removed
int addNodes::postFrame(void) {

  return DTK_CONTINUE ;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new addNodes(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
