/*

  a DIVERSE DSO to set the background and fog

  John Kelso, 5/04, kelso@nist.gov

  some of the code was based on work by Eric Tester at Virginia Tech

  the background and fog color (red, green, blue) and fog density are gotten
  from four floats in the dtk shared memory file "backgroundFog"

 */

#include <Performer/pf/pfEarthSky.h>
#include <Performer/pr/pfFog.h>

#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

// Variables allocated in Performer shared memory so they can be seen by
// the draw process
static float *fogData = NULL ; // red, green, blue, far
static pfFog *fog = NULL ; // pointer to fog object

class backgroundFog : public dtkAugment {
public:
  backgroundFog(dpf *app) ;
  ~backgroundFog(void) ;
  int preConfig(void) ;
  int postConfig(void) ;
  int postFrame(void) ;

private:
  dpfDisplay *display ;
  u_int64_t writeCount ;
  dtkSharedMem *shm ;
} ;

////////////////////////////////////////////////////////////////////////
//  Draw callback to set the fog distance and color.  Fog values may only
//  be changed in the draw process
int preDraw(pfTraverser* trav, void* data) {

  //pfChannel* channel = (pfChannel*) data ;

  static bool firstFrame= true;
  //  Set up the fog on the first preDraw call
  if (firstFrame) {
    pfEnable(PFEN_FOG);
    fog = new pfFog();
    fog->setFogType(PFFOG_VTX_EXP2);
    fog->apply();
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON);
    firstFrame= false;
  }
  
  if (fogData[3]>0.f) {
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_OFF) ;
    //  Turn on fog, set color and range
    pfEnable(PFEN_FOG) ;
    fog->setColor(fogData[0], fogData[1], fogData[2]) ;
    float density = (1.f-fogData[3])*10.f ;
    //float density = fogData[3] ;
    fog->setRange(0.f, density) ;
    fog->apply() ;
    pfOverride(PFSTATE_FOG | PFSTATE_ENFOG, PF_ON) ;
    
  } else {
    //  Turn off fog
    pfDisable(PFEN_FOG);
  }
  
  return PFTRAV_CONT ;
}

////////////////////////////////////////////////////////////////////////
backgroundFog::backgroundFog(dpf *app) :
  dtkAugment("backgroundFog") {

  display = app->display() ;

  //get DTK shared memory
  // use existing segment, or create if it doesn't exist
  shm = new dtkSharedMem(sizeof(float)*4, "backgroundFog", DTK_GET) ;
  if (!shm || shm->isInvalid()) {
    dtkMsg.add(DTKMSG_ERROR,"backgroundFog::backgroundFog() failed: "
	       " unable to create or get shared memory\n") ;
    return ;
  } 
  
  if(shm->writeCount(1)) return ;
  writeCount = shm->getWriteCount() ;
  shm->setAutomaticByteSwapping(sizeof(float)) ;
  
  setDescription("sets color of background and fog, and fog near/far") ;
  validate() ;
} ;

////////////////////////////////////////////////////////////////////////
backgroundFog::~backgroundFog(void) {
}

////////////////////////////////////////////////////////////////////////
// set up variables in shared memory
int backgroundFog::preConfig(void) {
  
  fogData = (float *) pfMalloc(sizeof(float)*4) ;
  
  return dtkAugment::REMOVE_CALLBACK ; 
}

////////////////////////////////////////////////////////////////////////
int backgroundFog::postConfig(void) {
  
   pfEarthSky *esky = 
     display->first()->first()->first()->pfchan(0)->getESky();
   if (!esky) {
     // create an earthSky for master channel and hope channel share mask works
     pfEarthSky *esky = new pfEarthSky();
     display->first()->first()->first()->pfchan(0)->setESky(esky);
   }

  // Data to the callback is a pointer to the first channel
  // which contains pointers to eSky & fog
  display->world()->
    setTravData(PFTRAV_DRAW, 
		display->first()->first()->first()->pfchan(0)) ;
  display->world()->setTravFuncs(PFTRAV_DRAW, preDraw, NULL) ;
  
  return dtkAugment::REMOVE_CALLBACK ; 
}

////////////////////////////////////////////////////////////////////////
int backgroundFog::postFrame(void) {
  
  if (shm->read(fogData)) {
    dtkMsg.add(DTKMSG_ERROR,"backgroundFog::postFrame() failed: "
	       " unable to read shared memory segment %s\n","backgroundFog") ;
    return dtkAugment::ERROR_ ;
  }
  
  //  Set the earth sky to be the same color as fog
  pfEarthSky* eSky = display->first()->first()->first()->pfchan(0)->getESky();
  if (eSky) {
    eSky->setMode(PFES_BUFFER_CLEAR, PFES_FAST) ;
    eSky->setColor(PFES_CLEAR, fogData[0], fogData[1], fogData[2], 1.f) ;
  } else { 
    dtkMsg.add(DTKMSG_ERROR,"backgroundFog::postFrame() failed: "
	       "pfGetEsky() returned null\n") ;
  }
  return dtkAugment::CONTINUE ; 
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p) {
  dpf *app = static_cast<dpf *>(manager) ;
  return new backgroundFog(app) ;
}

static int dtkDSO_unloader(dtkAugment *augment) {
  delete augment ;
  return DTKDSO_UNLOAD_CONTINUE ;
}
