//////////////////////////////////////////////////////////////////////////
//
// boxClip: Clip world node objects to an axis aligned box.
//
// Overview:
//
//    A box (axis-aligned in world coordinates) is provided through a
//    pointer obtained by a call to a function provided by another DSO.
//
//    We use GL calls in a preDraw callback attached to the world node
//    to clip all objects under world against the box.
//
//////////////////////////////////////////////////////////////////////////
//
// Box comes in in world coordinates.
//
// NOTE: I use doubles for a lot of the coordinate stuff here
//		because the plane equation coefficients that
//		we pass to GL have to be doubles.  But, if it
//		makes a difference to performance, we can change
//		most of the doubles to floats.
//
//////////////////////////////////////////////////////////////////////////
//
// To Do:
//
//	- frame accuracy
//	- add shm commands to turn it on and off
//	- clip other nodes than world?
//	- combine with boxCull DSO???????
//	- remove the use of the variable names like planeEqXneg
//	- make file global variables into class variables?
//	- use pfBox rather than Box??
//
//////////////////////////////////////////////////////////////////////////
//
// The plane equations for the clipping planes have this form:
//      Ax + By + Cz + D = 0
//
// The half space that is retained is:
//      Ax + By + Cz + D > 0
//
// The half space that is removed is:
//      Ax + By + Cz + D < 0
//
//////////////////////////////////////////////////////////////////////////



#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

typedef double Box[2][3];

Box *getRubberbandBoxPtr ();

//
// A comment about variable names:  
//    The variables planeEqXneg, planeEqXpos, etc versus planeEq[][]:
//    They are the same.  I started using the explicit naming
//    when I was working out in my head what the values and use
//    of these plane equations should be.  But the planeEq[][]
//    form should probably be used and the explicit names should
//    be removed.

static double *planeEqXneg = NULL;
static double *planeEqXpos = NULL;
static double *planeEqYneg = NULL;
static double *planeEqYpos = NULL;
static double *planeEqZneg = NULL;
static double *planeEqZpos = NULL;
//!static double *planeEq[2][3] ;  // these will be set to the values above



//
// Here's the main class
//
class boxClip : public dtkAugment {

public:

  // The required public methods
  boxClip (dpf *app) ;
  int preConfig (void);
  int postConfig (void);
  int preFrame (void);
  int postFrame (void);
  ~boxClip ();


private:


  dpf* app;
  dpfDisplay *display;

  // the box we are clipping against
  Box *currentBox;

} ;


//////////////////////////////////////////////////////////////////////
//  
//
// preDraw and postDraw are the callback function attached to the world node.
//
// We enable the clipping before drawing the world node and we
// disable after drawing the world node.  That way, we only 
// clip the stuff under the world node.
//
static int 
preDraw (pfTraverser *trav, void *data) {


    glClipPlane (GL_CLIP_PLANE0, planeEqXneg);
    glClipPlane (GL_CLIP_PLANE1, planeEqXpos);
  
    glClipPlane (GL_CLIP_PLANE2, planeEqYneg);
    glClipPlane (GL_CLIP_PLANE3, planeEqYpos);

    glClipPlane (GL_CLIP_PLANE4, planeEqZneg);
    glClipPlane (GL_CLIP_PLANE5, planeEqZpos);

    glEnable (GL_CLIP_PLANE0);
    glEnable (GL_CLIP_PLANE1);
    glEnable (GL_CLIP_PLANE2);
    glEnable (GL_CLIP_PLANE3);
    glEnable (GL_CLIP_PLANE4);
    glEnable (GL_CLIP_PLANE5);

  return PFTRAV_CONT;
}  // end of preDraw



//  Disable the clip plane in post draw
int 
postDraw (pfTraverser *trav, void *data) {

  // we disable here so that we don't clip non-world objects

  glDisable (GL_CLIP_PLANE0);
  glDisable (GL_CLIP_PLANE1);
  glDisable (GL_CLIP_PLANE2);
  glDisable (GL_CLIP_PLANE3);
  glDisable (GL_CLIP_PLANE4);
  glDisable (GL_CLIP_PLANE5);

  return PFTRAV_CONT;
}  // end of postDraw


//  
//////////////////////////////////////////////////////////////////////




//////////////////////////////////////////////////////////////////////

// Constructor

// create object and set description.
// be sure to validate it!
boxClip::boxClip (dpf *app_in) :
  dtkAugment ("boxClip")
{
  setDescription (
    "boxClip DSO: Clips world node objects outside of an axis-aligned \n"
  "  clipping box provided by the rubberbandBox DSO.\n") ;

  app = app_in;

  display = NULL;

  currentBox = NULL;

  validate ();
} // end of constructor


boxClip::~boxClip ()
{
	pfFree (planeEqXneg);
	pfFree (planeEqXpos);
	pfFree (planeEqYneg);
	pfFree (planeEqYpos);
	pfFree (planeEqZneg);
	pfFree (planeEqZpos);

}  // end of destructor
//////////////////////////////////////////////////////////////////////









//////////////////////////////////////////////////////////////////////

//
// preConfig is called only once by dpf, after pfInit() and 
//  dpfDisplay:preConfig() but before pfConfig()
//
int boxClip::preConfig (void) {



  //  Allocate plane equation out of Performer shared memory
  //  Apparently we need to do this in preConfig, not sure why.
  planeEqXneg = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());
  planeEqXpos = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());

  planeEqYneg = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());
  planeEqYpos = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());

  planeEqZneg = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());
  planeEqZpos = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());

  //!planeEq[0][0] = planeEqXneg;
  //!planeEq[1][0] = planeEqXpos;
  //!planeEq[0][1] = planeEqYneg;
  //!planeEq[1][1] = planeEqYpos;
  //!planeEq[0][2] = planeEqZneg;
  //!planeEq[1][2] = planeEqZpos;

  int i;
  for (i = 0; i < 4; i++)
	{
  	planeEqXneg[i] = 0;
  	planeEqXpos[i] = 0;
  	planeEqYneg[i] = 0;
  	planeEqYpos[i] = 0;
  	planeEqZneg[i] = 0;
  	planeEqZpos[i] = 0;
	}

  planeEqXneg[0] =  1;
  planeEqXpos[0] = -1;
  planeEqYneg[1] =  1;
  planeEqYpos[1] = -1;
  planeEqZneg[2] =  1;
  planeEqZpos[2] = -1;

  // At this point, the all of the planes pass through the origin
  // and each is perpendicular to the apprpriate axis.

  // We are clipping out everything at this point.


  return dtkAugment::CONTINUE ;
}   // end of preConfig


//
// postConfig is called after pfConfig
//
int boxClip::postConfig (void) {



  //  Set the preDraw and postDraw functions set up the clipping planes
  // app->display()->world()->setTravFuncs (PFTRAV_CULL, preDraw, postDraw);
  app->display()->world()->setTravFuncs (PFTRAV_DRAW, preDraw, postDraw);

  return dtkAugment::REMOVE_CALLBACK ;
}  // end of postConfig

//
//////////////////////////////////////////////////////////////////////



//
// We're calling preFrame only once
//
int boxClip::preFrame (void)
{
  display = app->display ();
  if (!display)
    {
    dtkMsg.add(DTKMSG_ERROR,
                 "boxClip::preFrame() failed:"
                 "Unable to get display pointer.\n");
    return dtkAugment::ERROR_;
    }

  // Get the pointer to the changeable box.
  currentBox = (Box *) getRubberbandBoxPtr ();

  return dtkAugment::REMOVE_CALLBACK;
}    // end of preFrame


//
// We call postFrame every frame to set the
// clip plane equations in accordance with the box.
//
int boxClip::postFrame (void) {


	// Change the plane equations to match the faces of the box.

        planeEqXneg[3] = - (*currentBox)[0][0];
        planeEqXpos[3] =   (*currentBox)[1][0];

        planeEqYneg[3] = - (*currentBox)[0][1];
        planeEqYpos[3] =   (*currentBox)[1][1];

        planeEqZneg[3] = - (*currentBox)[0][2];
        planeEqZpos[3] =   (*currentBox)[1][2];


  return dtkAugment::CONTINUE ; 

}    // end of postFrame




//////////////////////////////////////////////////////////////////////




/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new boxClip (app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
