//////////////////////////////////////////////////////////////////////////
//
// boxCull : This DSO culls to an axis-aligned box 
//
// Overview:
// 
//    This DSO is intended to enhance the performance of the boxClip
//    DSO, but enabling Performer to do proper culling against the
//    the same box against which we are clipping.  Perhaps this DSO
//    should be folded into the boxClip DSO.
//
//    A box (axis-aligned in world coordinates) is provided through a
//    pointer obtained by a call to a function provided by another DSO.
//
//    Commands provided through dtk shared memory tell the dso to 
//    enable or disable culling to the provided box.
//
//////////////////////////////////////////////////////////////////////////
//
// Box comes in in world coordinates.
//
// The culling is done through pre-cull callbacks associated with pfNodes.
//
// The commands take the form:
//     cull nodeName
//     uncull nodeName
//
//////////////////////////////////////////////////////////////////////////
//
// To Do:
//
//	- frame accuracy
//	- combine routines preCull and cullInner
//	- combine firstTime and PointersSet variables
//	- combine this DSO with boxClip?????
//	- move file global variables into class?
//      - use pfBox rather than Box
//      - destructor should free stuff
//	- is it correct to search for nodes only under scene node?
//            (see setupCullCallback )
//
//////////////////////////////////////////////////////////////////////////



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <dpf.h>
#include <dtk/dtkDSO_loader.h>



typedef double Box[2][3];

// Get pointer to the changeable box.
Box *getRubberbandBoxPtr ();


static int PointersSet = 0;
static Box *CurrentBox;

// A matrix that gives the current dpf to world transformation
static pfMatrix *DpfToWorld;

//
// Here's the main class
//
class boxCull : public dtkAugment {

public:

  // The required public methods

  boxCull (dpf *app) ;
  int preConfig (void);
  int postConfig (void);
  int preFrame (void);
  int postFrame (void);
  ~boxCull ();


private:


  dpf* app;
  dpfDisplay *display;

  // the matrix at the world node
  pfMatrix worldMat;

  dtkSharedMem * shmCmd;

  // Pointer to the current box
  Box *currentBox;

  int setupCullCallback (char *);
  int setupCullCallback (pfNode *);

  int removeCullCallback (char *);
  int removeCullCallback (pfNode *);

  static int preCull (pfTraverser *, void *);
  static int postCull (pfTraverser *, void *);

  // int preCullInner (pfMatrix &, pfNode *);

} ;


//////////////////////////////////////////////////////////////////////



#if 0
//
// Convert a point from dpf to world coordinates
//
static void dpfToWorld (pfVec3 *pt)
  {
  float x, y, z;


  x = 
	(*pt)[0]*(*DpfToWorld)[0][0] +
	(*pt)[1]*(*DpfToWorld)[1][0] +
	(*pt)[2]*(*DpfToWorld)[2][0] +
	         (*DpfToWorld)[3][0] ;

  y =
	(*pt)[0]*(*DpfToWorld)[0][1] +
	(*pt)[1]*(*DpfToWorld)[1][1] +
	(*pt)[2]*(*DpfToWorld)[2][1] +
	         (*DpfToWorld)[3][1] ;

  z =
	(*pt)[0]*(*DpfToWorld)[0][2] +
	(*pt)[1]*(*DpfToWorld)[1][2] +
	(*pt)[2]*(*DpfToWorld)[2][2] +
	         (*DpfToWorld)[3][2] ;

  (*pt)[0] = x / (*DpfToWorld)[3][3];
  (*pt)[1] = y / (*DpfToWorld)[3][3];
  (*pt)[2] = z / (*DpfToWorld)[3][3];

  }  // end of dpfToWorld
#endif

//////////////////////////////////////////////////////////////////////
//  

static int preCullInner (pfMatrix &mat, pfNode *node);



//
// The preCull callback.  The return value of this routine tells the
//  cull traversal whether or not to cull out this branch of the
//  the scene graph.
//
int 
boxCull::preCull (pfTraverser *trav, void *data) {

  if ( ! PointersSet )
    {
    PointersSet = 1;
    CurrentBox = (Box *) getRubberbandBoxPtr ();
    }


  static pfMatrix mat;


  pfNode *node = trav->getNode ();
  
  trav->getMat (mat);
 

  return preCullInner (mat, node);
  }  // end of preCull





//
// This routine does the bulk of the cull calculation for the given pfNode.
// The matrix is the current transformation from the root of the scene graph.
//

static int preCullInner (pfMatrix &mat, pfNode *node)
  {

  // printf (">>>>>>>>   preCullInner called for node %p\n", node);

  ///////////////////////////////////////////////////////////////////////
  //
  //  get node's bounding sphere, transform by mat, This gives bounding 
  //    sphere in dpf coords
  //
  //  Transform dpfToWorld
  //
  //  intersect with bc->currentBox (which is in world coords)
  //
  //  return one of the following:
  //    PFTRAV_CONT - continue the traversal
  //    PFTRAV_PRUNE - skip this node
  //    PFTRAV_TERM - skip the rest of the traversal 
  //      (this box culling should not do _TERM )
  //
  ///////////////////////////////////////////////////////////////////////

  static int firstTime = 1;
  static pfSphere *bsph, *xbsph;
  static pfBox *pbox;
  static pfVec3 pt;

  // printf ("\n\n\nBeginning of preCullInner node %p\n", node);


  if (firstTime)
    {
    firstTime = 0;
    bsph = new pfSphere;
    xbsph = new pfSphere;
    pbox = new pfBox;
    }


  node->getBound (bsph);  // should we test the return value of getBound?

  // transform the sphere by the incoming matrix.  xbsph is now in dpf coords
  xbsph->orthoXform (bsph, mat);

  // now transform xbsph to world coords
  xbsph->orthoXform (xbsph, *DpfToWorld);

  // make the pfBox that we are culling against 
  pbox->min.set ((*CurrentBox)[0][0], (*CurrentBox)[0][1], (*CurrentBox)[0][2]);
  pbox->max.set ((*CurrentBox)[1][0], (*CurrentBox)[1][1], (*CurrentBox)[1][2]);


  // return value based on intersection of box and sphere

  return (pbox->contains (xbsph) & PFIS_MAYBE) ? PFTRAV_CONT : PFTRAV_PRUNE;

}  // end of preCullInner



//   We don't actually use this.  I include it to maintain a skeleton
//   of callback routines.
int 
boxCull::postCull (pfTraverser *trav, void *data) {


  return PFTRAV_CONT;
}  // end of postCull


//  
//////////////////////////////////////////////////////////////////////




//////////////////////////////////////////////////////////////////////

// Constructor

// create object and set description.
// be sure to validate it!
boxCull::boxCull (dpf *app_in) :
  dtkAugment ("boxCull")
{
  setDescription (
    "boxCull DSO: Culls specified objects based an axis-aligned clipping\n"
  "  box provided by the rubberbandBox DSO.\n"
  );





  app = app_in;

  display = NULL;

  shmCmd = NULL;


  worldMat.makeIdent();

  currentBox = CurrentBox = NULL;
  DpfToWorld = NULL;

  validate ();
} // end of constructor


boxCull::~boxCull ()
{
// should free stuff here
}  // end of destructor
//////////////////////////////////////////////////////////////////////









//////////////////////////////////////////////////////////////////////

//
// preConfig is called only once by dpf, after pfInit() and 
//  dpfDisplay:preConfig() but before pfConfig()
//
// More initialization happens in preFrame
//
int boxCull::preConfig (void) {


  DpfToWorld = (pfMatrix *) pfMalloc(sizeof(pfMatrix)) ;
  DpfToWorld->makeIdent ();





  return dtkAugment::REMOVE_CALLBACK;
}   // end of preConfig


//
// postConfig is called after pfConfig
//
int boxCull::postConfig (void) {


  // Get pointer to changeable box
  CurrentBox = currentBox = (Box *) getRubberbandBoxPtr ();

  // printf ("postConfig: CurrentBox = %p    %f\n", 
  //         CurrentBox, (*CurrentBox)[0][0]);


  return dtkAugment::REMOVE_CALLBACK ;
}  // end of postConfig

//
//////////////////////////////////////////////////////////////////////









//
// We're calling preFrame only once
//
int boxCull::preFrame (void)
{
  display = app->display ();
  if (!display)
    {
    dtkMsg.add(DTKMSG_ERROR,
                 "boxCull::preFrame() failed:"
                 "Unable to get display pointer.\n");
    return dtkAugment::ERROR_;
    }

  char *shmName = getenv ("BOX_CULL_CMD_SHM");
  if (shmName == NULL)
	{
	shmName = "boxCullCmd";
	}

  shmCmd = new dtkSharedMem(sizeof(char)*1000, shmName);
  if ( shmCmd == NULL)
	{
	// error
	dtkMsg.add(DTKMSG_ERROR,
                 "boxCull::preFrame() failed:"
                 " Couldn't open shared memory segment\"%s\"\n",
                 shmName) ;
        return dtkAugment::ERROR_;
	}
  else if (shmCmd->isInvalid())
	{
	// error 
	dtkMsg.add(DTKMSG_ERROR,
                 "boxCull::preFrame() failed:"
                 " Couldn't open shared memory segment\"%s\"\n",
                 shmName) ;
        return dtkAugment::ERROR_;
	}
  else
	{
	shmCmd->flush ();
	}


  CurrentBox = currentBox = (Box *) getRubberbandBoxPtr ();
  // printf ("CurrentBox = %p    %f\n", CurrentBox, (*CurrentBox)[0][0]);


  display->world()->getMat(worldMat) ;

  return dtkAugment::REMOVE_CALLBACK;
}    // end of preFrame


//
// We call postFrame every frame to read commands from the queue
// and to setup or remove the cull callback as directed.
//
// We also get the current navigation matrix and calc the dpf to world matrix.
//
int boxCull::postFrame (void) 
{

  char line[1000];
  char operation[1000];
  char name[1000];
  static int frame = 0;

  frame++;


  if (shmCmd->qread (line))
	{


	if (sscanf (line, "%s %s", operation, name) != 2)
		{
		//  error parsing cmd
    		dtkMsg.add(DTKMSG_INFO,
                 "boxCull::postFrame():"
                 "Received bad command through shared memory: < %s >.\n", line);
		}
	else if (strcasecmp ("cull", operation) == 0)
		{
		setupCullCallback (name);
		}
	else if (strcasecmp ("uncull", operation) == 0)
		{
		removeCullCallback (name);
		}
	else
		{
		// error
    		dtkMsg.add(DTKMSG_INFO,
                 "boxCull::postFrame():"
                 "Received bad command through shared memory: < %s >.\n", line);
		}
		
	
	}




  // Get the navigation matrix and generate the current dpf to world matrix

  pfMatrix navMat ;
  display->nav()->getMat(navMat) ;

  *DpfToWorld = worldMat * navMat;
  DpfToWorld->invertAff (*DpfToWorld);


  // printf ("Exiting postFrame %d\n", frame );

  return dtkAugment::CONTINUE ; 
}    // end of postFrame



//
// This attachs the cull callback to the node and recursed to attach
// the callback to all of its children.
//
int boxCull::setupCullCallback (pfNode *node)
{

  // add the cull callback
  node->setTravFuncs (PFTRAV_CULL, preCull, NULL);
  node->setTravData  (PFTRAV_CULL, this);

  // printf ("setupCullCallback at node %p\n", node);

  // If it's a group, descend and do the children
  if (node->isOfType (pfGroup::getClassType()))
    {
    pfNode *child;
    pfGroup *group = (pfGroup *)node;

    int i;
    int n = group->getNumChildren ();
    for (i = 0; i < n; i++)
      {
      if ( (child = group->getChild (i)) == NULL)
        {
        // error
	dtkMsg.add (DTKMSG_INFO,
                 "boxCull::setupCullCallback():"
                 "Error while descending scene graph\n");
	// What else can we do?  Don't want to stop traversal.
        }
      else
        {
	// recurse
        setupCullCallback (child);
        }
      }


    }

  return 0;
}  // end of setupCullCallback (pfNode *)


//
// This attachs the cull callback to the named node and all descendants.
//
int boxCull::setupCullCallback (char *nodeName)
{

  // Get pointer to named pfNode

  // should we search under world or scene? or what? JGH
  pfNode * node = 
    (pfNode *) display->scene()->find (nodeName, pfNode::getClassType());

  if (node == NULL)
    {
    // error
    dtkMsg.add (DTKMSG_ERROR,
                 "boxCull::setupCullCallback():"
                 "Unable to find node < %s >.\n", nodeName);
    return -1;
    }
  else
    {
    // setup callback based on pointer
    setupCullCallback (node);
    }

  // printf ("End of setupCullCallback %s\n", nodeName);
  return 0;
}  // end of setupCullCallback (char *)






//
// remove callback on indicated node and all descendants
//
int boxCull::removeCullCallback (pfNode *node)
{

  // set callbacks to NULL
  node->setTravFuncs (PFTRAV_CULL, NULL, NULL);

  // printf ("removeCullCallback at node %p\n", node);

  // If it's a group, descend to children
  if (node->isOfType (pfGroup::getClassType()))
    {
    pfNode *child;
    pfGroup *group = (pfGroup *)node;

    int i;
    int n = group->getNumChildren ();
    for (i = 0; i < n; i++)
      {
      if ( (child = group->getChild (i)) == NULL)
        {
        // error
	dtkMsg.add (DTKMSG_INFO,
                 "boxCull::removeCullCallback():"
                 "Error while descending scene graph\n");
        }
      else
        {
	// recurse
        removeCullCallback (child);
        }
      }


    }

  return 0;
}  // end of removeCullCallback (pfNode *)


//
// remove cull callback from named node and descendants.
//
int boxCull::removeCullCallback (char *nodeName)
{

  // should it be world or scene?
  pfNode * node = 
    (pfNode *) display->scene()->find (nodeName, pfNode::getClassType());

  if (node == NULL)
    {
    // error
    dtkMsg.add (DTKMSG_ERROR,
                 "boxCull::removeCullCallback():"
                 "Unable to find node < %s >.\n", nodeName);
    return -1;
    }
  else
    {
    removeCullCallback (node);
    }

  // printf ("End of removeCullCallback %s\n", nodeName);
  return 0;
}  // end of removeCullCallback (char *)





//////////////////////////////////////////////////////////////////////




/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new boxCull (app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
