//////////////////////////////////////////////////////////////////////////
//
// boxVolSlice - Creates texture mapped box and cross-sections of a 
//               3D volume of data.
//
// Overview:
//
//    The user provides a 3D volume of data that serves as the texture.
//    This is provided through an enviroment variable.
//
//    A box (axis-aligned in world coordinates) is provided through a
//    pointer obtained by a call to a function provided by another DSO.
//
//    Through the volume file the user can specify whether he wants to see:
//      * the sides of the box,
//	* cross-sections through the far sides (relative to viewer) of the box
//	* both box and cross-sections.
//
//    For each frame, the dso gets the box, then updates the box and/or 
//    cross-sections, and the visibility of the cross-sections (based on head 
//    position).  The idea is that the cross-section polygons on the far side
//    of the box will be visible, providing a backdrop or visual context for 
//    the box.
//
//    This was originally tested in conjunction with the rubberbandBox DSO
//    which provided the box.  But it should work with any other DSO that
//    can provide the box.
//
//    NOTE:  	The positions of the various texture mapped polygons are 
//		updated at every frame by changing the vertex data in memory.
//
//////////////////////////////////////////////////////////////////////////
//
//  Format of the .vol file: 
//
//      dim             X Y Z
//      voxelFN         voxel_filename
//      lutFN           lut_filename
//      numComponents   1 | 2 | 3 | 4
//      interpMethod    POINT | LINEAR | BILINEAR | TRILINEAR | QUADLINEAR
//      whichSides      BOX | CROSS_SECTIONS | BOX&CROSS_SECTIONS
//
//  keywords are case-insensitive
//
//  voxel data is in binary 
//
//  Currently can only handle numComponents 1
//
//  If lutFN is not present then it uses a idenity lut.  LUT file is
//  in ascii: 256 entries of form r g b a.
//  
//  interpMethods correspond to PFTEX_ identifiers.
//
//  Anything after (and including) a "#" is considered a comment and is
//  discarded.
//
//
//////////////////////////////////////////////////////////////////////////



/*

  To Do:

	- Why is shading of texture mapped polys different than savg-cube polys?
	- Should handle all possibilities inherent in the .vol file format
	- Will this work if the volume is placed under a non-identity DCS?
	- frame accuracy
	- Make all file global variables (e.g. sliceVertex) into class vars?
	- Shared memory command mechanism for turning things on and off?
	- More flexibility in specifying visibility of slices & box?
	- Destructor should free things.

*/





#include <dpf.h>
#include <GL/gl.h>
#include <Performer/pr/pfTexture.h>
#include <Performer/pr/pfMaterial.h>
#include <dtk/dtkDSO_loader.h>

// data type for incoming box data
typedef double Box[2][3];

// Function for getting the current box
Box *getRubberbandBoxPtr ();

//
// These give the underlying geometry for each of the texture mapped polygons.
//
// Terminology: slice polygon - an axis aligned cross-section polygon through 
//				the whole volume
//		box polygon   - an polygon corresponding to a face of an
//				axis aligned box.
//
// All coordinates are in world coordinates.
//
// Items dimensioned [2][3] are arranged such that [0][*] are
// the minimum xyz, and [1][*] are the maximum xyz.
//
static pfVec3 *sliceVertex[2][3];
static pfVec3 *boxVertex[2][3];
static pfVec3 *sliceNormal[2][3];
static pfVec3 *boxNormal[2][3];

//
// Performer geosets for the slice and box polygons
static pfGeoSet *sliceGSet[2][3];
static pfGeoSet *boxGSet[2][3];

//
// Indicates whether box polys and slice polys are to be displayed.
//
static int BoxOn = 0;
static int SlicesOn = 0;


// Which slices are currently visible.  This depends of head position.
// Code depends on only one of each pair of slices being visible at any time
static int currentSliceVis[2][3];




//
// Class for the dso.  
//
// Perhaps some of the static globals above should be class variables.
//
class boxVolSlice : public dtkAugment {

public:

  // required public methods
  boxVolSlice (dpf *app) ;
  int preConfig (void);
  int postConfig (void);
  int preFrame (void);
  int postFrame (void);
  ~boxVolSlice ();


private:

  dpf* app;
  dpfDisplay *display;
  dtkInLocator *head;
  Box *currentBox;

  pfTexture *texture;
  pfGeode *sliceNode;
  pfGeode *boxNode;

  // make Performer texture mapped polygons
  int makePF_TMSlices ();

  // visibility based on head
  void adjustSliceVisibility ();

  // create the pfTexture
  int makeTexture (char *volFN);

} ;









//////////////////////////////////////////////////////////////////////

// Constructor

// Set description and set object vars
// be sure to validate it!
boxVolSlice::boxVolSlice (dpf *app_in) :
  dtkAugment ("boxVolSlice")
{

  // Most of the real initialization happens in preFrame.

  setDescription (
    "boxVolSlice DSO: Provides axis-aligned texture mapped polygons based \n"
  "  on a 3D texture and a box provided by the another DSO.\n");

  app = app_in;

  display = NULL;
  head = NULL;
  currentBox = NULL;
  texture = NULL;
  sliceNode = NULL;
  boxNode = NULL;

  // be sure to validate it!
  validate ();
} // end of constructor


boxVolSlice::~boxVolSlice ()
{

// JGH should pfFree things here

}  // end of destructor
//////////////////////////////////////////////////////////////////////









//////////////////////////////////////////////////////////////////////

//
// preConfig is called only once by dpf, after pfInit() and 
//  dpfDisplay:preConfig() but before pfConfig()
//
int boxVolSlice::preConfig (void) {

  // Apparently nothing should be done here.  We are postponing
  // initializations as long as possible.  See preFrame.

  return dtkAugment::CONTINUE ;
}   // end of preConfig


//
// postConfig is called after pfConfig
//
int boxVolSlice::postConfig (void) {


  head = (dtkInLocator *) app->get("head", DTKINLOCATOR_TYPE);
  if(!head || head->isInvalid())
    {
    head = NULL ;
    dtkMsg.add (DTKMSG_NOTICE,
        "rubberbandBox::postConfig: "
        "Unable to get valid head position pointer.\n");
    }


  return dtkAugment::REMOVE_CALLBACK ;
}  // end of postConfig

//
//////////////////////////////////////////////////////////////////////








//////////////////////////////////////////////////////////////////////

//
// Some utility routines
//


//
// Puts the indicated node under the named parent node.
//
static int
putInSceneGraph (dpfDisplay *display, pfNode *node, char *parentName)
  {
  pfGroup *parent = NULL;

  parent = (pfGroup*) display->world()->
                         lookup (parentName, pfGroup::getClassType()) ;
  if (!parent)
      {
      parent= (pfGroup*) display->world()->
			 lookup (parentName, pfDCS::getClassType()) ;
      }

  if (!parent)
      {
      parent= (pfGroup*) display->world()->
			 lookup (parentName, pfSCS::getClassType()) ;
      }

  if (!parent)
      {
      return -1;
      }


  parent->addChild (node);
  return 0;

  }  // end of putInSceneGraph



//
// Trim whitespace from the beginning and end of a string
//
static void trim (char *s)
  {
  int start, len, i;

  start = strspn (s, " \n\t\v\f\r");

  len = strcspn (s+start, " \n\t\v\f\r");

  for (i = 0; i < len; i++)
	{
	s[i] = s[i+start];
	}
  s[len] = 0;

  }  // end of trim


//
// these defines identify the two types of texture mapped polygons.
//
#define SIDES_BOX 1
#define SIDES_CROSS_SECTIONS 2


//
// This data type contains all of the info derived from the .vol file.
//
typedef struct
	{
	int dim[3];
	int numComponents;
	char voxelFN[300];
	char lutFN[300];
	char interpMethod[30];
	int interpTypePF;
	int whichSides;
	} VolumeInfo;

#if 0
//
// Debugging routine:
//
static void
printVolInfo (FILE *fp, VolumeInfo *volInfo)
{

  fprintf (fp, "VolumeInfo at %p\n", volInfo);
  fprintf (fp, "  dim = %d %d %d\n", 
			volInfo->dim[0],
			volInfo->dim[1],
			volInfo->dim[2]);

  fprintf (fp, "  numComponents = %d\n", 
			volInfo->numComponents);

  fprintf (fp, "  voxelFN = %s\n", 
			volInfo->voxelFN);

  fprintf (fp, "  lutFN = %s\n", 
			volInfo->lutFN);

  fprintf (fp, "  interpMethod = %s\n", 
			volInfo->interpMethod);

  fprintf (fp, "  interpTypePF = %d\n", 
			volInfo->interpTypePF);

  fprintf (fp, "  whichSides = %d\n", 
			volInfo->whichSides);

  fprintf (fp, "\n");


}  // end of printVolInfo
#endif

//
// Trims string at first "#"
//
static void 
trimComment (char *c)
  {
  while (*c)
    {
    if (*c == '#')
	{
	*c = 0;
	return;
	}
    c++;
    }
  }  // end of trimComment


//
// converts all chars in string to upper case
//
static void
cvtToUpper (char *c)
  {
  while (*c)
	{
	*c = toupper (*c);
	c++;
	}
  }  // end of cvtToUpper




//
// Reads the contents of a .vol file into a VolumeInfo data structure.
// This code parses the format described above.
//
static int
getVolFileInfo (char *volFN, VolumeInfo *volInfo)
{
  int lineNum;
  char keyword[1000];
  char valueToken[1000];
  int x, y, z;

  // set defaults:
  volInfo->dim[0] = 0;
  volInfo->numComponents = 0;
  volInfo->voxelFN[0] = 0;
  volInfo->lutFN[0] = 0;
  volInfo->interpMethod[0] = 0;
  volInfo->interpTypePF = PFTEX_POINT;
  volInfo->whichSides = 0;

  FILE *fp;

  char line[1000];

  if ( (fp = fopen (volFN, "r")) == NULL)
	{
        dtkMsg.add (DTKMSG_ERROR,
          "boxVolSlice::getVolFileInfo failed: "
          "Unable to open volume file < %s >.\n", volFN);
	return -1;
	}

  lineNum = 1;
  while (fgets (line, sizeof(line), fp) != NULL) 
	{

	// printf ("got line <%s>\n", line);

	trimComment (line);

	if (sscanf (line, "%s", keyword) == 1)
	  {


	  if (strcasecmp (keyword, "dim") == 0)
	    {
	    if (sscanf (line, "%s %d %d %d", keyword, &x, &y, &z) != 4)
		{
        	dtkMsg.add (DTKMSG_INFO,
          	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
		}
	    else
		{
		volInfo->dim[0] = x;
		volInfo->dim[1] = y;
		volInfo->dim[2] = z;
		}
	    }
	  else if (strcasecmp (keyword, "voxelFN") == 0)
	    {
	    if (sscanf (line, "%s %s", keyword, valueToken) != 2)
		{
        	dtkMsg.add (DTKMSG_INFO,
          	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
		}
	    else
		{
		strcpy (volInfo->voxelFN, valueToken);
		}
	    }
	  else if (strcasecmp (keyword, "lutFN") == 0)
	    {
	    if (sscanf (line, "%s %s", keyword, valueToken) != 2)
		{
        	dtkMsg.add (DTKMSG_INFO,
          	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
		}
	    else
		{
		strcpy (volInfo->lutFN, valueToken);
		}
	    }
	  else if (strcasecmp (keyword, "numComponents") == 0)
	    {
	    if (sscanf (line, "%s %d", keyword, &x) != 2)
		{
        	dtkMsg.add (DTKMSG_INFO,
          	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
		}
	    else
		{
		volInfo->numComponents = x;
		}
	    }
	  else if (strcasecmp (keyword, "interpMethod") == 0)
	    {
	    if (sscanf (line, "%s %s", keyword, valueToken) != 2)
		{
        	dtkMsg.add (DTKMSG_INFO,
          	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
		}
	    else
		{
		strcpy (volInfo->interpMethod, valueToken);
		if (strcasecmp (valueToken, "POINT") == 0)
			{
			volInfo->interpTypePF = PFTEX_POINT;
			}
		else if (strcasecmp (valueToken, "LINEAR") == 0)
			{
			volInfo->interpTypePF = PFTEX_LINEAR;
			}
		else if (strcasecmp (valueToken, "BILINEAR") == 0)
			{
			volInfo->interpTypePF = PFTEX_BILINEAR;
			}
		else if (strcasecmp (valueToken, "TRILINEAR") == 0)
			{
			volInfo->interpTypePF = PFTEX_TRILINEAR;
			}
		else if (strcasecmp (valueToken, "QUADLINEAR") == 0)
			{
			volInfo->interpTypePF = PFTEX_QUADLINEAR;
			}
		else
			{
        		dtkMsg.add (DTKMSG_INFO,
          	  	  "boxVolSlice::getVolFileInfo: "
		  	  "Unrecognized interpolation method on "						  "line %d from volume file.\n", lineNum);
			}
		}
	    }
	  else if (strcasecmp (keyword, "whichSides") == 0)
	    {
	    if (sscanf (line, "%s %s", keyword, valueToken) != 2)
		{
        	dtkMsg.add (DTKMSG_INFO,
          	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
		}
	    else
		{
		cvtToUpper (valueToken);
		volInfo->whichSides = 
		  (strstr(valueToken, "BOX") ? SIDES_BOX : 0) |
		  (strstr(valueToken, "CROSS_SECTIONS") ? SIDES_CROSS_SECTIONS : 0) ;

		if (volInfo->whichSides == 0)
			{
        		dtkMsg.add (DTKMSG_INFO,
          	  	  "boxVolSlice::getVolFileInfo: "
		  	  "Error reading line %d from volume file < %s >.\n", 
		  	  lineNum, volFN);
			}

		}

	    }
	  else
	    {
	    dtkMsg.add (DTKMSG_INFO,
            	  "boxVolSlice::getVolFileInfo: "
		  "Error reading line %d from volume file < %s >.\n", 
		  lineNum, volFN);
	    }




	  }

	lineNum++;
	}

  fclose (fp);


  
  return 0;
}  // end of getVolFileInfo

//
// This routine gets the value of the env var BOX_VOL_SLICE_FILE and
// parses it into three parts.  The format of the string is:
//
//          volumeFilename:boxSliceNodeName:parentName
// where
//
//	volumeFilename is the name of the file describing the volume
//	boxSliceNodename is the name to be given to the pfNode containing
//		the box and slice polygons
//	parentName is the name of the node under which the box slice node
//		is to be placed
//
// Defaults if not specified:
//	volumeFilename - no default, must be specified
//	boxSliceNodename - "boxVolSlice"
//	parentName - "world"
//

static int
getEnvVarInfo (char *volFN, char *boxSliceNodeName, char *parentName)
{
  char *es;
  char *ci, *co;

  if ( (es = getenv ("BOX_VOL_SLICE_FILE")) == NULL)
	{
	fprintf (stderr, "No volume env var present.\n");
	return -1;
	}

  // march though the string using ci as pointer to current char
  ci = es;

  // co is pointer to where we are putting the chars we get out of env str
  // copy chars until ':' or end of string
  co = volFN;
  while ((*ci != ':') && (*ci != 0))
	{
	*co++ = *ci++;
	}
  *co = 0;

  // skip :
  if (*ci == ':')
	{
	ci++;
	}

  co = boxSliceNodeName;
  while ((*ci != ':') && (*ci != 0))
	{
	*co++ = *ci++;
	}
  *co = 0;

  if (*ci == ':')
	{
	ci++;
	}

  co = parentName;
  while ((*ci != ':') && (*ci != 0))
	{
	*co++ = *ci++;
	}
  *co = 0;


  // trim off white space
  trim (volFN);
  trim (boxSliceNodeName);
  trim (parentName);

  // vol file name is the only one we absolutely have to have.
  if (strlen (volFN) == 0)
	{
	fprintf (stderr, "No volume file name present.\n");
	return -1;
	}

  // If others aren't present, then we provide defaults

  if (strlen (boxSliceNodeName) == 0)
	{
	strcpy (boxSliceNodeName, "boxVolSlice");
	}

  if (strlen (parentName) == 0)
	{
	strcpy (parentName, "world");
	}

  return 0;

}  // end of getEnvVarInfo





//
// Make the 3D pfTexture based on the contents of the vol file.
//
int boxVolSlice::makeTexture (char *volFN)
{
  VolumeInfo volInfo;


  // read the vol file
  if (getVolFileInfo (volFN, &volInfo))
	{
	// error
	return -1;
	}

  //  debug
  // printVolInfo (stdout, &volInfo);


  // Which polys are to be displayed:
  SlicesOn = volInfo.whichSides & SIDES_CROSS_SECTIONS;
  BoxOn = volInfo.whichSides & SIDES_BOX;

  if (volInfo.numComponents != 1)
	{
	fprintf (stderr, "Error: Volume numComponents must be 1.\n");
	dtkMsg.add (DTKMSG_ERROR,
            	  "boxVolSlice::makeTexture: "
		  "Currently number of components in volume data must be 1.\n");
	return -1;
	}


  // Create the pfTexture
  texture = new pfTexture;

  // set the filter based on the contents of the vol file
  texture->setFilter (PFTEX_MINFILTER, volInfo.interpTypePF);
  texture->setFilter (PFTEX_MAGFILTER, volInfo.interpTypePF);


  // The total number of voxels:
  unsigned int imgDataLen = volInfo.dim[0] * volInfo.dim[1] * volInfo.dim[2];

  // alloc pf memory for texture image

  unsigned int *txtrImg = (unsigned int *) pfMalloc (3*imgDataLen);
  if (txtrImg == NULL)
	{
	dtkMsg.add (DTKMSG_ERROR,
            	  "boxVolSlice::makeTexture: "
		  "Unable to pfMalloc sufficient memory for texture image.\n");
	return -1;
	}

  FILE *fp;


  // alloc pf heap memory for raw texture 
  static unsigned char *imgData = (unsigned char *) malloc ( imgDataLen );
  if (imgData == NULL)
	{
	// error 
	dtkMsg.add (DTKMSG_ERROR,
            	  "boxVolSlice::makeTexture: "
		  "Unable to malloc sufficient memory for texture data.\n");
	return -1;
	}




  // read in the raw texture data

  if ( (fp = fopen (volInfo.voxelFN, "rb")) == NULL )
	{
	dtkMsg.add (DTKMSG_ERROR,
            	  "boxVolSlice::makeTexture: "
		  "Unable to open texture voxel file < %s >.\n", volInfo.voxelFN);
	return -1;
	}

  if (fread (imgData, 1, imgDataLen, fp) != imgDataLen)
	{
	dtkMsg.add (DTKMSG_ERROR,
            	  "boxVolSlice::makeTexture: "
		  "Unable to read texture voxel file < %s >.\n", volInfo.voxelFN);
	fclose (fp);
	return -1;
	}

  fclose (fp);

  // Now move the raw texture data to the Performer texture image by
  // running it through the looup table.

  unsigned char *c_out = (unsigned char *) txtrImg;
  unsigned char *c_in = imgData;


  // First make the lookup table


  float lut[256][4];

  unsigned int ii, jj;

  if (strlen (volInfo.lutFN) == 0)
    {

    // if the file name is not there, just make a standard grey LUT

    for (jj = 0; jj < 256; jj++)
	{
        for (ii = 0; ii < 3; ii++)
		{
		lut[jj][ii] = jj/255.0;
		}
	lut[jj][3] = 1.0;
	}
    }
  else
    {

    // read the lookup table file

    if ( (fp = fopen (volInfo.lutFN, "r")) == NULL )
	{
	fprintf (stderr, "Error openning texture file.\n");
	return -1;
	}

    for (jj = 0; jj < 256; jj++)
	{
        for (ii = 0; ii < 4; ii++)
		{
		if (fscanf (fp, "%f", &(lut[jj][ii])) != 1)
		  {
		  dtkMsg.add (DTKMSG_ERROR,
            	    "boxVolSlice::makeTexture: "
		    "Unable to read lookup table file < %s >.\n", volInfo.lutFN);
    		  fclose (fp);
		  return -1;
		  }
		
		}
	}
    fclose (fp);
    }


  c_out = (unsigned char *) txtrImg;
  c_in = imgData;
  for (ii = 0; ii < imgDataLen; ii++)
	{
	// do the lookup
	*(c_out++) = (unsigned char) (255*lut[*c_in][0]);
	*(c_out++) = (unsigned char) (255*lut[*c_in][1]);
	*(c_out++) = (unsigned char) (255*lut[*c_in][2]);
	c_in++;
	}


  // set the texture image
  texture->setImage (txtrImg, 3, 
			volInfo.dim[0], volInfo.dim[1], volInfo.dim[2]);

  return 0;

}  // end of boxVolSlice::makeTexture









//
// makePF_TMSlices makes the Performer objects that contain
//   the texture mapped slice and box polygons.  The variables
//   that point to these objects are either class variables or
//   file global variables.
//
int boxVolSlice::makePF_TMSlices ()
  {

  // printf ("Beginning of makePF_TMSlices.\n");

  int i, j;
  char volFN [300];
  char boxSliceNodeName [100];
  char parentName [100];

  // Get the volume file name, etc. from the environment.
  if (getEnvVarInfo (volFN, boxSliceNodeName, parentName))
        {
	// Error message generated inside call
printf ("env var info\n");
        return -1;
        }

#if 0
  printf ("volFN = < %s >\n", volFN);
  printf ("boxSliceNodeName = < %s >\n", boxSliceNodeName );
  printf ("parentName = < %s >\n", parentName);
#endif

  // make the pfTexture 
  if (makeTexture (volFN))
	{
printf ("texture\n");
	return -1;
	}







  // Allocate performer memory for the vertices and normals
  for (j = 0; j < 2; j++)
    {
    for (i = 0; i < 3; i++)
      {
      sliceVertex[j][i] = (pfVec3*) pfMalloc(sizeof(pfVec3)*(4)) ;
      boxVertex[j][i] = (pfVec3*) pfMalloc(sizeof(pfVec3)*(4)) ;
      sliceNormal[j][i] = (pfVec3*) pfMalloc(sizeof(pfVec3)) ;
      boxNormal[j][i] = (pfVec3*) pfMalloc(sizeof(pfVec3)) ;
      }
    }



  // Note that because the polygons are always axis aligned, the
  // normals never change.

  sliceNormal[0][0]->set ( 1.0, 0.0, 0.0);
  sliceNormal[0][1]->set ( 0.0, 1.0, 0.0);
  sliceNormal[0][2]->set ( 0.0, 0.0, 1.0);
  sliceNormal[1][0]->set (-1.0, 0.0, 0.0);
  sliceNormal[1][1]->set ( 0.0,-1.0, 0.0);
  sliceNormal[1][2]->set ( 0.0, 0.0,-1.0);

  boxNormal[0][0]->set (-1.0, 0.0, 0.0);
  boxNormal[0][1]->set ( 0.0,-1.0, 0.0);
  boxNormal[0][2]->set ( 0.0, 0.0,-1.0);
  boxNormal[1][0]->set ( 1.0, 0.0, 0.0);
  boxNormal[1][1]->set ( 0.0, 1.0, 0.0);
  boxNormal[1][2]->set ( 0.0, 0.0, 1.0);


  // Now the colors.  Even though the polygons are texture mapped, 
  // we still seem to need to specify colors.  
  //
  // This bears some additional investigation.
  //
  pfVec4 *colors = (pfVec4*) pfMalloc(sizeof(pfVec4)) ;
  *colors = pfVec4(1.0, 1.0, 1.0, 1.0) ;

  pfTexEnv *tev;
  tev = new pfTexEnv;
  // tev->setBlendColor (1.0, 1.0, 1.0, 1.0);
  // here is where we specify how the texture color/alpha
  // and the geometry color/alpha are combined to result
  // in the color/alpha of the texture-mapped geometry
  // tev->setMode(PFTE_BLEND);



  // We create a pfGeoState that is used for all of the polygons
  pfGeoState *gstate = new pfGeoState;
  gstate->setAttr(PFSTATE_TEXTURE, texture);
  gstate->setMode(PFSTATE_ENTEXTURE,PF_ON);
  gstate->setMode(PFSTATE_ENLIGHTING,PF_ON);
  // gstate->setMode(PFSTATE_CULLFACE,PFCF_OFF);
  gstate->setAttr(PFSTATE_TEXENV, tev);


#if 0

    // Apparently this is not needed.

    pfMaterial *mat = new pfMaterial;
    mat->setSide (PFMTL_BOTH);
    gstate->setAttr(PFSTATE_FRONTMTL, mat); 
    gstate->setAttr(PFSTATE_BACKMTL, mat); 
#endif

    // the pfTexgen that we create here means that the texture coordinates
    // will be taken as identical with the vertex coordinates.
    pfTexGen *texgen = new pfTexGen; 
    gstate->setMode(PFSTATE_ENTEXGEN, PF_ON); 
    gstate->setAttr(PFSTATE_TEXGEN, texgen); 

    texgen->setMode(PF_S, PFTG_OBJECT_LINEAR); 
    texgen->setMode(PF_T, PFTG_OBJECT_LINEAR); 
    texgen->setMode(PF_R, PFTG_OBJECT_LINEAR); 
    texgen->setMode(PF_Q, PFTG_OFF); 
    texgen->setPlane(PF_S, 1.0f, 0.0f, 0.0f, 0.0f); 
    texgen->setPlane(PF_T, 0.0f, 1.0f, 0.0f, 0.0f); 
    texgen->setPlane(PF_R, 0.0f, 0.0f, 1.0f, 0.0f);



  // create separate nodes for the cross sections and the box polygons
  sliceNode = new pfGeode;
  boxNode = new pfGeode;



  // Create one pfGeoSet for each of the polygons

  for (j = 0; j < 2; j++)
    {
    for (i = 0; i < 3; i++)
      {

      // sliceGSet[j][i] = (pfGeoSet *) pfMalloc (sizeof(pfGeoSet));
      sliceGSet[j][i] = new pfGeoSet;
      sliceGSet[j][i]->setPrimType(PFGS_QUADS) ;
      sliceGSet[j][i]->setNumPrims(1) ;
      sliceGSet[j][i]->
	setAttr (PFGS_COORD3, PFGS_PER_VERTEX, sliceVertex[j][i], NULL);
      sliceGSet[j][i]->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, colors, NULL) ;
      sliceGSet[j][i]->
	setAttr(PFGS_NORMAL3,  PFGS_PER_PRIM, sliceNormal[j][i], NULL) ;
      sliceGSet[j][i]->setGState (gstate);
  
      // boxGSet[j][i] = (pfGeoSet *) pfMalloc (sizeof(pfGeoSet));
      boxGSet[j][i] = new pfGeoSet;
      boxGSet[j][i]->setPrimType(PFGS_QUADS) ;
      boxGSet[j][i]->setNumPrims(1) ;
      boxGSet[j][i]->
	setAttr (PFGS_COORD3, PFGS_PER_VERTEX, boxVertex[j][i], NULL);
      boxGSet[j][i]->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, colors, NULL) ;
#if 1
      boxGSet[j][i]->
	setAttr(PFGS_NORMAL3,  PFGS_PER_PRIM, boxNormal[j][i], NULL) ;
#endif
      boxGSet[j][i]->setGState (gstate);
      // All of the box polygons are placed into the boxNode
      boxNode->addGSet (boxGSet[j][i]);
      }
   }

  // Only three of the slice (cross-section) polygons are placed into
  // the sliceNode.  Only one member of each pair of parallel slices
  // are made visible at any one time.  The decision as to which one
  // is visible is made based on the head position.
  // 
  // Because we don't have the head postion right now, 
  // We'll start out with the "lower" slices visible.
  sliceNode->addGSet (sliceGSet[0][0]);
  sliceNode->addGSet (sliceGSet[0][1]);
  sliceNode->addGSet (sliceGSet[0][2]);

  // This indicates which slices are currently visible.
  currentSliceVis[0][0] = 1;
  currentSliceVis[0][1] = 1;
  currentSliceVis[0][2] = 1;
  currentSliceVis[1][0] = 0;
  currentSliceVis[1][1] = 0;
  currentSliceVis[1][2] = 0;
  

  // We now create a group node to hold one or both of these.
  pfGroup *group = new pfGroup;
  if (SlicesOn)
    {
    group->addChild (sliceNode);
    }

  if (BoxOn)
    {
    group->addChild (boxNode);
    }

  // we give the node its name
  group->setName (boxSliceNodeName);

  // and we place the group under the spcified parent node
  putInSceneGraph (display, group, parentName);

  return 0;
  }  // end of boxVolSlice::makePF_TMSlices




//
// This define clamps the vertex coordinates to the range (0,1).
// Because of a depth-fighting-like display when we are at
// 0 or 1, we actually clamp to values to (0.001, 0.999).
//
#define CLAMP(a)   (((a)<=0)?(0.001):(((a)>=1)?(0.999):(a)))



//
// Make the slice and box polygon vertex data match the current
// box data.
//
static void
matchBox (double box[2][3], pfVec3 *sliceVertex[2][3], pfVec3 *boxVertex[2][3])
	{

	float minX, minY, minZ, maxX, maxY, maxZ;
	
	// Clamp the box corners to the appropriate coordinate range.
	minX = CLAMP (box[0][0]);
	minY = CLAMP (box[0][1]);
	minZ = CLAMP (box[0][2]);
	maxX = CLAMP (box[1][0]);
	maxY = CLAMP (box[1][1]);
	maxZ = CLAMP (box[1][2]);

	// Slice polygonss cut through the entire volume
	sliceVertex[0][0][0].set (minX, 0.0, 0.0);
	sliceVertex[0][0][1].set (minX, 1.0, 0.0);
	sliceVertex[0][0][2].set (minX, 1.0, 1.0);
	sliceVertex[0][0][3].set (minX, 0.0, 1.0);

	sliceVertex[0][1][0].set (0.0, minY, 0.0);
	sliceVertex[0][1][1].set (1.0, minY, 0.0);
	sliceVertex[0][1][2].set (1.0, minY, 1.0);
	sliceVertex[0][1][3].set (0.0, minY, 1.0);

	sliceVertex[0][2][0].set (0.0, 0.0, minZ);
	sliceVertex[0][2][1].set (1.0, 0.0, minZ);
	sliceVertex[0][2][2].set (1.0, 1.0, minZ);
	sliceVertex[0][2][3].set (0.0, 1.0, minZ);

	sliceVertex[1][0][0].set (maxX, 0.0, 0.0);
	sliceVertex[1][0][1].set (maxX, 1.0, 0.0);
	sliceVertex[1][0][2].set (maxX, 1.0, 1.0);
	sliceVertex[1][0][3].set (maxX, 0.0, 1.0);

	sliceVertex[1][1][0].set (0.0, maxY, 0.0);
	sliceVertex[1][1][1].set (1.0, maxY, 0.0);
	sliceVertex[1][1][2].set (1.0, maxY, 1.0);
	sliceVertex[1][1][3].set (0.0, maxY, 1.0);

	sliceVertex[1][2][0].set (0.0, 0.0, maxZ);
	sliceVertex[1][2][1].set (1.0, 0.0, maxZ);
	sliceVertex[1][2][2].set (1.0, 1.0, maxZ);
	sliceVertex[1][2][3].set (0.0, 1.0, maxZ);



	// Box polygons match the extent of the (clamped) box.
	boxVertex[0][0][3].set (minX, minY, minZ);
	boxVertex[0][0][2].set (minX, maxY, minZ);
	boxVertex[0][0][1].set (minX, maxY, maxZ);
	boxVertex[0][0][0].set (minX, minY, maxZ);

	boxVertex[0][1][0].set (minX, minY, minZ); 
	boxVertex[0][1][1].set (maxX, minY, minZ);
	boxVertex[0][1][2].set (maxX, minY, maxZ);
	boxVertex[0][1][3].set (minX, minY, maxZ);

	boxVertex[0][2][0].set (minX, minY, minZ);
	boxVertex[0][2][1].set (maxX, minY, minZ);
	boxVertex[0][2][2].set (maxX, maxY, minZ);
	boxVertex[0][2][3].set (minX, maxY, minZ);

	boxVertex[1][0][0].set (maxX, minY, minZ);
	boxVertex[1][0][1].set (maxX, maxY, minZ);
	boxVertex[1][0][2].set (maxX, maxY, maxZ);
	boxVertex[1][0][3].set (maxX, minY, maxZ);

	boxVertex[1][1][0].set (minX, maxY, minZ); 
	boxVertex[1][1][1].set (maxX, maxY, minZ);
	boxVertex[1][1][2].set (maxX, maxY, maxZ);
	boxVertex[1][1][3].set (minX, maxY, maxZ);

	boxVertex[1][2][0].set (minX, minY, maxZ);
	boxVertex[1][2][1].set (maxX, minY, maxZ);
	boxVertex[1][2][2].set (maxX, maxY, maxZ);
	boxVertex[1][2][3].set (minX, maxY, maxZ);

	// Force the pfGeoSets to recalculate their bounding boxes
	// to ensure proper culling.

        boxGSet[0][0]->setBound (NULL, PFBOUND_DYNAMIC);
        boxGSet[0][1]->setBound (NULL, PFBOUND_DYNAMIC);
        boxGSet[0][2]->setBound (NULL, PFBOUND_DYNAMIC);
        boxGSet[1][0]->setBound (NULL, PFBOUND_DYNAMIC);
        boxGSet[1][1]->setBound (NULL, PFBOUND_DYNAMIC);
        boxGSet[1][2]->setBound (NULL, PFBOUND_DYNAMIC);

        sliceGSet[0][0]->setBound (NULL, PFBOUND_DYNAMIC);
        sliceGSet[0][1]->setBound (NULL, PFBOUND_DYNAMIC);
        sliceGSet[0][2]->setBound (NULL, PFBOUND_DYNAMIC);
        sliceGSet[1][0]->setBound (NULL, PFBOUND_DYNAMIC);
        sliceGSet[1][1]->setBound (NULL, PFBOUND_DYNAMIC);
        sliceGSet[1][2]->setBound (NULL, PFBOUND_DYNAMIC);


	}  // end of matchBox





//
// We're calling preFrame only once: to get display and setup performer
// data structures.
//
int boxVolSlice::preFrame (void)
{
  display = app->display ();

  if (!display) 
    {
    dtkMsg.add(DTKMSG_ERROR,
                 "boxVolSlice::preFrame() failed:"
                 "Unable to get display pointer.\n");
    return dtkAugment::ERROR_;
    }

  // get the pointer to the current box from the rubberbandBox DSO
  currentBox = (Box *) getRubberbandBoxPtr ();

  // make the Performer objects
  if (makePF_TMSlices ())
    {
    dtkMsg.add(DTKMSG_ERROR,
                 "boxVolSlice::preFrame() failed:"
                 "Unable to make Performer data structures.\n");
    return dtkAugment::ERROR_;
    }


  // match the texture mapped poly vertices to the current box.
  matchBox (*currentBox, sliceVertex, boxVertex);

  // force boxNode to recalc it's bounding volume.
  boxNode->setBound (NULL, PFBOUND_DYNAMIC);

  return dtkAugment::REMOVE_CALLBACK;
}    // end of preFrame


//
// We call postFrame every frame to adjust the box and set
//  the visibility of the slices
//
int boxVolSlice::postFrame (void) {


  matchBox (*currentBox, sliceVertex, boxVertex);
  boxNode->setBound (NULL, PFBOUND_DYNAMIC);

  adjustSliceVisibility ();

  return dtkAugment::CONTINUE ; 
}    // end of postFrame


// This routine looks at the head position and decides which
// of the cross-section polygons should be visible.
//
// The idea is that the cross-section polygons on the far side
// of the box will be visible, providing a backdrop or visual 
// context for the box.
//
void boxVolSlice::adjustSliceVisibility () {

  float headData[6];
  static pfCoord headCoord;
  int i;

  if (head == NULL)
    {
    return;
    }

   head->read (headData);

   headCoord.xyz.set (headData[0], headData[1], headData[2]);
   headCoord.hpr.set (0.0, 0.0, 0.0);
   display->dpfToWorld (&headCoord);


  // Only one of each pair of parallel slices is displayed at one time.

  // Once a slice is made visible, we switch to the other (parallel) slice
  // only when we cross over to the back side of the visible slice.
  //
  // This means that when we are between the slices, we still only see
  // one of the two slices, and we hold onto the currently visible
  // slice as long as we are on the right side of it.

  for (i = 0; i < 3; i++)
    { 

    // For each pair, we check if we are still on the "front" slide
    // of the currently visible member of the pair.  If so, then
    // no change is required.  If not, we switch to the other slice.

    if (currentSliceVis[0][i])
	{
	if ( headCoord.xyz[i] < (*currentBox)[0][i] )
		{
		currentSliceVis[0][i] = 0;
    		sliceNode->removeGSet (sliceGSet[0][i]);
		currentSliceVis[1][i] = 1;
    		sliceNode->addGSet (sliceGSet[1][i]);
		}
	}
    else  // currentSliceVis[1][i] is true
	{
	if ( headCoord.xyz[i] > (*currentBox)[1][i] )
		{
		currentSliceVis[1][i] = 0;
    		sliceNode->removeGSet (sliceGSet[1][i]);
		currentSliceVis[0][i] = 1;
    		sliceNode->addGSet (sliceGSet[0][i]);
		}
	}
    }



}  // end of adjustSliceVisibility








//////////////////////////////////////////////////////////////////////




/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new boxVolSlice (app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
