// This loads defines a clipping plane according to the wand position
//
// DIVERSE graphics interface -- Performer ---- version 2.2.0
// released: Fri Oct 18 09:27:29 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DGIPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//

#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <fstream.h>

static double* planeEq;
// shared memory for plane normal (from wand)
dtkSharedMem* shmSlice;
// shared memory for place distance from wand
dtkSharedMem* shmDist;
//  Shared memory segment for the on-off button
dtkSharedMem* shm2;
//  Shared memory segment for the spin rate
dtkInLocator *wand ;
// Shared memory for wandPointer
dtkSharedMem* shm4;
unsigned char selected;
float wandData[6];
pfCoord wandCoord;
pfVec3 check;

class clipWand : public dtkAugment {
public:

  clipWand(dpf *app) ;
  int preConfig(void);
  int postConfig(void);
  int preFrame(void);
  int postFrame(void);
  ~clipWand();

private:
  dpf* app;
  dpfDisplay *display;
} ;

//  Enable the clipWand plane in pre draw
int preDraw (pfTraverser *trav, void *data) {

  // first see if the on button is set to on
  unsigned char onoff;  
  if (shm2) {
    shm2->read(&onoff);
  }

  if (onoff == 1) {
    float vals[3];
    if (shmSlice)
      shmSlice->read(vals);
    float valD;
    if (shmDist)
      shmDist->read(&valD);
    
    static void* firstChan = NULL;
    if (firstChan == NULL) 
      firstChan = trav->getChan();
    
    if (firstChan == trav->getChan()) {
      // printf ("calculating plane EQ\n");
      
      //  Set up the coefficients in the plane equation
      //  Ax + By + Cz = D
      
      //  A,B,C correspond to the normal vector coefficients
      int c;
      for (c=0; c<3; c++) {
	planeEq[c] = vals[c];
      }
      
      //  Plug in position to solve for D
      planeEq[3] = -valD;
    }

    //printf ("enabling clipWand plane, channel %x\n",trav->getChan());
    glClipPlane(GL_CLIP_PLANE0, planeEq);
    glEnable(GL_CLIP_PLANE0);
  }

  return PFTRAV_CONT;
}

//  Disable the clip plane in post draw
int postDraw (pfTraverser *trav, void *data) {
  //printf ("disabling clip plane, channel %x\n", trav->getChan());
  glDisable(GL_CLIP_PLANE0);

  return PFTRAV_CONT;
}

// create object and set description.
// be sure to validate it!
clipWand::clipWand(dpf *app_in) :
  dtkAugment("clipWand")
{
  app = app_in;
  setDescription("clipping plane") ;
  validate();

} ;


clipWand::~clipWand()
{
   delete shmSlice;
   delete shmDist;
   delete shm2;
   delete shm4;
}

// called only once by dpf, after pfInit() and dpfDisplay:preConfig()
// but before pfConfig()
int clipWand::preConfig(void) {
  //  Allocate plane equation out of Performer shared memory
  planeEq = (double*) pfMalloc (4 * sizeof(double), pfGetSharedArena());

  shmSlice = new dtkSharedMem(3 * sizeof(float), "slice");
  shmDist = new dtkSharedMem( sizeof(float), "dval");

  //  Check if there is a segment "doclip" the on-off button
  shm2 = new dtkSharedMem(sizeof(unsigned char), "doclip");
  if(!shm2 || shm2->isInvalid()) {
      shm2 = NULL;
  }

  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand || wand->isInvalid())
    wand = NULL ;

  shm4 = new dtkSharedMem(sizeof(unsigned char), "pointerOn");
  if(!shm4 || shm4->isInvalid()) {
    shm4 = NULL;
  }

  // turn button off to start
  selected = 0;
  if (shm2)
     shm2->write(&selected);
  
  // start out by hiding the wandPointer
  if (shm4) {
    selected = 1;
    shm4->write(&selected);
  }

  return DTK_CONTINUE ;
}


int clipWand::preFrame(void)
{
  display = app->display();
  if(!display) return DTK_CALLBACK_ERROR;

  return DTK_REMOVE_CALLBACK;
}

int clipWand::postConfig(void) {

  //  Set the preDraw and postDraw functions set up the clipping planes
  app->display()->world()->setTravFuncs (PFTRAV_DRAW, preDraw, postDraw);
  return DTK_REMOVE_CALLBACK ;
}

int clipWand::postFrame(void) {
  
  // read the wand direction
  if (wand)
    wand->read(wandData);

  // define the wand data in world coordinates
  wandCoord.xyz.set(wandData[0], wandData[1], wandData[2]);
  wandCoord.hpr.set(wandData[3], wandData[4], wandData[5]);
  display->dpfToWorld(&wandCoord);
  
  // create a matrix from the HPR position in world coordinates
  pfMatrix wandWorld;
  wandWorld.makeEuler(wandCoord.hpr[0],wandCoord.hpr[1],wandCoord.hpr[2]);
  
  // create a vector in the +y direction and move according to the wand
  pfVec3 upVector(0.f, 1.f, 0.f);
  check.xformVec(upVector, wandWorld);

  //write check values to shmSlice
  float vals[3];
  vals[0] = check[0];
  vals[1] = check[1];
  vals[2] = check[2];
  if (shmSlice)
    shmSlice->write(vals);

  return DTK_CONTINUE ; 
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new clipWand(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
