/*
 * dtk-StartUp.c - Start up trackd to drive the tracker and the interface
 *                 programs to DIVERSE
 *
 *
 * Steve Satterfield
 * NIST/ITL/MCSD/SAVG
 * steve@nist.gov
 *
 * June 13, 2002
 * June 4, 2003
 * December 17, 2003
 * May 12, 2004 - Added alternate calibration mode of trackd
 * June 2, 2004 - Removed "-status all" from the calibration mode.
 * June 17, 2004 - added the -k option to kill trackd, etc.
 * October 21, 2004 - Added support for dtk-getTrackd-`hostname`
 *                    so host specific getTrackd can be used.
 * November 3, 2004 - Added sleep after starting trackd
 * December 8, 2005 - Made changes for conversion to Prism
 * December 21, 2005 - replaced mysystem so it can do suid on Prism (JTK)
 * December 23, 2005 - finished convesion for suid on Prism
 *
 *
 * Notes: This program must be compiled and installed by root as it
 *        needs to run suid root.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

/* The maximum number of arguments in a single command */
#define MAX_ARGS (512)

/* define commands */
/* this is for Prism */
#ifdef __ia64__
#define KILL_CMD "/bin/kill"
#define RM_CMD "/bin/rm"
#define KILLALL_CMD "/usr/bin/killall"
#define IPCRM_CMD "/usr/bin/ipcrm"
#define IPCS_CMD "/usr/bin/ipcs | /bin/grep dtk | /bin/grep 887"
#endif

/* this is for IRIX */
#ifdef mips
#define KILL_CMD "/usr//bin/kill"
#define RM_CMD "/sbin/rm"
#define KILLALL_CMD "/sbin/killall"
#define IPCRM_CMD "/usr/sbin/ipcrm"
#define IPCS_CMD "/usr/sbin/ipcs | /sbin/grep dtk | /sbin/grep 887"
#endif

#define BG 0
#define FG 1


int verbose;

int
mysystem(char *command, int wait)
{
/*
   wait = 0 wait on command to complete
   wait = 1 do not wait on command to complete
*/
  char *argv[MAX_ARGS] ;
  int argc = 1 ;
  pid_t wgetPid ;

  char *dup = strdup(command) ;

  if (verbose) {
    fprintf(stderr,"Executing '%s'\n",command) ;
  }

  argv[0] = strtok(dup, " ") ;
  while ((argv[argc] = strtok(NULL, " "))) {
    argc++ ;
    if (argc==(MAX_ARGS)) {
      fprintf(stderr,"MAX_ARGS exceeded\n") ;
      return -1 ;
    }
  }
  argv[argc] = NULL ;

  if ((wgetPid=fork())==0) {
    // this is the child process
    execv(argv[0], argv) ;
    fprintf(stderr,"Unable to exec process\n") ;
    return -1 ;
  } else if (wgetPid==-1) {
    fprintf(stderr,"Unable to fork process\n") ;
    return -1 ;
  }
  if (wait) {
    if (verbose) fprintf(stderr,"waiting\n");
    waitpid(wgetPid) ;
  }

  return 0 ;
}

void
usage() 
{
      fprintf(stderr,"Usage: dtk-StartUp [-v] [-k | -c [configfilename]]\n");
      fprintf(stderr,"    -v  verbose\n");
      fprintf(stderr,"    -k  kill and do not restart trackd\n");
      fprintf(stderr,"    -c  invoke calibration mode\n");
      fprintf(stderr,"\n\n\n");
      fprintf(stderr,"  Normal operation:  trackd and dtk-getTrackd are started\n");
      fprintf(stderr,"                     for normal RAVE usage.\n");
      fprintf(stderr,"\n\n\n");
      fprintf(stderr,"  Calibration mode: invoke trackd with alternate conf\n");
      fprintf(stderr,"                    file that reads raw data with no internal correction.\n");
      fprintf(stderr,"                    dtk-getTrackd is not started.\n");
      fprintf(stderr,"                    Typicaly used with dtk-rawTrackd to put values to stdout.\n");
      fprintf(stderr,"                    An optional config filename can be specified for a user\n");
      fprintf(stderr,"                    supplied calibration config file\n");
      exit(1);
}



#define BUFSIZE 256
#define CMDSIZE 512

main (int argc, char **argv)
{
  FILE *ptr;


  /* Only delete the shared memory only for dtk.887 (user.group) */
  char *cmd=IPCS_CMD;

  char buf[BUFSIZE],type;
  int id;
  int i, calibrationMode,killMode;
  char *configfile,configcmd[CMDSIZE];
  char *configdefault="/usr/local/trackd/etc/trackd_raw.conf";
  char hostname[BUFSIZE], *p;
  char shellcmd[BUFSIZE];

  verbose = 0;
  killMode = 0;
  calibrationMode = 0;
  for (i=1; i<argc; i++) {
    if (argv[i][1] == 'v') verbose = 1;
    else if (argv[i][1] == 'k') killMode = 1;
    else if (argv[i][1] == 'c') {
                   calibrationMode = 1;
		   configfile=configdefault;
		   if (i+1 < argc) {
			        configfile=argv[i+1];
			        if (strlen(configfile)>CMDSIZE) {
			                  fprintf(stderr,"Alternate config file name too long\n");
					  exit(1);
			        }
		   }
		   if (verbose == 1) {
		                fprintf(stderr,"Using config file: '%s'\n",configfile);
                                sprintf(configcmd,"/usr/local/trackd/bin/runTrackd -file %10s",configfile);
		   } else {
                                sprintf(configcmd,"/usr/local/trackd/bin/runTrackdSillent -file %10s",configfile);
		   }
		   i++;
    }
    else usage();
  }


  gethostname (hostname, BUFSIZE);
  if (verbose) fprintf(stderr,"Long hostname= %s\n",hostname);
  p=strchr(hostname, '.');
  if (p != NULL) {
    *p='\0';  /* Convert long host name to short host name */
  }
  if (verbose) fprintf(stderr,"Short hostname= %s\n",hostname);

  fprintf(stderr, "dtk-StartUp initiating the tracker, please be patient\n");

  /* Kill any previous deamons running */
  mysystem(KILLALL_CMD" -9 trackd",FG);

  snprintf(shellcmd, BUFSIZE, KILLALL_CMD" -9 dtk-getTrackd-%s",hostname);
  mysystem(shellcmd,FG);

  /* Clear out DIVERSE shared memory */
  mysystem(RM_CMD" -rf /tmp/dtkSharedMem/* /dev/shm/dtkSharedMem/*",FG);

  /* Clear trackd shared memory segments */
  mysystem(IPCRM_CMD" -M 4126 -M 4127",FG);

  /* Clean up other shared memory pieces */
  if ((ptr = popen(cmd,"r")) != NULL)
    while (fgets(buf, BUFSIZE, ptr) != NULL) {
      type=buf[0];
      buf[0]=' '; buf[12]=0;
      sscanf(buf,"%d",&id);
      sprintf(buf,IPCRM_CMD" -%c %d",type,id);
      mysystem(buf,FG);
    }
  pclose(ptr);

  /* See if starting or only killing trackd */

  if (killMode == 0) {

      /* Now start everything */

      if (calibrationMode == 0) {

	  /* Norml production mode */
	  fprintf(stderr,"Starting Trackd\n");

	  if (verbose) {
	    mysystem("/usr/local/trackd/bin/runTrackd",BG);
	  } else {
	    mysystem("/usr/local/trackd/bin/runTrackdSilent",BG);
	  }

	  if (!shmCheck(4126, 1308, 2, 10, verbose) || !shmCheck(4127, 296, 2, 10, verbose)) {
            fprintf(stderr, "Could not get trackd shared memory, trackd probably failed to start.\n") ;
	    fprintf(stderr, "For additonal information try: dtk-StartUp -v\n");
            fprintf(stderr, "Exiting without starting tracking.\n");
	    exit(1);
	  }
	  
	  fprintf(stderr,"Starting dtk-getTrackd\n");
	  snprintf(shellcmd, BUFSIZE, "%s/bin/dtk-getTrackd-%s",getenv("HEVROOT"), hostname);
	  mysystem(shellcmd,BG);

      } else {

      /* Calibration mode */
      mysystem(configcmd,BG);
      }

  } else {
    fprintf(stderr,"trackd and dtk-getTrackd killed and not restarted\n");
  }

  fprintf(stderr,"dtk-StartUp procedures completed\n");
  exit(0);
}
