/*
 * DIVERSE Toolkit (DTK) -- version 2.0.7
 * released: May 21, 2002   08:23:20  EDT
 * DTK HOMPAGE is at: http://www.diverse.vt.edu/DTK/
 * 
 * The DIVERSE Toolkit
 * Copyright (C) 2000, 2001, 2002  Lance Arsenault
 * 
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License (GPL) as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software, in the top level source directory in a file
 * named "COPYING.GPL"; if not, see it at:
 * http://www.gnu.org/copyleft/gpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 *
 */
/******************** reads VRCO inc. trackd (tm) *******************/

/* This reads data from system V shared memory with the form from VRCO
 * inc. trackd.  That is writing CAVE (tm) devices. Then it stuffs it
 * into DTK shared memory, then it sleeps until the next cycle.  This
 * just does the head and wand trackers and the wand and joystick.  It
 * does this at a rate of about 100 Hz.  To see the data run this at
 * the same time:
 *
 * for head tracker:     dtk-readFloats head -r
 *
 * for wand tracker:     dtk-readFloats wand -r
 *
 * for joystick data:    dtk-readFloats joystick -r
 *
 * for buttons:          dtk-readButtons buttons -r
 *
 */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/stat.h>
#include <math.h>
#include <dtk.h>


#define PID_FILE    "/tmp/dtk-getTrackd_PID"

static int run = 1;
static int setHead = 1;

// on vista, one dpf unit is 42 inches, tracker spews feet at
// correct origin and orientation
#define SCALE (42.f/12.f)

static void SIGUSR1catcher(int sig)
{
  dtkMsg.add(DTKMSG_NOTICE,
	     "caught signal %d switching "
             "HEAD tracker to fixed.\n",sig);
  setHead = 0;
}

static void SIGUSR2catcher(int sig)
{
  dtkMsg.add(DTKMSG_NOTICE,
	     "caught signal %d switching "
             "HEAD tracker to moving.\n",sig);
  setHead = 1;
}

static void Exit_catcher(int sig)
{
  dtkMsg.add(DTKMSG_NOTICE,
	     "caught signal %d .. exiting.\n",
	     sig);
  run = 0;
}

int main(int argc, char **argv)
{
  dtkMsg.setPreMessage("%s%s%s", dtkMsg.color.vil,
                       "dtk-getTrackd", dtkMsg.color.end);

  dtkSharedMem head_shm(sizeof(float)*6,"head");
  if(head_shm.isInvalid())
    return 1;

  dtkSharedMem wand_shm(sizeof(float)*6,"wand");
  if(wand_shm.isInvalid())
    return 1;

  dtkSharedMem joystick_shm(sizeof(float)*2,"joystick");
  if(joystick_shm.isInvalid())
    return 1;

  dtkSharedMem button_shm(sizeof(unsigned char),"buttons");
  if(button_shm.isInvalid())
    return 1;

  {
    unlink(PID_FILE);
    umask(00);
    FILE *pid_file = fopen(PID_FILE, "w");
    if(!pid_file)
      {
	dtkMsg.add(DTKMSG_ERROR, 1,
		   "fopen(\""
		   PID_FILE
		   "\", \"w\") failed.\n");
	return 1;
      }
    else
      {
	fprintf(pid_file,"%d\n",getpid());
	fclose(pid_file);
      }
  }

  signal(SIGINT,Exit_catcher);
  signal(SIGTERM,Exit_catcher);
  signal(SIGUSR1,SIGUSR1catcher);
  signal(SIGUSR2,SIGUSR2catcher);

  dtkVRCOSharedMem sysV;
  if(sysV.isInvalid()) return 1;

  // run at 100 Hz => 1/100 second period
  dtkRealtime_init(0.01);

  for(;run;)
    {
      float h[6], w[6];

      /* here's the NULL configuation case */


      // Like Visage with scale for vista
      w[0] = (sysV.wand->x - 0.0f)/SCALE;
      w[1] = (-sysV.wand->z - 0.0f)/SCALE;
      w[2] = (sysV.wand->y - SCALE)/SCALE;
      w[3] = sysV.wand->h;
      w[4] = sysV.wand->p;
      w[5] = -sysV.wand->r;

      wand_shm.write(w);

#if 0
      printf("wand: (%0.3f %0.3f %0.3f) (%0.3f %0.3f %0.3f)  ",
	     w[0], w[1], w[2], w[3], w[4], w[5]) ;
#endif
     
      
      if (setHead) {
	// Like Visage with scale for vista
	h[0] = (sysV.head->x - 0.0f)/SCALE;
	h[1] = (-sysV.head->z - 0.0f)/SCALE;
	h[2] = (sysV.head->y - SCALE)/SCALE;
	h[3] = sysV.head->h;
	h[4] = sysV.head->p;
	h[5] = -sysV.head->r;


#if 0
	printf("head: (%0.3f %0.3f %0.3f) (%0.3f %0.3f %0.3f)\n",
	       h[0], h[1], h[2], h[3], h[4], h[5]) ;
#endif
      } else {
	h[0] = 0.f ;
	h[1] = 0.f;
	h[2] = 0.f;
	h[3] = 0.f;
	h[4] = 0.f;
	h[5] = 0.f;
#if 0
	printf("head no getting set!\n") ;
#endif
      }
      head_shm.write(h);

      unsigned char button_state = ((sysV.button[1])?01:0)
	| ((sysV.button[0])?02:0) | ((sysV.button[2])?04:0);

      if(button_shm.write(&button_state)) break; // error
      if(joystick_shm.write(sysV.val)) break; // error

      // sleep until next cycle.
      dtkRealtime_wait();
    }
  dtkRealtime_close();
  unlink(PID_FILE);

  return 0;
}
