// this DIVERSE DSO grabs the contents of the first window, and dumps the
// contents to a sequentially numbered file.  Itdoes this every frame,
// regardless of the frame rate.

// The window that's dumped to the files is pipe(0)->pwin(0)

// The screen area to be dumped is at window's original size and location.
// Move or resize it and you'll get whatever is being displayed where the
// window originally was

// the base of the output file name is specified by the environment variable
// HEV_FRAME_GRABBER_FILE, if set, otherwise the base of the file used is
// "/tmp/frameGrabber-$LOGNAME".  In either case the base has a six digit
// number appended to it, and the extension "rgb".

// if HEV_FRAME_GRABBER_NUM_FRAMES is set to a positive number, then once l
// is pressed, that many frames will be logged before logging is turned off.
// This can be useful to capture all frames of an animation sequence.  For
// bet results, the frameTimeStamp DSO should also be used to ensure no
// nodes of the sequence are skipped or duplicated

// if HEV_FRAME_GRABBER_WAIT_FRAMES is set to a positive number, then when
// the global frame counter equals this number, logging will automatically
// start.  This can be useful to start logging at the beginning of an
// animation sequence.  Timing issues will usually prevent you from setting
// this to 0 and getting the first frame.  Instead, if your sequence has "N"
// frames, try setting HEV_FRAME_GRABBER_WAIT_FRAMES to a N give or take a
// few frames until the first frame captured is the first frame of the
// sequence.

// a quick way to check your results is to use the ImageMagick animate
// command- something like:
//   animate -delay 33 /tmp/frameGrabber-${LOGNAME}-00*.rgb



// Author: John Kelso, kelso@nist.gov, February, 2005

//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/

// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

class frameGrabber : public dtkAugment {
  
public:
  
  frameGrabber (dpf *app) ;
  ~frameGrabber (void) ;
  int preConfig (void) ;
  int postConfig (void) ;
  int postFrame (void) ;
  
private:
  
  FILE* pathfile ;
  char* moviefilebase ;
  
  dpf *app ;
  pfPipeWindow *pfpwin ;
  int xo, yo, xs, ys ;
  
  dtkDequeuer *dequeuer;
  dtkInXKeyboard *keyboard;
  
  KeyCode xkc_l;
  KeyCode xkc_shift_l, xkc_shift_r;

  // how many frames to grab
  int numFrames ;

  // how many frames to wait before starting
  int waitFrames ;

  // file name of snap-shot of this frame
  char *frameFileName ;

  // flags whether we're logging frames
  bool logging ;

};

////////////////////////////////////////////////////////////////////////
frameGrabber::frameGrabber (dpf *app_in) :
  dtkAugment("frameGrabber") {
  app = app_in ;
  setDescription("creates a flip-book movie from a DIVERSE window") ;
  
  pfpwin = NULL ;
  numFrames = -1 ;
  waitFrames = -1 ;
  keyboard = NULL;
  dequeuer = NULL;
  frameFileName = NULL ;
  logging = false ;

  if(getenv("HEV_FRAME_GRABBER_FILE")) {
    moviefilebase = strdup(getenv("HEV_FRAME_GRABBER_FILE")) ;
  } else {
    char *format = "/tmp/frameGrabber-%s" ;
    char *logname = getenv("LOGNAME") ;
    if(!logname) logname = "nonknown_user" ;
    moviefilebase = (char *) malloc(strlen(format) + strlen(logname)) ;
    sprintf(moviefilebase, format, logname) ;
  }
  
  if (getenv("HEV_FRAME_GRABBER_NUM_FRAMES")) {
    if ((sscanf(getenv("HEV_FRAME_GRABBER_NUM_FRAMES"),"%d",&numFrames) != 1) ||
	numFrames<=0) {
      dtkMsg.add(DTKMSG_WARNING,"frameGrabber: "
		 "HEV_FRAME_GRABBER_NUM_FRAMES, \"%s\", is not a positive number, "
		 "ignoring\n", getenv("HEV_FRAME_GRABBER_NUM_FRAMES")) ;
    } else {
      dtkMsg.add(DTKMSG_WARNING,"frameGrabber: "
		 "HEV_FRAME_GRABBER_NUM_FRAMES set- "
		 "grabbing %d frames when l or L is pressed.\n", numFrames) ;
    }
  }

  if (getenv("HEV_FRAME_GRABBER_WAIT_FRAMES")) {
    if ((sscanf(getenv("HEV_FRAME_GRABBER_WAIT_FRAMES"),"%d",&waitFrames) != 1) ||
	waitFrames<0) {
      dtkMsg.add(DTKMSG_WARNING,"frameGrabber: "
		 "HEV_FRAME_GRABBER_WAIT_FRAMES, \"%s\", is not a non-negative number, "
		 "ignoring\n", getenv("HEV_FRAME_GRABBER_WAIT_FRAMES")) ;
    } else {
      dtkMsg.add(DTKMSG_WARNING,"frameGrabber: "
		 "HEV_FRAME_GRABBER_WAIT_FRAMES set- "
		 "will start grabbing frames when global frame counter equals %d.\n", 
		 waitFrames) ;
    }
  }

  // dtkAugment starts out invalid.
  validate();
}

////////////////////////////////////////////////////////////////////////
frameGrabber::~frameGrabber (void) {
  
  dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: "
	     "last frame filename = \"%s\"\n",
	     frameFileName) ;

  if (moviefilebase) {
    free(moviefilebase) ;
    moviefilebase = NULL ;
  }
  
  if (dequeuer) {
    delete dequeuer ;
    dequeuer = NULL ;
  }
  
}

////////////////////////////////////////////////////////////////////////
int frameGrabber::preConfig(void) {
  dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: setting pfMultiprocess(0)\n") ;
  pfMultiprocess(0) ;
  return dtkAugment::REMOVE_CALLBACK ;
}

////////////////////////////////////////////////////////////////////////
// initial setup
int frameGrabber::postConfig(void) {

  dequeuer = new dtkDequeuer(app->record());
  if(!dequeuer || dequeuer->isInvalid()) return ERROR_;
  
  keyboard = static_cast<dtkInXKeyboard *>(app->get("xkeyboard", 
						    DTKINXKEYBOARD_TYPE));
  if(!keyboard) return ERROR_;
  keyboard->queue();

  if(!(xkc_l = keyboard->getKeyCode(XK_l)) ||
     !(xkc_shift_l = keyboard->getKeyCode(XK_Shift_L)) ||
     !(xkc_shift_r = keyboard->getKeyCode(XK_Shift_R)))
    return ERROR_;
  
  // get the window to dump
  pfpwin = app->display()->first()->first()->pfpwin() ;

  if (!pfpwin) {
    dtkMsg.add(DTKMSG_ERROR,1,
	       "frameGrabber::postConfig() failed "
	       "to find pipe(0)->pwin(0)->pfpwin().\n") ;
    return dtkAugment::ERROR_ ;
  }

  if (waitFrames<0) {
    dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: move the window to the desired location \n") ;
    dtkMsg.add(DTKMSG_NOTICE,"    and size, then press the \"L\" key\n") ;
    dtkMsg.add(DTKMSG_NOTICE,"    to create the frame files.\n\n") ;
    dtkMsg.add(DTKMSG_NOTICE,"    WHEN LOGGING, DON'T MOVE OR RESIZE THE WINDOW!!\n") ;
  }

  return DTK_REMOVE_CALLBACK ; 
}


////////////////////////////////////////////////////////////////////////
// dump the current frame
int frameGrabber::postFrame (void) {

  // frame counter
  static int count = 0 ;
  static bool shiftPressed = false ;

  // for getting the "l" or "L" key
  dtkRecord_event* e;
  KeyCode key;

  // was there a key event?
  if ((e=dequeuer->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&key, e);
    if(key == xkc_shift_l || key == xkc_shift_r) {
      if (pressed) {
	shiftPressed = true ;
	//printf("shift pressed = true\n") ;
      } else {
	shiftPressed = false ;
	//printf("shift pressed = false\n") ;
      }
    } else if (pressed && key == xkc_l) {
      if (!logging) {
	// start logging
	pfpwin->getCurScreenOriginSize(&xo, &yo, &xs, &ys) ; 
	logging = true ;
	// added jtk 1/05- beats me!
	xo = yo = 0 ;
	if (shiftPressed) {
	  dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: appending to frame counter, "
		     "starting logging\n") ;
	} else {
	  count = 0 ;
	  dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: frame counter reset, "
		     "starting logging\n") ;
	}
      } else {
	// stop logging and reset
	logging = false ;
	dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: stopping logging, "
		   "last frame filename \"%s\"\n",
		   frameFileName) ;
	return DTK_CONTINUE ;
      }
    } else {  // not shift or l
      return DTK_CONTINUE ;
    }
  }

  if (!logging && (pfGetFrameCount())==waitFrames) {
    logging = true ;
    // autostart logging
    pfpwin->getCurScreenOriginSize(&xo, &yo, &xs, &ys) ; 
    xo = yo = 0 ;
    dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: autostarting logging\n") ;
  }
  
  if (logging) {
    // roll up a file name for this frame
    if (frameFileName) {
      free(frameFileName) ;
      frameFileName = NULL ;
    }
    frameFileName = (char *) malloc(strlen(moviefilebase) + 12) ;
    sprintf(frameFileName,"%s-%6.6d.rgb",moviefilebase,count) ;
    
    if (count == 0)
      dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: "
		 "first frame filename = \"%s\"\n",
		 frameFileName) ;
    
    dtkMsg.add(DTKMSG_INFO,"frameGrabber: "
	       "global frame count %d, log frame count %d\r",
	       pfGetFrameCount(),count) ;
    fflush(stdout) ;

    count++ ;
    if (numFrames>0 && count>=numFrames) {
      logging = false ;
      count = 0 ;
      dtkMsg.add(DTKMSG_NOTICE,"frameGrabber: stopping logging after %d frames, "
		 "last frame filename \"%s\"\n",
		 numFrames, frameFileName) ;
      return DTK_CONTINUE ;
    }
    
    pfuSaveImage(frameFileName, xo, yo, xs, ys, 0) ;
  }
  return DTK_CONTINUE ;
}

/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager) ;
  return new frameGrabber(app) ;
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment ;
  return DTKDSO_UNLOAD_CONTINUE ;
}


