//
// Copyright 2001, Silicon Graphics, Inc.
// ALL RIGHTS RESERVED
//
// UNPUBLISHED -- Rights reserved under the copyright laws of the
// United States.   Use of a copyright notice is precautionary only
// and does not imply publication or disclosure.
//
// U.S. GOVERNMENT RESTRICTED RIGHTS LEGEND:
// Use, duplication or disclosure by the Government is subject to
// restrictions as set forth in FAR 52.227.19(c)(2) or subparagraph
// (c)(1)(ii) of the Rights in Technical Data and Computer Software
// clause at DFARS 252.227-7013 and/or in similar or successor
// clauses in the FAR, or the DOD or NASA FAR Supplement.
// Contractor/manufacturer is Silicon Graphics, Inc.,
// 2011 N. Shoreline Blvd. Mountain View, CA 94039-7311.
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without
// fee, provided that (i) the above copyright notices and this
// permission notice appear in all copies of the software and related
// documentation, and (ii) the name of Silicon Graphics may not be
// used in any advertising or publicity relating to the software
// without the specific, prior written permission of Silicon Graphics.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
// DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
// THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
// OR PERFORMANCE OF THIS SOFTWARE.
//

//////////////////////////////////////////////////////////////////////
// IFLLoader.cxx
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// includes
//////////////////////////////////////////////////////////////////////

// C
#include <assert.h>

// C++
//#include <iostream.h>
#include <fstream>
using namespace std ;
#include <stdlib.h>

// IFL
#include <ifl/iflFile.h>

// Volumizer
#include <Volumizer2/ParameterVolumeTexture.h>

// local
#include "IFLLoader.h"

//////////////////////////////////////////////////////////////////////
// factory
//////////////////////////////////////////////////////////////////////

IFLLoader* 
IFLLoader::open(const char fileName[])
{
    iflFile* file = iflFile::open(fileName, O_RDONLY);

    if (file == NULL) {
        return NULL;
    } 
    iflSize dimensions;
    file->getDimensions(dimensions);
  
    // determine size 
    const int size[3] = { 
        dimensions.x,
        dimensions.y,
        dimensions.z
    }; 

    // determine format
    vzExternalTextureFormat format;
    switch (dimensions.c) {
    case 1:
        format = VZ_LUMINANCE;
        break;

    case 2:
        format = VZ_LUMINANCE_ALPHA;
        break;

    case 4:
        format = VZ_RGBA;
        break;

    default:

        file->close();
        return NULL;
    }

    // determine type
    vzTextureType type;
    switch (file->getDataType()) {
    case iflChar:
        type = VZ_BYTE;
        break;

    case iflUChar:
        type = VZ_UNSIGNED_BYTE;
        break;

    case iflShort:
        type = VZ_SHORT;
        break;

    case iflUShort:
        type = VZ_UNSIGNED_SHORT;
        break;

    case iflInt:
        type = VZ_INT;
        break;

    case iflUInt:
        type = VZ_UNSIGNED_INT;
        break;

    case iflFloat:
        type = VZ_FLOAT;
        break;

    default:

        file->close();
        return NULL;
    }
    return new IFLLoader(file, format, type, size);
}


//////////////////////////////////////////////////////////////////////
// constructor
//////////////////////////////////////////////////////////////////////

IFLLoader::IFLLoader(iflFile* file, vzExternalTextureFormat format,
    vzTextureType type, const int size[3])
    :
    m_file(file),
    m_format(format),
    m_type(type)
{
    m_size[0] = size[0];
    m_size[1] = size[1];
    m_size[2] = size[2];
}

//////////////////////////////////////////////////////////////////////
// destructor
//////////////////////////////////////////////////////////////////////

IFLLoader::~IFLLoader()
{
    m_file->close();
}

//////////////////////////////////////////////////////////////////////
// get format
//////////////////////////////////////////////////////////////////////

vzExternalTextureFormat
IFLLoader::getFormat() const
{
    return m_format;
}

//////////////////////////////////////////////////////////////////////
// get type
//////////////////////////////////////////////////////////////////////

vzTextureType
IFLLoader::getType() const
{
    return m_type; 
}

//////////////////////////////////////////////////////////////////////
// get size
//////////////////////////////////////////////////////////////////////

void
IFLLoader::getSize(int size[3]) const
{
    size[0] = m_size[0];
    size[1] = m_size[1];
    size[2] = m_size[2];
}

//////////////////////////////////////////////////////////////////////
// load brick
//////////////////////////////////////////////////////////////////////

bool 
IFLLoader::loadBrick(void* data, const int position[3],
    const int size[3])
{
    assert((position[0] >= 0) && (position[0] < m_size[0]));
    assert((position[1] >= 0) && (position[1] < m_size[1]));
    assert((position[2] >= 0) && (position[2] < m_size[2]));

    const int bytesPerVoxel =
        getNumChannels(m_format) *
        getBytesPerChannel(m_type);

    // clip brick at volume boundary
    const int clippedSize[3] = {
        size[0] <= m_size[0] - position[0]
            ? size[0] : m_size[0] - position[0],
        size[1] <= m_size[1] - position[1]
            ? size[1] : m_size[1] - position[1],
        size[2] <= m_size[2] - position[2]
            ? size[2] : m_size[2] - position[2]
    };

    const bool clipped =
        (clippedSize[0] != size[0]) ||
        (clippedSize[1] != size[1]) ||
        (clippedSize[2] != size[2]);

    unsigned char* target = (unsigned char*) data;
    unsigned char* source;

    // allocate temporary buffer
    if (clipped) {
        const int totalSize =
            clippedSize[0] *
            clippedSize[1] *
            clippedSize[2] *
            bytesPerVoxel;

        source = target;
        target = new unsigned char[totalSize];
    }

    // load brick
    iflConfig config(m_file->getDataType(), iflInterleaved);
    const iflStatus status = m_file->getTile(
        position[0], position[1], position[2],
        clippedSize[0], clippedSize[1], clippedSize[2],
        target, &config);

    if (status != iflOKAY) {
      //cerr << "error: cannot read brick" << endl;

        if (clipped)
            delete[] target;

        return false;
    }

    // copy from temporary buffer
    if (clipped) {
        const int totalSize =
            size[0] *
            size[1] *
            size[2] *
            bytesPerVoxel;

        // swap
        source = target;
        target = (unsigned char*) data;

        // erase brick
        memset(target, 0, totalSize);

        // copy from buffer
        int
            sourceIndex = 0,
            targetIndex = 0;

        for (int z = 0; z < clippedSize[2]; z++) {
            for (int y = 0; y < clippedSize[1]; y++) {
                memcpy(
                    &target[targetIndex * bytesPerVoxel],
                    &source[sourceIndex * bytesPerVoxel],
                    clippedSize[0] * bytesPerVoxel);

                sourceIndex += clippedSize[0];
                targetIndex += size[0];
            }
            targetIndex += (size[1] - clippedSize[1]) * size[0];
        }
      
        delete[] source;
    }
    
    return true;
}

//////////////////////////////////////////////////////////////////////
// load volume
//////////////////////////////////////////////////////////////////////

vzParameterVolumeTexture*
IFLLoader::loadVolume(vzInternalTextureFormat internalFormat)
{
    const int size =
        m_size[0] *
        m_size[1] *
        m_size[2] *
        getNumChannels(m_format) *
        getBytesPerChannel(m_type);
  
    vzMemoryAllocationCallback allocator;
    vzMemoryDeletionCallback deallocator;
    void* userData;

    vzMemory::getMemoryManagementCallbacks(allocator, deallocator,
        userData);

    unsigned char* data = (unsigned char*) allocator(size, userData);

    const int position[3] = {0, 0, 0};

    if (!loadBrick(data, position, m_size)) {
      //cerr << "error: cannot load volume" << endl;

        deallocator(data, userData);
        return NULL;
    }
    
    vzParameterVolumeTexture* texture =
        new vzParameterVolumeTexture(m_size, data,
            m_type, m_format, internalFormat);
    
    texture->addDeletionCallback(deleteCallback, data);
    
    return texture;
}
 
void
IFLLoader::deleteCallback(vzObject* /* object */, void* data)
{
#if 0
    delete[] (unsigned char*) data;
#else
    vzMemoryAllocationCallback allocator;
    vzMemoryDeletionCallback deallocator;
    void* userData;

    vzMemory::getMemoryManagementCallbacks(allocator, deallocator,
        userData);

    deallocator(data, userData);
#endif
}

//////////////////////////////////////////////////////////////////////
// IFLLoader.cxx - EOF
//////////////////////////////////////////////////////////////////////

