////////////////////////////////////////////////////////////////////////
// TFChannel.cxx
////////////////////////////////////////////////////////////////////////

// C
#include <math.h>

// C++
//#include <fstream.h>

#include <fstream>
using namespace std ;
// local
#include "TFChannel.h"

//////////////////////////////////////////////////////////////////////
// 2 constructors - 1 for stream, 1 just create without reading
//////////////////////////////////////////////////////////////////////

TFChannel::TFChannel(istream& stream, int length)
{
    char string[255];
    m_length = length;

    // interpolation - skip "Interpolation_mode:"
    int value;
    stream >> string >> value;
    //cout << "echo string: " << string << "  and value: " << value << endl;
    m_interpolation = Interpolation(value);
    
    // control points
    stream >> pointCount;
    //cout << "pointcount: " << pointCount << endl;

    inpoints = new int[pointCount];
    invalues = new float[pointCount];

    for (int i = 0; i < pointCount; i++) {
        stream >> inpoints[i];
        stream >> invalues[i];
	//cout << "next set of values: " << inpoints[i] << "  " << invalues[i] << endl;
    }

    // ramp
    stream >> m_rampCenter >> m_rampWidth;
    //cout << "center: " << m_rampCenter << "  and width: " << m_rampWidth << endl;
    m_values = new float[length];

    // fill in values
    if (m_interpolation == INTERPOLATION_PAINT) {
        for (int j = 0; j < length; j++)
            stream >> m_values[j];
    }
    else {
      switch (m_interpolation) {
      case INTERPOLATION_LINEAR:
        generateRamps();
        break;
	
      case INTERPOLATION_SPLINES:
        generateSplines();
        break;
	
      case INTERPOLATION_GAUSSIAN:
        generateGaussian();
        break;
	
      case INTERPOLATION_PAINT:
        // no evaulation required for paint interpolation
        break;
	
      case INTERPOLATION_WINDOW_LEVEL:
        generateLinearRamp();
        break;
      }

      m_dirty = false;
    }
}

TFChannel*
TFChannel::create(Interpolation interpolation, int length)
{
  Interpolation inter = INTERPOLATION_LINEAR;
  const int
        rampCenter = 127,
        rampWidth = 128;
  
  return new TFChannel(inter, length, rampCenter, rampWidth);
}

//////////////////////////////////////////////////////////////////////
// constructor
//////////////////////////////////////////////////////////////////////

TFChannel::TFChannel(Interpolation inter, int length, 
		     int rampCenter, int rampWidth) {
  m_interpolation = inter;
  m_length = length;
  m_values = new float[length];
  m_rampCenter = rampWidth;
  m_rampWidth = rampCenter;
}

//////////////////////////////////////////////////////////////////////
// destructor
//////////////////////////////////////////////////////////////////////

TFChannel::~TFChannel()
{
    delete[] m_values;
}

//////////////////////////////////////////////////////////////////////
// dirty
//////////////////////////////////////////////////////////////////////

void
TFChannel::setDirty(bool flag)
{
    m_dirty = flag;
}

bool
TFChannel::isDirty() const
{
    return m_dirty;
}

//////////////////////////////////////////////////////////////////////
// evaluate
//////////////////////////////////////////////////////////////////////

void TFChannel::evaluate()
{
    switch (m_interpolation) {
    case INTERPOLATION_LINEAR:
        generateRamps();
        break;
    
    case INTERPOLATION_SPLINES:
        generateSplines();
        break;
    
    case INTERPOLATION_GAUSSIAN:
        generateGaussian();
        break;
    
    case INTERPOLATION_PAINT:
        // no evaulation required for paint interpolation
        break;

    case INTERPOLATION_WINDOW_LEVEL:
        generateLinearRamp();
        break;
    }

    m_dirty = false;
}

//////////////////////////////////////////////////////////////////////
// values
//////////////////////////////////////////////////////////////////////

const float*
TFChannel::getValues()
{
    return m_values;
}

//////////////////////////////////////////////////////////////////////
// generate ramps
//////////////////////////////////////////////////////////////////////

void
TFChannel::generateRamps()
{
  m_values[0] = invalues[0];
  for ( int i = 1; i < pointCount; ++i ) {
    int npts = inpoints[i] - inpoints[i-1];
    float sval = (invalues[i] - invalues[i-1])/(float) (npts) ;

    for ( int j = inpoints[i-1]+1; j <= inpoints[i]; ++j ) {
      m_values[j] = m_values[j-1] + sval;
    }
  }
}

//////////////////////////////////////////////////////////////////////
// generate spline
//////////////////////////////////////////////////////////////////////

void
TFChannel::generateSplines()
{
  int num_cpoints = pointCount;
  // XXX memory leak?
  int*   cpoint_x = new int[num_cpoints];
  float* cpoint_y = new float[num_cpoints];
  float* sec_derivate = new float[num_cpoints];

  for (int i = 0; i < pointCount; i++) { 
    cpoint_x[i] = inpoints[i];
    cpoint_y[i] = invalues[i];
  }

  initSpline(cpoint_x, cpoint_y, num_cpoints, sec_derivate);

  for (int i = 0; i <= m_length - 1; i++) {
    const float value =
      evaluateSpline(cpoint_x, cpoint_y, sec_derivate, num_cpoints, i);
    
    m_values[i] = clamp(value);
  }
}

//////////////////////////////////////////////////////////////////////
// init spline
//////////////////////////////////////////////////////////////////////

void
TFChannel::initSpline(int x[], float y[], int n, float y2[])
{
    float p, qn, sig, un;

    // XXX memory leak?
    float* u = new float[n];
  
    y2[0] = u[0] = 0.0;
    for (int i = 1; i < n - 1; i++) {
        sig = (float) (x[i] - x[i - 1]) / (x[i + 1] - x[i - 1]);
        p = sig * y2[i - 1] + 2;

        y2[i] = (sig - 1) / p;

        u[i] =
            ((y[i + 1] - y[i]) / (x[i + 1] - x[i])) - 
            ((y[i] - y[i - 1]) / (x[i] - x[i - 1]));

        u[i] = (6.0 * u[i] / (x[i + 1] - x[i - 1]) - sig * u[i - 1]) / p;
    }
    qn = un = 0.0;

    y2[n - 1] = (un - qn * u[n - 2]) / (qn * y2[n - 2] + 1.0);
    for (int k = n - 2; k >= 0; k--) {
        y2[k] = y2[k] * y2[k + 1] + u[k];
    }
}

//////////////////////////////////////////////////////////////////////
// eval spline
//////////////////////////////////////////////////////////////////////

float
TFChannel::evaluateSpline(int xa[], float ya[], float y2a[],
    int n, int x)
{ 
    int klo, khi, k;
    float h, b, a;

    klo = 0;
    khi = n - 1;

    while (khi - klo > 1) {
        k = (khi + klo) >> 1;
        if (xa[k] > x) khi = k;
        else klo = k;
    }

    h = xa[khi] - xa[klo];

    a = (xa[khi] - x) / h;
    b = (x - xa[klo]) / h;

    return
        (a*ya[klo] + b*ya[khi] + ((a*a*a-a)*y2a[klo] +(b*b*b-b)*y2a[khi]) 
            * (h*h) / 6.0);
}

//////////////////////////////////////////////////////////////////////
// generate gaussian
//////////////////////////////////////////////////////////////////////

void
TFChannel::generateGaussian()
{
    for (int i = 0; i < m_length ; i++) {
      float value = 0;
      for (int j = 0; j < pointCount; ++j ) {
        // accumulate share of each control point
	const int x = inpoints[i];
	const float y = invalues[i];
	float normalizedX = (float) x / m_length;
	float normalizedIndex = (float) i / m_length;

	value += evaluateGaussian(normalizedIndex,
				  normalizedX, y);     
      }
        m_values[i] = clamp(value);
    } 
}

//////////////////////////////////////////////////////////////////////
// evaluate gaussian
//////////////////////////////////////////////////////////////////////

float
TFChannel::evaluateGaussian(float x, float n, float mu)
{
    // keep mu * sigma constant
    const float
        sigma = 1 / mu,
        X = n - x;

    // this is hacked to look good... should be modified
    return mu * exp(-(X * X) / (0.0008 * sigma));
}

//////////////////////////////////////////////////////////////////////
// generate linear ramp
//////////////////////////////////////////////////////////////////////

void
TFChannel::generateLinearRamp()
{
    for (int i = 0; i < m_length; i++) {
        const float value = (float)
            (i - (m_rampCenter - m_rampWidth / 2)) / m_rampWidth;

        m_values[i] = clamp(value);
    }
}

//////////////////////////////////////////////////////////////////////
// interpolation
//////////////////////////////////////////////////////////////////////

void
TFChannel::setInterpolation(Interpolation interpolation)
{
    if (m_interpolation == interpolation)
        return; 

    m_interpolation = interpolation;

    m_dirty = true;
}

TFChannel::Interpolation
TFChannel::getInterpolation()
{
    return m_interpolation;
}

//////////////////////////////////////////////////////////////////////
// length
//////////////////////////////////////////////////////////////////////

void
TFChannel::setLength(int length)
{
    if (m_length == length)
        return;

    delete[] m_values;

    m_values = new float[length];
    m_length = length;

    m_dirty = true;
}


int
TFChannel::getLength() const
{
    return m_length;
}

//////////////////////////////////////////////////////////////////////
// ramp center
//////////////////////////////////////////////////////////////////////

void
TFChannel::setRampCenter(int center)
{
    m_rampCenter = center;
    m_dirty = true;
}

int
TFChannel::getRampCenter() const
{
    return m_rampCenter;
} 

//////////////////////////////////////////////////////////////////////
// ramp width
////////////////////////////////////////////////////////////////////// 
void
TFChannel::setRampWidth(int width)
{
    m_rampWidth = width;
    m_dirty = true;
}

int
TFChannel::getRampWidth() const
{
    return m_rampWidth;
}

//////////////////////////////////////////////////////////////////////
// value
//////////////////////////////////////////////////////////////////////

void
TFChannel::setValue(int index, float value) 
{
    if ((index < 0) || (index >= m_length))
        return;

    m_values[index] = clamp(value);

    m_dirty = true;
}

////////////////////////////////////////////////////////////////////// 
// current control point
////////////////////////////////////////////////////////////////////// 

//////////////////////////////////////////////////////////////////////
// save
//////////////////////////////////////////////////////////////////////

void
TFChannel::saveToStream(ostream &stream) const
{
    // interpolation
    stream << "Interpolation_mode: " << m_interpolation << endl;

    // control points
    /*////////////////////////////////////////////////////////////////
      stream << m_points.size() << endl;
    for (TFControlPointList::const_iterator iter = m_points.begin();
        iter != m_points.end(); iter++)
    {
        stream << iter->m_index << " " << iter->m_value << endl;
    }
    ---------------------------------------------------------*/
    // ramp
    stream << m_rampCenter << " " << m_rampWidth << endl;

    // paint values
    if (m_interpolation == INTERPOLATION_PAINT)
        for (int i = 0; i < m_length; i++)
            stream << m_values[i] << endl;
}

//////////////////////////////////////////////////////////////////////
// clamp
//////////////////////////////////////////////////////////////////////

float
TFChannel::clamp(float value)
{
    if (value < 0)
        return 0;
    else if (value > 1)
        return 1;
    else
        return value;
}

//////////////////////////////////////////////////////////////////////
// TFChannel.cxx - EOF
//////////////////////////////////////////////////////////////////////

