
/*
  sets the hpr of a dcs so it "points" to a specified position.
  
  To use as a billboard, have the object under this node centered at the
  origin and facing (0, -1, 0)
  
  here's some example code:
  
    pfMatrix mat ;
    mat.makeTrans(x, y, z) ;
    pfSCS *scs ;
    scs = new pfSCS(mat) ;
    display->world()->addChild(scs) ;
    
    billboardNode *bbnode ;
    bbnode = new billboardNode(&(display->headPos()->xyz)) ;
    pfNode *foo = pfdLoadFile("foo.pfb") ;
    
    bbnode->addChild(foo) ;
    scs->addChild(bbnode) ;
  
Methods:
  billboardNode(pfVec3 *pos) - where pos is a pointer to a pfVec3 giving
                               the position of the viewer.  The pointer's
			       data can be modified to move the viewer.
  

  void axis(pfVec3 vec) - specifies the axis of rotation.  The default is
                          (0, 0, 1)

  void pointRotate(void) - specifies to change both heading around the rotation
                           axis, and pitch to point the node at the viewer.
			   This is the default.

  void axisRotate(void) - specifies to only change the heading around the 
                          rotation axis, This is useful for objects like
			  billboard trees.

*/


#include <stdio.h>
#include <unistd.h>

#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pr/pfLinMath.h>

// axis to rotate around
static pfVec3 _axis ;

// set to true if rotating about a point
static int _pointMode ;

// pointer to position to point at
static pfVec3 *_pos ;

static int appCallback(pfTraverser *trav, void *data) ;

class billboardNode: public pfDCS {
public:
  billboardNode(pfVec3 *pos) ;
  void axis(pfVec3 vec) ;
  void pointRotate(void) ;
  void axisRotate(void) ;
private:
} ;

billboardNode::billboardNode(pfVec3 *pos) {
  _pos = pos ;
  _axis.set(0.f, 0.f, 1.f) ;
  _pointMode =  true ;
  this->setTravFuncs(PFTRAV_APP, appCallback, NULL) ;
} ;

void billboardNode::axis(pfVec3 vec) {
  _axis = vec ;
  _axis.normalize() ;
} ;

void billboardNode::pointRotate(void) {
  _pointMode = true ;
} ;

void billboardNode::axisRotate(void) {
  _pointMode = false ;
} ;

static int appCallback(pfTraverser *trav, void *data) {

  ////////////////////////////////////////////////////////////////////////
  // get matrix of compound transformations from root down to and 
  // including this dcs node
  pfMatrix compoundMat ;
  trav->getMat(compoundMat) ;

  ////////////////////////////////////////////////////////////////////////
  // get the matrix of this dcs node
  pfDCS *dcsNode = (pfDCS*)trav->getNode();

  pfMatrix dcsMat ;
  dcsNode->getMat(dcsMat) ;

  ////////////////////////////////////////////////////////////////////////
  // get the transformation above this node

  pfMatrix dcsMatInv ;
  dcsMatInv.invertOrtho(dcsMat) ;

  pfMatrix aboveMat ;
  aboveMat =  dcsMatInv * compoundMat ;

  pfCoord above ;
  aboveMat.getOrthoCoord(&above) ;

  // set the DCS so the object's hpr is negated, then rotated by hpr
  pfMatrix mat ;
  mat.makeEuler(above.hpr[0], above.hpr[1], above.hpr[2]) ;
  mat.invertOrtho(mat) ;


  // the following algorithm is from
  // http://www.evl.uic.edu/yg/billboard.html, 
  // written by Alex Hill, ahill@evl.uic.edu
  // acording to Alex, the code is public domain
  // I have to admit step 4 loses me

  // 1- get a vector from the node to the position
  pfVec3 pos = *_pos - above.xyz ;
  pos.normalize();
  
  // 2- get the cross product of the vector going from the node to
  // the position and the rotation axis. 
  pfVec3 axisCrossPos;
  axisCrossPos.cross(_axis, pos);
  axisCrossPos.normalize();
  
  // 3- get the cross product of the front axis of the node, (0, -1, 0), and
  // the rotation axis
  pfVec3 axisCrossFront;
  axisCrossFront.cross(_axis, pfVec3(0.f, -1.f, 0.f));
  axisCrossFront.normalize();

  // 4a- get the cross of the axis and the axisCrossFront- supposed to be a
  // normal towards the viewer, but this one baffles me
  pfVec3 axisCross_AxisCrossFront;
  axisCross_AxisCrossFront.cross(_axis, axisCrossFront);
  axisCross_AxisCrossFront.normalize() ;

  // 4b- get the rotation of the axis to the viewer vector not sure what
  // right triangle we're using for the sides though- part of the mystery
  // above I guess.
  float opposite = axisCross_AxisCrossFront.dot(axisCrossPos);
  float adjacent = axisCrossFront.dot(axisCrossPos);
  float axisRot = pfArcTan2(opposite,adjacent);

  // 5- create a matrix- rotate around the axis
  // if the axis is the default, this changes heading
  pfMatrix rotMat;
  rotMat.makeRot(axisRot, _axis[0], _axis[1], _axis[2]);
  
  // if in point mode, add another rotation
  if (_pointMode)
    {
      //6- get the rotation from the axis to the vector to the head
      float crossRot;
      crossRot = pfArcSin(_axis.dot(pos));

      //7- rotate around the axis normal to the viewer vector and axis
      pfMatrix pitchMat;
      pitchMat.makeRot(-crossRot,axisCrossPos[0],axisCrossPos[1],axisCrossPos[2]);

      //add this rotation matrix to the existing rotation matrix
      rotMat.postMult(pitchMat);
    }
 
  // add this hpr offset to the "reset" DCS.  do a pre multiply so it'll use
  // the axis of the user, not the axis of the transformed world
  mat.preMult(rotMat) ;

  dcsNode->setMat(mat) ;

  //sleep(5) ;

  return PFTRAV_CONT ;
}

