// read a file- each line contains two to four "words", a label, a command
// to execute when button is pressed, and optionally one to execute when
// released.  If a release command is given, an optional fourth argument
// can specify the button's initial state.  by default all buttons are off

// radio button support added jtk 9/05.  all buttons between lines
// containing only "begin" and "end" will be in a group and be set as radio
// buttons.  The reset command can have the specal value "NULL" in order to
// initialize the radio button group but not have any reset command be
// executed.  Since only one button can be turned on in a group, the last
// "1" will be the value used.

// spacer added jtk 9/05.  a line with a blank label will produce a button
// sized space in the array

// each line creates a button

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Light_Button.H>
#include <FL/Fl_Button.H>

#include <dtk.h>
#include <dpf.h>

#include "getString.C"
#include "normalize.C"

// for sizing buttons and window, in pixels
#define BORDER (2) 
#define GAP (5)
#define BUTTON_WIDTH (100)
#define BUTTON_HEIGHT (25)

struct Item {
  Fl_Light_Button *button ;
  char *label ;
  bool initSet ;
  char *setCommand ;
  char *resetCommand ;
  // -1 if not in a group- a count to distingush which group of radio buttons we're in
  // this is also used as an index into the previousButton array
  int radioCount ;
  int n ;
} ;
static Item *item =NULL;

// an array of button indices.  The index is the group number.  The value is
// the button index of the last pressed button in the group, or -1 if no
// button pressed yet
static int *previousButton = NULL ;

// how many items, total
static int num = 0 ;

static bool allBlanks(char *s) {
  for (int i=0; s[i] != '\0'; i++) {
    if (s[i] != ' ')
      return false ;
  }
  return true ;
}

static void item_callback(Fl_Widget *w, void *data) {
  // get button number
  int n = ((Item*)data)->n ;
  //~printf("n = %d\n",n) ;

  if (item[n].radioCount>=0) {
    // radio buttons are messy
    // is there a previous unset command to do?
    int p = previousButton[item[n].radioCount] ;
    //~printf("p = %d\n",p) ;
    if (p>=0) {
      char *cmd = item[p].resetCommand ;
      if (cmd)
	system(cmd) ;
    }
    previousButton[item[n].radioCount] = n ;
    system(item[n].setCommand) ; 
  } else if (!item[n].resetCommand) { 
    // just do set function if button only has one state
    system(item[n].setCommand) ;
    item[n].button->clear() ;
  } else {
    // do either set or reset depending on state
    if (item[n].button->value()) {
      system(item[n].resetCommand) ;
      item[n].button->clear() ;
    } else {
      system(item[n].setCommand) ;
      item[n].button->set() ;
    }
  }
}

// setall will not execute the set function for radio or single value
// buttons, and only for buttons that aren't currently set
static void setCallback(Fl_Widget *w, void *data) {
  for (int i=0; i<num; i++) {
    if (item[i].label[0] == '\0') 
      continue ;
    if ( (item[i].radioCount<0 && 
	  !(item[i]).button->value() && 
	  (item[i]).resetCommand)) {
      //~printf("setting button \"%s\": ",(item[i]).label) ;
      //~printf("executing command \"%s\"\n",(item[i]).setCommand) ;
      system((item[i]).setCommand) ;
      (item[i]).button->set() ;
    }
  }
}

// will execute the reset function for those buttons that have it defined,
// and only for buttons that are set
// then sets buttons to their initial state, excuting their set command
static void resetCallback(Fl_Widget *w, void *data) {
  for (int i=0; i<num; i++) {
    if (item[i].label[0] == '\0') 
      continue ;
    if ((item[i]).button->value() && !(item[i]).initSet && (item[i]).resetCommand) {
      //~printf("resetting button \"%s\": ",(item[i]).label) ;
      //~printf("executing command \"%s\"\n",(item[i]).resetCommand) ;
      system((item[i]).resetCommand) ;
      (item[i]).button->clear() ;
    }
  }
  for (int i=0; i<num; i++) {
    if (item[i].label[0] == '\0') 
      continue ;
    if (!(item[i]).button->value() && (item[i]).initSet && (item[i]).setCommand) {
      //~printf("resetting button \"%s\": ",(item[i]).label) ;
      //~printf("executing command \"%s\"\n",(item[i]).resetCommand) ;
      system((item[i]).setCommand) ;
      (item[i]).button->set() ;
    }
  }
}

static void usage() {
  fprintf(stderr,"Usage: hev-buttonArray [[-c|--columns] c] [-r|--reset} [-a|-setall]\n") ;
}

int main(int argc, char **argv) {

  // default number of columns
  int cols = 3 ;

  // true if drawing "reset" button
  bool reset = false ;

  // true if drawing "set all" button
  bool setall = false ;

  // true if inside a group of radio buttons
  bool radio = false ;
  Fl_Group *g = NULL ;
  // counts the groups so can distinguish adjacent groups
  int radioCount = -1 ;

  int fltk_argc;
  char **fltk_argv;
  int other_argc;
  char **other_argv;
  // dtkFLTKOptions_get will pull out FLTK options into its return pointer.
  if(dtkFLTKOptions_get(argc, (const char **) argv,
                        &fltk_argc,  &fltk_argv,
                        &other_argc, &other_argv)) return 1;
  
  char *prog = other_argv[0] ;

  // parse other args
  {
    int o = 1 ; // skip argv[0]
    while (o<other_argc) {

      int l = strlen(argv[o]) ;
      if (l<3) l=3 ;

      if ((!strncasecmp(other_argv[o],"--columns",l)) ||
	  (!strcasecmp(other_argv[o],"-c"))) {
	o++ ;
	float foo ;
	if ((dpf_string_to_float(other_argv[o], &foo)) || (foo<1.f)) {
	  fprintf(stderr,"%s: invalid column number, \"%s\"\n",
		  prog,other_argv[o]) ;
	  usage() ;
	  return -1 ;
	}
	cols = (int) foo ;

      } else if ((!strncasecmp(other_argv[o],"--reset",l)) ||
	  (!strcasecmp(other_argv[o],"-r"))) {
	reset = true ;

      } else if ((!strncasecmp(other_argv[o],"--setall",l)) ||
	  (!strcasecmp(other_argv[o],"-a"))) {
	setall = true ;

      } else {
	fprintf(stderr,"%s: invalid parameter \"%s\"\n", prog, other_argv[o]) ;
	usage() ;
	return -1 ;
      }

      o++ ;
    }
  }

  // read in the data
  char *line = NULL ;
  int largc ;
  char **largv = NULL ;

  int lineNum = 1 ;
  while (getString(&line)>=0) {
    if (normalize(line)) {
      dpf_charsToArgcArgv(line, &largc, &largv) ;
      //~printf("largc = %d: ",largc) ;
      //~for (int i=0; i<largc; i++)
      //~printf("\"%s\" ",largv[i]) ;
      //~printf("\n") ;

      if (largc>4) {
	fprintf(stderr,"%s: invalid line, line %d, \"%s\"\n",
		prog,lineNum,line) ;
	usage() ;
	return -1 ;
      }

      // test for spacer
      if (largc==0 || allBlanks(largv[0])) {
	item = (Item*) realloc(item,sizeof(Item)*(num+1)) ;
	item[num].label = "";
	if (radio)
	  item[num].radioCount = radioCount;
	else
	  item[num].radioCount = -1;
	num++ ;
	
      // test for begin or end
      } else if (largc==1 && !strcasecmp(largv[0],"begin")) {
	if (radio) { // already in a radio group
	  fprintf(stderr,"%s: begin without an earlier end, line %d, \"%s\"\n",
		  prog,lineNum,line) ;
	  usage() ;
	  return -1 ;
	} else {
	  radio = true ;
	  radioCount++ ;
	  //~printf("radioCount = %d\n",radioCount) ;
	  previousButton = (int*) realloc(previousButton, 
					  (radioCount+1)*sizeof(int)) ;
	  previousButton[radioCount] = -1 ; 
	}
	  
      } else if (largc==1 && !strcasecmp(largv[0],"end")) {
	if (!radio) { // already NOT in a radio group
	  fprintf(stderr,"%s: end without an earlier begin, line %d, \"%s\"\n",
		  prog,lineNum,line) ;
	  usage() ;
	  return -1 ;
	} else {
	  radio = false ;
	}
	
      } else if (largc>=2) {
	item = (Item*) realloc(item,sizeof(Item)*(num+1)) ;
	item[num].label = strdup(largv[0]) ;
	item[num].setCommand = strdup(largv[1]) ;
	if (radio)
	  item[num].radioCount = radioCount;
	else
	  item[num].radioCount = -1;

	// NULL can be used to specify a NULL reset command, allowing an
	// initState to be specified for radio buttons with no desired reset
	// command
	if (largc>=3 && strcmp(largv[2],"NULL")) {
	  item[num].resetCommand = strdup(largv[2]) ;
	} else
	  item[num].resetCommand = NULL ;
	
	if (largc==4) {
	  if (!radio && !item[num].resetCommand) {
	    fprintf(stderr,"%s: attempt to initialize a single state button, line %d, \"%s\"\n",
		    prog,lineNum,line) ;
	    usage() ;
	    return -1 ;
	  }
	  if (!strcmp(largv[3],"1")) {
	    item[num].initSet = true ;
	    if (radio)
	      previousButton[radioCount] = num ;
	  }
	  else if (!strcmp(largv[3],"0"))
	    item[num].initSet = false ;
	  else {
	    fprintf(stderr,"%s: invalid line, line %d, \"%s\"\n",
		    prog,lineNum,line) ;
	    usage() ;
	    return -1 ;
	  }
	} else 
	  item[num].initSet = false ;
	num++ ;
      } else {
	fprintf(stderr,"%s: invalid line, line %d, \"%s\"\n",
		prog,lineNum,line) ;
	usage() ;
	return -1 ;
      }
      lineNum++ ;
    }
  }
  
  if (num<cols)
    cols = num ;
  
  // how many rows to push the other buttons down because of the reset and
  // setall buttons
  int pushdown = 0 ;
  if (reset ||  setall) {
    if (cols==1) {
      if (reset && setall)
	pushdown = 2 ;
      else
	pushdown = 1 ;
    } else
      pushdown = 1 ;
  }

  // number of rows
  int rows = (num+cols-1)/(cols) ;
  
  //~printf("n=%d rows=%d cols=%d\n",n,rows,cols) ;

  Fl_Window *window = new Fl_Window(cols*(BUTTON_WIDTH+GAP),
				    (rows+pushdown)*(BUTTON_HEIGHT+GAP)) ;

  // add reset and set all buttons
  if (reset) {
    Fl_Button *resetButton = 
      new Fl_Button(BORDER, BORDER,
		    BUTTON_WIDTH, BUTTON_HEIGHT, 
		    "RESET");
    resetButton->type(FL_NORMAL_BUTTON);
    resetButton->callback(resetCallback, (void *) item);
    resetButton->color(FL_CYAN) ;
  }

  if (setall) {
    Fl_Button *setButton ;
    if (reset) {
      if (cols>1) {
	setButton = 
	  new Fl_Button(BORDER+BUTTON_WIDTH+GAP, BORDER,
			BUTTON_WIDTH, BUTTON_HEIGHT, 
			"SET ALL");
	
      } else {
	setButton = 
	  new Fl_Button(BORDER, BORDER+BUTTON_HEIGHT+GAP,
			BUTTON_WIDTH, BUTTON_HEIGHT, 
			"SET ALL");
      }
    } else {
      setButton = new Fl_Button(BORDER, BORDER,
				BUTTON_WIDTH, BUTTON_HEIGHT, 
				"SET ALL");
    }
    setButton->type(FL_NORMAL_BUTTON);
    setButton->callback(setCallback, (void *) item);
    setButton->color(FL_YELLOW) ;
  }

  for (int i = 0; i<num; i++) {
      
    //~printf("i=%d\n",i) ;

    // c is which column
    int c = i%cols ;

    // r is which row
    int r = i/cols ;

    // a spacer?
    if (item[i].label[0] == '\0') {
      //~printf("spacer\n") ;
      continue ;
    }

    // start a group?
    if ( (i>0 && item[i].radioCount>=0 && item[i].radioCount!=item[i-1].radioCount) ||
	 (i==0 && item[i].radioCount>=0) ) {
      //~printf("\ncreating group\n") ;
      g = new Fl_Group(0, 0, 
		       cols*(BUTTON_WIDTH+GAP),
		       (rows+pushdown)*(BUTTON_HEIGHT+GAP)) ;
    }

    // stop a group?
    if (item[i].radioCount<0 && (i>0 && item[i-1].radioCount>=0)) {
      //~printf("ending group\n\n") ;
      g->end() ;
      g = NULL ;
    }

    //~printf("%d: label = \"%s\" setCommand = \"%s\" resetCommand = \"%s\"",
    //~i,item[i].label,item[i].setCommand,item[i].resetCommand) ;     

    item[i].n = i ; 
    item[i].button = new Fl_Light_Button(BORDER+c*(BUTTON_WIDTH+GAP), 
					 BORDER+(r+pushdown)*(BUTTON_HEIGHT+GAP),  
					 BUTTON_WIDTH, BUTTON_HEIGHT, 
					 item[i].label);
    if (item[i].radioCount>=0) {
      item[i].button->type(FL_RADIO_BUTTON);
      //~printf(", radio button\n") ;
    } else {
      if (item[i].resetCommand) {
	item[i].button->type(FL_NORMAL_BUTTON);
	//~printf(", normal button\n") ;
      } else {
	item[i].button->type(FL_TOGGLE_BUTTON);
	//~printf(", toggle button\n") ;
      }
    }
    item[i].button->callback(item_callback, (void *) &(item[i]));

    if (item[i].initSet) {
      if (item[i].radioCount>=0) {
	item[i].button->setonly() ;
      }
      else
	item[i].button->set() ;
    } else
      item[i].button->clear() ;

  }

  // in case the end got left off or the last button is part of a group
  if (g) {
    g->end() ;
    //~printf("ending group\n\n") ;
  }

  window->end();
  window->show(fltk_argc, fltk_argv);
  return Fl::run();
}
