#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <Performer/pf.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfGroup.h>

#include <Performer/pr.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfGeoState.h>
#include <Performer/pr/pfGeoMath.h>
#include <Performer/pr/pfMaterial.h>

#include <Performer/pfutil.h>
#include <Performer/pfdu.h>

#include "getString.C"
#include "normalize.C"
// this same function is in the dpf library, but copied here for use with file loaders
#include "parse.C"

// comment this out to get rid of spewage
//#define DEBUG

static void remove_commas(char *buffer) {
  int j;
  /* Remove commas, so the file can work either way */
  for (j=0; j<strlen(buffer); j++) {
    if (buffer[j] == ',') {
        buffer[j] = ' ';
    }
  }
}

extern "C" pfNode * pfdLoadFile_gtb (char *filename) {
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_gtb: %s", filename);
  FILE *fp = pfdOpenFile(filename);
  if (!fp)
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_gtb: unable to open %s for reading", filename);

  int notifyLevel = pfGetNotifyLevel() ;
  
  // input line number for error messages
  int lineNum = 0 ;

  // read in the file and store data
  char *line = NULL ;
  int argc = 0 ;
  char **argv = NULL ;

  // what section are we in?

  // section 0, starting, or between files
  // section 1, argc=2, one line, number of verts, number of polys
  // section 2, argc=3 or argc=7, "number of verts" lines
  // section 3, argc>2, "number of verts" polys
  int section = 0 ;

  // total number of polygons and vertices
  int numV ;
  int numP ;

  // how many verts and polys we're gotten so far
  int v ;
  int p ;

  // stores vertices, each entry is xyz rgba
  float **verts = NULL ;

  // default color if not specified
  float defColor[4] ;

  bool transparency = false ;

  // restore builder to initial state
  pfdResetBldrGeometry();
  pfdResetBldrState();
  
  // since normals aren't supplied, generate them
  pfdBldrMode(PFDBLDR_AUTO_NORMALS, PFDGBLDR_NORMALS_GENERATE) ;
  
  // make nice tristrips
  pfdBldrMode(PFDBLDR_BREAKUP, 1) ;
  pfdBldrMode(PFDBLDR_BREAKUP_SIZE, 0) ;
  pfdBldrMode(PFDBLDR_BREAKUP_BRANCH, 20) ;
  pfdBldrMode(PFDBLDR_BREAKUP_STRIP_LENGTH, 3) ;
  pfdBldrMode(PFDBLDR_MESH_RETESSELLATE, 1) ;
  
  // set materials
  pfMaterial *m = new pfMaterial();
  m->setColor(PFMTL_AMBIENT,  0.2f, 0.2f, 0.2f);
  m->setColor(PFMTL_DIFFUSE,  0.8f, 0.8f, 0.8f);
  m->setColor(PFMTL_SPECULAR, 0.0f, 0.0f, 0.0f);
  m->setColor(PFMTL_EMISSION, 0.0f, 0.0f, 0.0f);
  m->setColorMode(PFMTL_BOTH, PFMTL_CMODE_AMBIENT_AND_DIFFUSE);
  
  pfdBldrStateAttr(PFSTATE_FRONTMTL, m);	  
  pfdBldrStateAttr(PFSTATE_BACKMTL, m);
  pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_ON);


  // start reading in the data
  while (1) {
    if (notifyLevel >= PFNFY_INFO) {
      fprintf(stderr,"%8.8d   \r",lineNum) ;
      fflush(stderr) ;
    }
    if (section == 0) {
      v = p = 0 ;
      defColor[0] = defColor[1] = defColor[2] = defColor[3] = 0.f ;
      if (verts) {
#ifdef DEBUF
	printf("freeing %d vertices\n", numV) ;
#endif
	for (int i=0; i<numV; i++) 
	  free(verts[i]) ;
	free(verts) ;
	verts = NULL ;
      }
      section = 1 ;
    }
    if (notifyLevel >= PFNFY_INFO) {
      fprintf(stderr,"%8.8d   \r",lineNum) ;
      fflush(stderr) ;
    }
    lineNum++ ;
    if (getString(&line, fp)<0) {
      if (p!=0)
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_gtb: section 3 contains too few polygons") ;
      break ;
    } else {
#ifdef DEBUG
      printf("line = \"%s\"\n", line) ;
#endif

      /* Remove commas, so the file can work either way */
      remove_commas(line);

      if (normalize(line)) {
	if (dpf_charsToArgcArgv(line, &argc, &argv)>0) {
#ifdef DEBUG
	  printf("argc = %d, argv = ",argc) ;
	  for (int i=0; i<argc; i++) 
	    printf("\"%s\" ",argv[i]) ;
	  printf("\n") ;
#endif

	  if (section==1) {
	    if (argc!=2) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: section 1 doesn't contain two items") ;
	    } else {
	      if ((sscanf(argv[0],"%d",&numV)==1) &&
		  (sscanf(argv[1],"%d",&numP)==1) && 
		  numV>0 && numP>0) {
		section = 2 ;
		verts = (float **) malloc(sizeof(float*)*numV) ;
		//pfNotify(PFNFY_INFO, PFNFY_PRINT, 
#ifdef DEBUG
		printf(
			 "pfdLoadFile_gtb: section 1: %d vertices, %d polygons\n", numV, numP) ;
#endif
	      } else {
		pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
			 "pfdLoadFile_gtb: section 1 doesn't contain two positive integers") ;
	      }
	    }

	  } else if (section==2) {
	    if (v>=numV)
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: line %d, section 2, contains too many vertices",
		       lineNum) ;
	    if (argc==3 || argc==7) {
	      float *vert = (float *) malloc(sizeof(float)*7) ;
	      verts[v] = vert ;
	      if (argc==3 &&
		  (sscanf(argv[0],"%f",vert)==1) &&
		  (sscanf(argv[1],"%f",vert+1)==1) &&
		  (sscanf(argv[2],"%f",vert+2)==1)) {
		memcpy(vert+3, defColor, sizeof(float)*4) ;
	      } else if (argc==7 &&
			 (sscanf(argv[0],"%f",vert)==1) &&
			 (sscanf(argv[1],"%f",vert+1)==1) &&
			 (sscanf(argv[2],"%f",vert+2)==1) &&
			 (sscanf(argv[3],"%f",vert+3)==1) &&
			 (sscanf(argv[4],"%f",vert+4)==1) &&
			 (sscanf(argv[5],"%f",vert+5)==1) &&
			 (sscanf(argv[6],"%f",vert+6)==1) &&
			 *(vert+3)>=0.f && *(vert+3)<=1.f &&
			 *(vert+4)>=0.f && *(vert+4)<=1.f &&
			 *(vert+5)>=0.f && *(vert+5)<=1.f &&
			 *(vert+6)>=0.f && *(vert+6)<=1.f) {
		memcpy(defColor, vert+3, sizeof(float)*4) ;
		if (*(vert+6)<.9)
		  transparency = true ;
	      } else {
		pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
			 "pfdLoadFile_gtb: line %d, section 2, invalid line", lineNum) ;
		
	      }
	      v++ ;
	      if (v==numV)
		section = 3 ;
#ifdef DEBUG
	      printf("section 2, line %d, vert=%d, (%f %f %f) (%f %f %f %f)\n", lineNum, v,
		     *vert, *(vert+1), *(vert+2), *(vert+3), *(vert+4), *(vert+5), *(vert+6)) ;
#endif
	      if (transparency)
		pfdBldrStateMode(PFSTATE_TRANSPARENCY, PFTR_HIGH_QUALITY|PFTR_NO_OCCLUDE);
	      else
		pfdBldrStateMode(PFSTATE_TRANSPARENCY, PFTR_HIGH_QUALITY);
	      
	    } else {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: line %d, section 2, doesn't contain three or seven items",
		       lineNum) ;
	    }
	    
	  } else if (section==3) {
	    if (p>=numP)
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: line %d, section 3, contains too many polygons",
		       lineNum) ;
	    if (argc<2) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: line %d, section 3, contains too few numbers",
		       lineNum) ;
	    }
	    int nv ;
	    if (!(sscanf(argv[0],"%d",&nv)==1)) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: line %d, section 3, contains an invalid number",
		       lineNum) ;
	    }
	    if ((nv+1)!=argc) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_gtb: line %d, section 3, contains the wrong number of indices",
		       lineNum) ;
	    }
	    // process vertices

	    // for figuring if we can one color per poly or one per vertex
	    bool singleColor = true ;

	    // total number of vertices
	    pfdGeom *prim = pfdNewGeom(nv) ;

	    if (nv==1) 
	      prim->primtype = PFGS_POINTS ;
	    else if (nv==2)
	      prim->primtype = PFGS_LINES ;
	    else
	      prim->primtype = PFGS_POLYS ;
	    for (int i=0; i<nv; i++) {
	      int vi ; //vertex index
	      if (!((sscanf(argv[i+1],"%d",&vi)==1) &&
		  vi>=1 && vi<=(numV))) {
		pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
			 "pfdLoadFile_gtb: line %d, section 3, contains an invalid index",
			 lineNum) ;
	      }
	      vi-- ;
#ifdef DEBUG
	      printf("section 3, line %d, poly=%d, nv=%d, i=%d, index=%d\n", lineNum, p+1,
		     nv, i, vi+1) ;
	      printf("           referenced data = (%f %f %f) (%f %f %f %f)\n",
		     verts[vi][0], verts[vi][1], verts[vi][2], 
		     verts[vi][3], verts[vi][4], verts[vi][5], verts[vi][6]) ;
#endif
	      // copy vertex data and color
	      prim->coords[i] = pfVec3(verts[vi][0], verts[vi][1], verts[vi][2]) ;
	      prim->colors[i] = pfVec4(verts[vi][3], verts[vi][4], verts[vi][5], verts[vi][6]) ;

	      // see if we can use a single color for whole polygon
	      float initColor[4] ;
	      if (i==0) {
		memcpy(initColor, &(verts[vi][3]), sizeof(float)*4) ;
#ifdef DEBUG
		printf("setting initial color for polygon to (%f %f %f %f)\n",
		       initColor[0], initColor[1], initColor[2], initColor[3]) ;
#endif
	      } else if (singleColor && memcmp(initColor,&(verts[vi][3]), sizeof(float)*4)) {
#ifdef DEBUG
		printf("singleColor = false\n") ;
#endif
		singleColor = false ;
	      }
	    }

	    // turn off textures no matter what
	    for (int t=0; t<PF_MAX_TEXTURES; t++)
	      prim->tbind[t] = PFGS_OFF;

	    // set lighting type
	    if (singleColor)
	      prim->cbind = PFGS_PER_PRIM;
	    else
	      prim->cbind = PFGS_PER_VERTEX;

	    // add primitive to builder 
	    pfdAddBldrGeom(prim, 1);
  
	    // release primitive buffer
	    pfdDelGeom(prim);

	    p++ ;
	    if (p==numP) {
#ifdef DEBUG
	      printf("setting section to 0 in case there's more data\n") ;
#endif	      
	      section = 0 ;
	    }
	  } else {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_gtb: internal coding error, section = %d",
		     section);
	  }
	}
      } //if (normalize(line)) 
    } //if (getString(&line, fp)<0)
  } //while (1)
  
  if (!line)
    free(line) ;

  if (!argv) {
    for (int i=0; i<argc; i++)
      free(argv[i]) ;
    free(argv) ;
  }

  fclose(fp);
  
  // get a complete scene graph representing file's primitives
  pfNode *node = pfdBuild();

  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_gtb: rewritten 1/19/05, JTK");
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_gtb: finished reading %s", 
	   filename);
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_gtb: file contains %d lines, %d vertices, %d polygons", 
	   lineNum, numV, numP);
  
  return(node) ;
}

  

