/* 
   pfifl.c - instancing file loader with unique transformations
   
   usage:  pfNode *foo = pfdLoadFile("foo.ifl") ;
   
   A ifl file has the following format:
   
   # this is a comment
   
   # load a file named "filename.foo" under a node named "name" with an
   # optional transformation.  Use the name NULL to not assign a name to the
   # parent node

   # duplicate files are automatically instanced.  Each file is loaded or
   # instanced under either a pfGroup (identity transformation) or unique
   # pfSCS node.  If the node has a name, a pfGroup node with the name is
   # created under the pfSCS, with the loaded or instanced file under that.

   # duplicated lines are ignored

   filename.foo name
   filename.foo name tx ty tz
   filename.foo name tx ty tz h p r
   filename.foo name tx ty tz h p r s
   filename.foo name tx ty tz h p r sx sy sx
   
   transformed coordinate = scale*rotate*translate relative to a 
      fixed coordinate system

   see example3.ifl for a visual example

   for better culling, consider using the breakup pseudo-loader, such as
   perfly -W 1000 example.ifl.0,20,10.breakup

   John Kelso, kelso@nist.gov
   April 27, 2004
   
*/

//#define SPEW

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <Performer/pf.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfSCS.h>
#include <Performer/pfdu.h>

#include "getString.C"
#include "normalize.C"
#include "parse.C"

// function templates
void cleanup(void) ;

void loadfile(char *filename, char *nodename,
	      char *x=NULL, char *y=NULL, char*z=NULL, 
	      char *h=NULL, char *p=NULL, char *r=NULL,
	      char *sx=NULL, char *sy=NULL, char *sz=NULL) ;

// file name, pointer to pfNode where the file is loaded
struct fileNode {
  char *name ;
  pfNode *node ;
} ;

// name node, pointer to named pfGroup
struct nameNode {
  char *name ;
  pfGroup *node ;
} ;

// record the data for a single object- store indices of file, name, scs
struct objectData {
  int file ;
  char *name ;
  int scs ;
} ;

// evil global variables

// input line number for error messages
int lineNum ;

// an array of unique pfSCS nodes under which are loaded/instanced files or name pfGroups
pfSCS **scs ;
// how many of them?
int scsN ;
// stuff with no transformation goes under this node
pfGroup *ident ;

// an array of unique fileNodes
fileNode *file ;
// how many of them?
int fileN ;

// an array of unique objectData
objectData *object ;
// how many of them?
int objectN ;

pfGroup *namedNode ;

// group node to return ;
pfGroup *group ;

extern "C" pfNode * pfdLoadFile_ifl (char *filename) {
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_ifl: %s", filename);
  FILE *fp = fopen(filename, "r");
  if (!fp)
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_ifl: unable to open %s for reading", filename);
  
  int notifyLevel = pfGetNotifyLevel() ;

  // in a DSO you need to initialize with assignment statements
  lineNum = 0 ;

  file = NULL ;
  fileN = 0 ;

  namedNode = NULL ;

  scs = NULL ;
  scsN = 0 ;
  ident = NULL ;

  object = NULL ;
  objectN = 0 ;

  group = NULL ;

  // read in the file and store data
  char *line = NULL ;
  int argc ;
  char **argv = NULL ;

  // loop through the input
  while (1) {
    if (notifyLevel >= PFNFY_INFO) {
      fprintf(stderr,"%8.8d   \r",lineNum) ;
      fflush(stderr) ;
    }
    lineNum++ ;

    if (getString(&line, fp)<0) {
      cleanup() ;
      break ;
    } else {
      if (normalize(line)) {
	dpf_charsToArgcArgv(line, &argc, &argv) ;
	if (argc==2) // just a filename and nodename
	  loadfile(argv[0], argv[1]) ;
	else if (argc==5) // filename, nodename and translation
	  loadfile(argv[0], argv[1], 
		   argv[2],argv[3],argv[4]) ;
	else if (argc==8) // filename, nodename, translation and rotation
	  loadfile(argv[0], argv[1], 
		   argv[2],argv[3],argv[4], 
		   argv[5],argv[6],argv[7]) ;
	else if (argc==9) // filename, nodename, translation, rotation
	                  // and overall scale
	  loadfile(argv[0], argv[1], 
		   argv[2],argv[3],argv[4], 
		   argv[5],argv[6],argv[7], 
		   argv[8]) ;
	else if (argc==11) // filename, nodename, translation, rotation
	                   // and individual scales
	  loadfile(argv[0], argv[1], 
		   argv[2],argv[3],argv[4], 
		   argv[5],argv[6],argv[7], 
		   argv[8],argv[9],argv[10]) ;
      } //if (normalize(line))
    } //if (getString(&line, fp)<0)
  } // while (1)

  fclose(fp);
  return(group) ;
}

void cleanup(void) {
  int i ;
  for (i=0; i<fileN; i++) {
    free(file[i].name) ;
  }
  if (file) 
    free (file) ;
  for (i=0; i<objectN; i++) {
    free(object[i].name) ;
  }
  if (object) 
    free (object) ;
}

void loadfile(char *filename, char *nodename,
	      char *x, char *y, char*z, 
	      char *h, char *p, char *r,
	      char *sx, char *sy, char *sz) {

#ifdef SPEW
  printf("loadfile %s %s (%s %s %s) (%s %s %s) (%s %s %s)\n",
	 filename,  nodename,
	 x,   y,   z, 
	 h,   p,   r,
	 sx,  sy,  sz) ;
#endif

  // see if filename has already been loaded
  int thisFile ;
  for (thisFile=0; thisFile<fileN; thisFile++) {
    if (!strcmp(file[thisFile].name,filename)) {
#ifdef SPEW
      printf("File %s, %d, already loaded\n",filename, thisFile) ;
#endif
      break ;
    }
  }
  // didn't find it, so add it
  if (thisFile>=fileN) { 
    thisFile = fileN ;
    fileN++ ;
    file = (fileNode *) realloc(file, sizeof(fileNode)*fileN) ;
    file[thisFile].name = strdup(filename) ;
#ifdef SPEW
    printf("Trying to load file %s, %d\n",filename,thisFile) ;
#endif
    if (!(file[thisFile].node = pfdLoadFile(filename))) // unable to load file
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_ifl: line %d: unable to load file \"%s\"", 
	       lineNum, filename);
    // do this here so we can return a NULL if no files loaded
    if (!group)
      group = new pfGroup() ;
  }
  
  // do they want a node name, and if so, is it unique?
  if (strcmp("NULL",nodename)) {
    namedNode = new pfGroup ;
    if (!(namedNode->setName(nodename))) // name already exists!
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_ifl: line %d: unable to name node \"%s\"", 
	       lineNum, nodename);
#ifdef SPEW
    printf("New named node %s\n",nodename) ;
#endif
  } else {
    namedNode = NULL ;
#ifdef SPEW
    printf("Using unnamed node\n") ;
#endif
  }
  

  pfCoord coord ;
  pfVec3 scale ;

  // initialize to identity
  coord.xyz = pfVec3(0.f, 0.f, 0.f) ;
  coord.hpr = pfVec3(0.f, 0.f, 0.f) ;
  scale = pfVec3(1.f, 1.f, 1.f) ;

  if (x && y && z) {
    if ((sscanf(x,"%f",&coord.xyz[0])!=1) ||
	(sscanf(y,"%f",&coord.xyz[1])!=1) ||
	(sscanf(z,"%f",&coord.xyz[2])!=1))
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_ifl: line %d, "
	       "invalid number", lineNum);
    if (h && p && r) {
      if ((sscanf(h,"%f",&coord.hpr[0])!=1) ||
	  (sscanf(p,"%f",&coord.hpr[1])!=1) ||
	  (sscanf(r,"%f",&coord.hpr[2])!=1))
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_ifl: line %d, "
		 "invalid number", lineNum);
      if (sx) {
	if ((sscanf(sx,"%f",&scale[0])!=1))
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_ifl: line %d, "
		   "invalid number", lineNum);
	if (sy && sz) {
	  if ((sscanf(sy,"%f",&scale[1])!=1) ||
	      (sscanf(sz,"%f",&scale[2])!=1))
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_ifl: line %d, "
		     "invalid number", lineNum);
	} else {
	  scale[1] = scale[2] = scale[0] ;
	} //if (sy && sz)
      } //if (sx)
    } //if (h && p && r) 
  } //if (x && y && z)
  
  int thisSCS ;
  // see if it's an identity matrix
  if (!x ||
      (coord.xyz[0] == 0.f && coord.xyz[1] == 0.f && coord.xyz[2] == 0.f && 
       coord.hpr[0] == 0.f && coord.hpr[1] == 0.f && coord.hpr[2] == 0.f && 
       scale[0] == 1.f && scale[1] == 1.f && scale[2] == 1.f)) { 
    thisSCS = -1 ;
#ifdef SPEW
    printf("Identity transformation\n") ;
#endif
    if (!ident) {
      ident = new pfGroup() ;
      group->addChild(ident) ;
    }
  } else {
    pfMatrix matrix ;
    matrix.makeCoord(&coord) ;
    matrix.preScale(scale[0], scale[1], scale[2], matrix) ;
    for (thisSCS=0; thisSCS<scsN; thisSCS++) {
      if (*(scs[thisSCS]->getMatPtr()) == matrix) {
#ifdef SPEW
	printf("Reusing pfSCS %d\n",thisSCS) ;
#endif
	break ;
      }
    }
    // didn't find it, so add it
    if (thisSCS>=scsN) { 
      thisSCS = scsN ;
      scsN++ ;
      scs = (pfSCS **) realloc(scs, sizeof(pfSCS *)*scsN) ;
      scs[thisSCS] = new pfSCS(matrix) ;
#ifdef SPEW
      printf("Creating new pfSCS %d\n",thisSCS) ;
#endif
      group->addChild(scs[thisSCS]) ;
    }
  }

  // see if it's a unique object - to be so it needs to have same file name,
  // same node name and same orientation
  int thisObject ;
  for (thisObject=0; thisObject<objectN; thisObject++) {
    if ( (thisFile == object[thisObject].file) &&
	 (!strcmp(object[thisObject].name,nodename)) &&
	 (thisSCS == object[thisObject].scs)) {
      pfNotify(PFNFY_NOTICE, PFNFY_PRINT, 
	       "pfdLoadFile_ifl: %s, line %d, "
	       "duplicate object detected, skipping", 
	       filename, lineNum);
      break ;
    }
  }

  // didn't find it, so add it to list of objects
  if (thisObject>=objectN) { 
    thisObject = objectN ;
    objectN++ ;
    object = (objectData *) realloc(object, sizeof(objectData)*objectN) ;
    object[thisObject].file = thisFile ;
    object[thisObject].name = strdup(nodename) ;
    object[thisObject].scs = thisSCS ;
#ifdef SPEW
    printf("new object %d loaded\n",objectN) ;
#endif

    if (!namedNode) {       // no node name so connect directly 
                            //under scs[] of ident
      if (thisSCS == -1) {  // identity
	ident->addChild(file[thisFile].node) ;
      } else {
	scs[thisSCS]->addChild(file[thisFile].node) ;
      }
    } else {                // connect under named node
      if (thisSCS == -1) {  // identity
	ident->addChild(namedNode) ;
      } else {
	scs[thisSCS]->addChild(namedNode) ;
      }
      namedNode->addChild(file[thisFile].node) ;
    }
  } //if (thisObject>=objectN)
}
