/*

 libpfm4.C - ".m4" pseudo-loader.  preprocesses a file though m4, load the
 m4 output file, then deletes the output file

 Example, to load an ifl file with m4 proprocessor commands named foo.ifl,
 use:
    pfdLoadFile("foo.ifl.m4") ;

 Set the environment variable PFM4TMPDIR to specify an existing direcotory
 to hold the output file

  Author of this version:
      John Kelso
      Scientific Applications & Visualization Group
           SAVG web page: http://math.nist.gov/mcsd/savg/
      Mathematical & Computational Science Division
           MCSD web page: http://math.nist.gov/mcsd/
      National Institute of Standards and Technology
           NIST web page: http://www.nist.gov/

  For information about this software contact John Kelso, kelso@nist.gov

  "This software was developed at the National Institute of Standards and
  Technology by employees of the Federal Government in the course of their
  official duties. Pursuant to title 17 Section 105 of the United States
  Code this software is not subject to copyright protection and is in the
  public domain. This is an experimental system.  NIST assumes no
  responsibility whatsoever for its use by other parties, and makes no
  guarantees, expressed or implied, about its quality, reliability, or any
  other characteristic.

 We would appreciate acknowledgement if the software is used."

*/

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

#include <Performer/pfdu.h>

#include <string>

using namespace std ;

extern "C" pfNode * pfdLoadFile_m4 (char *c_filename) {
    
  pfNotify(PFNFY_INFO, PFNFY_USAGE,
	   "pfdLoadFile_m4: processing %s", c_filename);

  FILE *fp = pfdOpenFile(c_filename);
  if (!fp) {
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "  pfdLoadFile_m4: unable to open %s for reading", c_filename);
    return NULL ;
  }
  fclose(fp);
  
  if (strcmp(".m4",c_filename+strlen(c_filename)-3)) {
    pfNotify(PFNFY_FATAL, PFNFY_PRINT,
	     "pfdLoadFile_m4: file %s does not end in \".m4\"", c_filename);
    return NULL ;
  }
  
  // return values from system calls
  int ret ;
  
  // start having fun with STL
  string filename = c_filename ;

  // base of m4 output file
  string base ;

  // is there a directory in filename?
  string::size_type dir = filename.rfind('/') ;
  if (dir == string::npos)
    base = filename ;
  else 
    base = filename.substr(dir+1) ;
  
  // get temp dir
  string tmpdir ;
  if (getenv("PFM4TMPDIR")) 
    tmpdir = getenv("PFM4TMPDIR") ;
  else
    tmpdir = "/var/tmp" ;
  DIR *dirp ;
  if (!(dirp=opendir(tmpdir.c_str()))) {
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "  pfdLoadFile_m4: tmpdir \"%s\" is not a directory", 
	     tmpdir.c_str());
    return NULL ;
  }
  closedir(dirp) ;
  
  // extract ".m4"
  base = base.substr(0,base.rfind('.')) ;
  //cout << "base = " << base << endl ;
  
  // try to find a unused output file name
  string out ;
  // increment this until it's unique
  int ispin = 0 ;
  while(1) {
    // come up with a tempname
    char c_pid[11] ;
    sprintf(c_pid,"%10.10d",(int)getpid()) ;
    string pid = c_pid ;
    
    char c_spin[6] ;
    sprintf(c_spin,"%5.5d",ispin) ;
    string spin = c_spin ;

    out = tmpdir + "/libpfm4-" + pid + "-" + spin + "-" + base ;
    //cout << "out = " << out << endl ;
    
    struct stat buf ;
    if (-1==stat(out.c_str(),&buf)) 
      break ;
    ispin++ ;
    break ;
  }
  
  fp = fopen(out.c_str(), "w");
  if (!fp) {
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "  pfdLoadFile_m4: unable to open %s for writing", out.c_str());
    return NULL ;
  }
  fclose(fp);

  string cmd = "m4<" + filename + ">" + out ;
  //cout << "cmd = " << cmd << endl ;
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT,
	   "pfdLoadFile_m4: executing command:\n      %s", cmd.c_str());

  ret = system(cmd.c_str()) ;
  pfNode *node = NULL ;
  if (ret==0) 
    node = pfdLoadFile(out.c_str()) ;

  // delete out
  unlink(out.c_str()) ;

  return node ;
  
}
