#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <Performer/pf.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfGroup.h>

#include <Performer/pr.h>
#include <Performer/pr/pfFont.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfGeoState.h>
#include <Performer/pr/pfGeoMath.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pr/pfString.h>
#include <Performer/pf/pfText.h>
#include <Performer/pr/pfTexture.h>

#include <Performer/pfdu.h>
#include <Performer/pfutil.h>

#include "getString.C"
#include "normalize.C"
#include "parse.C"

// function templates
static void processPrimitive(void) ;
static void resetState(void) ;
static void addCoord(pfVec3 v) ;
static void addNorm(pfVec3 no) ;
static void addColor(pfVec4 c) ;
static void addTex2d(pfVec2 t) ;
static void addTex3d(pfVec3 t) ;
static void printPrimitive(void) ;
static void printMaterial(void) ;
static void doTri(void) ;
static void doPolyOrLineOrPoint(void) ;
static void doPolyN(void) ;
static void doMaterial(void) ;
static void breakupNodes(pfNode **node, pfGroup *parent) ;
static void shrinkVerts(int n, pfVec3 *verts, float scale) ;

static bool parsePreserveplanarquads(int argc, char **argv, int l) ;
static bool parsePlanarpolygonchecking(int argc, char **argv, int l) ;
static bool parseVerbosity(int argc, char **argv, int l) ;
static bool parseIndexed(int argc, char **argv, int l) ;
static bool parseTexture(int argc, char **argv, int l) ;
static bool parseText(int argc, char **argv, int l) ;
static bool parseTransparency(int argc, char **argv, int l) ;
static bool parseLighting(int argc, char **argv, int l) ;
static bool parseColormode(int argc, char **argv, int l) ;
static bool parseAlpha(int argc, char **argv, int l) ;
static bool parseShininess(int argc, char **argv, int l) ;
static bool parseStyle(int argc, char **argv, int l) ;
static bool parseStriplength(int argc, char **argv, int l) ;
static bool parsePixelsize(int argc, char **argv, int l) ;
static bool parseShrinkage(int argc, char **argv, int l) ;
static bool parseOctree(int argc, char **argv, int l) ;
static bool parseColorMode(int argc, char **argv, int l) ;
static bool parsePrimitive(int argc, char **argv, int l) ;
static bool parseTexturedPrimitive(int argc, char **argv, int l) ;
static bool parseData(int argc, char **argv, int l) ;
static bool parse2dTexturedData(int argc, char **argv, int l) ;
static bool parse3dTexturedData(int argc, char **argv, int l) ;

static bool parseNumbers(char **argv, int *i0) ;
static bool parseNumbers(char **argv, float *f0) ;
static bool parseNumbers(char **argv, float *f0, float *f1) ;
static bool parseNumbers(char **argv, float *f0, float *f1, float *f2) ;
static bool parseNumbers(char **argv, float *f0, float *f1, float *f2, float *f3) ;


// can set this to 0 to test SGI's pfdBreakup
#define SAVGBREAKUP 1

#if SAVGBREAKUP
pfNode *savgPfdBreakup(pfGeode *geode, float geodeSize, 
		       int stripLength, int geodeChild) ;
#endif

// evil global state variables

// since memory is reused for each file loaded, can't initialize them here-
// must do it in assign statements!!

// for verbosity level
bool verbose ;
bool debug ;

// for indexing of geometry
bool indexed ;

// input line number for error messages
int lineNum ;

// how many primitives have been processed
int primNum ;

// type of primitive we're creating
enum {END, LINE, POLY, POLYN, TRI, POINT} primitive ;
// this one doesn't change- it's just for spewage printing
const char *primitiveName[] = {"END", "LINE", "POLY", "POLYN", "TRI", "POINT"} ;

// color per primitive, or color per vertex
enum {PRIM, VERTEX} color ;
// the single primitive color value
pfVec4 primColor ;

// used to specify state of transparency and lighting
enum {ON, OFF, DEFAULT, AUTO} lightingState, transparencyState ;

// true if color data includes transparency
bool transparency ;

// how big does alpha get before we call it opaque?
#define OPAQUEALPHA (.9)

// true if normals data are being supplied
bool normals ;

// number of sides in a polygon if parameter is supplied
int sides ;

// number of data lines in current primitive
int n ;

// true if we pass colors to builder- i.e. colormode != off
bool passColors ;

// set to the type of texturing we've loaded
//   2==2d, 3==3d, else none
int textureFile ;

// set to the type of texturing we're doing in the current polygon
//   2==2d, 3==3d, else none
int textureType ;
pfTexture *texture ;
pfTexEnv *textureEnv ;

// array of vertices
pfVec3 *coords ;
// number of elements in coords array
int nCoords ;

// array of 2d textures
pfVec2 *texCoords ;
// number of elements in texCoords array
int nTexCoords ;

// array of normals
pfVec3 *norms ;
// number of elements in norms array
int nNorms ;

// array of colors
pfVec4 *colors ;
// number of elements in colors array
int nColors ;

// how many elements to increase each array by
#define BUMP (1048576)

// stores current material
pfMaterial *mat ;

// default values- values as set when loader starts
struct materials {
  int transparency ;
  int lighting ;
  int colormode ;
  float ambient[3] ;
  float diffuse[3] ;
  float emission[3] ;
  float specular[3] ;
  float alpha ;
  float shininess ;
} ;

// hang text and geometry under this node
pfGroup *group ;

// stores information about text strings
struct {
  pfFont *font ;
  char fontName[PF_MAXSTRING] ;
  int justification ;
  int style ;
  bool flatten ;
} textMode ;

// holds default values- values as set when loader starts
materials defaults ;

// pixelsize
float pixelsize ;

// shrinkage
float shrinkage ;

// style
enum {FILLED, WIREFRAME} style ;

// values to be passed to savgPfdBreakup
float geodeSize ;
int stripLength ;
int geodeChild ;

// do we retessellate planar quads?
bool retessellate ;

// check for planar polygons?
bool planarpolygoncheck ;
pfPlane *plane ;

extern "C" pfNode * pfdLoadFile_savg (char *filename) {
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_savg: %s", filename);
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_savg: compiled on %s at %s", __DATE__,__TIME__);
  FILE *fp = pfdOpenFile(filename);
  if (!fp) {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: unable to open %s for reading", filename);
  }

  // initialize global state variables
  debug = false ;
  verbose = false ;
  indexed = false ;
  n = 0 ;
  lineNum = 0 ;
  primNum = 0 ;
  primColor = pfVec4(1.f, 1.f, 1.f, 1.f) ;
  coords = NULL ;
  nCoords = 0 ;
  texCoords = NULL ;
  nTexCoords = 0 ;
  norms = NULL ;
  nNorms = 0 ;
  colors = NULL ;
  nColors = 0 ;
  pixelsize = 1.f ;
  shrinkage = 1.f ;
  style = FILLED ;
  retessellate = true ;
  planarpolygoncheck = false ;
  lightingState = AUTO ;
  transparencyState = AUTO ;
  plane = new pfPlane ;
  textureType = 0 ;
  textureFile = 0 ;
  texture = NULL ;
  textureEnv = NULL ;

  group = NULL ;

  textMode.font = NULL ;
  strcpy(textMode.fontName,"Times-Elfin") ;
  textMode.flatten = false ;
  textMode.justification = PFSTR_LEFT ;
  textMode.style = PFDFONT_EXTRUDED ;

  // pfdbreakup says good values are 1000, 20, 10
  // pfdBuilder uses 1000, 10, 4 for default values
  // pfdMesher uses 1000 for default strip length- go figure...
  // I'm making these work for best quality rendering and reasonable culling
  geodeSize = 0.f;
  stripLength = 1 ; // stripLength is number of triangles per geoset
  geodeChild = 20 ;

  resetState() ;

  // restore builder to initial state
  pfdResetBldrGeometry();
  pfdResetBldrState();

  mat = new pfMaterial ;

  // apply materials to both sides.  Need to have two sided lights too!  
  // See pfLightModel::setTwoSide
  mat->setSide(PFMTL_BOTH) ;

  {
    // set up default material
    pfMaterial *initMat = new pfMaterial ;
    defaults.colormode = initMat->getColorMode(PFMTL_FRONT) ;
    if (defaults.colormode == PFMTL_CMODE_OFF) 
      passColors = false ;
    else 
      passColors = true ;
    
    initMat->getColor(PFMTL_CMODE_AMBIENT, 
		      defaults.ambient, defaults.ambient+1, defaults.ambient+2) ;
    initMat->getColor(PFMTL_CMODE_DIFFUSE, 
		      defaults.diffuse, defaults.diffuse+1, defaults.diffuse+2) ;
    initMat->getColor(PFMTL_CMODE_SPECULAR, 
		      defaults.specular, defaults.specular+1, defaults.specular+2) ;
    initMat->getColor(PFMTL_CMODE_EMISSION, 
		      defaults.emission, defaults.emission+1, defaults.emission+2) ;
    defaults.alpha = initMat->getAlpha() ;
    defaults.shininess = initMat->getShininess() ;
    pfDelete(initMat) ;

    // set up default state
    pfGeoState *initGState = new pfGeoState ;
    defaults.transparency = initGState->getMode(PFSTATE_ENLIGHTING) ;
    defaults.lighting = initGState->getMode(PFSTATE_ENLIGHTING) ;
    pfDelete(initGState) ;
  }

  // to get this, init verbose & debug to true
  if (debug)  {
     printf("pfdLoadFile_savg debug: default values:\n") ;
     if (passColors)
       printf("pfdLoadFile_savg debug:   passing color information to the builder\n") ;
     else
       printf("pfdLoadFile_savg debug:   NOT passing color information to the builder\n") ;
     printf("pfdLoadFile_savg debug:   lighting = %d\n",defaults.lighting) ;
     printf("pfdLoadFile_savg debug:   transparency = %d\n",defaults.transparency) ;
     printf("pfdLoadFile_savg debug:   PFMTL_CMODE_AMBIENT = %f %f %f\n",
	   defaults.ambient[0], defaults.ambient[1], defaults.ambient[2]) ;
     printf("pfdLoadFile_savg debug:   PFMTL_CMODE_DIFFUSE = %f %f %f\n",
	   defaults.diffuse[0], defaults.diffuse[1], defaults.diffuse[2]) ;
     printf("pfdLoadFile_savg debug:   PFMTL_CMODE_EMISSION = %f %f %f\n",
	   defaults.emission[0], defaults.emission[1], defaults.emission[2]) ;
     printf("pfdLoadFile_savg debug:   PFMTL_CMODE_SPECULAR = %f %f %f\n",
	   defaults.specular[0], defaults.specular[1], defaults.specular[2]) ;
     printf("pfdLoadFile_savg debug:   Alpha = %f\n",defaults.alpha) ;
     printf("pfdLoadFile_savg debug:   Shininess = %f\n",defaults.shininess) ;
     printf("pfdLoadFile_savg debug:   default colormode = ") ;	 
    if (defaults.colormode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_AMBIENT_AND_DIFFUSE\n") ; 
    else if (defaults.colormode == PFMTL_CMODE_AMBIENT)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_AMBIENT\n") ; 
    else if (defaults.colormode == PFMTL_CMODE_DIFFUSE)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_DIFFUSE\n") ; 
    else if (defaults.colormode == PFMTL_CMODE_EMISSION)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_EMISSION\n") ; 
    else if (defaults.colormode == PFMTL_CMODE_SPECULAR)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_SPECULAR\n") ; 
    else if (defaults.colormode == PFMTL_CMODE_OFF)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_OFF\n") ; 
    else if (defaults.colormode == PFMTL_CMODE_COLOR)
       printf("pfdLoadFile_savg debug: PFMTL_CMODE_COLOR\n") ; 
    else 
       printf("pfdLoadFile_savg debug: unknown!!\n") ; 
}     


  // to get this, init verbose & debug to true
  if (debug) {
    printf("pfdLoadFile_savg debug: after resetting builder: \n") ;
    printf("pfdLoadFile_savg debug:   pfdGetBldrMode(PFDBLDR_MESH_SHOW_TSTRIPS) = %d\n",
	     pfdGetBldrMode(PFDBLDR_MESH_SHOW_TSTRIPS)) ;
    printf("pfdLoadFile_savg debug:   pfdGetBldrMode(PFDBLDR_MESH_INDEXED) = %d\n",
	     pfdGetBldrMode(PFDBLDR_MESH_INDEXED)) ;
    printf("pfdLoadFile_savg debug:   pfdGetBldrMode(PFDGBLDR_SHARE_INDEX_LISTS) = %d\n",
	     pfdGetBldrMode(PFDGBLDR_SHARE_INDEX_LISTS)) ;
    printf("pfdLoadFile_savg debug:   pfdGetBldrMode(PFDBLDR_MESH_MAX_TRIS) = %d\n",
	   pfdGetBldrMode(PFDBLDR_MESH_MAX_TRIS)) ;
     printf("pfdLoadFile_savg debug:   pfdGetBldrMode(PFDBLDR_MESH_RETESSELLATE) = %d\n",
	   pfdGetBldrMode(PFDBLDR_MESH_RETESSELLATE)) ;
  }

  int notifyLevel = pfGetNotifyLevel() ;

  // read in the file and store data
  char *line = NULL ;
  int argc ;
  char **argv = NULL ;
  while (1) {
    if (notifyLevel >= PFNFY_INFO) {
      fprintf(stderr,"%8.8d   \r",lineNum) ;
      fflush(stderr) ;
    }
    lineNum++ ;
    if (getString(&line, fp)<0) {
      processPrimitive() ;
      break ;
    } else {
      if (normalize(line)) {
	dpf_charsToArgcArgv(line, &argc, &argv) ;
	if (debug)
	  printf("pfdLoadFile_savg debug: line: \"%s\" argc=%d\n",line,argc) ;
	// how many characters to match
  	int l = strlen(argv[0]) ;
	if (l<3) l=3 ;

	// start parsing non-blank lines
	// unadvertised special
	if (argc==1 && (!(strncasecmp(argv[0],"end",l)))) { // end
	  if (debug)
	    printf("pfdLoadFile_savg debug: command: end\n") ;
	  processPrimitive() ;
	  
	} else if (parsePreserveplanarquads(argc, argv, l)) {
	  
	} else if (parsePlanarpolygonchecking(argc, argv, l)) {

	} else if (parseVerbosity(argc, argv, l)) {
	  
	} else if (parseIndexed(argc, argv, l)) {
	  
	} else if (parseTexture(argc, argv, l)) {
	  
	} else if (parseText(argc, argv, l)) {

	} else if (parseTransparency(argc, argv, l)) {

	} else if (parseLighting(argc, argv, l)) {

	} else if (parseColormode(argc, argv, l)) {

	} else if (parseAlpha(argc, argv, l)) {

	} else if (parseShininess(argc, argv, l)) {

	} else if (parseStyle(argc, argv, l)) {

	} else if (parseStriplength(argc, argv, l)) {

	} else if (parsePixelsize(argc, argv, l)) {

	} else if (parseShrinkage(argc, argv, l)) {

	} else if (parseOctree(argc, argv, l)) {

	} else if (parseColorMode(argc, argv, l)) {

	} else if (parsePrimitive(argc, argv, l)) {

	} else if (parseTexturedPrimitive(argc, argv, l)) {

	} else if (parseData(argc, argv, l)) {

	} else if (parse2dTexturedData(argc, argv, l)) {

	} else if (parse3dTexturedData(argc, argv, l)) {
	  
	} else {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid line", lineNum);
	} // if (argc==1 && (!(strncasecmp(argv[0],"end",l)))) {
      } // if (normalize(line)) {
    } // if (notifyLevel >= PFNFY_INFO) {
  } //while (1)
  
  if (!line)
    free(line) ;
  
  fclose(fp);
  
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_savg: finished reading %s", 
	   filename);
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_savg: file contains %d lines, %d primitives", 
	   lineNum-1, primNum);
  
  if (coords) {
    free(coords) ;
    coords = NULL ;
  }

  if (texCoords) {
    free(texCoords) ;
    texCoords = NULL ;
  }

  if (colors) {
    free(colors) ;
    colors = NULL ;
  }

  if (norms) {
    free(norms) ;
    norms = NULL ;
  }

  // enable indexing if requested
  if (indexed)
    pfdBldrMode(PFDBLDR_MESH_INDEXED, 1) ;
  else
    pfdBldrMode(PFDBLDR_MESH_INDEXED, 0) ;

  // set strip length in mesher too in case retessellation is turned off
  pfdBldrMode(PFDBLDR_MESH_MAX_TRIS, stripLength) ;

  if (retessellate) 
    pfdBldrMode(PFDBLDR_MESH_RETESSELLATE, 1) ;
  else
    pfdBldrMode(PFDBLDR_MESH_RETESSELLATE, 0) ;

  if (primNum>0) {
    // get a complete scene graph representing file's primitives
    pfNode *node = pfdBuild();
    
    if (node) {
      breakupNodes(&node, NULL) ;
      if (!node) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: pfdBreakup returned NULL- "
		 "check your file for unusual octree and striplength values.");
      }
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: pfdBuilder returned NULL- "
	       "check your file for vertex geometry errors.");
    }
    if (!group)
      return(node) ;
    else
      group->addChild(node) ;
  }
  return(group) ;
}

////////////////////////////////////////////////////////////////////////
void processPrimitive(void) {
  if (n>0) {
    if (debug)
      printPrimitive() ;
    
    primNum++ ;
    doMaterial() ;
    if (primitive==TRI) // turn tristrip into a bunch of triangles
      doTri() ;
    else if (primitive==POLYN)
      doPolyN() ;
    else if ((primitive==POLY) || (primitive==POINT) || (primitive==LINE))
      doPolyOrLineOrPoint() ;
    else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: processPrimitive, line %d, "
	       "BUG! unexpected or unassigned primitive type", lineNum);
    }

    // get ready for next one
    resetState() ;
  } // if (n>0)
}

////////////////////////////////////////////////////////////////////////
// rearrange the data from a tristrip into a bunch of triangles
// why can't I pass tristrips to the builder???
void doTri(void) {

  int i ; //again!

  // n-2 triangles from a tristrip with n points

  pfdGeom *prim = pfdNewGeom(3) ;
  prim->pixelsize = pixelsize ;
  prim->primtype = PFGS_POLYS ;
  
  // turn off normals if not supplied per vertex
  if (normals)
    prim->nbind = PFGS_PER_VERTEX;
  else
    prim->nbind = PFGS_OFF;
  
  // assign color(s)
  if (passColors) {
    if (color==VERTEX)
      prim->cbind = PFGS_PER_VERTEX;
    else {
      prim->cbind = PFGS_PER_PRIM;
      (prim->colors)[0] = primColor ;
    }
  } else {
    prim->cbind = PFGS_OFF;
    if (debug)
      printf("pfdLoadFile_savg debug: not passing colors\n") ;
  }
  
  // turn on and set first texture if data given
  if (textureType==0)
    prim->tbind[0] = PFGS_OFF;
  else if (textureType==2)
    prim->tbind[0] = PFGS_PER_VERTEX;
  else {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: line %d, "
	     "3DTEXTURE not supported yet", lineNum);
  }
  // turn off textures 1-PF_MAX_TEXTURES, no matter what
  for (i=1; i<PF_MAX_TEXTURES; i++)
    prim->tbind[i] = PFGS_OFF;

  // spew out as separate triangle polygons- need to watch orientation
  for (i=0; i<(n-2); i++) {
    if (i%2) { // flip order
      // copy vertices
      (prim->coords)[0] = coords[i] ;
      (prim->coords)[1] = coords[i+2] ;
      (prim->coords)[2] = coords[i+1] ;
      if (shrinkage != 1.f) 
	shrinkVerts(3, prim->coords, shrinkage) ;

      // copy texture vertices, if any
      if (textureType==2) {
	(prim->texCoords[0])[i] = texCoords[i] ;
	(prim->texCoords[1])[i] = texCoords[i+2] ;
	(prim->texCoords[2])[i] = texCoords[i+1] ;
      }

      // copy normals, if any
      if (normals) {
	(prim->norms)[0] = norms[i] ;
	(prim->norms)[1] = norms[i+2] ;
	(prim->norms)[2] = norms[i+1] ;
      }
      // copy colors
      if (passColors && color==VERTEX) {
	(prim->colors)[0] = colors[i] ;
	(prim->colors)[1] = colors[i+2] ;
	(prim->colors)[2] = colors[i+1] ;
      }
    } else {
      // copy vertices
      (prim->coords)[0] = coords[i] ;
      (prim->coords)[1] = coords[i+1] ;
      (prim->coords)[2] = coords[i+2] ;
      if (shrinkage != 1.f) 
	shrinkVerts(3, prim->coords, shrinkage) ;

      // copy texture vertices, if any
      if (textureType==2) {
	(prim->texCoords[0])[i] = texCoords[i] ;
	(prim->texCoords[1])[i] = texCoords[i+1] ;
	(prim->texCoords[2])[i] = texCoords[i+2] ;
      }

      // copy normals, if any
      if (normals) {
	(prim->norms)[0] = norms[i] ;
	(prim->norms)[1] = norms[i+1] ;
	(prim->norms)[2] = norms[i+2] ;
      }
      // copy colors
      if (passColors && color==VERTEX) {
	(prim->colors)[0] = colors[i] ;
	(prim->colors)[1] = colors[i+1] ;
	(prim->colors)[2] = colors[i+2] ;
      }
    } // if (i%2)

    // print the pfdGeom
    //printGeom(prim) ;

    // add triangle to builder 
    pfdAddBldrGeom(prim, 1);

  } //for (i=0; i<(n-2); i++)

  // release primitive buffer
  pfdDelGeom(prim);
  
}

////////////////////////////////////////////////////////////////////////
// add a polygon, linestrip or point to the builder.  For polygon, the
// number of sides equals the number of vertices
void doPolyOrLineOrPoint(void) { 

  //printf("doPolyOrLineOrPoint\n") ;

  // total number of vertices
  pfdGeom *prim = pfdNewGeom(n) ;
  prim->pixelsize = pixelsize ;

  if (primitive==POLY)
    prim->primtype = PFGS_POLYS ;
  else if (primitive==LINE)
    prim->primtype = PFGS_LINESTRIPS ;
  else if (primitive==POINT)
    prim->primtype = PFGS_POINTS ;
  else {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: doPolyOrLineOrPoint, line %d, "
	     "BUG! unexpected or unassigned primitive type", lineNum);
  }

  int i ; //again!

  if (primitive==POLY && planarpolygoncheck && n>3) {
    plane->makePts(coords[0], coords[1], coords[2]) ;
    for (i=3; i<n; i++) {
      if (!plane->contains(coords[i])) {
	pfNotify(PFNFY_NOTICE, PFNFY_PRINT, 
		 "pfdLoadFile_savg: non-planar vertex, line %d = (%f, %f, %f)\n",
	       lineNum, (coords[i])[PF_X], (coords[i])[PF_Y], (coords[i])[PF_Z]) ;
      }
    }
  }

  // turn off normals if not supplied per vertex
  if (normals)
    prim->nbind = PFGS_PER_VERTEX;
  else
    prim->nbind = PFGS_OFF;

  // assign color(s)
  if (passColors) {
    if (color==VERTEX)
      prim->cbind = PFGS_PER_VERTEX;
    else {
      prim->cbind = PFGS_PER_PRIM;
      prim->colors[0] = primColor ;
    } 
  } else {
    prim->cbind = PFGS_OFF ;
    if (debug)
      printf("pfdLoadFile_savg debug: not passing colors\n") ;
  }

  // turn on and set first texture if data given
  if (textureType==0)
    prim->tbind[0] = PFGS_OFF;
  else if (textureType==2)
    prim->tbind[0] = PFGS_PER_VERTEX;
  else {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: line %d, "
	     "3DTEXTURE not supported yet", lineNum);
  }
  // turn off textures 1-PF_MAX_TEXTURES, no matter what
  for (i=1; i<PF_MAX_TEXTURES; i++)
    prim->tbind[i] = PFGS_OFF;

  // copy vertices
  for(i=0; i<n; i++)
    prim->coords[i] = coords[i] ;

  if (shrinkage!=1.f) 
    shrinkVerts(n, prim->coords, shrinkage) ;

  // copy texture vertices, if any
  if (textureType==2) {
    for(i=0; i<n; i++) 
      (prim->texCoords[0])[i] = texCoords[i] ;
  }

  // copy normals, if any
  if (normals) {
    for(i=0; i<n; i++)
      prim->norms[i] = norms[i] ;
  }
  
  // copy colors, if any
  if (passColors && color==VERTEX) {
    for(i=0; i<n; i++)
      prim->colors[i] = colors[i] ;
  }

  // print the pfdGeom
  //printGeom(prim) ;

  // add primitive to builder 
  pfdAddBldrGeom(prim, 1);
  
  // release primitive buffer
  pfdDelGeom(prim);
  
}

////////////////////////////////////////////////////////////////////////
// create (n/sides) polygons of n sides each
void doPolyN(void) { 

  //printf("doPoly\n") ;

  if (n%sides != 0) {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: line %d, "
	     "Number of vertices, %d, is not divisible by number of sides, %d",
	     lineNum, n, sides);
  }

  // (n/sides) polygons of n sides each
  pfdGeom *prim = pfdNewGeom(sides) ;
  prim->pixelsize = pixelsize ;
  
  prim->primtype = PFGS_POLYS ;
  
  // turn on and set first texture if data given
  if (textureType==0)
    prim->tbind[0] = PFGS_OFF;
  else if (textureType==2)
    prim->tbind[0] = PFGS_PER_VERTEX;
  else {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: line %d, "
	     "3DTEXTURE not supported yet", lineNum);
  }

  int i ; //again!

  // turn off textures 1-PF_MAX_TEXTURES, no matter what
  for (i=1; i<PF_MAX_TEXTURES; i++)
    prim->tbind[i] = PFGS_OFF;

  // turn off normals if not supplied per vertex
  if (normals)
    prim->nbind = PFGS_PER_VERTEX;
  else
    prim->nbind = PFGS_OFF;
  
  // assign color(s)
  if (passColors) {
    if (color==VERTEX)
      prim->cbind = PFGS_PER_VERTEX;
    else {
      prim->cbind = PFGS_PER_PRIM;
      prim->colors[0] = primColor ;
    } 
  } else {
    prim->cbind = PFGS_OFF ;
    if (debug) 
      printf("pfdLoadFile_savg debug: not passing colors\n") ;
  }

  // spew out as separate polygon primitives
  int j = 0 ;
  for (i=0; i<n; i+=sides) {
    for (j=0; j<sides; j++) {
      
      // copy vertices
      prim->coords[j] = coords[i+j] ;

      // copy texture vertices, if any
      if (textureType==2) 
	  (prim->texCoords[0])[j] = texCoords[i+j] ;

      // copy normals, if any
      if (normals)
	prim->norms[j] = norms[i+j] ;

      // copy colors
      if (passColors && color==VERTEX)
	prim->colors[j] = colors[i+j] ;

    } // for (j=0; j<sides; j++)

    if (shrinkage!=1.f) 
      shrinkVerts(sides, prim->coords, shrinkage) ;

    // add triangles to builder 
    pfdAddBldrGeom(prim, 1);

  } // for (i=0; i<n; i+=sides)
  
  // release primitive buffer
  pfdDelGeom(prim);
  
}

////////////////////////////////////////////////////////////////////////
void doMaterial(void) {

#if 0
  static bool first = true ;
  if (first) {
    first = false ;
    pfdBldrMode(PFDBLDR_AUTO_NORMALS, 1) ;
    printf("pfdGetBldrMode(PFDBLDR_AUTO_NORMALS) = %d\n",
	   pfdGetBldrMode(PFDBLDR_AUTO_NORMALS)) ;
  }
#endif

  pfdBldrStateAttr(PFSTATE_FRONTMTL, mat);	  
  pfdBldrStateAttr(PFSTATE_BACKMTL, mat);	  

  if (lightingState == AUTO) {
    // turn off lighting for 1d or 0d objects, on otherwise
    if (primitive == POINT || primitive == LINE || style == WIREFRAME) 
      pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_OFF);
    else
      pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_ON);
  } else if (lightingState == DEFAULT) {
    pfdBldrStateMode(PFSTATE_ENLIGHTING, defaults.lighting);
  } else if (lightingState == ON) {
    pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_ON);
  } else if (lightingState == OFF) {
    pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_OFF);
  } else {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: logic error in doMaterial, lighting property unknown");
  }

  if (textureFile == 2) {
    pfdBldrStateAttr(PFSTATE_TEXTURE, texture) ;
    pfdBldrStateMode(PFSTATE_ENTEXTURE, PF_ON) ;
    pfdBldrStateAttr(PFSTATE_TEXENV, textureEnv);
  } else {
    pfdBldrStateAttr(PFSTATE_TEXTURE, NULL) ;
    pfdBldrStateMode(PFSTATE_ENTEXTURE, PF_OFF) ;
  }

  if (style==WIREFRAME)
    pfdBldrStateMode(PFSTATE_ENWIREFRAME, PFTR_ON) ;
  else
    pfdBldrStateMode(PFSTATE_ENWIREFRAME, PFTR_OFF) ;

  if (transparencyState == AUTO) {
    if (transparency)
      pfdBldrStateMode(PFSTATE_TRANSPARENCY, 
		       PFTR_HIGH_QUALITY | PFTR_NO_OCCLUDE);
    else
      pfdBldrStateMode(PFSTATE_TRANSPARENCY, PFTR_OFF);
  } else if (transparencyState == DEFAULT) {
    pfdBldrStateMode(PFSTATE_TRANSPARENCY, defaults.transparency);
  } else if (transparencyState == ON) {
    pfdBldrStateMode(PFSTATE_TRANSPARENCY, 
		     PFTR_HIGH_QUALITY | PFTR_NO_OCCLUDE);
  } else if (transparencyState == OFF) {
    pfdBldrStateMode(PFSTATE_TRANSPARENCY, PFTR_OFF);
  } else {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_savg: logic error in doMaterial, transparency property unknown");
  }

  if (debug)
    printMaterial() ;
}

////////////////////////////////////////////////////////////////////////
void printPrimitive(void) {
  if (n>0) {
    printf("pfdLoadFile_savg debug: primitive %d = %s\n",primNum,primitiveName[primitive]) ;
    printf("  %d vertices with: vertices, ",n) ;
    if (color==PRIM) 
      printf("primitive color (%4.2f %4.2f %4.2f %4.2f)",
	     primColor[0], primColor[1], 
	     primColor[2], primColor[3]) ;
    else
      printf("per-vertex colors ") ;
    if (transparency) printf("(transparency) ") ;
    if (normals) printf("normals ") ;
    printf("\n  ") ;
    int i ;
    for (i=0; i<n; i++) {
      printf("%4.4d: ", i+1) ;
      if (sides>0)
	printf("(%4.4d) ",(n%sides)+1) ;
      printf("%+5.3f %+5.3f %+5.3f    ", coords[i][0], coords[i][1], coords[i][2]) ;
      if (color==VERTEX)
	printf("%4.2f %4.2f %4.2f %4.2f   ",
	       colors[i][0], colors[i][1], colors[i][2], colors[i][3]) ;
      if (normals)
	printf("%+5.3f %+5.3f %+5.3f", norms[i][0], norms[i][1], norms[i][2]) ;
      printf("\n  ") ;
    }
    printf("nVerts = %d, nColors = %d, nNorms = %d\n",
	   nCoords, nColors, nNorms) ;
  } // if (n>0)
}

////////////////////////////////////////////////////////////////////////
void printMaterial(void) {

  if (n>0) {
    printf("pfdLoadFile_savg debug: materials and states for above primitive:\n") ;
    int val ;

    val =  pfdGetBldrStateMode(PFSTATE_ENLIGHTING);
    printf("  lighting = %d, ", val) ; 
    if (val == PF_OFF) {
      printf("off\n") ;
    } else if (val == PF_ON) {
      printf("on\n") ;
    } else {
      printf("UNKNOWN VALUE!\n") ;
    }

    val =  pfdGetBldrStateMode(PFSTATE_ENWIREFRAME);
    printf("  wireframe = %d, ", val) ; 
    if (val == PF_OFF) {
      printf("off\n") ;
    } else if (val == PF_ON) {
      printf("on\n") ;
    } else {
      printf("UNKNOWN VALUE!\n") ;
    }

    val =  pfdGetBldrStateMode(PFSTATE_TRANSPARENCY);
    printf("  transparency = 0x%x, ", val) ; 
    if (val == PFTR_OFF)
      printf("off (0x%x) ", PFTR_OFF) ;
    else if (val & PFTR_ON) {
      printf("on (0x%x) ", PFTR_ON) ;
      if ((val^PFTR_NO_OCCLUDE) == PFTR_FAST)
	printf("fast (0x%x) ", PFTR_FAST) ;
      else if ((val^PFTR_NO_OCCLUDE) == PFTR_HIGH_QUALITY)
	printf("high_quality (0x%x) ", PFTR_HIGH_QUALITY) ;
      else if ((val^PFTR_NO_OCCLUDE) == PFTR_BLEND_ALPHA)
	printf("blend_alpha (0x%x) ", PFTR_BLEND_ALPHA) ;
      else if ((val^PFTR_NO_OCCLUDE) == PFTR_MS_ALPHA)
	printf("ms_alpha (0x%x) ", PFTR_MS_ALPHA) ;
      else if ((val^PFTR_NO_OCCLUDE) == PFTR_MS_ALPHA_MASK)
	printf("ms_alpha_mask (0x%x) ", PFTR_MS_ALPHA_MASK) ;
      if (val & PFTR_NO_OCCLUDE)
	printf("no_occlude (0x%x) ", PFTR_NO_OCCLUDE) ;
    }
    printf("\n") ;
    
    pfMaterial *bldrMat = (pfMaterial*) pfdGetBldrStateAttr(PFSTATE_FRONTMTL);	  
    
    if (!bldrMat)
      printf("  NULL material returned by builder!!\n") ;
    else {
      printf("  shininess = %f\n", bldrMat->getShininess()) ;
      printf("  alpha = %f\n", bldrMat->getAlpha()) ;
      
      int colormode = bldrMat->getColorMode(PFMTL_FRONT) ;
      printf("  colormode = ") ;
      if (colormode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
	printf("PFMTL_CMODE_AMBIENT_AND_DIFFUSE\n") ; 
      else if (colormode == PFMTL_CMODE_AMBIENT)
	printf("PFMTL_CMODE_AMBIENT\n") ; 
      else if (colormode == PFMTL_CMODE_DIFFUSE)
	printf("PFMTL_CMODE_DIFFUSE\n") ; 
      else if (colormode == PFMTL_CMODE_EMISSION)
	printf("PFMTL_CMODE_EMISSION\n") ; 
      else if (colormode == PFMTL_CMODE_SPECULAR)
	printf("PFMTL_CMODE_SPECULAR\n") ; 
      else if (colormode == PFMTL_CMODE_OFF)
	printf("PFMTL_CMODE_OFF\n") ; 
      else if (colormode == PFMTL_CMODE_COLOR)
	printf("PFMTL_CMODE_COLOR\n") ; 
      else 
	printf("unknown!!\n") ; 
      
      float c[3];
      bldrMat->getColor(PFMTL_CMODE_AMBIENT, c, c+1, c+2) ;
      printf("  ambient = %f %f %f\n",c[0],c[1],c[2]) ;
      bldrMat->getColor(PFMTL_CMODE_DIFFUSE, c, c+1, c+2) ;
      printf("  diffuse = %f %f %f\n",c[0],c[1],c[2]) ;
      bldrMat->getColor(PFMTL_CMODE_EMISSION, c, c+1, c+2) ;
      printf("  emission = %f %f %f\n",c[0],c[1],c[2]) ;
      bldrMat->getColor(PFMTL_CMODE_SPECULAR, c, c+1, c+2) ;
      printf("  specular = %f %f %f\n",c[0],c[1],c[2]) ;
    }
  }
}
////////////////////////////////////////////////////////////////////////
void resetState(void) {
  primitive=END ;
  sides = 0 ;
  color = PRIM ;
  if (primColor[3]<OPAQUEALPHA)
    transparency = TRUE ;
  else
    transparency = FALSE ;
  normals = FALSE ;
  n = 0 ;
  textureType=0 ;
}

////////////////////////////////////////////////////////////////////////
static void addCoord(pfVec3 v) {
  if (n>nCoords) {
    nCoords+=BUMP ;
    coords = (pfVec3*) realloc(coords,sizeof(pfVec3)*nCoords) ;
  }
  coords[n-1] = v ;
}

////////////////////////////////////////////////////////////////////////
static void addTex2d(pfVec2 t) {
  if (n>nTexCoords) {
    nTexCoords+=BUMP ;
    texCoords = (pfVec2*) realloc(texCoords,sizeof(pfVec2)*nTexCoords) ;
  }
  texCoords[n-1] = t ;
}

////////////////////////////////////////////////////////////////////////
static void addTex3d(pfVec3 t) {
  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	   "pfdLoadFile_savg: addTex3d called");
}

////////////////////////////////////////////////////////////////////////
static void addNorm(pfVec3 no) {
  normals = true ;
  if (n>nNorms) {
    nNorms+=BUMP ;
    norms = (pfVec3*) realloc(norms,sizeof(pfVec3)*nNorms) ;
  }
  norms[n-1] = no ;
}

////////////////////////////////////////////////////////////////////////
static void addColor(pfVec4 c) {
  color = VERTEX ;
  if (n>nColors) {
    nColors+=BUMP ;
    colors = (pfVec4*) realloc(colors,sizeof(pfVec4)*nColors) ;
  }
  colors[n-1] = c ;
  if (c[3]<OPAQUEALPHA)
    transparency = true ;
  primColor = c ;
}

////////////////////////////////////////////////////////////////////////
bool parsePreserveplanarquads(int argc, char **argv, int l) {
  if ((argc==1) && (!(strncasecmp(argv[0],"preserveplanarquads",l)))) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: preserveplanarquads\n") ;
    processPrimitive() ;
    retessellate = false ;
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
bool parsePlanarpolygonchecking(int argc, char **argv, int l) {
  if ((argc==1) && (!(strncasecmp(argv[0],"preserveplanarquads",l)))) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: planarpolygonchecking\n") ;
    processPrimitive() ;
    planarpolygoncheck = true ;
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
bool parseVerbosity(int argc, char **argv, int l) {
  if ((argc==2) && (!(strncasecmp(argv[0],"verbosity",l)))) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: verbosity\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    if (!(strncasecmp(argv[1],"on",l2))) {
      printf("pfdLoadFile_savg debug: verbosity on\n") ;
      verbose = true; debug = false ;
    } else if (!(strncasecmp(argv[1],"off",l2))) {
      printf("pfdLoadFile_savg debug: verbosity off\n") ;
      verbose = false; debug = false ;
    } else if (!(strncasecmp(argv[1],"debug",l2))) {
      printf("pfdLoadFile_savg debug: verbosity debug\n") ;
      verbose = true; debug = true ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid parameter", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
bool parseIndexed(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"indexed",l))) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: indexed\n") ;
    processPrimitive() ;
    indexed = true ;
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
bool parseTexture(int argc, char **argv, int l) {
  if ((argc==2) && (!(strncasecmp(argv[0],"texture",l)))) {
    processPrimitive() ;
    textureFile = 2 ;
    
    texture = new pfTexture ;
    textureEnv = new pfTexEnv;
    textureEnv->setMode(PFTE_DECAL);
    textureEnv->setBlendColor(0.0f, 0.0f, 0.0f, 1.f);
    if (texture->loadFile(argv[1])) {
      if (debug)
	printf("pfdLoadFile_savg debug: command: texture %s\n", argv[1]) ;
      // see if the image has an alpha channel
      int nc ;
      texture->getImage(NULL, &nc, NULL, NULL, NULL);
      if (nc != 3)
	transparency = true ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "could not load texture file \"%s\"", lineNum, argv[1]);
      
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
bool parseText(int argc, char **argv, int l) {
  if ((argc>2) && (!(strncasecmp(argv[0],"text",l)))) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: text\n") ;
    processPrimitive() ;
    // copy the defaults into a new text structure
    int largc = 1 ;
    // stores the compound transformation
    pfMatrix m ;
    m.makeIdent() ;
    // stores xyz, sx, sy, sz, hpr before being added to matrix
    float x, y, z, sx, sy, sz, h, p, r ;

    pfString *string = NULL ;
    while (largc<argc) {
      int l = strlen(argv[largc]) ;
      if (l<2) l=2 ;
      
      if (!(strncasecmp(argv[largc],"string",l)) && 
	  (largc+1<argc)) {   // at least one more argument
	
	// only send out a text string if you hit a STRING keyword
	string = new pfString ;
	string->setString(argv[largc+1]) ;
	//~printf("    string \"%s\"\n",argv[largc+1]) ;
	largc+=2 ;
	
      } else if (!(strncasecmp(argv[largc],"font",l)) && 
		 (largc+1<argc)) {   // at least one more argument
	strcpy(textMode.fontName,argv[largc+1]) ;
	//~printf("    font \"%s\"\n",argv[largc+1]) ;
	largc+=2 ;
	
      } else if (!(strncasecmp(argv[largc],"flatten",l)) && 
		 (largc+1<argc)) {   // at least one more argument
	int l2 = strlen(argv[largc+1]) ;
	if (l<2) l2=2 ;
	if (!(strncasecmp(argv[largc+1],"on",l2))) {
	  textMode.flatten = true ;
	} else if (!(strncasecmp(argv[largc+1],"off",l2))) {
	  textMode.flatten = false ;
	} else {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid TEXT FLATTEN value \"%s\"", lineNum, argv[largc+1]);
	}
	//~printf("    flatten \"%s\"\n",argv[largc+1]) ;
	largc+=2 ;
	
      } else if (!(strncasecmp(argv[largc],"justification",l)) && 
		 (largc+1<argc)) {   // at least one more argument
	int l2 = strlen(argv[largc+1]) ;
	if (l<1) l2=1 ;
	if (!(strncasecmp(argv[largc+1],"left",l2))) {
	  textMode.justification = PFSTR_LEFT ;
	} else if (!(strncasecmp(argv[largc+1],"right",l2))) {
	  textMode.justification = PFSTR_RIGHT ;
	} else if (!(strncasecmp(argv[largc+1],"center",l2))) {
	  textMode.justification = PFSTR_CENTER ;
	} else {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid TEXT JUSTIFICATION value \"%s\"", lineNum, argv[largc+1]);
	}
	//~printf("    justification \"%s\"\n",argv[largc+1]) ;
	largc+=2 ;
	
      } else if (!(strncasecmp(argv[largc],"style",l)) && 
		 (largc+1<argc)) {   // at least one more argument
	int l2 = strlen(argv[largc+1]) ;
	if (l<1) l2=1 ;
	if (!(strncasecmp(argv[largc+1],"textured",l2))) {
	  textMode.style = PFDFONT_TEXTURED ;
	} else if (!(strncasecmp(argv[largc+1],"outlined",l2))) {
	  textMode.style = PFDFONT_OUTLINED ;
	} else if (!(strncasecmp(argv[largc+1],"filled",l2))) {
	  textMode.style = PFDFONT_FILLED ;
	} else if (!(strncasecmp(argv[largc+1],"extruded",l2))) {
	  textMode.style = PFDFONT_EXTRUDED ;
	} else {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid TEXT STYLE value \"%s\"", lineNum, argv[largc+1]);
	}
	//~printf("    justification \"%s\"\n",argv[largc+1]) ;
	largc+=2 ;
	
      } else if (!(strncasecmp(argv[largc],"xyz",l)) &&  
		 (largc+3<argc) &&  // at least three more arguments
		 parseNumbers(&(argv[largc+1]), 
			      &x, &y, &z)) {
	printf("    xyz = %f %f %f\n",x, y, z) ;
	//m.preTrans(x, y, z, m) ;
	m.postTrans(m, x, y, z) ;
	largc+=4 ;
	
      } else if (!(strncasecmp(argv[largc],"sxyz",l)) &&  
		 (largc+3<argc) &&  // at least three more arguments
		 parseNumbers(&(argv[largc+1]), &sx, &sy, &sz)) {
	printf("    sxyz = %f %f %f\n",sx, sy, sz) ;
	//m.preScale(sx, sy, sz, m) ;
	m.postScale(m, sx, sy, sz) ;
	largc+=4 ;
	
      } else if (!(strncasecmp(argv[largc],"s",l)) &&  
		 (largc+1<argc) &&  // at least one more arguments
		 parseNumbers(&(argv[largc+1]), &sx)) {
	printf("    s = %f\n",sx) ;
	//m.preScale(sx, sx, sx, m) ;
	m.postScale(m, sx, sx, sx) ;
	largc+=2 ;
	
      } else if (!(strncasecmp(argv[largc],"hpr",l)) &&  
		 (largc+3<argc) &&  // at least three more arguments
		 parseNumbers(&(argv[largc+1]), &h, &p, &r)) {
	printf("    hpr = %f %f %f\n",h, p, r) ;
	//m.preRot(h, 0.f, 0.f, 1.f, m) ;
	//m.preRot(p, 1.f, 0.f, 0.f, m) ;
	//m.preRot(r, 0.f, 1.f, 0.f, m) ;
	m.postRot(m, h, 0.f, 0.f, 1.f) ;
	m.postRot(m, p, 1.f, 0.f, 0.f) ;
	m.postRot(m, r, 0.f, 1.f, 0.f) ;
	largc+=4 ;
	
      } else if (!(strncasecmp(argv[largc],"rgba",l)) &&  
		 (largc+4<argc) &&  // at least four more arguments
		 parseNumbers(&(argv[largc+1]), &(primColor[0]), &(primColor[1]), 
			     &(primColor[2]), &(primColor[3]))) {
	//~printf("    rgba = %f %f %f %f\n",primColor[0], primColor[1], primColor[2], primColor[3]) ;
	if (primColor[3]<OPAQUEALPHA)
	  transparency = true ;
	largc+=5 ;
	
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid text keyword \"%s\"", lineNum, argv[largc]);
      }
      
    } // while (largc<argc) {
    // send out text if string keyword encountered
    if (string) {
      if (!group)
	group = new pfGroup ;
      textMode.font = pfdLoadFont_type1(textMode.fontName, textMode.style);
      if (!textMode.font) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "can't load font \"%s\" style %d ", 
		 lineNum, textMode.fontName, textMode.style);
      }
      string->setFont(textMode.font) ;
      if (textMode.flatten)
	string->flatten() ;
      string->setMode(PFSTR_JUSTIFY, textMode.justification) ;
      string->setColor(primColor[0], primColor[1], primColor[2], 
		       primColor[3]) ;

      // set alpha and material
      pfGeoState *gState = new pfGeoState ;
      gState->setAttr(PFSTATE_FRONTMTL, mat);	  
      gState->setAttr(PFSTATE_BACKMTL, mat);	  
      if (transparency) {
	gState->setMode(PFSTATE_TRANSPARENCY, 
			PFTR_HIGH_QUALITY | PFTR_NO_OCCLUDE);
      }
      // need to do this to get transparency in strings
      textMode.font->setAttr(PFFONT_GSTATE, gState) ;
      string->setGState(gState) ;
      string->setMat(m) ;

      pfText *text = new pfText ;
      text->addString(string);
      group->addChild(text) ;
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseTransparency(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"transparency",l)) &&
      (argc==2)) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: transparency\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    if (!(strncasecmp(argv[1],"on",l2)))
      transparencyState = ON ;
    else if (!(strncasecmp(argv[1],"off",l2)))
      transparencyState = OFF ;
    else if (!(strncasecmp(argv[1],"default",l2)))
      transparencyState = DEFAULT ;
    else if (!(strncasecmp(argv[1],"auto",l2)))
      transparencyState = AUTO ;
    else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid parameter", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseLighting(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"lighting",l)) &&
      (argc==2)) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: lighting\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    if (!(strncasecmp(argv[1],"on",l2)))
      lightingState = ON ;
    else if (!(strncasecmp(argv[1],"off",l2)))
      lightingState = OFF ;
    else if (!(strncasecmp(argv[1],"default",l2)))
      lightingState = DEFAULT ;
    else if (!(strncasecmp(argv[1],"auto",l2)))
      lightingState = AUTO ;
    else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid parameter", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseColormode(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"colormode",l)) &&
      (argc==2)) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: colormode\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    if (!(strncasecmp(argv[1],"ambient",l2))) {
      mat->setColorMode(PFMTL_BOTH, PFMTL_CMODE_AMBIENT) ;
      passColors = true ;
    } else if (!(strncasecmp(argv[1],"diffuse",l2))) {
      mat->setColorMode(PFMTL_BOTH, PFMTL_CMODE_DIFFUSE) ;
      passColors = true ;
    } else if (!(strncasecmp(argv[1],"emission",l2))) {
      mat->setColorMode(PFMTL_BOTH, PFMTL_CMODE_EMISSION) ;
      passColors = true ;
    } else if (!(strncasecmp(argv[1],"specular",l2))) {
      mat->setColorMode(PFMTL_BOTH, PFMTL_CMODE_SPECULAR) ;
      passColors = true ;
    } else if (!(strncasecmp(argv[1],"ambient_and_diffuse",l2))) {
      mat->setColorMode(PFMTL_BOTH, PFMTL_CMODE_AMBIENT_AND_DIFFUSE) ;
      passColors = true ;
    } else if (!(strncasecmp(argv[1],"off",l2))) {
      mat->setColorMode(PFMTL_BOTH, PFMTL_CMODE_OFF) ;
      passColors = false ;
    } else if (!(strncasecmp(argv[1],"default",l2))) {
      mat->setColorMode(PFMTL_BOTH, defaults.colormode) ;
      passColors = true ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid parameter", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseAlpha(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"alpha",l)) &&
      (argc==2)) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: alpha\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    float alpha ;
    if (!(strncasecmp(argv[1],"default",l2))) {
      mat->setAlpha(defaults.alpha) ;
    } else if (parseNumbers(&(argv[1]), &alpha)) {
      if (alpha<0.f|| alpha>128.f) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number range", lineNum);
      } else
	mat->setAlpha(alpha) ; 
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseShininess(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"shininess",l)) &&
      (argc==2)) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: shininess\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    float shininess ;
    if (!(strncasecmp(argv[1],"default",l2))) {
      mat->setShininess(defaults.shininess) ;
    } else if (parseNumbers(&(argv[1]), &shininess)) {
      if (shininess<0.f|| shininess>128.f) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number range", lineNum);
      } else
	mat->setShininess(shininess) ; 
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseStyle(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"style",l)) &&
      (argc==2)) {
    if (debug)
      printf("pfdLoadFile_savg debug: command: style\n") ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    if (l2<3) l2=3 ;
    if (!(strncasecmp(argv[1],"filled",l2)))
      style=FILLED ;
    else if (!(strncasecmp(argv[1],"wireframe",l2)))
      style=WIREFRAME ;
    else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid parameter", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseStriplength(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"striplength",l)) &&
      (argc==2)) { // striplength n
    if (debug)
      printf("pfdLoadFile_savg debug: command: striplength\n") ;
    processPrimitive() ;
    if (parseNumbers(&(argv[1]), &stripLength)) {
      if (stripLength<0) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "stripLength must be non-negative", lineNum);
      }
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parsePixelsize(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"pixelsize",l)) &&
      (argc==2)) { // pixelsize n
    if (debug)
      printf("pfdLoadFile_savg debug: command: pixelsize\n") ;
    processPrimitive() ;
    if (parseNumbers(&(argv[1]), &pixelsize)) {
      if (pixelsize<0) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "pixelsize must be non-negative", lineNum);
      }
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseShrinkage(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"shrinkage",l)) &&
      (argc==2)) { // shrinkage n
    if (debug)
      printf("pfdLoadFile_savg debug: command: shrinkage\n") ;
    processPrimitive() ;
    if (parseNumbers(&(argv[1]), &shrinkage)) {
      if (shrinkage==0.f) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "shrinkage must be non-negative", lineNum);
      }
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseOctree(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"octree",l)) &&
      (argc==3)) { // octree s c
    if (debug)
      printf("pfdLoadFile_savg debug: command: octree\n") ;
    processPrimitive() ;
    if (parseNumbers(&(argv[1]), &geodeSize) &&
	(parseNumbers(&(argv[2]), &geodeChild))) {
      if (geodeSize<0.f) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "octree geodeSize must be non-negative", lineNum);
      }
      if (geodeChild<1) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "octree geodeChild must be positive", lineNum);
      }
      if (geodeSize<1.f && geodeChild==1) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "if size is less than one, child must be greater than one",
		 lineNum);
      }
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parsePrimitive(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"tristrips",l)) ||
      !(strncasecmp(argv[0],"lines",l)) ||
      !(strncasecmp(argv[0],"polygons",l)) ||
      !(strncasecmp(argv[0],"points",l))) { 
    // return false if texture specified
    if (argc>1) {
      int l2 = strlen(argv[1]) ;
      if (!(strncasecmp(argv[1],"3dtexture",l2)) ||
	  !(strncasecmp(argv[1],"2dtexture",l2)))
	return false ;
    }
    if (debug)
      printf("pfdLoadFile_savg debug: command: primitive = %s\n",argv[0]) ;
    processPrimitive() ;
    if (!(strncasecmp(argv[0],"tristrips",l))) {
      primitive = TRI ;
    } else if (!(strncasecmp(argv[0],"lines",l))) {
      primitive = LINE ;
    } else if (!(strncasecmp(argv[0],"polygons",l))) {
      primitive = POLY ;
    } else if (!(strncasecmp(argv[0],"points",l))) {
      primitive = POINT ;
    }
    
    if (argc==1) {
      return true ;
    } else if (argc==5) { // process r g b a
      if (parseNumbers(&(argv[1]),
		       &(primColor[0]), &(primColor[1]), 
		       &(primColor[2]), &(primColor[3]))) {
	if (primColor[3]<OPAQUEALPHA)
	  transparency = true ;
      } 
    } else if ((primitive==POLY) && (argc==2)) { // poly n
      if (parseNumbers(&(argv[1]), &sides)) {
	if (sides<=0) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid number of polygon sides", lineNum);
	}
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      }
      primitive = POLYN ;
    } else if ((primitive==POLY) && (argc==6)) { // poly r g b a n
      if (parseNumbers(&(argv[1]),
		       &(primColor[0]), &(primColor[1]), 
		       &(primColor[2]), &(primColor[3])) &&
	  parseNumbers(&(argv[5]), &sides)) {
	if (sides<=0) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid number of polygon sides", lineNum);
	}
	if (primColor[3]<OPAQUEALPHA)
	  transparency = true ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      }
      primitive = POLYN ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseTexturedPrimitive(int argc, char **argv, int l) {
  if (!(strncasecmp(argv[0],"tristrips",l)) ||
      !(strncasecmp(argv[0],"lines",l)) ||
      !(strncasecmp(argv[0],"polygons",l)) ||
      !(strncasecmp(argv[0],"points",l))) { 
    
    if (argc==1) {
      return false ;
    } else {
      int l2 = strlen(argv[1]) ;
      if (!(strncasecmp(argv[1],"3dtexture",l2))) {
	if (textureFile!=3) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "3DTEXTURE specified, but no 3d texture file loaded", 
		   lineNum);
	}
	textureType = 3 ;
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "3DTEXTURE not supported yet", lineNum);
      } else if (!(strncasecmp(argv[1],"2dtexture",l2))) {
	if (textureFile!=2) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "2DTEXTURE specified, but no 2d texture file loaded", 
		   lineNum);
	}
	textureType = 2 ;
      } else  // return false if texture NOT specified
	return false ;
    }
    
    if (debug)
      printf("pfdLoadFile_savg debug: command: primitive = %s\n",argv[0]) ;
    { // kludge to save textureType, as this resets it
      int foo = textureType ;
      processPrimitive() ;
      textureType = foo ;
    }

    if (!(strncasecmp(argv[0],"tristrips",l))) {
      primitive = TRI ;
    } else if (!(strncasecmp(argv[0],"lines",l))) {
      primitive = LINE ;
    } else if (!(strncasecmp(argv[0],"polygons",l))) {
      primitive = POLY ;
    } else if (!(strncasecmp(argv[0],"points",l))) {
      primitive = POINT ;
    }
    
    if (argc==2) {
      return true ;
    } else if (argc==6) { // process r g b a
      if (parseNumbers(&(argv[2]),
		       &(primColor[0]), &(primColor[1]), 
		       &(primColor[2]), &(primColor[3]))) {
	if (primColor[3]<OPAQUEALPHA)
	  transparency = true ;
      } 
    } else if ((primitive==POLY) && (argc==3)) { // poly n
      if (parseNumbers(&(argv[2]), &sides)) {
	if (sides<=0) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid number of polygon sides", lineNum);
	}
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      }
      primitive = POLYN ;
    } else if ((primitive==POLY) && (argc==7)) { // poly r g b a n
      if (parseNumbers(&(argv[2]),
		       &(primColor[0]), &(primColor[1]), 
		       &(primColor[2]), &(primColor[3])) &&
	  parseNumbers(&(argv[6]), &sides)) {
	if (sides<=0) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid number of polygon sides", lineNum);
	}
	if (primColor[3]<OPAQUEALPHA)
	  transparency = true ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      }
      primitive = POLYN ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid format", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parse2dTexturedData(int argc, char **argv, int l) {
  if (textureType==2) {
    if (argc==5) { // coord uv
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uv coord\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      pfVec3 v ;
      pfVec2 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1]))) {
	if (n>0 && ((color==VERTEX) || normals)) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex2d(t) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else if (argc==8) { // coord uv normal
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uv coord xyz normal\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      if (primitive==LINE || primitive==POINT) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "primitive can't have normals", lineNum);
      }
      pfVec3 v, no ;
      pfVec2 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1])) &&
	  parseNumbers(&(argv[5]), &(no[0]), &(no[1]), &(no[2]))) {
	if (n>0 && ((color==VERTEX) || (!normals))) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex2d(t) ;
	addNorm(no) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else if (argc==9) { // coord uv color
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uv coord rgba color\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      pfVec3 v ;
      pfVec4 c ;
      pfVec2 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1])) &&
	  parseNumbers(&(argv[5]), &(c[0]), &(c[1]), &(c[2]), &(c[3]))) {
	if (n>0 && ((color==PRIM) || normals)) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex2d(t) ;
	addColor(c) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else if (argc==12) { // coord uv color normal
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uv coord xyz normal rgba color\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      if (primitive==LINE) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "lines can't have normals", lineNum);
      }
      pfVec3 v, no ;
      pfVec4 c ;
      pfVec2 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1])) &&
	  parseNumbers(&(argv[5]), &(c[0]), &(c[1]), &(c[2]), &(c[3])) &&
	  parseNumbers(&(argv[9]), &(no[0]), &(no[1]), &(no[2]))) {
	if (n>0 && ((color==PRIM) || (!normals))) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex2d(t) ;
	addColor(c) ;
	addNorm(no) ;
	
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number of arguments", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parse3dTexturedData(int argc, char **argv, int l) {
  if (textureType==3) {
    if (argc==6) { // coord uvw
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uvw coord\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      pfVec3 v ;
      pfVec3 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1]), &(t[2]))) {
	if (n>0 && ((color==VERTEX) || normals)) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex3d(t) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else if (argc==9) { // coord uvw normal
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uvw coord xyz normal\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      if (primitive==LINE || primitive==POINT) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "primitive can't have normals", lineNum);
      }
      pfVec3 v, no ;
      pfVec3 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1]), &(t[2])) &&
	  parseNumbers(&(argv[6]), &(no[0]), &(no[1]), &(no[2]))) {
	if (n>0 && ((color==VERTEX) || (!normals))) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex3d(t) ;
	addNorm(no) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else if (argc==10) { // coord uvw color
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uvw coord rgba color\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      pfVec3 v ;
      pfVec4 c ;
      pfVec3 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1]), &(t[2])) &&
	  parseNumbers(&(argv[6]), &(c[0]), &(c[1]), &(c[2]), &(c[3]))) {
	if (n>0 && ((color==PRIM) || normals)) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex3d(t) ;
	addColor(c) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else if (argc==13) { // coord uvw color normal
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord uvw coord xyz normal rgba color\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      if (primitive==LINE) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "lines can't have normals", lineNum);
      }
      pfVec3 v, no ;
      pfVec4 c ;
      pfVec3 t ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(t[0]), &(t[1]), &(t[2])) &&
	  parseNumbers(&(argv[6]), &(c[0]), &(c[1]), &(c[2]), &(c[3])) &&
	  parseNumbers(&(argv[10]), &(no[0]), &(no[1]), &(no[2]))) {
	if (n>0 && ((color==PRIM) || (!normals))) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addTex3d(t) ;
	addColor(c) ;
	addNorm(no) ;
	
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number of arguments", lineNum);
    }
    
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseData(int argc, char **argv, int l) {
  if (textureType==0) {
    if (argc==3) { // coord
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      pfVec3 v ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2]))) {
	if (n>0 && ((color==VERTEX) || normals)) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } //if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2]))) { 
      
    } else if (argc==6) { // coord normal
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord xyz normal\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      if (primitive==LINE || primitive==POINT) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "primitive can't have normals", lineNum);
      }
      pfVec3 v, no ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(no[0]), &(no[1]), &(no[2]))) {
	if (n>0 && ((color==VERTEX) || (!normals))) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addNorm(no) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2]))) {
      
    } else if (argc==7) { // coord color
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord rgba color\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      pfVec3 v ;
      pfVec4 c ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(c[0]), &(c[1]), &(c[2]), &(c[3]))) {
	if (n>0 && ((color==PRIM) || normals)) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addColor(c) ;
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } //if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) && 
      
    } else if (argc==10) { // coord color normal
      if (debug)
	printf("pfdLoadFile_savg debug: data: xyz coord xyz normal rgba color\n") ;
      if (primitive==END) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "data given without a primitive type", lineNum);
      }
      if (primitive==LINE) {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "lines can't have normals", lineNum);
      }
      pfVec3 v, no ;
      pfVec4 c ;
      if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
	  parseNumbers(&(argv[3]), &(c[0]), &(c[1]), &(c[2]), &(c[3])) &&
	  parseNumbers(&(argv[7]), &(no[0]), &(no[1]), &(no[2]))) {
	if (n>0 && ((color==PRIM) || (!normals))) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "inconsistent data format for primitive", lineNum);
	}
	n++ ;
	addCoord(v) ;
	addColor(c) ;
	addNorm(no) ;
	
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      } // if (parseNumbers(&(argv[0]), &(v[0]), &(v[1]), &(v[2])) &&
      
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid number of arguments", lineNum);
    }
    
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
static bool parseColorMode(int argc, char **argv, int l) {
  if ((argc==2 || argc==4) // keyword [ r g b ] | [ default ]
      && (!(strncasecmp(argv[0],"ambient",l))
	  || !(strncasecmp(argv[0],"diffuse",l))
	  || !(strncasecmp(argv[0],"specular",l))
	  || !(strncasecmp(argv[0],"emission",l)))) {
    
    if (debug)
      printf("pfdLoadFile_savg debug: command: color = %s\n",argv[0]) ;
    processPrimitive() ;
    int l2 = strlen(argv[1]) ;
    float rc, gc, bc ;
    bool defc ;
    defc = false ;
    if (argc==4) { // process r g b 
      if (parseNumbers(&(argv[1]), &rc, &gc, &bc)) {
	if (rc<0.f || rc>1.f ||
	    gc<0.f || gc>1.f ||
	    bc<0.f || bc>1.f) {
	  pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		   "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_savg: line %d, "
		   "invalid number range", lineNum);
	}
      } else {
	pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
		 "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_savg: line %d, "
		 "invalid number format", lineNum);
      }
    } else if (!(strncasecmp(argv[1],"default",l2))) {
      defc = true ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid format", lineNum);
    }
    
    char *colortype = strdup(argv[0]) ;
    if (!(strncasecmp(argv[0],"ambient",l))) {
      if (defc)
	mat->setColor(PFMTL_CMODE_AMBIENT,
		      defaults.ambient[0], defaults.ambient[1], defaults.ambient[2]) ;
      else
	mat->setColor(PFMTL_CMODE_AMBIENT,rc, gc, bc) ;
    } else if (!(strncasecmp(argv[0],"diffuse",l))) {
      if (defc)
	mat->setColor(PFMTL_CMODE_DIFFUSE,
		      defaults.diffuse[0], defaults.diffuse[1], defaults.diffuse[2]) ;
      else
	mat->setColor(PFMTL_CMODE_DIFFUSE,rc, gc, bc) ;
    } else if (!(strncasecmp(argv[0],"emission",l))) {
      if (defc)
	mat->setColor(PFMTL_CMODE_EMISSION,
		      defaults.emission[0], defaults.emission[1], defaults.emission[2]) ;
      else
	mat->setColor(PFMTL_CMODE_EMISSION,rc, gc, bc) ;
    } else if (!(strncasecmp(argv[0],"specular",l))) {
      if (defc)
	mat->setColor(PFMTL_CMODE_SPECULAR,
		      defaults.specular[0], defaults.specular[1], defaults.specular[2]) ;
      else
	mat->setColor(PFMTL_CMODE_SPECULAR,rc, gc, bc) ;
    } else {
      pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	       "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_savg: line %d, "
	       "invalid keyword", lineNum);
    }
    return true ;
  } else
    return false ;
}

////////////////////////////////////////////////////////////////////////
// simple wrapper function to see if a string is an integer
bool isInt(char *s, int *i) {
  char *e ;
  long int l = strtol(s, &e, 10) ;
  if (s+strlen(s) == e) {
    *i = (int) l ;
    return true ;
  } else {
    return false ;
  }
}

////////////////////////////////////////////////////////////////////////
// simple wrapper function to see if a string is a float
bool isFloat(char *s, float *f) {
  char *e ;
  double d = strtod(s, &e) ;
  if (s+strlen(s) == e) {
    *f = (float) d ;
    return true ;
  } else {
    return false ;
  }
}

////////////////////////////////////////////////////////////////////////
// parse one int into f0 from argv[0]
bool parseNumbers(char **argv, int *i0) {
  if (isInt(argv[0],i0))
    return true ;
  else
    return false ;
}

////////////////////////////////////////////////////////////////////////
// parse one float into f0 from argv[0]
bool parseNumbers(char **argv, float *f0) {
  if (isFloat(argv[0],f0))
    return true ;
  else
    return false ;
}

////////////////////////////////////////////////////////////////////////
// parse two floats into f0, f1 from argv[0], argv[1]
bool parseNumbers(char **argv, float *f0, float *f1) {
  if (isFloat(argv[0],f0) &&
      isFloat(argv[1],f1))
    return true ;
  else
    return false ;
}

////////////////////////////////////////////////////////////////////////
// parse three floats into f0, f1, f2 from argv[0], argv[1], argv[2] 
bool parseNumbers(char **argv, float *f0, float *f1, float *f2) {
  if (isFloat(argv[0],f0) &&
      isFloat(argv[1],f1) &&
      isFloat(argv[2],f2))
    return true ;
  else
    return false ;
}

////////////////////////////////////////////////////////////////////////
// parse four floats into f0, f1, f2, f3 from argv[0], argv[1], argv[2], argv[3] 
bool parseNumbers(char **argv, float *f0, float *f1, float *f2, float *f3) {
  if (isFloat(argv[0],f0) &&
      isFloat(argv[1],f1) &&
      isFloat(argv[2],f2) &&
      isFloat(argv[3],f3))
    return true ;
  else
    return false ;
}








////////////////////////////////////////////////////////////////////////
// traverse the scenegraph, breaking up geodes as we go
// stolen from pfdbreakup pseudo-loader
static void breakupNodes(pfNode **node, pfGroup *parent) {

  if ((*node)->isExactType (pfGeode::getClassType())) {
#if SAVGBREAKUP
    pfNode *newNode = savgPfdBreakup((pfGeode *) *node,
		 geodeSize, stripLength, geodeChild);
#else
    pfNode *newNode = pfdBreakup((pfGeode *) *node,
		 geodeSize, stripLength, geodeChild);
#endif
    if (parent) 
      parent->replaceChild(*node, newNode) ;
    pfDelete(*node) ;
    if (!parent) 
      *node = newNode ;
  } else if ((*node)->isOfType (pfGroup::getClassType())) {
    pfGroup *group = (pfGroup *) *node ;
    int i ; 
    int n = group->getNumChildren() ;
    for (i=n-1 ; i>=0 ; i--) {
      pfNode *child = group->getChild(i);
      breakupNodes(&child, group) ;
    }
  } else {
    printf("Surprise!! unexpected node type returned by pfdBuilder- ignored\n") ;
  }
  
}
////////////////////////////////////////////////////////////////////////
// shrink a triangle towards its center by a certain percentage
// shrinkage is done in place
// scale needs to be between 0 and 1
void shrinkVerts(int n, pfVec3 *verts, float scale) {
  int i ;
  pfVec3 center = verts[0] ;
  for (i=1; i<n; i++) 
    center = center+verts[i] ;
  center = center/n ;

  for (i=0; i<n; i++) 
    verts[i].combine(scale, verts[i], (1.f-scale), center) ;
}

////////////////////////////////////////////////////////////////////////
// NIST SAVG, kelso 1/04
// a modified version of /usr/share/Performer/src/lib/libpfdu/pfdBreakup.c
//
// mods:
//   converted to C++
//   stringLength means number of triangles/geoset
//   stringLength can be odd- for tristrips this means you have to
//      flip orientation of every other tristrip
//   if stripLength == 0, skip stripLength processing
//   comments and error checking, mom and apple pie

// I can't figure out why if size--0, then childern must be >1, so I just
// test for it

/*
 * Copyright 1995, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * This source code ("Source Code") was originally derived from a
 * code base owned by Silicon Graphics, Inc. ("SGI")
 * 
 * LICENSE: SGI grants the user ("Licensee") permission to reproduce,
 * distribute, and create derivative works from this Source Code,
 * provided that: (1) the user reproduces this entire notice within
 * both source and binary format redistributions and any accompanying
 * materials such as documentation in printed or electronic format;
 * (2) the Source Code is not to be used, or ported or modified for
 * use, except in conjunction with OpenGL Performer; and (3) the
 * names of Silicon Graphics, Inc.  and SGI may not be used in any
 * advertising or publicity relating to the Source Code without the
 * prior written permission of SGI.  No further license or permission
 * may be inferred or deemed or construed to exist with regard to the
 * Source Code or the code base of which it forms a part. All rights
 * not expressly granted are reserved.
 * 
 * This Source Code is provided to Licensee AS IS, without any
 * warranty of any kind, either express, implied, or statutory,
 * including, but not limited to, any warranty that the Source Code
 * will conform to specifications, any implied warranties of
 * merchantability, fitness for a particular purpose, and freedom
 * from infringement, and any warranty that the documentation will
 * conform to the program, or any warranty that the Source Code will
 * be error free.
 * 
 * IN NO EVENT WILL SGI BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT
 * LIMITED TO DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES,
 * ARISING OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THE
 * SOURCE CODE, WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT OR
 * OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED BY PERSONS OR
 * PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED FROM,
 * OR AROSE OUT OF USE OR RESULTS FROM USE OF, OR LACK OF ABILITY TO
 * USE, THE SOURCE CODE.
 * 
 * Contact information:  Silicon Graphics, Inc., 
 * 1600 Amphitheatre Pkwy, Mountain View, CA  94043, 
 * or:  http://www.sgi.com
 */

/*
 * pfdBreakup.c
 *
 * $Revision: 1.23 $
 * $Date: 2003/10/17 23:32:34 $
 *
 * Input: A pfGeoSet that:
 * 	1. Is non-indexed.
 *	2. Is PFGS_STRIPS
 *	Input is NOT deleted, this is left to the caller
 * 
 * Output: 
 *      1: The root of a subgraph
 *      2: The subgraph is octree-like, with geodes
 *         of no larger than geodeSize and strips no longer 
 *         than stripLength.
 */

#ifdef mips
#include <bstring.h>
#else
#include <string.h>
#endif

#include <Performer/pf.h>
#include <Performer/pfdu-DLL.h>

#define INDEX ushort

static pfNode *subdivide(pfList *gsetList, float geodeSize, int geodeChild);

// NIST SAVG added kelso 1/04
void flipTri2(int n, pfVec2 *tri) ;
void flipTri3(int n, pfVec3 *tri) ;
void flipTri4(int n, pfVec4 *tri) ;

pfNode *savgPfdBreakup(pfGeode *geode, float geodeSize, 
		       int stripLength, int geodeChild) {
  
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, 
  "savgPfdBreakup: geodeSize = %f, stripLength = %d, geodeChild = %d\n",
	 geodeSize, stripLength, geodeChild) ;

  // NIST SAVG- added param check, kelso 1/04
  if (stripLength<0) {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "savgPfdBreakup: stripLength must be non-negative");
  }
  if (geodeSize<0.f) {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "savgPfdBreakup: geodeSize must be non-negative");
  }
  if (geodeChild<1) {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "savgPfdBreakup: geodeChild must be positive");
  }
  if (geodeSize<1.f && geodeChild==1) {
    pfNotify(PFNFY_ALWAYS, PFNFY_PRINT, 
	     "pfdLoadFile_savg: exiting at %s line %d", __FILE__, __LINE__);
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "savgPfdBreakup: if size is less than one, child must be greater than one");
  }
  int numGSet = geode->getNumGSets();
  
  /* striplength must be even */
  /* NIST SAVG- got rid of this requirement- kelso 1/04
  if (stripLength &1)
    stripLength++;
  */
  
  // list of chpped up geosets to pass to "subdivide"
  pfList *gsetList = new pfList(sizeof(pfGeoSet*), 100);
    
  // loop through geosets
  int igs;
  for (igs = 0 ; igs < numGSet ; igs++) {

    pfGeoSet *gs = geode->getGSet(igs);
    
    // NIST SAVG added code, kelso 1/04
    // if stripLength zero, just skip it
    if(stripLength == 0) {
      pfGeoSet *newgs = new pfGeoSet() ;
      pfCopy(newgs, gs);
      gsetList->add(newgs);
      // go on to next geoset in list
      continue;
    }

    /* 
     * break single geoset with many strips into multiple geosets
     * each of stripLength TRIANGLES 
     */
	
    // this'll get decremented to zero in while loop below
    int numPrims = gs->getNumPrims();

    INDEX  *verti, *normi, *colori, *tcoordi;

    pfGeoState *gstate = gs->getGState();

    pfVec3 *coords;
    gs->getAttrLists(PFGS_COORD3, (void **)&coords, &verti);

    pfVec3 *norms;
    gs->getAttrLists(PFGS_NORMAL3, (void **)&norms, &normi);

    pfVec4 *colors;
    gs->getAttrLists(PFGS_COLOR4, (void **)&colors, &colori);

    pfVec2 *tcoords;
    gs->getAttrLists(PFGS_TEXCOORD2, (void **)&tcoords, &tcoordi);
	
    int primType = gs->getPrimType();

    if ((primType != PFGS_TRISTRIPS &&
	 primType != PFGS_FLAT_TRISTRIPS &&
	 primType != PFGS_TRIS) ||
	verti != NULL ||
	normi != NULL || 
	colori != NULL ||
	tcoordi  != NULL) {
      pfGeoSet *newgs = new pfGeoSet() ;
      const char *primTypeNames[] = { "POINTS",  "LINES", "LINESTRIPS",  "TRIS",  
				      "QUADS",  "TRISTRIPS",  "FLAT_LINESTRIPS",  
				      "FLAT_TRISTRIPS",  "POLYS",  "TRIFANS",  
				      "FLAT_TRIFANS" } ;
      pfNotify(PFNFY_INFO, PFNFY_USAGE, 
	       "savgPfdBreakup: primitive \"%s\" is not a set of non-indexed tstrips or "
	       "non-indexed triangles",primTypeNames[primType]);
      pfCopy(newgs, gs);
      gsetList->add(newgs);
      // go on to next geoset in list
      continue;
    }
	
    if(primType == PFGS_TRIS) {
      //printf("\n\n!!!!!!!triangles\n\n\n") ;
      int nbind = gs->getAttrBind(PFGS_NORMAL3) ;
      int cbind = gs->getAttrBind(PFGS_COLOR4);
      int tbind = gs->getAttrBind(PFGS_TEXCOORD2);
	
      // indices into next data element
      int	nextVert, nextColor, nextNorm;
      nextVert = nextColor = nextNorm = 0;

      /* split tri lists */
      // loop through triangles- one triangle per primitive
      while (numPrims > 0) {

	int startColor, startNorm;
	int numColors, numNorms;
	pfVec3 *newCoords, *newNorms;
	pfVec4 *newColors;
	pfVec2 *newTCoords;

	int startVert = nextVert;

	// here stripLength is used as number of triangles!!
	int num = numPrims < stripLength ? numPrims : stripLength;

	/* 
	 * when near the end, split in half so we don't 
	 * end up with a silly small geoset on the end
	 */
	int remainNum = numPrims;
	if (remainNum < 2*num && remainNum > num)
	  num = remainNum/2;

	// 3 vertices per primitive
	int numVerts = num*3;

	nextVert += numVerts;
	
	numPrims -= num;

	pfGeoSet *newGSet = new pfGeoSet() ;
	gsetList->add(newGSet);
	newGSet->setPrimType(primType);
	newGSet->setNumPrims(num);
	newGSet->setGState(gstate);

	/* 
	 * copy coords
	 */
	newCoords = (pfVec3*)pfMalloc((uint)(numVerts*sizeof(pfVec3)));
	bcopy(&coords[startVert], newCoords, 
	      (uint)(numVerts * sizeof(pfVec3)));
	newGSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, newCoords, NULL);
		
	/* 
	 * copy normals 
	 */
	switch (nbind) {
	case PFGS_PER_VERTEX:
	  startNorm = nextNorm;
	  numNorms = numVerts;
	  nextNorm += numNorms;
	  break;
	case PFGS_PER_PRIM:
	  startNorm = nextNorm;
	  nextNorm += num;
	  numNorms = num;
	  break;
	case PFGS_OVERALL:
	  startNorm = nextNorm = 0;
	  numNorms = 1;
	  break;
	case PFGS_OFF:
	default:
	  numNorms = 0;
	  break;
	}
	newNorms = NULL;
	if (numNorms > 0) {
	  newNorms = (pfVec3 *)pfMalloc((uint)(numNorms * sizeof(pfVec3))) ;
	  bcopy(&norms[startNorm], newNorms, 
		(uint)(numNorms*sizeof(pfVec3)));
	}
	newGSet->setAttr(PFGS_NORMAL3, nbind, newNorms, NULL);
	
	/* 
	 * copy colors
	 */
	switch (cbind) {
	case PFGS_PER_VERTEX:
	  startColor = nextColor;
	  numColors = numVerts;
	  nextColor += numColors;
	  break;
	case PFGS_PER_PRIM:
	  startColor = nextColor;
	  numColors = num;
	  nextColor += numColors;
	  break;
	case PFGS_OVERALL:
	  startColor = nextColor = 0;
	  numColors = 1;
	  break;
	case PFGS_OFF:
	default:
	  numColors = 0;
	  break;
	}
	newColors = NULL;
	if (numColors > 0) {
	  newColors = (pfVec4 *)pfMalloc((uint)(numColors*sizeof(pfVec4))) ;
	  bcopy(&colors[startColor], newColors,
		(uint)(numColors*sizeof(pfVec4)));
	}
	newGSet->setAttr(PFGS_COLOR4, cbind, newColors, NULL);
	/*
	 * copy tcoords 
	 */
	switch (tbind) {
	case PFGS_PER_VERTEX:
	  newTCoords = (pfVec2 *)pfMalloc((uint)(numVerts*sizeof(pfVec2))) ;
	  bcopy(&tcoords[startVert], newTCoords, 
		(uint)(numVerts*sizeof(pfVec2)));
	  newGSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, newTCoords, NULL);
	  break;
	case PFGS_OFF:
	default:
	  newGSet->setAttr(PFGS_TEXCOORD2, PFGS_OFF, NULL, NULL);
	  break;
	}
      }
      // go on to next geoset in list
      continue;


    } else {  //if(primType == PFGS_TRIS)
      
      // tristrip processing

      //printf("\n\n!!!!!!!tristrips\n\n\n") ;
      int *plens = gs->getPrimLengths();
      
      int nbind = gs->getAttrBind(PFGS_NORMAL3);
      int cbind = gs->getAttrBind(PFGS_COLOR4);
      int tbind = gs->getAttrBind(PFGS_TEXCOORD2);
      
      int	nextVert, nextColor, nextNorm;
      nextVert = nextColor = nextNorm = 0;
      int 	i;
      for (i = 0 ; i < numPrims ; i++) {

	// index of first vertex in next primitive
	int nextPrimVert = nextVert + plens[i];

	// NIST SAVG- added for flip test, kelso 1/04
	// index of first vertex in this primitive
	int startPrimVert = nextVert ;

	/* starting tristrip for this geoset */
	// always two more vertices than triangles in a tristrip
	while (nextVert < nextPrimVert - 2) {

	  // index of start of new geoset
	  int startVert = nextVert;
	  
	  /* 
	   * when near the end, split in half so we don't 
	   * end up with a silly small geoset on the end
	   */
	  int remainVert = nextPrimVert - nextVert;
	  // NIST SAVG- stripLength is triangles- kelso 1/04
	  if (remainVert-2 < 2*stripLength && remainVert-2 > stripLength) {
	    nextVert += remainVert/2;
	    /* NIST SAVG, got rid of this requirement, kelso 1/04
	    if ((nextVert - startVert) & 1)
	      nextVert++;
	    */
	  } else
	    nextVert += stripLength;
	  
	  if (nextVert > (nextPrimVert-2))
	    nextVert = nextPrimVert - 2;
	  
	  // NIST SAVG- added kelso 1/04
	  // do we flip our triangle?
	  bool flip = ((startVert-startPrimVert) & 1) ;

	  int numVerts = nextVert - startVert + 2;
	  
	  pfGeoSet *newGSet = new pfGeoSet() ;
	  gsetList->add(newGSet);
	  newGSet->setPrimType(primType);
	  /* XXX - for better geoset size, should group multiple 
	   * strips up to striplength verts*/
	  newGSet->setNumPrims(1);
	  newGSet->setGState(gstate);

	  int *pl = (int *)pfMalloc(sizeof(int), 
				    pfGetSharedArena());
	  pl[0] = numVerts;
	  newGSet->setPrimLengths(pl);

	  /* 
	   * copy coords
	   */
	  pfVec3 *newCoords = (pfVec3*)pfMalloc((uint)(numVerts*sizeof(pfVec3))) ;
	  bcopy(&coords[startVert], newCoords, 
		(uint)(numVerts * sizeof(pfVec3)));
	  // NIST SAVG added by kelso 1/04
	  // flip triangles if starting at an odd vertex in this prim
   	  if (flip) 
	    flipTri3(numVerts, newCoords) ;
	  newGSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, newCoords, NULL);
	  /* 
	   * copy normals 
	   */
	  int numNorms ;
	  int startNorm ;
	  switch (nbind) {
	  case PFGS_PER_VERTEX:
	    startNorm = nextNorm;
	    if (primType == PFGS_FLAT_TRISTRIPS) {
	      numNorms = numVerts - 2;
	      nextNorm += numNorms;
	    } else {
	      numNorms = numVerts;
	      nextNorm += (numNorms-2);
	    }
	    break;
	  case PFGS_PER_PRIM:
	    startNorm = nextNorm = i;
	    numNorms = 1;
	    break;
	  case PFGS_OVERALL:
	    startNorm = nextNorm = 0;
	    numNorms = 1;
	    break;
	  case PFGS_OFF:
	  default:
	    numNorms = 0;
	    break;
	  }
	  pfVec3 *newNorms = NULL;
	  if (numNorms > 0) {
	    newNorms = (pfVec3 *)pfMalloc((uint)(numNorms * sizeof(pfVec3))) ;
	    bcopy(&norms[startNorm], newNorms, 
		  (uint)(numNorms*sizeof(pfVec3)));
	    // NIST SAVG added by kelso 1/04
	    // flip triangles if starting at an odd vertex in this prim
	    if (flip) 
	      flipTri3(numNorms, newNorms) ;
	  }
	  newGSet->setAttr(PFGS_NORMAL3, nbind, newNorms, NULL);
	      
	  /* 
	   * copy colors
	   */
	  int startColor;
	  int numColors ;
	  switch (cbind) {
	  case PFGS_PER_VERTEX:
	    startColor = nextColor;
	    if (primType == PFGS_FLAT_TRISTRIPS) {
	      numColors = numVerts - 2;
	      nextColor += numColors;
	    } else {
	      numColors = numVerts;
	      nextColor += (numColors-2);
	    }
	    break;
	  case PFGS_PER_PRIM:
	    startColor = nextColor = i;
	    numColors = 1;
	    break;
	  case PFGS_OVERALL:
	    startColor = nextColor = 0;
	    numColors = 1;
	    break;
	  case PFGS_OFF:
	  default:
	    numColors = 0;
	    break;
	  }
	  pfVec4 *newColors = NULL;
	  if (numColors > 0) {
	    newColors = (pfVec4 *)pfMalloc((uint)(numColors*sizeof(pfVec4))) ;
	    bcopy(&colors[startColor], newColors,
		  (uint)(numColors*sizeof(pfVec4)));
	    // NIST SAVG added by kelso 1/04
	    // flip triangles if starting at an odd vertex in this prim
	    if (flip) 
	      flipTri4(numColors, newColors) ;
	  }
	  newGSet->setAttr(PFGS_COLOR4, cbind, newColors, NULL);

	  /*
	   * copy tcoords 
	   */

	  pfVec2 *newTCoords;
	  switch (tbind) {
	  case PFGS_PER_VERTEX:
	    newTCoords = (pfVec2 *)pfMalloc((uint)(numVerts*sizeof(pfVec2))) ;
	    bcopy(&tcoords[startVert], newTCoords, 
		  (uint)(numVerts*sizeof(pfVec2)));
	    // NIST SAVG added by kelso 1/04
	    // flip triangles if starting at an odd vertex in this prim
	    if ((startVert-startPrimVert) && 1) 
	      flipTri2(numVerts, newTCoords) ;
	    newGSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, newTCoords, NULL);
	    break;
	  case PFGS_OFF:
	  default:
 	    newGSet->setAttr(PFGS_TEXCOORD2, PFGS_OFF, NULL, NULL);
	    break;
	  }
	} //  while (nextVert < nextPrimVert - 2)
	nextVert = nextPrimVert;
	if (nbind == PFGS_PER_VERTEX &&
	    primType != PFGS_FLAT_TRISTRIPS)
	  nextNorm += 2;
	if (cbind == PFGS_PER_VERTEX &&
	    primType != PFGS_FLAT_TRISTRIPS)
	  nextColor += 2;
      } //for (i = 0 ; i < numPrims ; i++)
    } // else
  } // for (igs = 0 ; igs < numGSet ; igs++)
    /* 
     * Take array of geosets and place into a rough octree 
     * based upon bbox centers 
     */
  pfNode *node = subdivide(gsetList, geodeSize, geodeChild);
  pfDelete(gsetList);
  return node;
}

static pfNode *
subdivide(pfList *gsetList, float geodeSize, int geodeChild) {
  
  int ngset = gsetList->getNum();
  pfBox bigbox ;
  bigbox.makeEmpty() ;
  
  /* 
   * compute center 
   */
  int i ;
  for (i = 0 ; i < ngset ; i++) {
    pfBox gsbbox;
    pfGeoSet *gset = (pfGeoSet *) gsetList->get(i);
    gset->getBound(&gsbbox);
    bigbox.extendBy(&gsbbox) ;
  }
  pfVec3 center ;
  center.combine(0.5f, bigbox.min, 0.5f, bigbox.max);
  
  float size2 = PFSQR_DISTANCE_PT3(bigbox.min, bigbox.max);
  
  /* 
   * if number of geosets < children, or smaller than specified
   * geode dimension, return a single geode
   */
  if (ngset <= geodeChild || size2 < geodeSize*geodeSize) {
    pfGeode *geode = new pfGeode();
    for (i = 0 ; i < ngset ; i++) {
      pfGeoSet *gset = (pfGeoSet *) gsetList->get(i);
      geode->addGSet(gset) ;
    }
    return (pfNode*)geode;
  }
  
  pfList *lists[8];
  for (i = 0 ; i < 8 ; i++)
    lists[i] = new pfList(sizeof(pfGeoSet*), 10);
  
  pfVec3  norms[3] = { pfVec3(1.0, 0.0, 0.0), 
		       pfVec3(0.0, 1.0, 0.0), 
		       pfVec3(0.0, 0.0, 1.0) };

  pfPlane planes[3];
  for (i = 0 ; i < 3 ; i++)
    planes[i].makeNormPt(norms[i], center);

  /* 
   * place geosets on one of eight lists based on center
   */
  for (i = 0 ; i < ngset ; i++) {
    pfBox bbox;
    pfVec3 gscenter;
    pfGeoSet *gset;
    int which = 0;
    
    gset = (pfGeoSet *) gsetList->get(i);
    gset->getBound(&bbox);
    gscenter.combine(0.5f, bbox.min, 0.5f, bbox.max);

    int j ;
    for (j = 0 ; j < 3 ; j++) {
      if (planes[j].contains(gscenter))
	which |= (1<<j);
    }
    lists[which]->add(gset);
  }
  /* 
   * check for number of active octants
   */
  int active = 0;
  for (i = 0 ; i < 8 ; i++) {
    if (lists[i]->getNum() > 0)
      active++;
  }
  /*
   * if only one octant active, arbitrarily subdivide
   */
  pfGroup *grp = new pfGroup() ;
  
  if (active < 2) {
    int nper = PF_MAX2(ngset/4, 3);
    int which = 0;
    for (i = 0 ; i < 8 ; i++)
      lists[i]->reset();
    
    for (i = 0 ; i < ngset ; i++) {
      pfGeoSet *gset = (pfGeoSet *) gsetList->get(i);
      if (i > (which+1)*nper)
	which++;
      lists[which]->add(gset);
    }
  }
  /* 
   * add recursion result to current group &  cleanup
   */
  for (i = 0 ; i < 8 ; i++) {
    if (lists[i]->getNum() > 0) {
      pfNode *child = subdivide(lists[i], geodeSize, geodeChild);
      if (child)
	grp->addChild(child);
    }
    pfDelete(lists[i]);
  }
  return (pfNode *)grp;
}

////////////////////////////////////////////////////////////////////////
// given an array of n pfVec2s representing a tristrip coord, 
// reverse hand-ed-ness in place
void flipTri2(int n, pfVec2 *tri) {
  int i ;
  pfVec2 holder ;
  for (i=0; i<n-2; i+=3) {
    holder = tri[i+1] ;
    tri[i+1] = tri[i+2] ;
    tri[i+2] = holder ;
  }
}

////////////////////////////////////////////////////////////////////////
// given an array of n pfVec3s representing a tristrip coord, 
// reverse hand-ed-ness in place
void flipTri3(int n, pfVec3 *tri) {
  int i ;
  pfVec3 holder ;
  for (i=0; i<n-2; i+=3) {
    holder = tri[i+1] ;
    tri[i+1] = tri[i+2] ;
    tri[i+2] = holder ;
  }
}

////////////////////////////////////////////////////////////////////////
// given an array of n pfVec4s representing a tristrip color, 
// reverse hand-ed-ness in place
void flipTri4(int n, pfVec4 *tri) {
  int i ;
  pfVec4 holder ;
  for (i=0; i<n-2; i+=3) {
    holder = tri[i+1] ;
    tri[i+1] = tri[i+2] ;
    tri[i+2] = holder ;
  }
}

