/*
 * pfseq.c - the animation sequence loader based on lobpfpoly
 *            the Prisims ".poly" file loader
 *
 * Steve Satterfield
 * Original implementation: April 19, 2000
 * Revised: February 23, 2001
 * Revised: February 21, 2002
 * Revised: January 27, 2005
 */

/*
 * pfpoly.c: $Revision: 1.15 $ $Date: 1995/11/29 16:02:54 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Performer/pf.h>
#include <Performer/pfutil.h>
#include <Performer/pfdu.h>


/*
 *
 * pfdLoadFile_seq (char *filename)
 *
 *
 * 
 * Read a sequence file which contains a list of keywords defining a time step
 * sequence and builds a data structure with a pfSequence node at the top so
 * that the frames are animated.
 *
 * .seq file format contains a list of keywords followed by parameters
 *
 *    Keyword       Parameters        Purpose
 *    -------       ----------        -------
 *
 *     #            anything          comment line
 *
 *     time         float             Indicates the time between frames
 *                                    in seconds. Should be > 0.0.
 *                                    .1 works well.
 *
 *     frame                          Indicates the beginning of a new 
 *                                    animation frame defined by the following
 *                                    keywords.
 *
 *     file         filename          filename can be any valid performer
 *                                    file format.
 *                                    This file is added to the current frame
 *
 *     object       filename          Similar to file, but is not immediately
 *                                    added to the current frame. Instead, the
 *                                    display sturcture is defined to be the
 *                                    current object and is saved for later
 *                                    use by the "translate" keyword.
 *
 *     objectname   name              assigns a name (ASCII string) to be the
 *                                    name of the current object. This name
 *                                    can be used later to refer to this
 *                                    object.
 *
 *     useobject    name              make the object previously assigned
 *                                    name to to the current object.
 *
 *     quat         w x y z           Specifies a current quaternion rotation.
 *
 *     scale        x y z             Specifies a current scale.
 *
 *     translate    x y z             Specifies a current translation to and
 *                                    places the most recent specified object
 *                                    at the location specified by the x,y,z
 *                                    position (translate) in the current
 *                                    scale and quaternion rotation.
 *
 *
 *  Example 1:
  *
 *     # This is an example of two frames each containing 3 files
 *     # Each frame is displayed for .2 seconds
 *     time .2
 *     frame
 *     file object1.poly
 *     file object2.poly
 *     file object3.poly
 *     frame
 *     file object4.poly
 *     file object5.poly
 *     file object6.poly
 *
 *  Example 2 (./example2.seq):
 *
 *    # example2.seq - using scale, object, translate keywords
 *    time 1
 *    object esprit.flt
 *    quat 0 0 0 0
 *    frame
 *    scale 1 1 1
 *    translate 0 0 0
 *    frame
 *    scale .5 .5 .5
 *    translate 10 0 0
 *    frame
 *    scale .25 .25 .25
 *    translate 20 0 0
 *
 *
 */


/*
 * pfdLoadFile_seq -- Load ".seq" files into IRIS Performer
 */

#define S_MAX 80


/*
#define THREE_SCS 0
*/


extern pfNode *
pfdLoadFile_seq (char *filename)
{
    FILE *fp;
    char input[S_MAX],keyword[S_MAX],objname[S_MAX];
    pfSequence *seq_node;
    pfNode *model;
    pfGroup *group;
#ifdef THREE_SCS
    pfSCS *scale_scs,*rot_scs,*trans_scs;
#endif
    pfSCS *combined_scs;
    double time;
    int i;


    pfNode *object;
    float qw,qx,qy,qz;   /* Current Quaternion rotation */
    pfQuat q;                        /* pfQuat represents a quaternion as the
					four floating point values (x, y, z, w)
					of a pfVec4. */
    float sx,sy,sz;      /* Current scale */
    float tx,ty,tz;

    pfMatrix m;      /* transformation matrix */
    pfMatrix trans_mat,scal_mat,quat_mat;

    qw=1.; qx=0.; qy=0.; qz=0.;
    sx=1.; sy=1.; sz=1.;
    tx=0.; ty=0.; tz=0.;

    time = 1.0;

    /* Build a pfSequence node and add the models from the list of files */
    seq_node = pfNewSeq();


    fp = pfdOpenFile(filename);
    while (fgets(input, S_MAX, fp) != NULL) {
      i = (int)strlen(input);

      if (i > 1) {       /* skip blank lines */
	if (input[0] != '#') {   /* skip comment lines */
	  sscanf(input, "%s\n",keyword);
	  /* fprintf(stderr,"keyword= %s\n",keyword); */

	  if (strcmp("file",keyword) == 0) {
	    sscanf(&input[4],"%s",objname);
	    /* fprintf(stderr,"Filename = '%s'\n",objname); */

	    if (!pfdInitConverter(objname)) {
	      exit(1);
	    }

	    model = pfdLoadFile(objname);
	    pfAddChild(group, model);
	  }
	  else if (strcmp("object",keyword) == 0) {
	    sscanf(&input[6],"%s",objname);
	    /* fprintf(stderr,"object = '%s'\n",objname); */
	    if (!pfdInitConverter(objname)) {
	      exit(1);
	    }

	    object = pfdLoadFile(objname);
	  }
	  else if (strcmp("objectname",keyword) == 0) {
	    sscanf(&input[10],"%s",objname);
	    /* fprintf(stderr,"objectname = '%s'\n",objname); */
	    pfNodeName(object, objname);

	  }
	  else if (strcmp("useobject",keyword) == 0) {
	    sscanf(&input[9],"%s",objname);
	    /* fprintf(stderr,"useobject = '%s'\n",objname); */

	    object = (pfNode*) pfFindNode(seq_node, objname,		   \
                                      pfGetNodeClassType());		   \
	    /*								   \
	    if (object == NULL) {					   \
	         fprintf(stderr," Searching group node for %s\n",objname); \
	         object = (pfNode*) pfFindNode(group, objname,		   \
                                      pfGetNodeClassType());		   \
	         fprintf(stderr," object ptr %d\n",object);		   \
	    }								   \
	    */
	    if (object == NULL) 
               pfNotify(PFNFY_NOTICE, PFNFY_PRINT,
                      "Warning name '%s' not found", objname);


	  }
	  else if (strcmp("frame",keyword) == 0) {
	    /* fprintf(stderr,"Create new frame\n"); */
	    fprintf(stderr,".");
	    group = pfNewGroup();
	    pfAddChild(seq_node, group);
	  }
	  else if (strcmp("quat",keyword) == 0) {
	    sscanf(&input[4],"%g %g %g %g",&qw,&qx,&qy,&qz);
	    /* fprintf(stderr,"qw=%g qx=%g qy=%g qz=%g\n",qw,qx,qy,qz); */
	    /* pfSetVec4(q, 0.5,0.5,.5, 1.); */
	  }
	  else if (strcmp("scale",keyword) == 0) {
	    sscanf(&input[5],"%g %g %g",&sx,&sy,&sz);
	    /* fprintf(stderr,"scale = %g %g %g\n",sx,sy,sz); */
	  }
	  else if (strcmp("translate",keyword) == 0) {
	    sscanf(&input[9],"%g %g %g",&tx,&ty,&tz);
	    /* fprintf(stderr,"translate= %g %g %g\n",tx,ty,tz); */

	    /*
             * Now what I want is to scale the object,
             * rotate it and then move to position.
	     */

#ifdef THREE_SCS
	    /* Use th SCS nodes to accomplish this */
	    pfMakeTransMat(m, tx,ty,tz);
	    trans_scs=pfNewSCS(m);

	    pfMakeScaleMat(m, sx,sy,sz);
	    scale_scs=pfNewSCS(m);

	    pfSetVec4(q, qx,qy,qz, qw);
	    pfMakeQuatMat(m, q);
	    /* pfMakeRotMat(m, 30., 0., 1., 0.); */
	    rot_scs=pfNewSCS(m);

	    pfAddChild(scale_scs, object);
	    pfAddChild(rot_scs, scale_scs);
	    pfAddChild(trans_scs, rot_scs);
	    
	    pfAddChild(group, trans_scs);
#else
	    /* Do multiply here and use only one SCS */
	    pfMakeTransMat(trans_mat, tx,ty,tz);
	    pfMakeScaleMat(scal_mat, sx,sy,sz);
	    pfSetVec4(q, qx,qy,qz, qw);
	    pfMakeQuatMat(quat_mat, q);

       
	    pfMultMat(m, scal_mat, quat_mat);  /* m = scale_mat*quat_mat */
	    pfMultMat(m, m, trans_mat);        /* m = m*trans_mat */
                        	    /* m = scale_mat * quat_mat * trans_mat */

	    combined_scs = pfNewSCS(m);

	    pfAddChild(combined_scs, object);
	    pfAddChild(group, combined_scs);
#endif
	  }
	  else if (strcmp("time",keyword) == 0) {
	    sscanf(&input[4],"%lg",&time);
	    /* fprintf(stderr,"Time step = %lg\n",time); */
	  }
	}
      }
    }

    pfSeqMode(seq_node, PFSEQ_START);
    pfSeqTime(seq_node, -1, time);
    pfSeqInterval(seq_node, PFSEQ_CYCLE, 0, -1);
    pfSeqDuration(seq_node, 1, -1);

    fclose(fp);
    fprintf(stderr,"\n");
    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "  pfdLoadFile_seq: %s", filename);
    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "  Created (by SGS), compiled %s %s",__DATE__, __TIME__);
    return((pfNode*)seq_node);

}
