/*
  
pfsge.c - a scenegraph loader and editor for file suffix "sge"

usage:  pfNode *foo = pfdLoadFile("foo.sge") ;

Adele Peskin
peskin@boulder.nist.gov
January 2005
  
A pfGroup node is created- under it are loaded nodes and files

A sge file has the following format:

# this is a comment
# keyword matching is case insensative

# create a pfGroup, pfDCS or pfSCS node named "nodeName."
# An optional transformation can be
# given.  A node name can be left unspecified by using the keyword NULL

GROUP groupname
DCS   dcsname [ x y z [ h p r [ s [ sy sz ] ] ] ] 
SCS   scsname [ x y z [ h p r [ s [ sy sz ] ] ] ]
LOAD  filename nodename
UNLOAD nodename
ON nodename groupname
OFF nodename
CLONE nodename clonename
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Performer/pf.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfSCS.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pfutil.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfSequence.h>
#include "getString.C"
#include "normalize.C"
// this same function is in the dpf library, but copied here for use with file loaders
#include "parse.C"

#define MAX_STRLEN (2048)

// function templates
void cleanup(void) ;

pfMatrix getMatrix(int argc, char **argv) ;

// file name and pointer to pfNode where the file is loaded
struct fileNode {
  char *name ;
  pfNode *node ;
} ;

// evil global variables

// input line number for error messages
int lineNum ;

// an array of unique fileNodes
fileNode *file ;

// how many of them?
int files ;

// group node to return ;
pfGroup *topgroup ;

// handy identity matrix
pfMatrix identity ;

// active node- might be a pfSCS or pfDCS cast to a pfGroup
pfGroup *active ;

extern "C" pfNode * pfdLoadFile_sge (char *filename) {
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_sge: %s", filename);
  FILE *fp = pfdOpenFile(filename);
  if (!fp)
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_sge: unable to open %s for reading", filename);
  
  int notifyLevel = pfGetNotifyLevel() ;

  // in a DSO you need to initialize with assignment statements
  lineNum = 0 ;
  file = NULL ;
  files = 0 ;
  topgroup = new pfGroup() ;
  identity.makeIdent() ;

  // read in the file and store data
  char *line = NULL ;
  char **argv = NULL ;
  int argc ;
  double time = 0.1;

  // loop through the input
  while (1) {
    if (notifyLevel >= PFNFY_INFO) {
      fprintf(stderr,"%8.8d   \r",lineNum) ;
      fflush(stderr) ;
    }
    lineNum++ ;

    if (getString(&line, fp)<0) {
      cleanup() ;
      break ;
    } else {
      if (normalize(line)) {
	dpf_charsToArgcArgv(line, &argc, &argv) ;

	// how many characters to match
  	int l = strlen(argv[0]) ;
	if (l<3) l=3 ;
	// start parsing non-blank lines

	// command = load objectname filename
	if (!(strncasecmp(argv[0],"load",l))) {
	  // see if filename has already been loaded
	  int thisFile ;
	  for (thisFile=0; thisFile<files; thisFile++) {
	    if (!strcmp(file[thisFile].name,argv[1]))
	      break ;
	  }
	    
	  // didn't find it, so add it
	  if (thisFile>=files) { 
	    thisFile = files ;
	    files++ ;
	    file = (fileNode *) realloc(file, sizeof(fileNode)*files) ;
	    file[thisFile].name = strdup(argv[1]) ;
	    file[thisFile].node = pfdLoadFile(argv[1]) ;

	    if (!file[thisFile].node) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to load file %s", lineNum, argv[2]);
	    }
	    else {
	      // see if there is name
	      if (argc > 2) {
		file[thisFile].node->setName(argv[2]);
	      }
	      else {
		pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
			 "pfdLoadFile_sge: line %d, "
			 "load command requires nodename and filename: ", lineNum);
	      }
	    }
	  } //if (thisFile>=files)
	}

	// command: unload objectname
	
	else if (!(strncasecmp(argv[0],"unload",l))) {
	  if (argc < 2) {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		     "unload command requires nodename: ", lineNum);
	  } 
	  else {
	    // find the node to remove
	    pfNode *node = NULL;
	    node = topgroup->lookup(argv[1], pfGroup::getClassType()) ;
	    if (!node) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to find node \"%s\"", lineNum, argv[1]);
	    } 
	    else {
	      // delete the node
	      int pcnt = node->getNumParents();
	      if (pcnt > 0) {
		// remove node
		pfGroup* parentgroup = node->getParent(0);
		
		if (node != NULL && parentgroup != NULL) 
		  parentgroup->removeChild(node);
	      }
	      pfDelete(node);
	    }
	  }
	} 

	// command: on objectname nodename
	else if (!(strncasecmp(argv[0],"on",l))) {
	  //printf("spit out line: on %s %s\n",argv[1],argv[2]);
	  if (argc < 3) {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		     "on command requires nodename and groupname ", lineNum);
	  } 
	  else {
	    pfNode *node = NULL;
	    node = topgroup->lookup(argv[1], pfNode::getClassType()) ;
	    if (!node) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to find node \"%s\"", lineNum, argv[1]);
	    } 
	    else {
	      pfGroup *grp = NULL ;
	      grp = (pfGroup*)topgroup->lookup(argv[2], pfGroup::getClassType()) ;
	      if (!grp)
		grp = (pfGroup*)topgroup->lookup(argv[2], pfDCS::getClassType()) ;
	      if (!grp)
		grp = (pfGroup*)topgroup->lookup(argv[2], pfSCS::getClassType()) ;
	      if (!grp) {
		// see if this node is a sequence node
		//printf("see if this is a sequence\n");
		pfSequence *sq = (pfSequence*)topgroup->lookup(argv[2], pfSequence::getClassType()) ;
		if (sq == NULL) 
		  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
			   "pfdLoadFile_sge: line %d, "
			   "unable to find group\"%s\"", lineNum, argv[2]);
		else
		  // if there is a seqence node add the group
		  sq->addChild(node);
	      }
	      else {
		grp->addChild(node);
	      }
	    }
	  }
	} 

	// command: off objectname
	else if (!(strncasecmp(argv[0],"off",l))) {
	  if (argc < 2) {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		     "off command requires nodename: ", lineNum);
	  } 
	  else {
	    // find the node to remove
	    pfNode *node = NULL;
	    node = topgroup->lookup(argv[1], pfGroup::getClassType()) ;
	    if (!node) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to find node \"%s\"", lineNum, argv[1]);
	    } 
	    else {
	      // find parent and remove
	      int pcnt = node->getNumParents();
	      if (pcnt > 0) {
		pfGroup *grp = node->getParent(0);
		if (grp != NULL)
		  grp->removeChild(node);
		else {
		  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to remove node\"%s\"", lineNum, argv[1]);
		}
	      }
	    }
	  }
	} 

	// command: clone objectname newobjectname
	else if (!(strncasecmp(argv[0],"clone",l))) {
	  if (argc < 3) {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		     "clone command requires nodename and a new object name ", lineNum);
	  } 
	  else {
	    // find the node to clone
	    pfNode *node = NULL;
	    node = topgroup->lookup(argv[1], pfGroup::getClassType()) ;
	    if (!node) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to find node \"%s\"", lineNum, argv[1]);
	    } 
	    else {
	      // clone it
	      pfNode* node2 = node->clone(0);
	      node2->setName(argv[2]);
	    }
	  }
	} 

	// command: unload objectname
	// DCS node
	else if (!(strncasecmp(argv[0],"dcs",l))) {
	  pfMatrix mat ;
	  if (argc>=3)
	    mat = getMatrix(argc-2, &argv[2]) ;
	  else 
	    mat = identity ;
	  pfDCS *node = new pfDCS() ;
	  node->setMat(mat) ;

	  // name is required
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to name node \"%s\"", lineNum, argv[1]);
	  }
	  else {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		       "dcs requires a name ", lineNum);
	  }

	// SCS node
	} else if (!(strncasecmp(argv[0],"scs",l))) {
	  pfMatrix mat ;
	  if (argc>=3)
	    mat = getMatrix(argc-2, &argv[2]) ;
	  else 
	    mat = identity ;
	  pfGroup *node;

	  if (mat == identity)
	    node = new pfGroup() ;
	  else
	    node = new pfSCS(mat) ;
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to name node \"%s\"", lineNum, argv[1]);
	  }
	  else {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		       "scs requires a name ", lineNum);
	  }
	}
	// group node
	else if (!(strncasecmp(argv[0],"group",l))) {
	  //printf("spit out line: group %s \n",argv[1]);
	  pfGroup *node = new pfGroup() ;
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to name node \"%s\"", lineNum, argv[1]);
	  }
	  else {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		       "group requires a name ", lineNum);
	  }
	}

	// sequence node
	else if (!(strncasecmp(argv[0],"SEQ",l))) {
	  //printf("spit out line: seq %s %s\n",argv[1], argv[2]);
	  pfSequence *seq_node = new pfSequence();
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(seq_node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sge: line %d, "
		       "unable to name sequence \"%s\"", lineNum, argv[1]);
	  }
	  if (argc>2 && (strcasecmp(argv[1],"NULL")))
	    time = (double)(atof(argv[2]));
	  seq_node->setTime(-1, time);
	  seq_node->setMode(PFSEQ_START);
	  seq_node->setInterval(PFSEQ_CYCLE, 0, -1);
	  seq_node->setDuration(1, -1); 
	}
	else {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		     "missing node name", lineNum);
	}
      } //if (normalize(line))
    } //if (getString(&line, fp)<0)
  } // while (1)

  fclose(fp);
  return(topgroup) ;
}


void cleanup(void) {
}

pfMatrix getMatrix(int argc, char **argv) {
  pfCoord coord ;
  pfVec3  scale ;

  // initialize to identity
  coord.xyz = pfVec3(0.f, 0.f, 0.f) ;
  coord.hpr = pfVec3(0.f, 0.f, 0.f) ;
  scale = pfVec3(1.f, 1.f, 1.f) ;

  if (argc!=3 && argc!=6 && argc!=7 && argc!=9)
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_sge: line %d, "
	     "invalid number of parameters", lineNum);

  if (argc >= 3) {
    if ((sscanf(argv[0],"%f",&coord.xyz[0])!=1) ||
	(sscanf(argv[1],"%f",&coord.xyz[1])!=1) ||
	(sscanf(argv[2],"%f",&coord.xyz[2])!=1))
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_sge: line %d, "
	       "invalid number", lineNum);
    else if (argc >= 6) {
      if ((sscanf(argv[3],"%f",&coord.hpr[0])!=1) ||
	  (sscanf(argv[4],"%f",&coord.hpr[1])!=1) ||
	  (sscanf(argv[5],"%f",&coord.hpr[2])!=1))
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_sge: line %d, "
		 "invalid number", lineNum);
      
      else if (argc >= 7) {
	if ((sscanf(argv[6],"%f",&scale[0])!=1))
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_sge: line %d, "
		   "invalid number", lineNum);
	
	else if (argc == 9) {
	  if ((sscanf(argv[7],"%f",&scale[1])!=1) ||
	      (sscanf(argv[8],"%f",&scale[2])!=1))
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sge: line %d, "
		     "invalid number", lineNum);
	} else {
	  scale[1] = scale[2] = scale[0] ;
	} // if (argc == 9)
      } // if (argc >= 7)
    } // if (argc >= 6)
  } // if (argc >= 3)

  pfMatrix matrix ;
  matrix.makeCoord(&coord) ;
  matrix.preScale(scale[0], scale[1], scale[2], matrix) ;

  return (matrix) ;
}
