// Copyright  2002 Silicon Graphics, Inc.  All Rights Reserved.
// 
// The recipient ("Recipient") of this software, including as modified
// ("Software") may reproduce, redistribute, use, and derive works from the
// Software without restriction, subject to the following conditions:
//
//  Redistribution of the Software in any form must reproduce this entire
// notice, including as modified in accordance with these provisions
// ("Notice");
//
//  Any Recipient who modifies and subsequently redistributes the Software
// shall add information to this Notice to sufficiently identify the
// Recipient's modifications;
//
//  Recipient may not use the name(s) of any previous Recipient to endorse
// or promote any products derived from the Software without prior express
// written permission from such previous Recipient.
//
// THE SOFTWARE IS PROVIDED "AS IS," WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTY OR CONDITION, INCLUDING WITHOUT LIMITATION ANY WARRANTIES OR
// CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A
// PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  PATENT LICENSES, IF ANY,
// PROVIDED HEREIN DO NOT APPLY TO COMBINATIONS OF THIS PROGRAM WITH OTHER
// SOFTWARE, OR ANY OTHER PRODUCT WHATSOEVER.  IN NO EVENT WILL THE
// ORIGINATOR OR SUBSEQUENT RECIPIENT OF THE SOFTWARE BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES ARISING IN RELATION TO THE SOFTWARE, ITS USE, OR THESE
// PROVISIONS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR NEGLIGENCE OR OTHER TORT, EVEN IF
// PRE-ADVISED OF THE PROSPECT OF SUCH DAMAGES.
//
// This Software was developed at private expense; if acquired under an
// agreement with the USA government or any contractor thereto, it is
// acquired as "commercial computer software" subject to the provisions of
// this license agreement, as specified in (a) 48 CFR 12.212 of the FAR;
// or, if acquired for Department of Defense units, (b) 48 CFR 227-7202 of
// the DoD FAR Supplement; or sections succeeding thereto.
// 
// Originator:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy, Mountain
// View, CA  94043.  http://www.sgi.com

//////////////////////////////////////////////////////////////////////
// MiscUtil.cxx
//////////////////////////////////////////////////////////////////////

// local
#include "MiscUtil.h"

// C/ C++
#include <stdio.h>
#include <iostream.h>
#include <string.h>
#include <limits.h>
#include <math.h>

// GL
#include <GL/gl.h>
#include <GL/glu.h>

// X
#include <X11/X.h>
#include <GL/glx.h>

#define SQR(X) ((X)*(X))

//////////////////////////////////////////////////////////////////////
// Return the number of bytes per channel for a given texture type 
//////////////////////////////////////////////////////////////////////
int getBytesPerChannel(vzTextureType type)
{
  int bpc = 0;
  switch (type) 
    {
    case VZ_BYTE:
      bpc = sizeof (GLbyte);
      break;
    case VZ_UNSIGNED_BYTE:
      bpc = sizeof (GLubyte);
      break;
    case VZ_FLOAT: 
      bpc = sizeof (GLfloat);
      break;
    case VZ_UNSIGNED_SHORT:
      bpc = sizeof (GLushort);
      break;
    case VZ_SHORT:
      bpc = sizeof (GLshort);
      break; 
     case VZ_UNSIGNED_INT:
      bpc = sizeof (GLuint);
      break;
    case VZ_INT:
      bpc = sizeof (GLint);
      break; 
   default:
      cerr<<"getBytesPerChannel: Unsupported texture type"<<endl;
      return 0;
    };
  return bpc;
  
}

//////////////////////////////////////////////////////////////////////
// Return the number of channels for a given texture format 
//////////////////////////////////////////////////////////////////////
int getNumChannels(vzExternalTextureFormat format)
{
  int numchannels = 0;
  switch (format) 
    {
    case VZ_RGBA:
      numchannels = 4;
      break;
    case VZ_LUMINANCE_ALPHA:
      numchannels = 2;
      break;
    case VZ_LUMINANCE:
      numchannels = 1;
      break;
    default:
      cerr<<"getNumChannels: Unsupported texture format!"<<endl;
      return 0;
    };
  return numchannels;
}

//////////////////////////////////////////////////////////////////////
// Return the number of bytes per texel given the 
// vzParameterVolumeTexture 
//////////////////////////////////////////////////////////////////////
int getBytesPerTexel(vzParameterVolumeTexture *tex)
{
  vzExternalTextureFormat externalFormat = tex->getDataFormat();
  vzTextureType textureType = tex->getDataType();
  return( getBytesPerChannel(textureType) * getNumChannels(externalFormat) );
}

//////////////////////////////////////////////////////////////////////
// Return the texture type, given a string name
//////////////////////////////////////////////////////////////////////
vzTextureType getTextureType(const char *texTypeName)
{
    vzTextureType texType = VZ_UNSIGNED_BYTE;
    
    if(strcmp(texTypeName,"ubyte") == 0 ||
       strcmp(texTypeName,"UNSIGNED_BYTE") == 0 ||
       strcmp(texTypeName,"VZ_UNSIGNED_BYTE") == 0) {
      texType = VZ_UNSIGNED_BYTE;
      cerr<<"Type = "<<texTypeName<<" VZ_UNSIGNED_BYTE"<<endl;
      
    } else if(strcmp(texTypeName,"byte") == 0 ||
	      strcmp(texTypeName,"BYTE") == 0 ||
	      strcmp(texTypeName,"VZ_BYTE") == 0) {
      texType = VZ_BYTE;
      cerr<<"Type = "<<texTypeName<<" VZ_BYTE"<<endl;

    } else if(strcmp(texTypeName,"ushort") == 0 ||
	      strcmp(texTypeName,"UNSIGNED_SHORT") == 0 ||
	      strcmp(texTypeName,"VZ_UNSIGNED_SHORT") == 0) {
      texType = VZ_UNSIGNED_SHORT;
      cerr<<"Type = "<<texTypeName<<" VZ_UNSIGNED_SHORT"<<endl;

    } else if(strcmp(texTypeName,"short") == 0 ||
	      strcmp(texTypeName,"SHORT") == 0 ||
	      strcmp(texTypeName,"VZ_SHORT") == 0) {
      texType = VZ_SHORT;
      cerr<<"Type = "<<texTypeName<<" VZ_SHORT"<<endl;
      
    } else if(strcmp(texTypeName,"uint") == 0 ||
	      strcmp(texTypeName,"UNSIGNED_INT") == 0 ||
	      strcmp(texTypeName,"VZ_UNSIGNED_INT") == 0) {
      texType = VZ_UNSIGNED_INT;
      cerr<<"Type = "<<texTypeName<<" VZ_UNSIGNED_INT"<<endl;

    } else if(strcmp(texTypeName,"int") == 0 ||
	      strcmp(texTypeName,"INT") == 0 ||
	      strcmp(texTypeName,"VZ_INT") == 0) {
      texType = VZ_INT;
      cerr<<"Type = "<<texTypeName<<" VZ_INT"<<endl;
      
    } else if(strcmp(texTypeName,"float") == 0 ||
	      strcmp(texTypeName,"FLOAT") == 0 ||
	      strcmp(texTypeName,"VZ_FLOAT") == 0) {
      texType = VZ_FLOAT;
      cerr<<"Type = "<<texTypeName<<" VZ_FLOAT"<<endl;
      
    } else {
      cerr<<"Unrecognized texture type "<<texTypeName<<endl;
    }

    return texType;
}

//////////////////////////////////////////////////////////////////////
// Return the texture format, given a string name
//////////////////////////////////////////////////////////////////////
vzExternalTextureFormat getDataFormat(const char *formatName)
{
    vzExternalTextureFormat format = VZ_LUMINANCE;
    
    if(strcmp(formatName,"rgba") == 0 || 
       strcmp(formatName,"RGBA") == 0 ||
       strcmp(formatName,"VZ_RGBA") == 0) {
      format = VZ_RGBA;
      cerr<<"Format = "<<formatName<<" VZ_RGBA"<<endl;
      
    } else if(strcmp(formatName,"luminance") == 0 || 
	      strcmp(formatName,"LUMINANCE") == 0 ||
	      strcmp(formatName,"VZ_LUMINANCE") == 0) {
      format = VZ_LUMINANCE;
      cerr<<"Format = "<<formatName<<" VZ_LUMINANCE"<<endl;
      
    } else if(strcmp(formatName,"luminance_alpha") == 0 ||
	      strcmp(formatName,"LUMINANCE_ALPHA") == 0 ||
	      strcmp(formatName,"VZ_LUMINANCE_ALPHA") == 0) {
      format = VZ_LUMINANCE_ALPHA;
      cerr<<"Format = "<<formatName<<" VZ_LUMINANCE_ALPHA"<<endl;

    } else {
      cerr<<"Unrecognized texture format "<<formatName<<endl;
    }
    
    return format;
}

//////////////////////////////////////////////////////////////////////
// Return the volume texture, given the shape
//////////////////////////////////////////////////////////////////////
vzParameterVolumeTexture *getVolumeTexture(vzShape *shape)
{
  vzParameter *parameter = shape->getAppearance()->getParameter("volume");
  if(!parameter)
    {
      cerr<<"Parameter 'volume' not found "<<endl;
      return(NULL);
    }
  
  return(dynamic_cast<vzParameterVolumeTexture *> (parameter));
}

//////////////////////////////////////////////////////////////////////
// Return the texture table, given the shape
//////////////////////////////////////////////////////////////////////
vzParameterLookupTable *getLookupTable(vzShape *shape)
{
  vzParameter *parameter = shape->getAppearance()->getParameter("lookup_table");
  if(!parameter)
    {
      cerr<<"Parameter 'lookup_table' not found "<<endl;
      return(NULL);
    }
  
  return(dynamic_cast<vzParameterLookupTable *> (parameter));
}

//////////////////////////////////////////////////////////////////////
// Querry the graphics hardware for the amount of texture memory available
//////////////////////////////////////////////////////////////////////
unsigned int getTextureMemorySize(Platforms *platform)
{
    unsigned int computeTextureMemorySize(Platforms *platform);
    void createDummyContext(Display *&display, GLXContext &context);
  
    GLXContext currContext = glXGetCurrentContext();
    
    Display *display;
    GLXContext newContext;
    
    if(currContext  == NULL ) {
	
	vzError::message(2,"NULL GLXContext. Creating dummy context!!!");
	createDummyContext(display, newContext);
    }
    
    // Get the amount of texture memory available on the system 
    // Account for no interleaving 
    unsigned long texMemSize = computeTextureMemorySize(platform);
    
    // Destroy the new context
    if( currContext == NULL ) {
	glXMakeCurrent(display, None, NULL);
	glXDestroyContext(display, newContext);
	XCloseDisplay(display);
    }
    
    return texMemSize;
}

//////////////////////////////////////////////////////////////////////
// Querry the graphics hardware for the amount of texture memory available
//////////////////////////////////////////////////////////////////////
unsigned int computeTextureMemorySize(Platforms *platform)
{
    // The following does not use proxy textures to determine the
    // amount of texture memory available on a given graphics sub-system.
    unsigned int memorySize = 0;
    *platform = IMPACT;
    
    // We must have a valid GLContext here
    const GLubyte *renderer = glGetString(GL_RENDERER);
    if(renderer == 0) {
	cout<<"NULL renderer string returned "<<endl;
	return(0);
    }
    
    cout<<"Renderer string = "<<((const char *)renderer)<<endl;
    
#ifndef __linux__
    
    // Infinite Reality ??
    char *ptr = strstr((const char *) renderer, "IR");
    if(ptr != 0) {
	int ge, rm, texsz;
	char type, pixd;
	
	// Format => InfiniteReality: "<type>/<pixd>/<nrm>/<texsz>/<nge>"
	sscanf((const char *)renderer,"IR%c\/%c\/%d\/%d\/%d",&type, &pixd, &rm, &texsz, &ge);
	cout<<"Type = IR"<<type<<", PIXD = "<<pixd<<", RM = "<<rm<<", TEXSZ = "<<texsz<<", GE = "<<ge<<endl;
	
	// InfiniteReality - find the amount of texture memory
	if (strstr ((const char *) renderer, "256")) 
	    memorySize = 256 * 1024 * 1024;
	else
	    memorySize = 64 * 1024 * 1024;
	
	*platform = IR;
	
	return(memorySize);
    }
    
    // IMPACT ??
    ptr = strstr((const char *) renderer, "IMPACT");
    if(ptr != 0) {
	
	int ge, rss, texsz;
	
	// Format => Solid Impact, High Impact and Maximum Impact: 
	// "IMPACT/<nge>/<nrss>/<texsz>"
	sscanf((const char *)renderer,"IMPACT\/%d\/%d\/%d",&ge, &rss, &texsz);
	cout<<"GE = "<<ge<<", RSS = "<<rss<<", TEXSZ = "<<texsz<<endl;
	
	memorySize = texsz * 1024 * 1024;
	
	*platform = IMPACT;
	
	return(memorySize);
    }
    
    // Octane2 VPRO ??
    ptr = strstr((const char *) renderer, "VPRO");
    if(ptr != 0) {
	
	int memsz;
	char rev;
	
	// Format => Octane2 VPro: "VPRO/<rev>/<memsz>
	sscanf((const char *)renderer,"VPRO\/%c\/%d",&rev, &memsz);
	cout<<"VPRO:, Rev = "<<rev<<", MEMSZ = "<<memsz<<endl;
	
	memorySize = memsz * 1024 * 1024;
	
	*platform = VPRO;
	
	// XXX Use 75 % of the total graphics memory for textures 
	return(memorySize * 0.75);
    }
#else
    char *ptr = strstr((const char *) renderer, "FireGL");
    if(ptr != 0) {
	
	cout<<"FireGL graphics"<<endl;
	
	// Assume 64 MB of texture memory
	return 64 * 1024 * 1024;
    }
#endif
    
    return(0);
}

//////////////////////////////////////////////////////////////////////
// create a dummy GLX context to querry OpenGL renderer string
//////////////////////////////////////////////////////////////////////
void createDummyContext(Display *&display, GLXContext &context) 
{
    Window dummyWin;
    
    display = XOpenDisplay(NULL);
    
    if( display == NULL ) {
	vzError::error(VZ_INVALID_VALUE, "RoamVolume::cannot open display");
	return;
    }
    
    static int attr[] = {GLX_RGBA, None}; 
    XVisualInfo *info = glXChooseVisual(display, DefaultScreen(display), attr);
    
    if( info == NULL ) {
	vzError::error(VZ_INVALID_VALUE, 
		       "RoamVolume::cannot choose visual");
	return;
    }
    context = glXCreateContext(display, info, NULL, True);
    
    // create a dummy window so we can make current
    Colormap dummyCmap = XCreateColormap(display, RootWindow(display, DefaultScreen(display)), 
					 info->visual, AllocNone);
    
    XSetWindowAttributes swa;
    swa.colormap = dummyCmap;
    swa.border_pixel = 0;
    dummyWin = XCreateWindow(display, RootWindow(display, DefaultScreen(display)),
			     0, 0, 300, 300, 0,
			     info->depth, InputOutput, info->visual,
			     CWBorderPixel | CWColormap,
			     &swa);
    XFree(info);
    
    glXMakeCurrent(display, dummyWin, context);
}

//////////////////////////////////////////////////////////////////////
// Render the faces of the given brick 
//////////////////////////////////////////////////////////////////////
void renderCube(double coords[6])
{
  // Using GL_QUADS for each of the brick faces
  // To render wireframe, set the glPolygonMode appropriately
  glBegin(GL_QUADS);
  
  // Face 0
  glVertex3d(coords[0], coords[1], coords[2]);
  glVertex3d(coords[3], coords[1], coords[2]);
  glVertex3d(coords[3], coords[4], coords[2]);
  glVertex3d(coords[0], coords[4], coords[2]);

  // Face 1 
  glVertex3d(coords[0], coords[1], coords[5]);
  glVertex3d(coords[3], coords[1], coords[5]);
  glVertex3d(coords[3], coords[4], coords[5]);
  glVertex3d(coords[0], coords[4], coords[5]);

  // Face 2
  glVertex3d(coords[0], coords[1], coords[2]);
  glVertex3d(coords[0], coords[1], coords[5]);
  glVertex3d(coords[0], coords[4], coords[5]);
  glVertex3d(coords[0], coords[4], coords[2]);

  // Face 3
  glVertex3d(coords[3], coords[1], coords[2]);
  glVertex3d(coords[3], coords[1], coords[5]);
  glVertex3d(coords[3], coords[4], coords[5]);
  glVertex3d(coords[3], coords[4], coords[2]);

  // Face 4
  glVertex3d(coords[0], coords[4], coords[2]);
  glVertex3d(coords[3], coords[4], coords[2]);
  glVertex3d(coords[3], coords[4], coords[5]);
  glVertex3d(coords[0], coords[4], coords[5]);

  // Face 5
  glVertex3d(coords[0], coords[1], coords[2]);
  glVertex3d(coords[3], coords[1], coords[2]);
  glVertex3d(coords[3], coords[1], coords[5]);
  glVertex3d(coords[0], coords[1], coords[5]);

  glEnd();
}

int isPowerOfTwo (int value) 
{
    return ( (value&(value-1)) == 0 );
}

int augmentPowerOfTwo(int x)
{
    if (isPowerOfTwo(x)) return x;
    else return truncatePowerOfTwo (x)*2;
}

int truncatePowerOfTwo(int value)
{
    if (value <= 0) return 0;
    unsigned int i, p;
    for (i=value, p=1; i>1; i>>=1, p<<=1) ;
    return p;
}
 
