// Copyright  2002 Silicon Graphics, Inc.  All Rights Reserved.
// 
// The recipient ("Recipient") of this software, including as modified
// ("Software") may reproduce, redistribute, use, and derive works from the
// Software without restriction, subject to the following conditions:
//
//  Redistribution of the Software in any form must reproduce this entire
// notice, including as modified in accordance with these provisions
// ("Notice");
//
//  Any Recipient who modifies and subsequently redistributes the Software
// shall add information to this Notice to sufficiently identify the
// Recipient's modifications;
//
//  Recipient may not use the name(s) of any previous Recipient to endorse
// or promote any products derived from the Software without prior express
// written permission from such previous Recipient.
//
// THE SOFTWARE IS PROVIDED "AS IS," WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTY OR CONDITION, INCLUDING WITHOUT LIMITATION ANY WARRANTIES OR
// CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A
// PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  PATENT LICENSES, IF ANY,
// PROVIDED HEREIN DO NOT APPLY TO COMBINATIONS OF THIS PROGRAM WITH OTHER
// SOFTWARE, OR ANY OTHER PRODUCT WHATSOEVER.  IN NO EVENT WILL THE
// ORIGINATOR OR SUBSEQUENT RECIPIENT OF THE SOFTWARE BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES ARISING IN RELATION TO THE SOFTWARE, ITS USE, OR THESE
// PROVISIONS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR NEGLIGENCE OR OTHER TORT, EVEN IF
// PRE-ADVISED OF THE PROSPECT OF SUCH DAMAGES.
//
// This Software was developed at private expense; if acquired under an
// agreement with the USA government or any contractor thereto, it is
// acquired as "commercial computer software" subject to the provisions of
// this license agreement, as specified in (a) 48 CFR 12.212 of the FAR;
// or, if acquired for Department of Defense units, (b) 48 CFR 227-7202 of
// the DoD FAR Supplement; or sections succeeding thereto.
// 
// Originator:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy, Mountain
// View, CA  94043.  http://www.sgi.com


//////////////////////////////////////////////////////////////////////
// TFEditor.cxx
//////////////////////////////////////////////////////////////////////

// C
#include <math.h>
#include <stdio.h>
#include <string.h>

// C++
#include <iostream.h>
#include <fstream.h>

// local
#include "TFEditor.h"
#include "TFManager.h"

// Motif
#include <Xm/MwmUtil.h>

//////////////////////////////////////////////////////////////////////
// constants
//////////////////////////////////////////////////////////////////////

const float
    TFEditor::s_height = 0.60,
    TFEditor::s_width  = 0.90,
    TFEditor::s_top    = 0.05,
    TFEditor::s_bottom = 0.05,
    TFEditor::s_left   = 0.05,
    TFEditor::s_right  = 0.05;

//////////////////////////////////////////////////////////////////////
// constructor
//////////////////////////////////////////////////////////////////////

TFEditor::TFEditor(Widget parent, int width, TFFormat format)
    :
    m_manager(TFManager::create(format, width)),
    m_parent(parent),
    m_windowLevelEnabled(false),
    m_userFunction(NULL),
    m_paintBrushUp(true),
    m_levelValue(width / 2),
    m_windowValue(width),
    m_lastIndex(0),
    m_lastValue(0.0),
    m_popped(false)
{
    m_viewport[0] = 0;
    m_viewport[1] = 0;
    m_viewport[2] = 400;
    m_viewport[3] = 200;

    create();

    // Luminance is the current channel by default
    if (format == TF_RGBA) {
        setCurrentChannel(TFManager::CHANNEL_RED);
    } else {
        setCurrentChannel(TFManager::CHANNEL_LUMINANCE);
    }
}

//////////////////////////////////////////////////////////////////////
// destructor
//////////////////////////////////////////////////////////////////////

TFEditor::~TFEditor()
{
    delete m_manager;
}

//////////////////////////////////////////////////////////////////////
// create
//////////////////////////////////////////////////////////////////////

void TFEditor::create()
{
    // font
    XFontStruct* font = XLoadQueryFont(XtDisplay(m_parent),
        "-*-helvetica-medium-r-normal--*-100-*-*-*-*-*-*");

    XmFontList fontList = XmFontListCreate(font,
        XmSTRING_DEFAULT_CHARSET);

    // shell
    m_shell = XtVaAppCreateShell("TFEditor", "TFEditor",
        topLevelShellWidgetClass, XtDisplay(m_parent), 
        XmNwidth,           400,
        XmNheight,          300,
        XtNminHeight,       200,
        XtNminWidth,        380,
        XmNdefaultFontList, fontList,
        XmNdeleteResponse, XmDO_NOTHING,
        NULL);

    // handle delete window
    Atom deleteWindow = XmInternAtom(XtDisplay(m_shell),
        "WM_DELETE_WINDOW", False);

    // set callbacks
    XmAddWMProtocolCallback(m_shell, deleteWindow,
        quitCallback, this);

    // form 
    Widget shellForm = XtVaCreateManagedWidget("shell_form",
        xmFormWidgetClass, m_shell,
        NULL);
  
    XmString
        FILE_str   = XmStringCreateSimple("File"),
        MODE_str   = XmStringCreateSimple("Interpolation"),
        FORMAT_str = XmStringCreateSimple("Format"),
	BUILTIN_str= XmStringCreateSimple("Built-in LUTs");
 
    // menu bar 
    Widget menuBar = XmVaCreateSimpleMenuBar(shellForm,
        "menu_bar",
        XmVaCASCADEBUTTON,  FILE_str,   XK_F,
        XmVaCASCADEBUTTON,  MODE_str,   XK_I,
        XmVaCASCADEBUTTON,  FORMAT_str, XK_o,
        XmVaCASCADEBUTTON,  BUILTIN_str,XK_b,
        XmNtopAttachment,   XmATTACH_FORM,
        XmNleftAttachment,  XmATTACH_FORM,
        XmNrightAttachment, XmATTACH_FORM,
        NULL);
  
    XmStringFree(FILE_str);
    XmStringFree(MODE_str);
    XmStringFree(FORMAT_str);
    XmStringFree(BUILTIN_str);

    XmString
        CLOSE_str = XmStringCreateSimple("Close"),
        SAVE_str  = XmStringCreateSimple("Save"),
        LOAD_str  = XmStringCreateSimple("Load");
 
    // pull down menu 
    Widget fileMenu = XmVaCreateSimplePulldownMenu(menuBar,
        "file_pulldown", 0,
        filePulldownCallback,
        XmVaPUSHBUTTON, LOAD_str,  NULL, NULL, NULL,
        XmVaPUSHBUTTON, SAVE_str,  NULL, NULL, NULL, 
        XmVaPUSHBUTTON, CLOSE_str, NULL, NULL, NULL,
        XmNuserData,    this,
        NULL);   		       
  
    XmStringFree(CLOSE_str);
  
    XmString
        window_level_str = XmStringCreateSimple("Window/Level"),
        linear_str       = XmStringCreateSimple("linear"),
        spline_str       = XmStringCreateSimple("splines"),
        gaussian_str     = XmStringCreateSimple("gaussian"),
        paint_str        = XmStringCreateSimple("paint");

    // interpolation mode menu
    m_interpolation = XmVaCreateSimplePulldownMenu(menuBar,
        "interp_mode_pulldown", 1, setTFModeCallback,
        XmVaRADIOBUTTON,   window_level_str, NULL, NULL, NULL,
        XmVaRADIOBUTTON,   linear_str,       NULL, NULL, NULL,
        XmVaRADIOBUTTON,   spline_str,       NULL, NULL, NULL,
        XmVaRADIOBUTTON,   gaussian_str,     NULL, NULL, NULL,
        XmVaRADIOBUTTON,   paint_str,        NULL, NULL, NULL,
        XmNradioBehavior,  TRUE,
        XmNradioAlwaysOne, TRUE,
        XmNuserData,       this,
        NULL);   	       
  
    XmStringFree(window_level_str);
    XmStringFree(linear_str);	
    XmStringFree(spline_str); 
    XmStringFree(gaussian_str);       
    XmStringFree(paint_str);  
  
    XmString
        luminance_str = XmStringCreateSimple("Luminance"),
        la_str        = XmStringCreateSimple("Luminance - Alpha"),
        rgba_str      = XmStringCreateSimple("RGBA");

    // format pulldown menu
    m_format = XmVaCreateSimplePulldownMenu(menuBar,
        "format_mode_pulldown", 2, setTFFormatCallback,
        XmVaRADIOBUTTON,   luminance_str, NULL, NULL, NULL,
        XmVaRADIOBUTTON,   la_str,        NULL, NULL, NULL,
        XmVaRADIOBUTTON,   rgba_str,      NULL, NULL, NULL,
        XmNradioBehavior,  TRUE,
        XmNradioAlwaysOne, TRUE,
        XmNuserData,       this,
        NULL);   	

    // initialize the motif state for the default mode:
    Widget button;
    switch (m_manager->getFormat()) {
    case TF_RGBA:
        m_currentChannel = TFManager::CHANNEL_RED;
        button = XtNameToWidget(m_format, "button_0");
        break;

    case TF_LUMINANCE_ALPHA:
        m_currentChannel = TFManager::CHANNEL_LUMINANCE;
        button = XtNameToWidget(m_format, "button_1");
        break;

    case TF_LUMINANCE:
        m_currentChannel = TFManager::CHANNEL_LUMINANCE;
        button = XtNameToWidget(m_format, "button_2");
        break;  
    }
    
    XtVaSetValues(button, XmNset, True, NULL);

    XmStringFree(luminance_str);
    XmStringFree(la_str);
    XmStringFree(rgba_str);
    
    XmString
        gray_scale_str = XmStringCreateSimple("Gray Scale"),
        hot_metal_str  = XmStringCreateSimple("Hot Metal"),
        spectral_str   = XmStringCreateSimple("Spectral"),
        contour_str    = XmStringCreateSimple("Contour");
    
    // format pulldown menu
    m_builtIn = XmVaCreateSimplePulldownMenu(menuBar,
        "format_mode_pulldown", 3, useBuiltInCallback,
        XmVaRADIOBUTTON,   gray_scale_str, NULL, NULL, NULL,
        XmVaRADIOBUTTON,   hot_metal_str, NULL, NULL, NULL,
        XmVaRADIOBUTTON,   spectral_str,  NULL, NULL, NULL,
        XmVaRADIOBUTTON,   contour_str,   NULL, NULL, NULL,
        XmNradioBehavior,  TRUE,
        XmNradioAlwaysOne, TRUE,
        XmNuserData,       this,
        NULL); 
    
    // initialize the motif state for the default mode:
    Widget builtInButton = XtNameToWidget(m_builtIn, "button_0");
    XtVaSetValues(builtInButton, XmNset, True, NULL);

    XmStringFree(gray_scale_str);
    XmStringFree(hot_metal_str);
    XmStringFree(spectral_str);
    XmStringFree(contour_str);
    
    XtManageChild(m_interpolation);
    XtManageChild(m_format);
    XtManageChild(m_builtIn);
    XtManageChild(fileMenu);
    XtManageChild(menuBar);
  
    // plot_area 
    GLint visualattribs[] = {GLX_RGBA, GLX_DOUBLEBUFFER, None};
    XVisualInfo* visinfo = glXChooseVisual(XtDisplay(m_shell), 0,
        visualattribs);

    if (!visinfo)
        cerr << "Bad Visual" << endl;
  
    Widget mainPanedWindow = XtVaCreateWidget("main_paned_window", 
        xmPanedWindowWidgetClass, shellForm,
        XmNleftAttachment,   XmATTACH_FORM,
        XmNrightAttachment,  XmATTACH_FORM,
        XmNtopWidget,        menuBar,
        XmNtopAttachment,    XmATTACH_WIDGET,
        XmNbottomAttachment, XmATTACH_FORM,
        NULL);
    
    Widget upperRowCol = XtVaCreateManagedWidget("upper_rowcol",
        xmRowColumnWidgetClass, mainPanedWindow,
        XmNpaneMinimum, 40,
        XmNpaneMaximum, 40,
        XmNheight,      40,
        XmNorientation, XmHORIZONTAL,
        NULL);

    XColor color, unused;
    XAllocNamedColor(XtDisplay(m_parent),
        DefaultColormapOfScreen(XtScreen(m_parent)), "grey20",
            &color, &unused);

    Pixel darkgrey = color.pixel;

    Widget plotAreaFrame = XtVaCreateManagedWidget("plot_area_frame",
        xmFrameWidgetClass, mainPanedWindow,
        XmNwidth,      400,
        XmNheight,     205,
        XmNshadowType, XmSHADOW_IN,
        XmNbackground, darkgrey,
        NULL);

    m_plotArea = XtVaCreateWidget("plot_area",
        glwMDrawingAreaWidgetClass, plotAreaFrame,
        GLwNvisualInfo, visinfo,
        XmNwidth,       400,
        XmNheight,      200,
        XmNbackground,  darkgrey,
        NULL);

    XtVaCreateManagedWidget("Channel:",
        xmLabelWidgetClass, upperRowCol,
        NULL);
  
    XtManageChild(m_plotArea);
    XtManageChild(mainPanedWindow);
  
    Widget channelForm = XtVaCreateWidget("channel_form",
        xmFormWidgetClass, upperRowCol,
        NULL);
  
    XtManageChild(channelForm);	       
  
    // rgba_radiobuttons
    XmString
        red_button_str       = XmStringCreate("Red",
            "-*-helvetica-medium-r-normal--*-100-*-*-*-*-*-*"),
        green_button_str     = XmStringCreateSimple("Green"),
        blue_button_str      = XmStringCreateSimple("Blue"),
        luminance_button_str = XmStringCreateSimple("Luminance"),
        alpha_button_str     = XmStringCreateSimple("Alpha");
  
    m_l = XmVaCreateSimpleRadioBox(channelForm,
        "l_radiobuttons", 0, setCurrentChannelCallback,
        XmVaRADIOBUTTON, luminance_button_str, NULL, NULL, NULL,
        XmNuserData,     this,
        XmNorientation,  XmHORIZONTAL,
        NULL);  

    m_la = XmVaCreateSimpleRadioBox(channelForm,
        "la_radiobuttons", 0, setCurrentChannelCallback,
        XmVaRADIOBUTTON, luminance_button_str, NULL, NULL, NULL,
        XmVaRADIOBUTTON, alpha_button_str,     NULL, NULL, NULL,
        XmNuserData,     this,
        XmNorientation,  XmHORIZONTAL,
        NULL);   
  
    m_rgba = XmVaCreateSimpleRadioBox(channelForm,
        "rgba_radiobuttons", 0, setCurrentChannelCallback,
        XmVaRADIOBUTTON, red_button_str,   NULL, NULL, NULL,
        XmVaRADIOBUTTON, green_button_str, NULL, NULL, NULL,
        XmVaRADIOBUTTON, blue_button_str,  NULL, NULL, NULL,
        XmVaRADIOBUTTON, alpha_button_str, NULL, NULL, NULL,
        XmNuserData,     this,
        XmNorientation,  XmHORIZONTAL,
        NULL);   

    // rowColumn_window_level - containing scale_window, scale_level 
    XmString
        window_str = XmStringCreateSimple("Window"),
        level_str  = XmStringCreateSimple("Level");
  
    m_windowLevelShell = XtVaAppCreateShell("WindowLevel",
        "WindowLevel", topLevelShellWidgetClass, XtDisplay(m_shell),
        XmNwidth,     225,
        XmNheight,    150,
        XmNminWidth,  225,
        XmNminHeight, 150,
        XmNdefaultFontList, fontList, 
        NULL);

    // allow resizing and moving, but not closing of the window.
    int funcs = MWM_FUNC_MOVE | MWM_FUNC_RESIZE;
    XtVaSetValues (m_windowLevelShell, XmNmwmFunctions, funcs, NULL);

    Widget windowLevelForm = XtVaCreateManagedWidget("shell_form",
        xmFormWidgetClass, m_windowLevelShell,
        XmNwidth,  225,
        XmNheight, 150,
        NULL);      

    Widget windowLevelPanedWindow = XtVaCreateWidget("main_paned_window", 
        xmPanedWindowWidgetClass, windowLevelForm,
        XmNleftAttachment,   XmATTACH_FORM,
        XmNrightAttachment,  XmATTACH_FORM,
        XmNtopAttachment,    XmATTACH_FORM,
        XmNbottomAttachment, XmATTACH_FORM,
        NULL);  

    XtManageChild(windowLevelPanedWindow);

    Widget rowColumnWindowLevel = XtVaCreateWidget("rowColumn_window_level",
        xmRowColumnWidgetClass,
        windowLevelPanedWindow, 
        XmNorientation, XmHORIZONTAL,
        NULL);

    XmStringFree(red_button_str); 
    XmStringFree(green_button_str);
    XmStringFree(blue_button_str);
    XmStringFree(alpha_button_str);   
    XmStringFree(luminance_button_str); 

    /* scales for window/level lookup*/
    m_scaleWindow = XtVaCreateWidget("scale_window",
        xmScaleWidgetClass,
        rowColumnWindowLevel,
        XmNorientation, XmVERTICAL,
        XmNtitleString, window_str,
        XmNmaximum,      256,
        XmNminimum,     -256,
        XmNvalue,        256,
        XmNshowValue,   True,
        XmNuserData,    this,
        NULL);     
       
    m_scaleLevel = XtVaCreateWidget("scale_level", 
        xmScaleWidgetClass, rowColumnWindowLevel,
        XmNorientation, XmVERTICAL,
        XmNtitleString, level_str,
        XmNmaximum,     256,
        XmNminimum,     1,
        XmNvalue,       127,
        XmNshowValue,   True,
        XmNuserData,    this,
        NULL);   

    XtManageChild(m_scaleLevel);
    XtManageChild(m_scaleWindow);
    XtManageChild(channelForm);
    XtManageChild(rowColumnWindowLevel);
  
    // now we must decide what format we're working with
    switch (m_manager->getFormat()) {
    case TF_LUMINANCE:
        XtManageChild(m_l);
        break;     
      
    case TF_LUMINANCE_ALPHA:
        XtManageChild(m_la);	
        break;     
      
    case TF_RGBA:		
        XtManageChild(m_rgba);
        break;
      
    default:
        break;
    }   

    // add OpenGL window callbacks    
    XtAddCallback(m_plotArea, GLwNginitCallback,  initCallback,   this);
    XtAddCallback(m_plotArea, GLwNexposeCallback, exposeCallback, this);
    XtAddCallback(m_plotArea, GLwNinputCallback,  inputCallback,  this); 
    XtAddCallback(m_plotArea, GLwNresizeCallback, resizeCallback, this);
   
    // scale callbacks
    XtAddCallback(m_scaleWindow,
        XmNdragCallback, scaleWindowCallback, 
        (XtPointer) &m_windowValue);
    XtAddCallback(m_scaleLevel,
        XmNdragCallback, scaleLevelCallback, 
        (XtPointer) &m_levelValue);
    XtAddCallback(m_scaleWindow,
        XmNvalueChangedCallback, scaleWindowCallback, 
        (XtPointer) &m_windowValue);
    XtAddCallback(m_scaleLevel,
        XmNvalueChangedCallback, scaleLevelCallback,
        (XtPointer) &m_levelValue);

    XtManageChild(shellForm);
    XtManageChild(windowLevelForm);
    XtPopdown(m_windowLevelShell);

    // initialize motif state for default interpolation mode
    setCurrentInterpolationType(getCurrentChannelPtr()->getInterpolation());
  
    XtSetKeyboardFocus(m_shell, m_plotArea);
}

//////////////////////////////////////////////////////////////////////
// file pulldown callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::filePulldownCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void* classPtr;  
    XtVaGetValues (XtParent(widget), XmNuserData, &classPtr, NULL);

    ((TFEditor*) classPtr)->filePulldown(widget, clientData,
        callData);
}

void
TFEditor::filePulldown(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    long which_button = (long) clientData;

    switch (which_button) {
    // load
    case 0:
        fileMenuLoad(m_shell);
        break;

    // save
    case 1:
        fileMenuSave(m_shell);
        break;

    // quit
    case 2:
        XtPopdown(m_shell);
        XtPopdown(m_windowLevelShell);
        break;

    default:
        break;
    }
}

//////////////////////////////////////////////////////////////////////
// quit callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::quitCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    ((TFEditor*) clientData)->popdown();

}

void
TFEditor::popdown()
{
    XtPopdown(m_shell); 
    XtPopdown(m_windowLevelShell);

    m_popped = false;
}

//////////////////////////////////////////////////////////////////////
// save
//////////////////////////////////////////////////////////////////////

void
TFEditor::fileMenuSave(Widget widget)
{
    Widget dialog = XmCreateFileSelectionDialog(widget, "fselbox",
        NULL, 0);
   
    XtManageChild(dialog);
    XtAddCallback(dialog, XmNokCallback, saveLUTCallback, this);
    XtAddCallback(dialog, XmNcancelCallback,
        (XtCallbackProc) XtUnmanageChild, NULL);
}

//////////////////////////////////////////////////////////////////////
// load
//////////////////////////////////////////////////////////////////////

void
TFEditor::fileMenuLoad(Widget widget)
{
    Widget dialog = XmCreateFileSelectionDialog (widget, "fselbox",
        NULL, 0);
    
    XtManageChild(dialog);
    XtAddCallback(dialog, XmNokCallback, loadLUTCallback, this);
    XtAddCallback(dialog, XmNcancelCallback,
        (XtCallbackProc) XtUnmanageChild, NULL);
}

//////////////////////////////////////////////////////////////////////
// mode callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::setTFModeCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void* classPtr;  
    XtVaGetValues(XtParent(widget), XmNuserData, &classPtr, NULL);

    ((TFEditor*) classPtr)->setTFMode(widget, clientData,
        callData);
}

void TFEditor::setTFMode(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    long which_button = (long) clientData;
    TFChannel::Interpolation interp =
        (TFChannel::Interpolation) which_button;
   
    setCurrentInterpolationType(interp);
    getCurrentChannelPtr()->setInterpolation(interp);

    draw();
    callUserFunction();
}

//////////////////////////////////////////////////////////////////////
// format callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::setTFFormatCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void *classPtr;  
    XtVaGetValues(XtParent(widget), XmNuserData, &classPtr, NULL);

    ((TFEditor*) classPtr)->setTFFormat(widget, clientData, callData);
}

void
TFEditor::setTFFormat(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    long which_button = (long) clientData;

    TFFormat format = TFFormat(which_button);
    if (format == m_manager->getFormat()) 
        return;

    setCurrentFormat(format);
    m_manager->setFormat(format);

    draw();
    callUserFunction();
}

//////////////////////////////////////////////////////////////////////
// built-in LUT callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::useBuiltInCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void *classPtr;  
    XtVaGetValues(XtParent(widget), XmNuserData, &classPtr, NULL);

    ((TFEditor*) classPtr)->useBuiltInLUT(widget, clientData, callData);
}

void
TFEditor::useBuiltInLUT(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    long which_button = (long) clientData;
    
    TFManager::Function function = (TFManager::Function) which_button;
    
    m_manager->createBuiltInLUT(function);

    setCurrentFormat(m_manager->getFormat());

    draw();
    callUserFunction();
}

//////////////////////////////////////////////////////////////////////
// current channel callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::setCurrentChannelCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void* classPtr;  
    XtVaGetValues (XtParent(widget), XmNuserData, &classPtr, NULL);

    ((TFEditor*) classPtr)->setCurrentChannel2(widget, clientData,
        callData);
}  

void
TFEditor::setCurrentChannel2(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    TFManager::Channel type;
    long which_button = (long) clientData;

    // RGBA
    if (m_manager->getFormat() == TF_RGBA) {
        switch (which_button) {
        case 0:
            type = TFManager::CHANNEL_RED;
            break;

        case 1:
            type = TFManager::CHANNEL_GREEN;
            break;

        case 2:
            type = TFManager::CHANNEL_BLUE;
            break;

        case 3:
            type = TFManager::CHANNEL_ALPHA;
            break;

        default:
            break;
        }

    // LUMINANCE or LUMINANCE_ALPHA
    } else {
        if (which_button == 0)
            type = TFManager::CHANNEL_LUMINANCE;
        else
            type = TFManager::CHANNEL_ALPHA;
    }

    setCurrentChannel(type);
    draw();
}
//////////////////////////////////////////////////////////////////////
// init callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::initCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{   
    ((TFEditor*) clientData)->init(widget, clientData,
        callData);
}

void
TFEditor::init(Widget widget, XtPointer clientData,
    XtPointer callData)
{   
    // get visual
    Arg arg;
    XVisualInfo* vi;
    XtSetArg(arg, GLwNvisualInfo, &vi);
    XtGetValues(widget, &arg, 1);

    // create context
    m_context = glXCreateContext(XtDisplay(widget), vi, NULL, GL_TRUE);

    draw();
}

//////////////////////////////////////////////////////////////////////
// expose callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::exposeCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    ((TFEditor*) clientData)->expose(widget, clientData, callData);
}

void
TFEditor::expose(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    GLwDrawingAreaCallbackStruct* csp = 
        (GLwDrawingAreaCallbackStruct*) callData;

    m_viewport[0] = 0;
    m_viewport[1] = 0;
    m_viewport[2] = (GLint) csp->width;
    m_viewport[3] = (GLint) csp->height;

    draw();
}

//////////////////////////////////////////////////////////////////////
// input
//////////////////////////////////////////////////////////////////////

void
TFEditor::inputCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    ((TFEditor*) clientData)->input(widget, clientData,
        callData);
}

void
TFEditor::input(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    XmDrawingAreaCallbackStruct* dacs =
        (XmDrawingAreaCallbackStruct *) callData;
 
    switch (dacs->event->type) {
    case ButtonPress: 
        GL_area_button_input((XButtonEvent*) dacs->event);
        break;
    
    case ButtonRelease:
        m_paintBrushUp = true;
        break;
    
    case MotionNotify:
        GL_area_motion_input((XMotionEvent *) dacs->event);
        break;
    
    case KeyPress:      
        break;
    
    case KeyRelease:
        return;  // do nothing on key release
    
    default:
        break;
    }
}

//////////////////////////////////////////////////////////////////////
// resize callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::resizeCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    ((TFEditor*) clientData)->resize(widget, clientData,
        callData);
}

void
TFEditor::resize(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    GLwDrawingAreaCallbackStruct* csp = 
        (GLwDrawingAreaCallbackStruct*) callData;

    m_viewport[0] = 0;
    m_viewport[1] = 0;
    m_viewport[2] = (GLint) csp->width;
    m_viewport[3] = (GLint) csp->height;
  
    XtVaSetValues(m_plotArea,
        XmNwidth,  m_viewport[2],
        XmNheight, m_viewport[3],
        NULL);

    XtVaSetValues(m_scaleWindow,
        XmNvalue, (int) m_windowValue,
        NULL);

    XtVaSetValues(m_scaleLevel,
        XmNvalue, (int) m_levelValue,
        NULL);
}

//////////////////////////////////////////////////////////////////////
// scale window callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::scaleWindowCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void* classPtr;  
    XtVaGetValues(widget, XmNuserData, &classPtr, NULL);

    ((TFEditor*) classPtr)->scaleWindow(widget, clientData,
        callData); 
}

void
TFEditor::scaleWindow(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    XmScaleGetValue(widget, &m_windowValue);
    getCurrentChannelPtr()->setRampWidth(m_windowValue);

    draw();
    callUserFunction();
}

//////////////////////////////////////////////////////////////////////
// scale level callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::scaleLevelCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    void* classPtr;  

    XtVaGetValues(widget, XmNuserData, &classPtr, NULL);
    ((TFEditor*)classPtr)->scaleLevel(widget, classPtr, callData); 
}

void
TFEditor::scaleLevel(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    XmScaleGetValue(widget, &m_levelValue);
    getCurrentChannelPtr()->setRampCenter(m_levelValue);

    draw();
    callUserFunction();
}

//////////////////////////////////////////////////////////////////////
// popup
//////////////////////////////////////////////////////////////////////

void
TFEditor::popup() 
{
    XtPopup(m_shell, XtGrabNone);
   
    if (m_windowLevelEnabled)
        XtPopup(m_windowLevelShell, XtGrabNone);
    
    m_popped = true;
}

//////////////////////////////////////////////////////////////////////
// print
//////////////////////////////////////////////////////////////////////

void
TFEditor::print()
{
    int num_channels;
    switch (m_manager->getFormat()) {
    case TF_LUMINANCE:
        num_channels = 1;
        break;

    case TF_LUMINANCE_ALPHA:
        num_channels = 2;
        break;
    
    case TF_RGBA:
        num_channels = 4;
        break;
    }

    const float
        length = m_manager->getLength(),
        *data = m_manager->getData();

    for (int i = 0; i < length; i++) {
        for (int j = 0; j < num_channels; j++)
            cerr << data[(i * num_channels) + j];

        cerr << endl;
    }
}

//////////////////////////////////////////////////////////////////////
// mouse button
//////////////////////////////////////////////////////////////////////

void
TFEditor::GL_area_button_input(XButtonEvent* event)
{
    int   newindex = x_plot_coord_2_tf_index(event->x); 
    float newvalue = y_plot_coord_2_tf_value(event->y); 
    int   xTolerance =
      (int) floor(6.0 / (float) m_viewport[2] * m_manager->getLength());
    float yTolerance = 6.0 / m_viewport[3];

    if (getCurrentChannelPtr()->getInterpolation() !=
        TFChannel::INTERPOLATION_PAINT)
    {
        TFChannel* channel = getCurrentChannelPtr();
        TFControlPointList::iterator point;

        switch (event->button) {
        // select existing control point
        case Button1:
            point = channel->findControlPoint(newindex, newvalue,
                xTolerance, yTolerance);

            channel->setCurrentControlPoint(point);
       
            // don't notify user 
            draw();
            return;

        // create new control point
        case Button2:
            // can't make a new control point with the same X value
            // as an existing control point (any y value is ok though)
            point = channel->findControlPoint(newindex, 0.5, 0, 1.0);
        
            // add control point and make it current
            if (point == NULL) {
                TFControlPoint* newpoint =
                    new TFControlPoint(newindex, newvalue);

                TFControlPointIterator newIter =
                    channel->addControlPoint(*newpoint);
                channel->setCurrentControlPoint(newIter);
            }
            break;

        // delete current control point
        case Button3:
            channel->deleteControlPoint();
            break;
        }
    }

    draw();
    callUserFunction();
} 

//////////////////////////////////////////////////////////////////////
// motion input callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::GL_area_motion_input(XMotionEvent* event)
{
    switch (event->state) {
    case Button1MotionMask:
        if (getCurrentChannelPtr()->getInterpolation() ==
            TFChannel::INTERPOLATION_PAINT)
        { 
            motionButton1Paint(event->x, event->y);
        } else {
            motionButton1(event->x, event->y);
        }
        break;

    case Button2MotionMask:
    case Button3MotionMask:
    default:
        break;
    }

    draw();
    callUserFunction();
}

//////////////////////////////////////////////////////////////////////
// mouse motion callback for non-paint interpolation
//////////////////////////////////////////////////////////////////////

void
TFEditor::motionButton1(int xpos, int ypos)
{
    int newindex = x_plot_coord_2_tf_index((float) xpos); 
    float newvalue = y_plot_coord_2_tf_value((float) ypos);

    TFChannel* channel = getCurrentChannelPtr();
    TFControlPointIterator current =
        channel->getCurrentControlPoint();

    // make sure current control point is not NULL
    if (current == NULL)
        return;

    if ((current->m_index > 0) &&
        (current->m_index < m_manager->getLength() - 1) )
    {
        TFControlPointIterator next, prev;
        next = current; next++;
        prev = current; prev--;

        if ((prev->m_index < newindex) && (newindex < next->m_index))
        {
            current->m_index = newindex;
            current->m_value = newvalue;

        } else if (newindex <= prev->m_index) {
            current->m_index = prev->m_index + 1;
            current->m_value = newvalue;

        } else if (next->m_index <= newindex ) {
            current->m_index = next->m_index - 1;
            current->m_value = newvalue;
        }

    } else if (channel->getCurrentControlPoint()->m_index == 0) {
        current->m_value = newvalue;

    } else if (channel->getCurrentControlPoint()->m_index == 
        m_manager->getLength() - 1)
    {
        current->m_value = newvalue;
    }
 
    channel->setDirty(true);
}

//////////////////////////////////////////////////////////////////////
// mouse motion callback for paint interpolation
//////////////////////////////////////////////////////////////////////

void
TFEditor::motionButton1Paint(int xpos, int ypos)
{
    int   newindex = x_plot_coord_2_tf_index((float) xpos); 
    float newvalue = y_plot_coord_2_tf_value((float) ypos);

    // if paint brush is currently up, put it down!
    if (m_paintBrushUp == true) {
        m_lastIndex = newindex;
        m_lastValue = newvalue;
        m_paintBrushUp = false;
    }

    int   low_index, hi_index;
    float low_value, hi_value;

    if (m_lastIndex < newindex) {
        low_index = m_lastIndex;
        low_value = m_lastValue;
        hi_index = newindex;
        hi_value = newvalue;

    } else {
        low_index = newindex;
        low_value = newvalue;
        hi_index = m_lastIndex;
        hi_value = m_lastValue;
    }
  
    if (low_index == hi_index) {
        getCurrentChannelPtr()->setValue(low_index, low_value);

    } else {
        for (int i=low_index; i<=hi_index; i++) {
            const float yval =
                (((hi_value - low_value) * 
                (((float) i - (float) low_index) /
                ((float) hi_index - (float) low_index))) + low_value);
            getCurrentChannelPtr()->setValue(i, yval);
        }
    }

    m_lastIndex = newindex;
    m_lastValue = newvalue;
} 

//////////////////////////////////////////////////////////////////////
// draw
//////////////////////////////////////////////////////////////////////

void
TFEditor::draw()
{ 
    // make context current
    glXMakeCurrent(XtDisplay(m_plotArea), XtWindow(m_plotArea),
        m_context);

    // viewport
    glViewport(
        m_viewport[0], m_viewport[1],
        m_viewport[2], m_viewport[3]);

    // clear
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, 1, 0, 1, -1, 1); 

    // modelview
    glMatrixMode(GL_MODELVIEW);

    //////////////////////////////////////////////////////////////////
    // draw channels
    //////////////////////////////////////////////////////////////////

    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glLineWidth(2);

    switch (m_manager->getFormat()) {
    case TF_LUMINANCE:
        drawChannel(TFManager::CHANNEL_LUMINANCE);
        break;

    case TF_LUMINANCE_ALPHA:
        drawChannel(TFManager::CHANNEL_LUMINANCE);
        drawChannel(TFManager::CHANNEL_ALPHA);

        drawChannel(getCurrentChannel()); // draw current on top
        break;

    case TF_RGBA:
        drawChannel(TFManager::CHANNEL_RED);
        drawChannel(TFManager::CHANNEL_GREEN);
        drawChannel(TFManager::CHANNEL_BLUE);
        drawChannel(TFManager::CHANNEL_ALPHA);

        drawChannel(getCurrentChannel()); // draw current on top
        break;
    
    default:
        break;
    }

    glDisable(GL_BLEND);
    glLineWidth(1);

    //////////////////////////////////////////////////////////////////
    // draw control points
    //////////////////////////////////////////////////////////////////

    switch (getCurrentChannelPtr()->getInterpolation()) {
    case TFChannel::INTERPOLATION_WINDOW_LEVEL:
    case TFChannel::INTERPOLATION_PAINT:
        break;
   
    default: 
        drawControlPoints(getCurrentChannel());
        break;
    }

    //////////////////////////////////////////////////////////////////
    // draw lookuptable
    //////////////////////////////////////////////////////////////////

    const float *r, *g, *b, *a, *l;

    r = m_manager->getChannel(TFManager::CHANNEL_RED)->getValues();
    g = m_manager->getChannel(TFManager::CHANNEL_GREEN)->getValues();
    b = m_manager->getChannel(TFManager::CHANNEL_BLUE)->getValues();
    a = m_manager->getChannel(TFManager::CHANNEL_ALPHA)->getValues();
    l = m_manager->getChannel(TFManager::CHANNEL_LUMINANCE)->getValues();

    glBegin(GL_QUAD_STRIP);

    for (int x = 0; x < m_manager->getLength(); x++) {
        switch (m_manager->getFormat()) {
        case TF_RGBA:
            glColor3f(r[x], g[x], b[x]);
            break;

        case TF_LUMINANCE:
            glColor3f(l[x], l[x], l[x]);
            break;

        case TF_LUMINANCE_ALPHA:
            glColor3f(l[x], l[x], l[x]);
            break;
        }

        glVertex2f(tf_index_2_x_plot_coord(x), .15);
        glVertex2f(tf_index_2_x_plot_coord(x), .225);
    }

    glEnd();

    //////////////////////////////////////////////////////////////////
    // draw lookup table multiplied with alpha
    //////////////////////////////////////////////////////////////////

    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glBegin(GL_QUAD_STRIP);

    for (int j = 0; j < m_manager->getLength(); j++) {
        switch (m_manager->getFormat()) {
        case TF_RGBA:
            glColor4f(r[j], g[j], b[j], a[j]);
            break;

        case TF_LUMINANCE:
            glColor4f(l[j], l[j], l[j], l[j]);
            break;

        case TF_LUMINANCE_ALPHA:
            glColor4f(l[j], l[j], l[j], a[j]);
            break;
        } 

        glVertex2f(tf_index_2_x_plot_coord(j), 0.075);
        glVertex2f(tf_index_2_x_plot_coord(j), 0.15);
    }

    glEnd();

    glDisable(GL_BLEND);
  
    //////////////////////////////////////////////////////////////////

    GLwDrawingAreaSwapBuffers(m_plotArea);
}

//////////////////////////////////////////////////////////////////////
// draw control points
//////////////////////////////////////////////////////////////////////

void
TFEditor::drawControlPoints(TFManager::Channel channel_type)
{
    TFChannel* channel = m_manager->getChannel(channel_type);

    TFControlPointList L = channel->getControlPointList();
    TFControlPointIterator start = L.begin();
    TFControlPointIterator stop  = L.end();

    glPointSize(5);
    glBegin(GL_POINTS); 

    // draw control points
    for (TFControlPointIterator i = start; i != stop; i++)
        glVertex2f(
            tf_index_2_x_plot_coord(i->m_index),
            tf_value_2_y_plot_coord(i->m_value));

    // highlight current control point
    TFControlPointIterator current = channel->getCurrentControlPoint();
    if (current != NULL) {
        glColor3f(0.7, 0.7, 0.0);
        glVertex2f(
            tf_index_2_x_plot_coord(current->m_index),
            tf_value_2_y_plot_coord(current->m_value));
    }

    glEnd();
    glPointSize(1);
}

//////////////////////////////////////////////////////////////////////
// draw channel
//////////////////////////////////////////////////////////////////////

void
TFEditor::drawChannel(TFManager::Channel channel_type)
{
    TFChannel* channel = m_manager->getChannel(channel_type);

    const float* channel_values;

    // set draw color
    switch (channel_type) {
    case TFManager::CHANNEL_RED:
        glColor4f(1, 0, 0, 1);
        break;

    case TFManager::CHANNEL_GREEN:
        glColor4f(0, 1, 0, 1);
        break;

    case TFManager::CHANNEL_BLUE:
        glColor4f(0.3, 0.3, 1, 1);
        break;
    
    case TFManager::CHANNEL_ALPHA:
        glColor4f(1, 1, 1, 1);
        break;
    
    case TFManager::CHANNEL_LUMINANCE:
        glColor4f(0.4, 0.4, 0.4, 1);
        break;
    
    default:
        break;
    }
  
    channel_values = channel -> getValues();
  
    // draw line strip
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i < m_manager->getLength(); i++)
        glVertex2f(
            tf_index_2_x_plot_coord(i),
            tf_value_2_y_plot_coord(channel_values[i]));
    glEnd();
}

//////////////////////////////////////////////////////////////////////
// transform function to plot area
//////////////////////////////////////////////////////////////////////

float
TFEditor::tf_index_2_x_plot_coord(int index)
{
    return ((float) index / m_manager->getLength()) * 
        s_width + s_left;
}

float
TFEditor::tf_value_2_y_plot_coord(float value)
{
    return (s_height * value) + s_bottom + 0.3;
}  

//////////////////////////////////////////////////////////////////////
// transform plot area to function
//////////////////////////////////////////////////////////////////////

int
TFEditor::x_plot_coord_2_tf_index(float x)
{
    x /= m_viewport[2]; 

    int index =
        (int) floor((((x - s_left) / s_width) *
        m_manager->getLength()) + 0.5);

    if (index < 0)
        index = 0;
    else if (index >= m_manager->getLength()) 
        index = m_manager->getLength() - 1;

    return index;
}

float
TFEditor::y_plot_coord_2_tf_value(float y)
{
    y = (m_viewport[3] - y) / m_viewport[3];

    float value = (y - s_bottom - 0.3) / s_height;

    // clamp
    if (value < 0) value = 0;
    else if (value > 1) value = 1;

    return value;
}

//////////////////////////////////////////////////////////////////////
// user function
//////////////////////////////////////////////////////////////////////

void
TFEditor::setUserFunction(TFUserFunctionPtr function, void* userData) 
{
    m_userFunction = function;
    m_userData = userData;
}

void
TFEditor::callUserFunction()
{
    if (m_userFunction == NULL)
        return;

    m_userFunction(m_manager->getData(), m_manager->getLength(),
        m_manager->getFormat(), m_userData);
}

//////////////////////////////////////////////////////////////////////

void
TFEditor::getTableInfo(void*& dataPtr, int& width, TFFormat& format)
{
    dataPtr = m_manager->getData();
    width   = m_manager->getLength();
    format  = m_manager->getFormat();
}

//////////////////////////////////////////////////////////////////////
// find widget
//////////////////////////////////////////////////////////////////////

Widget
TFEditor::interpolationTypeToWidget(TFChannel::Interpolation type)
{
    char* button_num;

    switch (type) {
    case TFChannel::INTERPOLATION_WINDOW_LEVEL:
        button_num = "button_0";
        break;
        
    case TFChannel::INTERPOLATION_LINEAR:
        button_num = "button_1";
        break;
        
    case TFChannel::INTERPOLATION_SPLINES:
        button_num = "button_2";
        break;
        
    case TFChannel::INTERPOLATION_GAUSSIAN:
        button_num = "button_3";
        break;
      
    case TFChannel::INTERPOLATION_PAINT:
        button_num = "button_4";
        break;
        
    default:
        cerr << "Invalid interpolation mode!" << endl;
        return NULL;
    };
    
    return XtNameToWidget(m_interpolation, button_num);
}

Widget
TFEditor::formatTypeToWidget(TFFormat format)
{
    switch (format) {
    case TF_RGBA:
         return m_rgba;

    case TF_LUMINANCE_ALPHA:
        return m_la;

    case TF_LUMINANCE:
        return m_l;

    default:
        cerr << "Invalid TF format!" << endl;
        return NULL;
    }
}

Widget
TFEditor::channelTypeToWidget(TFManager::Channel channel,
    TFFormat format)
{
    Widget parent = formatTypeToWidget(format);

    char* name;

    switch (channel) {
    case TFManager::CHANNEL_RED:
        name = "button_0";
        break;

    case TFManager::CHANNEL_GREEN:
        name = "button_1";
        break;

    case TFManager::CHANNEL_BLUE:
        name = "button_2";
        break;

    case TFManager::CHANNEL_ALPHA:
        if (parent == m_rgba) 
            name = "button_3";
        else
            name = "button_1";
        break;

    case TFManager::CHANNEL_LUMINANCE:
        name = "button_0";
        break;

    default:
        cerr << "Invalid channel ID!" << endl;
        return NULL;
    }
    
    return XtNameToWidget(parent, name);
}

//////////////////////////////////////////////////////////////////////
// current channel
//////////////////////////////////////////////////////////////////////

void
TFEditor::setCurrentChannel(TFManager::Channel type)
{
    char name[255];
    for (int i = 0; i < m_manager->getNumberOfChannels(); i++) {
        sprintf(name, "button_%d", i);
        Widget parent = formatTypeToWidget (m_manager->getFormat());
        Widget w = XtNameToWidget (parent, name);
        XtVaSetValues (w, XmNset, False, NULL);
    }

    Widget newwidget = channelTypeToWidget (type, m_manager->getFormat());
    XtVaSetValues (newwidget, XmNset, True, NULL);
    
    m_currentChannel = type;
    
    // update interpolation mode
    setCurrentInterpolationType (getCurrentChannelPtr()->getInterpolation());
}

TFManager::Channel
TFEditor::getCurrentChannel()
{
    return m_currentChannel;
}

TFChannel*
TFEditor::getCurrentChannelPtr()
{
    return m_manager->getChannel(m_currentChannel);
}

//////////////////////////////////////////////////////////////////////
// interpolation type
//////////////////////////////////////////////////////////////////////

void
TFEditor::setCurrentInterpolationType(TFChannel::Interpolation type)
{
    if (type == TFChannel::INTERPOLATION_WINDOW_LEVEL) {
        // update window-level sliders to the current values:
        XtVaSetValues(m_scaleWindow,
            XmNvalue, getCurrentChannelPtr()->getRampWidth(),
            NULL);
        
        XtVaSetValues(m_scaleLevel,
            XmNvalue, getCurrentChannelPtr()->getRampCenter(),
            NULL);
    }
    
    if ((type != TFChannel::INTERPOLATION_WINDOW_LEVEL) &&
        m_windowLevelEnabled)
    {
        XtPopdown(m_windowLevelShell);
        XtAddCallback(m_plotArea, GLwNinputCallback, inputCallback, this);
        m_windowLevelEnabled = false;
    }

    else if ((type == TFChannel::INTERPOLATION_WINDOW_LEVEL) &&
        !m_windowLevelEnabled)
    {
        // switching to window_level
        XtPopup(m_windowLevelShell, XtGrabNone);
        XtRemoveCallback(m_plotArea, 
            GLwNinputCallback, inputCallback, this);
        m_windowLevelEnabled = true;
    }

    char name[255];
    for (int i = 0; i < TFChannel::INTERPOLATION_COUNT; i++) {
        sprintf (name, "button_%d", i);
        Widget parent = m_interpolation;
        Widget w = XtNameToWidget (parent, name);
        XtVaSetValues (w, XmNset, False, NULL);
    }
    Widget wOn = interpolationTypeToWidget (type);
    XtVaSetValues (wOn, XmNset, True, NULL);    
}

//////////////////////////////////////////////////////////////////////
// save callback
//////////////////////////////////////////////////////////////////////

void
TFEditor::saveLUTCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    XmSelectionBoxCallbackStruct* data =
        (XmSelectionBoxCallbackStruct*) callData;
  
    char* filename;
    XmStringGetLtoR(data->value, XmSTRING_DEFAULT_CHARSET, &filename);
   
    TFEditor* editor = (TFEditor*) clientData;
    editor->saveToFile(filename);
 
    XtUnmanageChild(widget);
}

bool
TFEditor::saveToFile(const char filename[])
{
    ofstream stream(filename);

    if (!stream) {
        cerr
            << "TFEditor: Could not open file " << filename
            << " for writing!" << endl;

        return false;
    }

    m_manager->saveToStream(stream);
  
    return true;
}

//////////////////////////////////////////////////////////////////////
// load
//////////////////////////////////////////////////////////////////////

void
TFEditor::loadLUTCallback(Widget widget, XtPointer clientData,
    XtPointer callData)
{
    XmSelectionBoxCallbackStruct* data =
        (XmSelectionBoxCallbackStruct*) callData;

    char* filename;
    XmStringGetLtoR(data->value, XmSTRING_DEFAULT_CHARSET, &filename);

    TFEditor* editor = (TFEditor*) clientData;

    editor->loadFromFile(filename);

    XtUnmanageChild(widget);
} 

bool
TFEditor::loadFromFile(const char filename[])
{
    ifstream stream(filename);

    if (!stream) {
        cerr
            << "TFEditor: Could not read data from file " << filename
            << endl;

        return false;
    }
   
    TFManager* newTF = TFManager::createFromStream(stream);

    if (newTF == NULL)
        return false;

    delete m_manager;
    m_manager = newTF;

    setCurrentFormat(m_manager->getFormat());

    if(m_popped)
      draw();
    
    callUserFunction();
        
    return true;
}

//////////////////////////////////////////////////////////////////////
// current format
//////////////////////////////////////////////////////////////////////

void
TFEditor::setCurrentFormat(TFFormat format)
{
    XtUnmanageChild(m_rgba);
    XtUnmanageChild(m_la);
    XtUnmanageChild(m_l);

    switch (format) {
    case TF_LUMINANCE:
        XtManageChild(m_l);
        setCurrentChannel(TFManager::CHANNEL_LUMINANCE);
        break;

    case TF_LUMINANCE_ALPHA:
        XtManageChild (m_la);
        setCurrentChannel(TFManager::CHANNEL_LUMINANCE);
        break;

    case TF_RGBA:
        XtManageChild(m_rgba);
        setCurrentChannel(TFManager::CHANNEL_RED);
        break;

    default:
        cerr << "invalid format" << endl;
        return;
    };

    char name[255];
    Widget parent = m_format;
    for (int i = 0; i < 3; i++) {
        sprintf(name, "button_%d", i);
        
        Widget w = XtNameToWidget(parent, name);
        XtVaSetValues(w, XmNset, False, NULL);
    }

    sprintf(name, "button_%d", (int) format);
    Widget wOn = XtNameToWidget(parent, name);
    XtVaSetValues(wOn, XmNset, True, NULL);    
}

//////////////////////////////////////////////////////////////////////
// TFEditor.cxx - EOF
//////////////////////////////////////////////////////////////////////

