// Copyright  2002 Silicon Graphics, Inc.  All Rights Reserved.
// 
// The recipient ("Recipient") of this software, including as modified
// ("Software") may reproduce, redistribute, use, and derive works from the
// Software without restriction, subject to the following conditions:
//
//  Redistribution of the Software in any form must reproduce this entire
// notice, including as modified in accordance with these provisions
// ("Notice");
//
//  Any Recipient who modifies and subsequently redistributes the Software
// shall add information to this Notice to sufficiently identify the
// Recipient's modifications;
//
//  Recipient may not use the name(s) of any previous Recipient to endorse
// or promote any products derived from the Software without prior express
// written permission from such previous Recipient.
//
// THE SOFTWARE IS PROVIDED "AS IS," WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTY OR CONDITION, INCLUDING WITHOUT LIMITATION ANY WARRANTIES OR
// CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A
// PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  PATENT LICENSES, IF ANY,
// PROVIDED HEREIN DO NOT APPLY TO COMBINATIONS OF THIS PROGRAM WITH OTHER
// SOFTWARE, OR ANY OTHER PRODUCT WHATSOEVER.  IN NO EVENT WILL THE
// ORIGINATOR OR SUBSEQUENT RECIPIENT OF THE SOFTWARE BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES ARISING IN RELATION TO THE SOFTWARE, ITS USE, OR THESE
// PROVISIONS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR NEGLIGENCE OR OTHER TORT, EVEN IF
// PRE-ADVISED OF THE PROSPECT OF SUCH DAMAGES.
//
// This Software was developed at private expense; if acquired under an
// agreement with the USA government or any contractor thereto, it is
// acquired as "commercial computer software" subject to the provisions of
// this license agreement, as specified in (a) 48 CFR 12.212 of the FAR;
// or, if acquired for Department of Defense units, (b) 48 CFR 227-7202 of
// the DoD FAR Supplement; or sections succeeding thereto.
// 
// Originator:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy, Mountain
// View, CA  94043.  http://www.sgi.com


//////////////////////////////////////////////////////////////////////
// TFManager.cxx
//////////////////////////////////////////////////////////////////////

#include "TFManager.h"
#include "TFChannel.cxx"
#include <iostream.h>
#include <string.h>

//////////////////////////////////////////////////////////////////////
// format to channel count
//////////////////////////////////////////////////////////////////////

int
TFManager::formatToChannelCount(TFFormat format)
{
    switch (format) {
    case TF_LUMINANCE:
        return 1;

    case TF_LUMINANCE_ALPHA:
        return 2;

    case TF_RGBA:
        return 4;

    default:
        return 0;
    }
}

//////////////////////////////////////////////////////////////////////
// format to string
//////////////////////////////////////////////////////////////////////

const char*
TFManager::formatToString(TFFormat format)
{
    switch (format) {
    case TF_LUMINANCE:
        return "LUMINANCE";

    case TF_LUMINANCE_ALPHA:
        return "LUMINANCE_ALPHA";

    case TF_RGBA:
        return "RGBA";

    default:
        return NULL;
    }
}

//////////////////////////////////////////////////////////////////////
// factories
//////////////////////////////////////////////////////////////////////

TFManager*
TFManager::create(TFFormat format, int length)
{
    TFChannel** channels = new TFChannel*[5];
    int rampCenter = 127;
    int rampWidth = 128;
    for (int i = 0; i < 5; i++) {
        channels[i] = new TFChannel(TFChannel::INTERPOLATION_LINEAR, 
				    length, rampCenter, rampWidth);
    }

    return new TFManager(channels, format, length);
}

TFManager*
TFManager::createFromStream(istream& stream)
{
    char string[255];

    stream >> string;
    if (strcmp(string, "version_1.0") != 0) {
        cerr << "TFEditor: Invalid version string for LUT file." << endl;
        return NULL;
    }

    int length;
    stream >> length;
   
    TFFormat format; 
    stream >> string;
    if (strcmp(string, "LUMINANCE") == 0) {
        format = TF_LUMINANCE;

    } else if (strcmp(string, "LUMINANCE_ALPHA") == 0) {
        format = TF_LUMINANCE_ALPHA;

    } else if (strcmp(string, "RGBA") == 0) {
        format = TF_RGBA;

    } else {
        return NULL;
    }

    TFChannel** channels = new TFChannel*[5];
    for (int i = 0; i < 5; i++) {
      channels[i] = new TFChannel(stream, length);
    }

    return new TFManager(channels, format, length);
}

//////////////////////////////////////////////////////////////////////
// constructor
//////////////////////////////////////////////////////////////////////


TFManager::TFManager(TFChannel** channels, TFFormat format, int length)
    :
    m_channels(channels),
    m_format(format),
    m_length(length),
    m_table(new float[length * formatToChannelCount(format)]),
    m_dirty(true)
{
}

//////////////////////////////////////////////////////////////////////
// destructor
//////////////////////////////////////////////////////////////////////

TFManager::~TFManager()
{
    delete[] m_table;

    for (int i = 5; i < 5; i++)
        delete m_channels[i];
    delete[] m_channels;
}

//////////////////////////////////////////////////////////////////////
// format
//////////////////////////////////////////////////////////////////////

void
TFManager::setFormat(TFFormat format)
{
    if (m_format == format)
        return;

    m_format = format;

    delete[] m_table;
    m_table = new float[m_length * formatToChannelCount(format)];

    m_dirty = true;
}

TFFormat
TFManager::getFormat() const
{
    return m_format;
}

//////////////////////////////////////////////////////////////////////
// length
//////////////////////////////////////////////////////////////////////

void
TFManager::setLength(int length)
{
    for (int i = 0; i < 5; i++)
        m_channels[i]->setLength(length);
  
    delete [] m_table;
    m_table = new float[m_length * formatToChannelCount(m_format)];

    m_length = length;
  
    m_dirty = true;
}

int
TFManager::getLength() const
{
    return m_length;
}

//////////////////////////////////////////////////////////////////////
// data
//////////////////////////////////////////////////////////////////////

float*
TFManager::getData()
{
    // Poll the channels to see if they are dirty 
    // and need to be re-evaluated

    for (int i = 0; i < 5; i++)
        if (m_channels[i]->isDirty()) {
            m_channels[i]->evaluate();
            m_dirty = true;
        }

    // XXX hack
    m_dirty = true;

    // if the transfer function is dirty rebuild it
    if (m_dirty) {
        const float* values[5]; 

        int i;

        for (i = 0; i < 5; i++)
            values[i] = m_channels[i]->getValues();

        switch (m_format) {
        case TF_LUMINANCE:
            for (i = 0; i < m_length; i++)
                m_table[i] = values[0][i];
            break;
    
        case TF_LUMINANCE_ALPHA:
            for (i = 0; i < m_length; i++) {
                m_table[(i * 2) + 0] = values[0][i];
                m_table[(i * 2) + 1] = values[4][i];
            }
            break;
    
        case TF_RGBA:
            for (i = 0; i < m_length; i++) {
                m_table[(i * 4) + 0] = values[1][i];
                m_table[(i * 4) + 1] = values[2][i];
                m_table[(i * 4) + 2] = values[3][i];
                m_table[(i * 4) + 3] = values[4][i];
            }
            break;
        }

        m_dirty = false;
    }

    return m_table;
}

//////////////////////////////////////////////////////////////////////
// channels
//////////////////////////////////////////////////////////////////////

TFChannel*
TFManager::getChannel(Channel channel)
{
    switch (channel) {
    case CHANNEL_RED:
    case CHANNEL_GREEN:
    case CHANNEL_BLUE:
    case CHANNEL_ALPHA:
    case CHANNEL_LUMINANCE:
        return m_channels[(int) channel];

    default:
        return NULL;
    }
}

int
TFManager::getNumberOfChannels() const
{
    return formatToChannelCount(m_format);
}

//////////////////////////////////////////////////////////////////////
// channels
//////////////////////////////////////////////////////////////////////

void
TFManager::saveToStream(ostream& stream) const
{
    stream
        << "version_1.0" << endl
        << m_length << endl
        << formatToString(m_format) << endl;

    for (int i = 0; i < 5; i++)
        m_channels[i]->saveToStream(stream);
}

//////////////////////////////////////////////////////////////////////
// initialize channel from float points
//////////////////////////////////////////////////////////////////////
void
TFManager::initializeChannel(Channel channel, int slot, int num, float *controlPoints)
{
    delete m_channels[(int) channel];
    
    //const int
      //rampCenter = 127,
      //rampWidth = 128;
}

//////////////////////////////////////////////////////////////////////
// built in LUTs
//////////////////////////////////////////////////////////////////////

void
TFManager::createBuiltInLUT(Function function)
{
    int numPoints = 2;
    float *controlPoints = NULL;
    
    static float gray_scale_points[][4] =
    { {0.0, 0.0, 0.0, 0.0 },
      {1.0, 1.0, 1.0, 1.0 } };
    
    static float hot_metal_points[][4] = 
    { {0.0,  0.0, 0.0, 0.0 },
      {0.25, 0.5, 0.0, 0.0 },
      {0.5,  1.0, 0.5, 0.0 },
      {0.75, 1.0, 1.0, 0.5 },
      {1.0,  1.0, 1.0, 1.0 } };
	
    static float spectral_points[][4] =
    { { 0.00, 0.0000, 0.0000, 0.0000 },
      { 0.05, 0.4667, 0.0000, 0.5333 },
      { 0.10, 0.5333, 0.0000, 0.6000 },
      { 0.15, 0.0000, 0.0000, 0.6667 },
      { 0.20, 0.0000, 0.0000, 0.8667 },
      { 0.25, 0.0000, 0.4667, 0.8667 },
      { 0.30, 0.0000, 0.6000, 0.8667 },
      { 0.35, 0.0000, 0.6667, 0.6667 },
      { 0.40, 0.0000, 0.6667, 0.5333 },
      { 0.45, 0.0000, 0.6000, 0.0000 },
      { 0.50, 0.0000, 0.7333, 0.0000 },
      { 0.55, 0.0000, 0.8667, 0.0000 },
      { 0.60, 0.0000, 1.0000, 0.0000 },
      { 0.65, 0.7333, 1.0000, 0.0000 },
      { 0.70, 0.9333, 0.9333, 0.0000 },
      { 0.75, 1.0000, 0.8000, 0.0000 },
      { 0.80, 1.0000, 0.6000, 0.0000 },
      { 0.85, 1.0000, 0.0000, 0.0000 },
      { 0.90, 0.8667, 0.0000, 0.0000 },
      { 0.95, 0.8000, 0.0000, 0.0000 },
      { 1.00, 0.8000, 0.8000, 0.8000 }
    };
    
    static float contour_points[][4] =
    { {0.0,   0.0, 0.0, 0.3 },
      {0.166, 0.0, 0.0, 1.0 },
      {0.166, 0.0, 0.3, 0.3 },
      {0.333, 0.0, 1.0, 1.0 },
      {0.333, 0.0, 0.3, 0.0 },
      {0.5,   0.0, 1.0, 0.0 },
      {0.5,   0.3, 0.3, 0.0 },
      {0.666, 1.0, 1.0, 0.0 },
      {0.666, 0.3, 0.0, 0.0 },
      {0.833, 1.0, 0.0, 0.0 },
      {0.833, 0.3, 0.3, 0.3 },
      {1.0,   1.0, 1.0, 1.0 } };
    
    switch (function) {
	
    case GRAY_SCALE:
	numPoints = sizeof(gray_scale_points)/(4 * sizeof(float));
	controlPoints = &gray_scale_points[0][0];
	break;

    case HOT_METAL:
	numPoints = sizeof(hot_metal_points)/(4 * sizeof(float));
	controlPoints = &hot_metal_points[0][0];
	break;
	
    case SPECTRAL:
	numPoints = sizeof(spectral_points)/(4 * sizeof(float));
	controlPoints = &spectral_points[0][0];
	break;
	
    case CONTOUR:
	numPoints = sizeof(contour_points)/(4 * sizeof(float));
	controlPoints = &contour_points[0][0];
	break;
    }
    
    // Initialize the channels with the given data
    initializeChannel(CHANNEL_RED, 1, numPoints, controlPoints);
    initializeChannel(CHANNEL_GREEN, 2, numPoints, controlPoints);
    initializeChannel(CHANNEL_BLUE, 3, numPoints, controlPoints);
    initializeChannel(CHANNEL_ALPHA, 0, numPoints, controlPoints);
    
    setFormat(TF_RGBA);
    m_dirty = true;
}

//////////////////////////////////////////////////////////////////////
// TFEditor.cxx - EOF
//////////////////////////////////////////////////////////////////////

