// Copyright  2002 Silicon Graphics, Inc.  All Rights Reserved.
// 
// The recipient ("Recipient") of this software, including as modified
// ("Software") may reproduce, redistribute, use, and derive works from the
// Software without restriction, subject to the following conditions:
//
//  Redistribution of the Software in any form must reproduce this entire
// notice, including as modified in accordance with these provisions
// ("Notice");
//
//  Any Recipient who modifies and subsequently redistributes the Software
// shall add information to this Notice to sufficiently identify the
// Recipient's modifications;
//
//  Recipient may not use the name(s) of any previous Recipient to endorse
// or promote any products derived from the Software without prior express
// written permission from such previous Recipient.
//
// THE SOFTWARE IS PROVIDED "AS IS," WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTY OR CONDITION, INCLUDING WITHOUT LIMITATION ANY WARRANTIES OR
// CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A
// PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  PATENT LICENSES, IF ANY,
// PROVIDED HEREIN DO NOT APPLY TO COMBINATIONS OF THIS PROGRAM WITH OTHER
// SOFTWARE, OR ANY OTHER PRODUCT WHATSOEVER.  IN NO EVENT WILL THE
// ORIGINATOR OR SUBSEQUENT RECIPIENT OF THE SOFTWARE BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES ARISING IN RELATION TO THE SOFTWARE, ITS USE, OR THESE
// PROVISIONS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR NEGLIGENCE OR OTHER TORT, EVEN IF
// PRE-ADVISED OF THE PROSPECT OF SUCH DAMAGES.
//
// This Software was developed at private expense; if acquired under an
// agreement with the USA government or any contractor thereto, it is
// acquired as "commercial computer software" subject to the provisions of
// this license agreement, as specified in (a) 48 CFR 12.212 of the FAR;
// or, if acquired for Department of Defense units, (b) 48 CFR 227-7202 of
// the DoD FAR Supplement; or sections succeeding thereto.
// 
// Originator:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy, Mountain
// View, CA  94043.  http://www.sgi.com

//////////////////////////////////////////////////////////////////////
// pfVolume.C
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// includes
//////////////////////////////////////////////////////////////////////
#include "pfVolume.h"
#include <fstream.h>
#include "IFLLoader.cxx"
#include "TFManager.cxx"
#include "parse.C"
// Volumizer
#include <Volumizer2/ParameterVolumeTexture.h>
#include <ifl/iflFile.h>

//////////////////////////////////////////////////////////////////////
// macros
//////////////////////////////////////////////////////////////////////
#define FOR_ALL_SHAPES for(shapeID = 0; shapeID < _shape_set->getNumShapes(); shapeID++)

//////////////////////////////////////////////////////////////////////
// Symbols to be used by dlsym use to load the Volumizer file with
// the extension .vz. 
//////////////////////////////////////////////////////////////////////
pfType *pfVolume::pfvzClassType = 0;
int pfVolume::pfvzRenderActionSlot = 0;
int pfVolume::pfvzShapeSlot = 0;


extern "C" {
pfNode *pfdLoadFile_vzl( const char * );
int  pfdInitConverter_vzl(  void );
}

int pfdInitConverter_vzl( void )
{
    pfVolume::init();
    return TRUE;
}
 
pfNode *pfdLoadFile_vzl( const char *fname )
{
  //printf("start of vzl loader: %s\n",fname);
  pfVolume *vol = new pfVolume;
  if( !(vol->loadFile( fname )))
    return NULL;
  
  return vol;
}

//////////////////////////////////////////////////////////////////////
// Callbacks
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// new - use pfMalloc
//////////////////////////////////////////////////////////////////////
void *pfvzNew(size_t size, void */*userData*/) {

    return (pfMalloc(size, pfGetSharedArena()));
}

//////////////////////////////////////////////////////////////////////
// delete - use pfDelete
//////////////////////////////////////////////////////////////////////
void pfvzDelete(void *ptr, void */*userData*/) {

    pfFree(ptr);
}

#ifndef __linux__
//////////////////////////////////////////////////////////////////////
// Exeption hander to pass to vzError::setErrorHandler()
//////////////////////////////////////////////////////////////////////
void pfVolume::exceptionHandler( vzErrorSeverity severity, vzErrorType type,
				 const char */*format*/, va_list args, 
				 void* /*data*/)
{
     char message[200];
     sprintf(message, args);
     if(severity == VZ_ERROR)
       pfNotify( PFNFY_FATAL, PFNFY_PRINT, "pfVolume error type %d, severity %d, %s\n",
		 type, severity, message);
     else if(severity == VZ_WARNING)
       pfNotify( PFNFY_NOTICE, PFNFY_PRINT, "pfVolume error type %d, severity %d, %s\n",
		 type, severity, message);
       
}
#endif

//////////////////////////////////////////////////////////////////////
// pre APP callback
// Initializes the scene graph for the given pfPipeWindow, if necessary
//////////////////////////////////////////////////////////////////////
int pfVolume::preappCB (pfTraverser * trav, void *data)
{
     pfVolume *vol = (pfVolume *)data;
     
     pfPipeWindow *pw = trav->getChan()->getPWin();
     
     vol->preapp(pw);

     return PFTRAV_CONT;
}

//////////////////////////////////////////////////////////////////////
// pre DRAW callback
// Initializes the renderers for the given pfPipeWindow, if necessary
//////////////////////////////////////////////////////////////////////
int pfVolume::predrawCB (pfTraverser * trav, void *data)
{
     pfVolume *vol = (pfVolume *)data;
     
     pfPipeWindow *pw = trav->getChan()->getPWin();
     
     vzError::message(5, "pfVolume::predrawCB(): PID = %d, Data = %x, PipeWin = %x", getpid(), vol, pw);
     
     vol->predraw(pw);
     
     return PFTRAV_CONT;
}

//////////////////////////////////////////////////////////////////////
// Draw Callback used by the pfGeoSetCB for drawing the shape node
//////////////////////////////////////////////////////////////////////
void pfVolume::drawCB(pfGeoSet *gset, void *data)
{
     pfVolume *vol = (pfVolume *)gset->getUserData(pfvzShapeSlot);
     
     pfPipeWindow *pw = (pfPipeWindow *)data;
     
     vzError::message(5, "pfVolume::drawCB(): PID = %d, pfVolume = %x, pfPipeWindow = %x, Gset = %x", getpid(), vol, pw, gset);
     
     // Get the current pipe
     pfPipe *currentPipe = pfGetPipe(pfGetPipeNum());
     
     vzError::message(5, "pfVolume::drawCB(): Current Pipe = %x, GeoSet Pipe = %x", currentPipe, pw->getPipe());
     
     // Compare the current pipe with the pipe for the pfGeoSetCB and
     // draw the volume if we are on the same pipe only.
     if(currentPipe == pw->getPipe())
       vol->draw(pw);
}

//////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////
pfVolume::pfVolume( void )
{
    // Set default values
    _shape_set = NULL;
    vzError::setDebugLevel(1);
    
#ifndef __linux    
    vzError::setHandler(exceptionHandler);
#endif
    
    _sampling_rate[0] = _sampling_rate[1] = _sampling_rate[2] = 1.0;
    _initialized = 0;
}

//////////////////////////////////////////////////////////////////////
// Destructor
//////////////////////////////////////////////////////////////////////
pfVolume::~pfVolume( void )
{
     vzError::message(2,"pfVolume: Deleting the shape node");
     
     _shape_set->unref();
     pfFree(_geoset_cbs);
     pfFree(_windows);
}

//////////////////////////////////////////////////////////////////////
// Initialize Performer sub-class
//////////////////////////////////////////////////////////////////////
void pfVolume::init( void )
{
    if( pfvzClassType == 0 ) {

      // Set the new and delete operators to allocate memory from
      // pf shared arena.
      vzMemory::setMemoryManagementCallbacks(pfvzNew, pfvzDelete, NULL);

      pfNotify( PFNFY_INFO, PFNFY_PRINT, "Initializing Performer subclass 'pfVolume'" );
	
      // Initialize the pf class 
      pfObject::init();
	
      pfvzClassType = new pfType( pfNode::getClassType(), "pfVolume" );
    }
	    
    if(pfvzRenderActionSlot == 0) {
	pfvzRenderActionSlot = pfObject::getNamedUserDataSlot("pfvzRenderAction");
    }

    if(pfvzShapeSlot == 0) {
	pfvzShapeSlot = pfObject::getNamedUserDataSlot("pfvzShapeSlot");
    }

}


//////////////////////////////////////////////////////////////////////
// Load the given .vol file into a ParameterVolumeTexture and 
// initialize the shape nodeq
//////////////////////////////////////////////////////////////////////
int pfVolume::loadFile( const char *filename )
{
  //printf("start of loadFile\n");
  // Open the file for reading.
  FILE *fp = fopen(filename, "r");              
  if (!fp)
    printf("error opening input file %s\n",filename) ;

  char line[250];
  char volFilename[250];
  int pargc = 0;
  char **pargv ;

  // Set the preapp CB to initialize the pfGeoSetCB for the
  // current pfPipeWindow. This needs to be done in the APP
  // process
  setTravData(PFTRAV_APP, this );
  setTravFuncs(PFTRAV_APP, preappCB, NULL );
  
  // Set the predraw CB to initialize the pfvzRenderAction for
  // the current pfPipeWindow. This needs to be done in the
  // DRAW process
  setTravData(PFTRAV_DRAW, this );
  setTravFuncs(PFTRAV_DRAW, predrawCB, NULL );
  
  _shape_set = new vzShapeSet();
  
  float dim[3] = {10, 10, 10};
  float off[3] = {-5, -5, -5};
  double geometryROI[6] = { off[0], off[1], off[2], 
			    off[0] + dim[0], off[1] + 
			    dim[1], off[2] + dim[2]};

  vzParameterVolumeTexture *volume;
  vzShader *shader = new vzTMLUTShader();
  //vzTextureType type;
  //vzExternalTextureFormat format;
  int dimSize[3];

  // initialize info to make a new Texture, then read
  // each file and load up the texture data, then
  // define the vzParameterVolumeTexture
  // it needs size, data, type, format, internalFormat

  int pixelx = 0;
  int pixely = 0;
  int pixelz = 0;
  int place = 0;
  //int maxValue = 256;
  int scale = 1;
  //char lutname[250];

  // read the top line first: dimensions of brick and maximum value in file
  if ((fgets(line,250,fp)!=NULL)) {
    if (line != NULL) {
      dpf_charsToArgcArgv(line, &pargc, &pargv) ;
      if (pargc > 2) {
	pixelx = atoi(pargv[0]);
	pixely = atoi(pargv[1]);
	pixelz = atoi(pargv[2]);
      }
      else {
	vzError::warn(VZ_OPERATION_FAILED,"pfVolume: need to specify dimensions of brick");
	return 0;
      }
      if (pargc > 3) {
	scale = atoi(pargv[3]);
	//printf("new scale: %d\n",scale);
      }
    }
  }

  vzParameterLookupTable* table;
  //if (lutname == NULL)
  table = initializeLookupTable(256);
  //else
  //table = loadLookupTable(lutname);

  iflStatus sts;
  iflFile* file;
  iflSize dims;

  vzMemoryAllocationCallback allocator;
  vzMemoryDeletionCallback deallocator;
  void* userData;
  vzMemory::getMemoryManagementCallbacks(allocator, deallocator, userData);
  int size = pixelx*pixely*pixelz;
  unsigned char* newdata = (unsigned char*) allocator(size, userData);

  iflConfig cfg(iflUChar, iflInterleaved);
  unsigned char *data = new unsigned char[pixelx*pixely];
  //maxValue = 1;
  // loop through files: if first read everything
  while ((fgets(line,250,fp)!=NULL)) {
    //printf("print line from vzl file: %s\n",line);
    if (line != NULL) {
      dpf_charsToArgcArgv(line, &pargc, &pargv) ;
      //printf("pargc: %d\n",pargc);
      if (pargc > 0) {
	strcpy(volFilename,pargv[0]);
	//printf("next file: %s\n",volFilename);

	file = iflFile::open(volFilename, O_RDONLY, &sts);
	if (sts != iflOKAY) { /* handle the error */ }
	// read the entire image (just the first plane in z if image has depth)
	// into a buffer of unsiged chars
	file->getDimensions(dims);
	//printf("got dims: %d %d %d\n",dims.x,dims.y,dims.z);

        sts = file->getTile(0, 0, 0, pixelx, pixely, 1, data, &cfg);
	//int cnt = 0;
	for (int i = 0; i < pixelx*pixely; ++i) {
	  newdata[i+place] = scale * data[i];
	}
	place = place + pixelx*pixely;

	if (sts != iflOKAY) { /* handle error */ }

	// close the file
	file->close();
      }
    }
  }

  fclose(fp);

  // now define the volume with the big data set
  dimSize[0] = pixelx;
  dimSize[1] = pixely;
  dimSize[2] = pixelz;

  volume = new vzParameterVolumeTexture(dimSize, newdata,
            VZ_UNSIGNED_BYTE, VZ_LUMINANCE, VZ_DEFAULT_INTERNAL_FORMAT);   
  //volume->addDeletionCallback(deleteCallback, data);
  volume->setGeometryROI(geometryROI);
  
  // Initialize appearance
  _appearance = new vzAppearance(shader);
  _appearance->setParameter("volume", volume);
  volume->unref();
  
  _appearance->setParameter("lookup_table", table);
  table->unref();
  
  // Initialize geometry
  vzBlock *block = new vzBlock();
  
  block->setDimensions(dim);
  block->setOffsets(off);
  
  _geometry = block;
  
  // Initialize shape node
  _shape = new vzShape(_geometry, _appearance);
  _geometry->unref();
  _appearance->unref();
  
  // add the next shape to the shape set
  _shape_set->add(_shape);
  
  if (_shape_set == NULL) {
    vzError::warn(VZ_OPERATION_FAILED,"pfVolume: Could not load shape set from file %s", filename);
    return 0;
  }
  
  // Print some debug information
  vzError::message(2, "pfVolume: Loaded shape set = %x, PID = %d", _shape_set, getpid());
  
  // Number of shapes
  vzError::message(2,"pfVolume: Number of shapes = %d", _shape_set->getNumShapes());
  // Bounding Box
  float bbox[6];
  _shape_set->getBoundingBox(bbox);
  
  vzError::message(2,"pfVolume:: Bounding box = (%f, %f, %f) - (%f, %f, %f)",
		   bbox[0], bbox[1], bbox[2], bbox[3], bbox[4], bbox[5]);
  
  // Compute and set the bounding sphere for this node 
  pfSphere bs;
  bs.center.set( (bbox[0] + bbox[3])/2.0, 
		 (bbox[1] + bbox[4])/2.0, 
		 (bbox[2] + bbox[5])/2.0);
  bs.radius = (bbox[3] - bbox[0]) * sqrtf(3.0)/2.0;
  
  setBound( &bs, PFBOUND_STATIC );
  
  // Now initialize all Performer data structures
  initialize();
  
  return 1;
}

//////////////////////////////////////////////////////////////////////
// initialize the Performer data structures...
//////////////////////////////////////////////////////////////////////
void pfVolume::initialize()
{
    const int MAX_PIPES = 16;
    
    // Allocate memory for MAX_PIPES geosetCB's and pfPipeWindows
    _geoset_cbs = (pfGeoSetCB **) pfMalloc(sizeof(pfGeoSetCB *) * MAX_PIPES);
    
    _windows = (pfPipeWindow **) pfMalloc(sizeof(pfPipeWindow *) * MAX_PIPES);
    
    _num_windows = 0;
    
    // Create a pfGeoState
    _geo_state = new pfGeoState;
    
    _geo_state->makeBasic();
    
    // Request high-quality transparency. This geoset will be 
    // drawn in the transparent bin - following all opaque geometry.
    // In other words, the draw callback will be called after drawing all 
    // opaque geometry, ensuring correct alpha blending.
    _geo_state->setMode(PFSTATE_TRANSPARENCY, PFTR_HIGH_QUALITY);
    
    vzError::message(2, "pfVolume::initialize(): Create the pfGeoState %x, Num windows = %d", _geo_state, _num_windows);
}
 
//////////////////////////////////////////////////////////////////////
// preapp - If the given pfPipeWindow has not been seen before,
// add a pfGeoSetCB to the pfVolume geode
//////////////////////////////////////////////////////////////////////
void pfVolume::preapp(pfPipeWindow *pw)
{
     if(findPipeWindow(pw) < 0) 
	  initializeGeoSetCB(pw);
}

//////////////////////////////////////////////////////////////////////
// predraw - If the given pfPipeWindow has a NULL renderer, create
// a new renderer and set it as the user data for the pipe window
//////////////////////////////////////////////////////////////////////
void pfVolume::predraw(pfPipeWindow *pw)
{
    void *pWinData = pw->getUserData(pfvzRenderActionSlot);
    
    if(!pWinData) {
	 
      vzError::message(2, "pfVolume::predraw(): NULL data for pWin %x", pw);
      
      initializeRenderer(pw);
    }
}

//////////////////////////////////////////////////////////////////////
// find the given pfPipeWindow in the list of pfPipeWindow's
//////////////////////////////////////////////////////////////////////
int pfVolume::findPipeWindow(pfPipeWindow *pw)
{
     for(int i = 0; i < _num_windows; i++) {
	  
	  if(_windows[i] == pw) {
	       
	       // Found pfPipeWindow
	       return i;
	  }
     }
     return -1;
}

//////////////////////////////////////////////////////////////////////
// Draw the given pfGeoSetCB. Currently, there is one pfGeoSetCB for
// each pfPipeWindow. Depending upon the particular callback that is
// invoked and the user data, we draw the shape using the renderer
// for the current pfPipeWindow
//////////////////////////////////////////////////////////////////////
void pfVolume::draw(pfPipeWindow *pw) 
{
    if(!pw) {
	 vzError::warn(VZ_INVALID_VALUE, "pfVolume::draw(): NULL pWin, Num windows = %d", _num_windows);
	 return;
    }
    
    void *pWinData = pw->getUserData(pfvzRenderActionSlot);
    if(!pWinData) {
	 vzError::warn(VZ_INVALID_VALUE, "pfVolume::draw(): NULL data for pWin %x", pw);
	 return;
    }
    
    vzClipRenderAction *renderer = (vzClipRenderAction *) pWinData;
    
    vzError::message(5, "pfVolume::draw(): Shape Set = %x, Renderer = %x", _shape_set, renderer);
    
    // Push the current OpenGL state
    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT);
    
    // Set the appropriate GL state
    glDisable(GL_LIGHTING);
    
    // Disable back face culling
    glDisable(GL_CULL_FACE);
    
    glColor4f(1, 1, 1, 1);
    
    // set the sampling rate for the renderer
    renderer->setSamplingRate(_sampling_rate);
    
    // Draw in sorted order
    renderer->draw(_shape_set);
    
    // Restore the OpenGL state
    glPopAttrib();
    
}

//////////////////////////////////////////////////////////////////////
// create and set the render action for the pfPipeWindow
//////////////////////////////////////////////////////////////////////
void pfVolume::initializeRenderer(pfPipeWindow *pw)
{
    // Create the pfvzRenderAction
    pfvzRenderAction *renderer = new (pfGetSharedArena()) pfvzRenderAction();
    
    if(!renderer){
      vzError::error(VZ_OUT_OF_MEMORY,"Could not create render action for pipe window %x", pw);
      return;
    }
    
    vzError::message(2, "Setting user data for pWin %x to %x", pw, renderer);
    
    // Set the user data for this pfPipeWindow
    pw->setUserData(pfvzRenderActionSlot, renderer);
    
    vzError::message(2, "After setting user data for pWin %x to %x", pw, renderer);
}

//////////////////////////////////////////////////////////////////////
// create the pfGeoSetCB and set the draw callback for the class
//////////////////////////////////////////////////////////////////////
void pfVolume::initializeGeoSetCB(pfPipeWindow *pipe)
{
    float bbox[6];

    // Get the bounding box for this shape's geometry
    _shape_set->getBoundingBox(bbox);
    
    pfBox	box;
    box . min[0] = bbox[0];
    box . min[1] = bbox[1];
    box . min[2] = bbox[2];
    box . max[0] = bbox[3];
    box . max[1] = bbox[4];
    box . max[2] = bbox[5];
    
    // Must set a static bounding box on the geoset. This geoset has no 
    // explicit geometry so Performer has no way to compute its bounding
    // box.
    pfGeoSetCB	*gsetCB = new pfGeoSetCB;
    
    gsetCB->setBound (&box, PFBOUND_STATIC);
    gsetCB->setDrawCB(drawCB, pipe);
    gsetCB->setGState(_geo_state);
    
    // Store the pfVolume as the user data in the pfGeoSetCB.
    // This will be used to retrieve the appropriate shape node
    // in the drawCB.
    gsetCB->setUserData(pfvzShapeSlot, this);
    
    addGSet(gsetCB);
    
    _windows[_num_windows] = pipe;
    _geoset_cbs[_num_windows] = gsetCB;
    
    _num_windows++;
    
    vzError::message(2,"pfVolume::initializeGeoSetCB(): Created a new pfGeoSetCB %x. Number = %d, pipe index = %d", gsetCB, _num_windows, pipe->getPipeIndex());
}


//////////////////////////////////////////////////////////////////////
// initialize a LUMINANCE_ALPHA lookup table
//////////////////////////////////////////////////////////////////////
vzParameterLookupTable*
pfVolume::initializeLookupTable(int width)
{
    float* data = (float*) pfMalloc(sizeof(float) * width * 2, pfGetSharedArena());
    for (int i = 0; i < width; i++) {
	 const float value = (float) i / (width - 1);
	 const float value2 = (float) i / (4*(width - 1));
	 if (i < width/4)
	   data[i * 2 + 0] = value2; // luminance
	 else
	   data[i * 2 + 0] = 1.0; // luminance
	 data[i * 2 + 1] = value; // alpha
    }
    
    vzParameterLookupTable* table =
	 new vzParameterLookupTable(width, data, VZ_FLOAT,
				    VZ_LUMINANCE_ALPHA);
    
    table->addDeletionCallback(cleanupLookupTable, data);
    
    return table;
}


//////////////////////////////////////////////////////////////////////
// lookup table deletion callback
//////////////////////////////////////////////////////////////////////
void
pfVolume::cleanupLookupTable(vzObject* object, void* data)
{
    pfDelete(data);
}

//////////////////////////////////////////////////////////////////////
// load a lookup table from the given file
//////////////////////////////////////////////////////////////////////
vzParameterLookupTable*
pfVolume::loadLookupTable(char *lutFile)
{
  vzError::message(1, "Loading lut data from %s", lutFile);
    
  // Open the file for reading.
  ifstream stream(lutFile);
  
  if (!stream) {
    vzError::error(VZ_INVALID_VALUE, " pfVolume: Could not read data from file %s", lutFile);
    return 0;
  }
    
  // Create a new TFManager. Loads in the lookup table
  // from the given stream
  TFManager *tf = TFManager::createFromStream(stream);

  // Load the lut data
  float *lutData = tf->getData();
  TFFormat lutFormat = tf->getFormat();
  int lutLength = tf->getLength();
    
  vzExternalTextureFormat format;
  switch (lutFormat) {
  case TF_LUMINANCE:
    format = VZ_LUMINANCE;
    break;
  case TF_LUMINANCE_ALPHA:
    format = VZ_LUMINANCE_ALPHA;
    break;
    
  case TF_RGBA:
    format = VZ_RGBA;
    break;
    
  default:
    return 0;
  }
    
  int numChannels = tf->getNumberOfChannels();
  int size = sizeof(float) * lutLength * numChannels;
  float *allocData = (float *) pfMalloc(size, pfGetSharedArena());
    
  vzError::message(1, "Data = %x, length = %d, numChannels = %d", lutFile, lutLength, numChannels);
    
  for(int i = 0; i < lutLength; i++)
    for(int j = 0; j < numChannels; j++)
      allocData[i * numChannels + j] = lutData[i * numChannels + j];
    
  delete tf;
   
  vzParameterLookupTable* table = new vzParameterLookupTable(lutLength, allocData, VZ_FLOAT, format);
    
  table->addDeletionCallback(cleanupLookupTable, allocData);
   
  return table;
}
