/*
  Bugs:
    line destructor sometimes crashes in RAVE
    memory leak when label text gets updated

  Features to address:
    everything scales with navigation scale- make it fixed size

    can't easily measure a long distance, as you need to navigate.  Need to
    go modal with buttons instead?

    make it use slb instead of reading button 2 directly

    The highlighting for edit and delete and the highlighting for
    highlighting a range of measurments have become entangled.  If I
    highlight some measurments with the highlight range command, they can
    become unhighlighted when I'm in edit or delete mode and I bring the
    wand in proximity to an end point then move away from that end point.



*/

//////////////////////////////////////////////////////////////////////////
//
// measurement1D DSO - select 2 points and measure the distance between
// them
//
//////////////////////////////////////////////////////////////////////////

#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <Performer/pf/pfText.h>
#include <Performer/pr/pfGeoState.h>
#include <Performer/pr/pfString.h>
#include <Performer/pr/pfFont.h>
#include "billboardNode.C"

// modes read from shared memory to indicate current mode
#define CREATEMODE (1)
#define EDITMODE (2)
#define DELETEMODE (3)
#define INACTIVEMODE (4)
#define UNDOMODE (5)
#define SAVEMODE (6)
#define RESTOREMODE (7)
#define HIGHLIGHTONMODE (8)
#define HIGHLIGHTOFFMODE (9)
#define DELETEALLMODE (10)

// global variables, used by both classes, I'm lazy...

// store a copy of the inherited geoState
static pfGeoState *gState ;

// font for labels
static pfFont *font ;

// class for a line segment
class line {
public:  
  pfVec3 p[2];
  float distance ;
  pfDCS *lineDCS ;
  pfDCS *markerDCS[2] ;
  pfDCS *distDCS ;
  pfString *label ;
  bool hLight ;
  // print information about a line
  void print(void) {
    printf("line::print:\n") ;
    printf("  line = %p\n",this) ;
    {
      pfVec3 foo = p[0] ;
      printf("  p[0] = %f %f %f\n", foo[0], foo[1], foo[2]) ;
    }
    {
      pfVec3 foo = p[1] ;
      printf("  p[1] = %f %f %f\n", foo[0], foo[1], foo[2]) ;
    }
    {
      printf("  lineDCS = %p\n",lineDCS) ;
      pfCoord foo ;
      lineDCS->getMatPtr()->getOrthoCoord(&foo) ;
      dpf_printPfCoord(foo) ;
    }
    {
      printf("  markerDCS[0] = %p\n",markerDCS[0]) ;
      pfCoord foo ;
      markerDCS[0]->getMatPtr()->getOrthoCoord(&foo) ;
      dpf_printPfCoord(foo) ;
    }
    {
      printf("  markerDCS[1] = %p\n",markerDCS[1]) ;
      pfCoord foo ;
      markerDCS[1]->getMatPtr()->getOrthoCoord(&foo) ;
      dpf_printPfCoord(foo) ;
    }
    {
      printf("  distDCS = %p\n",distDCS) ;
      pfCoord foo ;
      distDCS->getMatPtr()->getOrthoCoord(&foo) ;
      dpf_printPfCoord(foo) ;
    }
    printf("  label = %p, string = \"%s\"\n",label,label->getString()) ;
        printf("  hLight = %d\n",hLight) ;
  }
  line(pfVec3 p0, pfVec3 p1) {
    p[0] = p0 ;
    p[1] = p1 ;
    lineDCS = new pfDCS ;
    markerDCS[0] = new pfDCS ;
    markerDCS[1] = new pfDCS ;
        distDCS = new pfDCS ;
    
    label = new pfString() ;
    label->setFont(font);
    label->setMode(PFSTR_JUSTIFY, PFSTR_CENTER);
    label->setColor(0.0f, .3f, 1.0f, 1.0f);
    label->setGState(gState) ;
    
    // try it with a local string
    char *foo = strdup("000000000") ;
    label->setString(foo) ;
    //~label->setString("000000000") ;
    pfMatrix mat ;
    mat.makeScale(.03f, .03f, .03f) ;
    label->setMat(mat) ;
    hLight = false ;
    
  }
  
  ~line(void) {
    //printf("line destructor called\n") ;
    for (int i=0; i<lineDCS->getNumChildren(); i++) {
      lineDCS->removeChild(lineDCS->getChild(i)) ;
    }
    pfDelete(lineDCS) ;
    for (int i=0; i<markerDCS[0]->getNumChildren(); i++) {
      markerDCS[0]->removeChild(markerDCS[0]->getChild(i)) ;
    }
    pfDelete(markerDCS[0]) ;
    for (int i=0; i<markerDCS[1]->getNumChildren(); i++) {
      markerDCS[1]->removeChild(markerDCS[1]->getChild(i)) ;
    }
    pfDelete(markerDCS[1]) ;
        pfDelete(label) ;
    for (int i=0; i<distDCS->getNumChildren(); i++) {
      distDCS->removeChild(distDCS->getChild(i)) ;
    }
    pfDelete(distDCS) ;
    //printf("line destructor exiting\n") ;
  }
  
} ;

class measurement1D : public dtkAugment {
  
public:
  measurement1D (dpf *app) ;
  int preFrame (void);
  int postFrame (void);
  ~measurement1D ();
  
private:
  dpf* app ;
  dpfDisplay *display ;

  //  Shared memory segment to put in measurement1D mode
  dtkSharedMem* shm_measure ;

  // current mode- see defines above
  unsigned char mode ;

  // mode from last time ;
  unsigned char lastMode ;

  // the data scale factor- specified in MEASUREMENT1D_SCALE envvar, or 1.f if
  // MEASUREMENT1D_SCALE isn't set
  float measurement1D_scale ;

  // the offset from the wand to the wand pointer
  // used in pointerVisibility() and pointerPosition()
  pfVec3 pointerOffset ;

  // used for line endpoints and wand pointer
  // used in pointerVisibility() and addLineToScenegraph()
  pfNode *markerNode ;

  ////////////////////////////////////////////////////////////////////////
  // scenegraph layout:
  //
  // a pointer to the node named "noClip" if it exists, or "world" if it doesn't 
  // all nodes are under this node
  pfGroup *world ;

  // this DCS has the wand transformation
  // under it is the wand pointer
  pfDCS *wandDCS;

  // also under the world node are all the line DCSs and rubberband line.
  // they are added in measurement1D::addLineToScenegraph

  // the pfText that displays the distance is under the world node
  // with a DCS to put it on the wand or the line
  // world->distanceDCS->billboardNode->pfText

  // the pointer is under and offset SCS whcih is under wandDCS
  //
  ////////////////////////////////////////////////////////////////////////

  // an array of line object pointers
  line **lines ;

  // and how many of them are in the array
  int numLines ;

  // simple wrapper function to see if a string is a float
  bool isFloat(char *s, float *f) ;

  // load a file specified by an envvar or by name if env==NULL or not set
  pfNode *loadFile(char *env, char *name) ;

  // add a line to the scenegraph
  void addLineToScenegraph(line *l) ;

  // turn wand pointer on or off- 
  // wand pointer is the same pointer as a line endpoint
  void pointerVisibility(bool s) ;

  // return true if button is pressed, via shared memory
  bool buttonReleased(void) ;

  // update the DCS in the wandDCS node based on the wand position
  void update_wandDCS(void) ;

  // return position of wand
  pfCoord wandCoord(void) ;

  // returns the position of the wand pointer
  pfCoord pointerPosition() ;

  // moves the distance label of a line to the indicated position
  void setLabelPosition(line *l, pfVec3 v);

  // add a new line to the lines array
  void allocLines(int n) ;

  // determine heading and pitch of rubberband line
  float findHeading(pfVec3 v);
  float findPitch(pfVec3 v);

  // if the pointer is near enough to a point to select a point it sets the
  // pointer to the line containing the point, and a pointer to the point in
  // the line
  void getClosestPoint(line **l, pfVec3 **p) ;

  // indicates that the point next to the pointer can be selected. if
  // visible, the DCS is updated every time, so this can be called to update
  // the position of the selection marker as well as turn it on/off
  void updateSelectionMarker(pfVec3 *p, bool s) ;

  // highlights a line
  void highlightLine(line *l, bool s) ;

  // deletes a line from the arry and removes it from the scenegraph
  void deleteLine(line *l) ;

  // sets a string to be the 9 digit representation of a float
  void float9(char *s, float f) ;

  // update the lines' 3 DCS nodes so the line goes through the points and the
  // marker objects are at the points
  // argument is index into lines array
  void updateLine(line *l) ;

  // handle each mode
  void createMode(void) ;
  void editMode(void) ;
  void deleteMode(void) ;
  void deleteAllMode(void) ;
  void inactiveMode(void) ;
  void undoMode(void) ;
  void saveMode(void) ;
  void restoreMode(void) ;
  void highlightOnMode(void) ;
  void highlightOffMode(void) ;
} ;

////////////////////////////////////////////////////////////////////////
measurement1D::measurement1D (dpf *app_in) :
  dtkAugment ("measurement1D") {

  setDescription ("measurement1D DSO: allows the user to measure"
		  "distance between points in space.\n");

  app = app_in ;
  display = app_in->display() ;

  // how many lines are stored in the arrays
  numLines = 0 ;
  lines = NULL ;

  // mode from last time
  lastMode = INACTIVEMODE ;

  // offset from wand to wand pointer
  pointerOffset = pfVec3(0.f, 0.1f, 0.f) ;

  // define shared memory for measurement1D mode
  shm_measure = new dtkSharedMem(sizeof(unsigned char), "measureMode");
  if(shm_measure->isInvalid())
    return ;
  else
    shm_measure->flush();
  
  // initial mode
  mode = INACTIVEMODE;
  shm_measure->write(&mode);

  // get data scaling factor
  {
    char *foo = getenv("MEASUREMENT1D_SCALE");
    if (foo) {
      if (!isFloat(foo, &measurement1D_scale)) {
	dtkMsg.add(DTKMSG_ERROR, "measurement1D::measurement1D() failed: "
		   " envvar MEASUREMENT1D_SCALE, \"%s\", is not a "
		   "valid point number\n",
		   foo) ;
	return ;
      } 
    }
    else 
      measurement1D_scale = 1.f ;
    printf("measurement1D::measurement1D: measurement1D_scale = %f\n",measurement1D_scale) ;
  }
  validate ();
} // end of constructor

////////////////////////////////////////////////////////////////////////
measurement1D::~measurement1D ()
{
  for (int i=0; i<numLines; i++) 
    delete(lines[i]) ;
}  // end of destructor

////////////////////////////////////////////////////////////////////////
int measurement1D::preFrame (void) {

  // get a copy of the geoState, for setting text
  gState = new pfGeoState ;

  // find our font
  font = pfdLoadFont_type1("Helvetica-Narrow-Bold", PFDFONT_FILLED);
  
  if (!font) {
    dtkMsg.add(DTKMSG_ERROR,
	       "measurement1D::preFrame() failed: "
	       "can't load font\n") ;
    return dtkAugment::ERROR_;
  }


  // load the objects for specifying endpoints and the wand pointer
  if (!(markerNode = loadFile("arrowObjectName", "bluesphere.savg"))) {
    return dtkAugment::ERROR_;
  }

  // look for the node to hang everything under
  // look for noClip first
  world = (pfGroup*) display->world()->lookup("noClip", pfDCS::getClassType()) ;
  if (!world) {
    printf("measurement1D::preFrame: noClip node not found, using world\n") ;
    world = (pfGroup*) display->world() ;
  }
  else
    printf("measurement1D::preFrame: using noClip node instead of world node\n") ;

  // settle for world if it isn't found
  if (!world) {
    dtkMsg.add(DTKMSG_ERROR,
	       "measurement1D::preFrame() failed: "
	       "can't find world node\n") ;
    return dtkAugment::ERROR_;
  }
  
  wandDCS = new pfDCS ;
  world->addChild(wandDCS) ;

  // preload the highlight objects to make timing more consistent
  highlightLine(NULL, false) ;

  return dtkAugment::REMOVE_CALLBACK;

}  // end of preFrame

////////////////////////////////////////////////////////////////////////
int measurement1D::postFrame (void) {

  // get the current mode
  shm_measure->read(&mode);

  switch (mode) {
  case CREATEMODE:
    createMode() ;
    break ;
  case EDITMODE:
    editMode() ;
    break ;
  case DELETEMODE:
    deleteMode() ;
    break ;
  case DELETEALLMODE:
    deleteAllMode() ;
    break ;
  case INACTIVEMODE:
    inactiveMode() ;
    break ;
  case UNDOMODE:
    undoMode() ;
    break ;
  case SAVEMODE:
    saveMode() ;
    break ;
  case RESTOREMODE:
    restoreMode() ;
    break ;
  case HIGHLIGHTONMODE:
    highlightOnMode() ;
    break ;
  case HIGHLIGHTOFFMODE:
    highlightOffMode() ;
    break ;
  }

  // for things like svae/restore/hightlight which only get done once
  // also for undo later on
  lastMode = mode ;
  
  return dtkAugment::CONTINUE ; 
  
}

////////////////////////////////////////////////////////////////////////
// simple wrapper function to see if a string is a float
bool measurement1D::isFloat(char *s, float *f) {
  char *e ;
  double d = strtod(s, &e) ;
  if (s+strlen(s) == e) {
    *f = (float) d ;
    return true ;
  } else {
    return false ;
  }
}

////////////////////////////////////////////////////////////////////////
// load a file specified by an envvar or by name if env==NULL or not set
pfNode *measurement1D::loadFile(char *env, char *name) {
  char *fname = name ;
  if (env) {
    fname = getenv(env) ;
    if (!fname) {
      fname = name ;
    }
  }
  printf("measurement1D::loadFile: loading \"%s\"\n",fname) ;
  return pfdLoadFile(fname) ;
}

////////////////////////////////////////////////////////////////////////
// turn wand pointer on or off- 
// wand pointer is the same pointer as a line endpoint
void measurement1D::pointerVisibility(bool s) {
  static bool v = false ;
  static pfSCS *offset = NULL ;
  if (!offset) {
    //printf("measurement1D::pointerVisibility: creating and setting pfSCS\n") ;
    pfMatrix mat;
    mat.makeTrans(pointerOffset[0],pointerOffset[1], pointerOffset[2]) ;
    offset = new pfSCS(mat) ;
    offset->addChild(markerNode) ;
  }
  if (v != s) {
    if (s) {
      //printf("  turning pointer on\n") ;
      wandDCS->addChild(offset) ;
    }
    else {
      //printf("  turning pointer off\n") ;
      wandDCS->removeChild(offset) ;
    }
    v = s ;
  }
  return ;
}

////////////////////////////////////////////////////////////////////////
// returns the position of the wand pointer
pfCoord measurement1D::pointerPosition(void) {
  pfCoord w = wandCoord() ;
  display->dpfToWorld(&w) ;
  pfMatrix wandHPR ;
  wandHPR.makeEuler(w.hpr[0], w.hpr[1],w.hpr[2]) ;
  pfVec3 p ;
  p.xformVec(pointerOffset, wandHPR);
  w.xyz += p ;
  return w ;
}

////////////////////////////////////////////////////////////////////////
// sets a string to be the 9 digit representation of a float
// s needs to be able to hold the result
void  measurement1D::float9(char*s, float f) {
  float a ; 
  a = f<0?-f:f ;
  
  if (a==0.0f) 
    sprintf(s,"0.0") ;
  else if (a>=10000000.f || a<=.001f)
    sprintf(s,"%8.2e",f) ;
  else if (a>1000000.f)
    sprintf(s,"%-8.0f",f) ;
  else if (a>100000.f)
    sprintf(s,"%-7.0f",f) ;
  else if (a>10000.f)
    sprintf(s,"%-6.0f",f) ;
  else if (a>1000.f)
    sprintf(s,"%-5.0f",f) ;
  else if (a>100.f)
    sprintf(s,"%-5.1f",f) ;
  else if (a>10.f)
    sprintf(s,"%-5.2f",f) ;
  else if (a>1.f)
    sprintf(s,"%-5.3f",f) ;
  else if (a>0.1f)
    sprintf(s,"%-4.4f",f) ;
  else if (a>0.01f) 
    sprintf(s,"%-5.5f",f) ;
  else if (a>0.001f) 
    sprintf(s,"%-6.6f",f) ;
  else if (a>0.0001f) 
    sprintf(s,"%-7.7f",f) ;
  else if (a>0.00001f) 
    sprintf(s,"%-8.8f",f) ;
  else if (a>0.000001f) 
    sprintf(s,"%-9.9f",f) ;

}

////////////////////////////////////////////////////////////////////////
// add a line to the scenegraph
void measurement1D::addLineToScenegraph(line *l) {

  // stores the generic line and highlighted line objects
  static pfNode* lineNode ;

  // load in the generic object for the connecting lines
  if (!lineNode &&
      !(lineNode = loadFile("MEASUREMENT1D_UNITLINE","unitLine.savg"))) {
    dtkMsg.add(DTKMSG_ERROR, "measurement1D::addLineToScenegraph() failed: "
	       "unable to load file \"%s\"\n","unitLine.savg") ;
    return ;
  }

  // add the line and endpoint markers to the scenegraph
  world->addChild(l->lineDCS) ;
  world->addChild(l->markerDCS[0]) ;
  world->addChild(l->markerDCS[1]) ;
  
  // load the line and endpoint objects under the DCSs
  l->lineDCS->addChild(lineNode) ;
  l->markerDCS[0]->addChild(markerNode) ;
  l->markerDCS[1]->addChild(markerNode) ;
  
  // set up the distance label
  world->addChild(l->distDCS) ;
  
  billboardNode *bb = new billboardNode(&(display->headPos()->xyz)) ;
  bb->pointRotate();
  l->distDCS->addChild(bb) ;
  
  pfText *text = new pfText ;
  text->addString(l->label) ;
  
  bb->addChild(text) ;
  
  return ;
}

////////////////////////////////////////////////////////////////////////
// return true if button 2 is released
// reads button data from shared memory- slb can tie this to a button
bool measurement1D::buttonReleased(void) {
  
  // shared memory that contains the button presses
  static dtkSharedMem *buttonShm = NULL ;
  static char *buttonShmName = NULL ;

  // get the memory if the first time
  if (!buttonShm) {

    buttonShmName = getenv ("MEASUREMENT1D_BUTTON_SHM");
    if (!buttonShmName)
      buttonShmName = "measurement1DButton" ;

    buttonShm =  new dtkSharedMem (buttonShmName);
    if (buttonShm->isInvalid()) {
      dtkMsg.add(DTKMSG_ERROR,
		 "measurement1D::buttonReleased: "
		 "unable to get button shared memory \"%s\"\n", buttonShmName) ;
      
      buttonShm = NULL ;
      return false ;
    }
    buttonShm->queue() ;
    buttonShm->flush() ;
  }

  static char last_buttonReleased = 0 ;
  char buttonReleased ;

  if (!buttonShm->qread (&buttonReleased)) {
    return false ;
  }

  bool retValue ;
  if (!buttonReleased && last_buttonReleased) {
    printf("measurement1D::buttonReleased: button released\n") ;
    retValue = true ;
  }
  else
    retValue = false ;

  last_buttonReleased = buttonReleased ;
  return retValue ;
}

////////////////////////////////////////////////////////////////////////
// update the DCS in the wandDCS node to point with the wand
void measurement1D::update_wandDCS(void) {
  pfCoord w = wandCoord() ;
  display->dpfToWorld(&w) ;
  /*
  printf("measurement1D::update_wandDCS:  wand at %f %f %f   %f %f %f\n",
	 w.xyz[0], w.xyz[1], w.xyz[2],
	 w.hpr[0], w.hpr[1], w.hpr[2]) ;
  */
  wandDCS->setCoord(&w) ;
}

////////////////////////////////////////////////////////////////////////
// return position of wand
pfCoord measurement1D::wandCoord(void) {
  static dtkInLocator *wand = NULL ;
  if (!wand) {
    wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
    if(!wand || wand->isInvalid()) {
      wand = NULL ;
      dtkMsg.add(DTKMSG_ERROR,
		 "measurement1D::wandCoord: Unable to get wand.\n");
    }
  }
  static int last_frameCount = -1 ;
  int frameCount = pfGetFrameCount() ;
  static pfCoord w ;
  if (last_frameCount != frameCount) {
    float wandData[6];
    wand->read(wandData);
    w.xyz.set(wandData[0], wandData[1], wandData[2]) ;
    w.hpr.set(wandData[3], wandData[4], wandData[5]) ;
  }
  last_frameCount = frameCount ;
  /*
  printf("measurement1D::wandCoord:  wand at %f %f %f   %f %f %f\n",
	 w.xyz[0], w.xyz[1], w.xyz[2],
	 w.hpr[0], w.hpr[1], w.hpr[2]) ;
  */	 
  return w ;
}

////////////////////////////////////////////////////////////////////////
void measurement1D::createMode(void) {

  update_wandDCS() ;
  pointerVisibility(true) ;

  // keep track of how many clicks we've had in this mode
  static int last_clicks = 0 ;
  static int clicks = 0 ;
  
  // did we get a button release?
  if (buttonReleased()) {
    clicks++ ;
    //printf("measurement1D::createMode: clicks = %d\n", clicks) ;
  }

  // first click
  if (last_clicks==0 && clicks==1) {
    //printf("measurement1D::createMode: first button release\n") ;
    // make sure there's room in the lines array
    numLines++ ;
    allocLines(numLines) ;

    // set endpoints and update DCSs
    // set endpoints to the wand pointer
    pfCoord foo = pointerPosition() ;
    pfVec3 p = foo.xyz ;
    printf("measurement1D::createMode: line started at %f %f %f\n",p[0],p[1],p[2]) ;
    int i = numLines-1 ; // for 0 indexing of arrays
    lines[i] = new line(p, p) ;
    updateLine(lines[i]) ;

    // add the line and endpoint markers to the scenegraph
    addLineToScenegraph(lines[i]) ;

    // put the distance label at the pointer
    setLabelPosition(lines[i], p);
  } 

  // we fixed the first point and move the second point
  else if (clicks==1) {
    pfCoord foo = pointerPosition() ;
    pfVec3 p = foo.xyz ;
    int i = numLines-1 ; // for 0 indexing of arrays
    lines[i]->p[1] = p ;
    updateLine(lines[i]) ;
    
    // put the distance label at the pointer
    setLabelPosition(lines[i], p);

  } 

  // second click
  else if (last_clicks==1 && clicks==2) {
    //printf("measurement1D::createMode: second button release\n") ;
    pfCoord foo = pointerPosition() ;
    pfVec3 p = foo.xyz ;
    int i = numLines-1 ; // for 0 indexing of arrays
    lines[i]->p[1] = p ;
    printf("measurement1D::createMode: line ended at %f %f %f\n",p[0],p[1],p[2]) ;
    updateLine(lines[i]) ;
    
    // put the distance label at the center of the line
    pfVec3 center ;
    center.scale(.5f,(lines[i]->p[0]+p)) ;
    setLabelPosition(lines[i], center);

    // get ready for next time
    clicks = 0 ;
  } 
  last_clicks = clicks ;
}

////////////////////////////////////////////////////////////////////////
void measurement1D::editMode(void) {

  update_wandDCS() ;
  pointerVisibility(true) ;

  static pfVec3 *lastSelectedPoint = NULL ;
  static line *lastSelectedLine= NULL ;

  // keep track of how many clicks we've had in this mode
  static int last_clicks = 0 ;
  static int clicks = 0 ;
  
  // did we get a button release?
  if (buttonReleased()) {
    clicks++ ;
    //printf("measurement1D::editMode: clicks = %d\n", clicks) ;
  }

  // no clicks- fishing for points
  if (clicks==0) {
    line *whichLine = NULL ;
    pfVec3 *whichPoint = NULL ;
    // see if the pointer is over a point
    getClosestPoint(&whichLine, &whichPoint) ;
    if (whichPoint && !lastSelectedPoint) {
      updateSelectionMarker(whichPoint, true) ;
    } 
    // if we didn't it one this time and did last time
    // turn off selection highlighting
    else if (!whichPoint && lastSelectedPoint) {
      updateSelectionMarker(lastSelectedPoint, false) ;
    }
    // did we hit one this time, hit one last time, but they're different points?
    else if (whichPoint && lastSelectedPoint && (whichPoint!=lastSelectedPoint)) {
      updateSelectionMarker(lastSelectedPoint, false) ;
      updateSelectionMarker(whichPoint, true) ;
    }
    lastSelectedPoint = whichPoint ;
    lastSelectedLine = whichLine ;

  }

  // first click move endpoint with pointer if one selected
  else if (clicks==1 && lastSelectedPoint) {
    // move the point
    pfCoord foo = pointerPosition() ;
    *lastSelectedPoint = foo.xyz ;
    // update the rubberband line and endpoints
    updateLine(lastSelectedLine) ;
    // put the distance label at the pointer
    setLabelPosition(lastSelectedLine, foo.xyz);
    // move the selection highlighting too
    updateSelectionMarker(lastSelectedPoint, true) ;
    // turn off the selection highlighting
    updateSelectionMarker(lastSelectedPoint, false) ;
  } 

  // second click
  // turn off selection highlighting
  else if (last_clicks==1 && clicks==2 && lastSelectedPoint) {
    // move the point
    pfCoord foo = pointerPosition() ;
    *lastSelectedPoint = foo.xyz ;
    // update the rubberband line and endpoints
    updateLine(lastSelectedLine) ;
    // put the distance label at the center of the line
    pfVec3 center ;
    center.scale(.5f,(lastSelectedLine->p[0]+lastSelectedLine->p[1])) ;
    setLabelPosition(lastSelectedLine, center);
    lastSelectedPoint = NULL ;
    lastSelectedLine= NULL ;

    // get ready for next time
    clicks = 0 ;
  }
  
  // if nothing selected, don't count clicks
  if (!lastSelectedPoint)
    clicks = 0 ;
    
  last_clicks = clicks ;

}

////////////////////////////////////////////////////////////////////////
void measurement1D::deleteMode(void) {

  update_wandDCS() ;
  pointerVisibility(true) ;

  static pfVec3 *lastSelectedPoint = NULL ;
  static line *lastSelectedLine= NULL ;

  // how long must the line be hghlighted before you can delete it
  const float wait = .0f ;

  // keep track of how many clicks we've had in this mode
  static int last_clicks = 0 ;
  static int clicks = 0 ;
  
  // did we get a button release?
  if (buttonReleased()) {
    clicks++ ;
    //printf("measurement1D::editMode: clicks = %d\n", clicks) ;
  }

  // always fishing for points
  line *whichLine = NULL ;
  pfVec3 *whichPoint = NULL ;
  // see if the pointer is over a point
  getClosestPoint(&whichLine, &whichPoint) ;
  if (whichPoint && !lastSelectedPoint) {
    updateSelectionMarker(whichPoint, true) ;
  } 
  // if we didn't it one this time and did last time
  // turn off selection highlighting
  else if (!whichPoint && lastSelectedPoint) {
    updateSelectionMarker(lastSelectedPoint, false) ;
    highlightLine(lastSelectedLine, false) ;
    pfInitClock(0.) ;
  }
  // did we hit one this time, hit one last time, but they're different points?
  else if (whichPoint && lastSelectedPoint && (whichPoint!=lastSelectedPoint)) {
    updateSelectionMarker(lastSelectedPoint, false) ;
    highlightLine(lastSelectedLine, false) ;
    updateSelectionMarker(whichPoint, true) ;
    pfInitClock(0.) ;
  }
  lastSelectedPoint = whichPoint ;
  lastSelectedLine = whichLine ;

  // if one selected, and button released, start a timer
  if (last_clicks==0 && clicks==1 && lastSelectedPoint) {
    pfInitClock(0.) ;
  }
  
  // if one selected, and no new clicks
  else if (clicks==1 && lastSelectedPoint) {
    if (pfGetTime()>wait) {
      highlightLine(lastSelectedLine, true) ;
    }
  }
  
  // second click, and one selected, nuke it!
  else if (clicks==2 && lastSelectedPoint && 
	   pfGetTime()>wait) {
    deleteLine(lastSelectedLine) ;
  }

  // reset if all else fails
  else
    clicks = 0 ;

  last_clicks = clicks ;
  
}

////////////////////////////////////////////////////////////////////////
void measurement1D::deleteAllMode(void) {
  //printf("measurement1D::deleteAllMode\n") ;
  pointerVisibility(false) ;

  int n = numLines ;

  for (int i=numLines-1; i>=0; i--) {
    deleteLine(lines[i]) ;
  }

  dtkMsg.add(DTKMSG_NOTICE,
	     "measurement1D::deletAllMode(): %d lines deleted\n",n) ;

  // only one deleteAll at a time
  mode = lastMode ;
  if (shm_measure) {
    shm_measure->write(&mode);
  }
}

////////////////////////////////////////////////////////////////////////
void measurement1D::inactiveMode(void) {
  //printf("measurement1D::inactiveMode\n") ;
  pointerVisibility(false) ;
}

////////////////////////////////////////////////////////////////////////
void measurement1D::undoMode(void) {
  printf("measurement1D::undoMode\n") ;
  pointerVisibility(false) ;

  // only one undo at a time
  mode = lastMode ;
  if (shm_measure) {
    shm_measure->write(&mode);
  }
}

////////////////////////////////////////////////////////////////////////
void measurement1D::saveMode(void) {
  //printf("measurement1D::saveMode\n") ;
  pointerVisibility(false) ;

  //  Shared memory segment used to enter a file name
  static dtkSharedMem* shm_filename = NULL ;
  
  // define shared memory for a filename to be defined
  if (!shm_filename) {
    shm_filename = new dtkSharedMem(sizeof(char)*1024, "filesname");
    if(shm_filename->isInvalid()) {
      dtkMsg.add(DTKMSG_ERROR, "measurement1D::saveMode() failed: "
		 "unable to open DTK shared mameory file \"%s\"\n",
		 "filesname") ;
      return ;
    }
    else
      shm_filename->flush();
  }
  
  // get filename from shared memory
  char filename[1024] ;
  if (shm_filename) {
    shm_filename->read(filename);
  }
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    dtkMsg.add(DTKMSG_ERROR,
	       "measurement1D::saveMode(): failed: "
	       "can't open file \"%s\" for writing\n",filename) ;
  }
  else {
    // start the file with a comment line
    fprintf(fp, 
	    "# This is an output file generated by the measurement1D DSO\n"
	    "# Written on %s at %s, by \"%s\"\n"
	    "# Each line contains xyz for 2 points (6 numbers)\n"
	    "# Each data pair was scaled by the value %f\n"
	    "# the scale value is specified by the MEASUREMENT1D_SCALE envvar\n",
	    __DATE__, __TIME__, getenv("LOGNAME"), measurement1D_scale);
    for (int i=0; i<numLines; i++) {
      pfVec3 p0 ;
      p0.scale(measurement1D_scale,lines[i]->p[0]) ;
      pfVec3 p1 ;
      p1.scale(measurement1D_scale,lines[i]->p[1]) ;
      fprintf(fp,
	      "%f %f %f    %f %f %f\n",
	      p0[0], p0[1], p0[2], p1[0], p1[1], p1[2]) ;
    }
    dtkMsg.add(DTKMSG_NOTICE,
	       "measurement1D::saveMode(): "
	       "%d lines written to \"%s\"\n",numLines,filename) ;
  }
  // only one save at a time
  mode = lastMode ;
  if (shm_measure) {
    shm_measure->write(&mode);
  }
  fclose(fp) ;
}

////////////////////////////////////////////////////////////////////////
void measurement1D::restoreMode(void) {
  //printf("measurement1D::restoreMode\n") ;
  pointerVisibility(false) ;

  //  Shared memory segment used to enter a file name
  static dtkSharedMem* shm_filename = NULL ;
  
  // define shared memory for a filename to be defined
  if (!shm_filename) {
    shm_filename = new dtkSharedMem(sizeof(char)*1024, "filesname");
    if(shm_filename->isInvalid()) {
      dtkMsg.add(DTKMSG_ERROR, "measurement1D::restoreMode() failed: "
		 "unable to open DTK shared mameory file \"%s\"\n",
		 "filesname") ;
      return ;
    }
    else
      shm_filename->flush();
  }
  
  // get filename from shared memory
  char filename[1024] ;
  if (shm_filename) {
    shm_filename->read(filename);
  }

  FILE *fp = fopen(filename, "r");
  if (!fp) {
    dtkMsg.add(DTKMSG_ERROR,
	       "measurement1D::restoreMode() failed: "
	       "can't open file \"%s\" for reading\n",filename) ;
  }
  else {
    char s[250] ;
    int pargc;
    char **pargv = NULL ;
    int n = 0 ;
    while ((fgets(s,250,fp)!=NULL)) {
      //printf("got the next line:\n   %s\n", s);
      if (s != NULL && strlen(s) != 0 &&
	  s[0] != '#') {
	//printf("parsing line\n") ;
	dpf_charsToArgcArgv(s, &pargc, &pargv) ;
	if (!(pargv == NULL || pargc == 0 || pargv[0] == '\0') && pargc==6) {
	  pfVec3 p0, p1 ;
	  p0[0] = atof(pargv[0])/measurement1D_scale ;
	  p0[1] = atof(pargv[1])/measurement1D_scale ; 
	  p0[2] = atof(pargv[2])/measurement1D_scale ;
	  p1[0] = atof(pargv[3])/measurement1D_scale ;
	  p1[1] = atof(pargv[4])/measurement1D_scale ;
	  p1[2] = atof(pargv[5])/measurement1D_scale ;
	  n++ ;
	  numLines++ ;
	  allocLines(numLines) ;
	  int i = numLines-1 ; // for 0 indexing of arrays
	  lines[i] = new line(p0, p1) ;
	  updateLine(lines[i]) ;
	  addLineToScenegraph(lines[i]) ;
	  // put the distance label at the center of the line
	  pfVec3 center ;
	  center.scale(.5f,(p0+p1)) ;
	  setLabelPosition(lines[i], center);
	}
      }
    }
    dtkMsg.add(DTKMSG_NOTICE,
	       "measurement1D::restoreMode(): "
	       "%d lines read from \"%s\"\n",n, filename) ;
  }
  
  // only one restore at a time
  mode = lastMode ;
  if (shm_measure) {
    shm_measure->write(&mode);
  }
  fclose(fp) ;
}

////////////////////////////////////////////////////////////////////////
void measurement1D::highlightOnMode(void) {
  //printf("measurement1D::highlightOnMode\n") ;
  pointerVisibility(false) ;

  //  Shared memory segment used to pass min/max values
  static dtkSharedMem* shm_minmax = NULL ;

  if (!shm_minmax) {
    shm_minmax = new dtkSharedMem(sizeof(float)*2, "minmax");
    if(shm_minmax->isInvalid()) {
      dtkMsg.add(DTKMSG_ERROR, "measurement1D::highlightOnMode() failed: "
		 "unable to open DTK shared mameory file \"%s\"\n",
		 "minmax") ;
      return ;
    }
    else
      shm_minmax->flush();
  }

  if (shm_minmax && shm_minmax->isValid()) {
    // get the minmax values from shared memory
    float minmax[2];
    if (shm_minmax)
      shm_minmax->read(minmax);
    
    // turn off highlighting on all other nodes
    for (int i=0; i<numLines; i++) 
      highlightLine(lines[i],false) ;
    
    int n = 0 ;
    // turn on highlighting on specified nodes
    for (int i=0; i<numLines; i++) {
      // modify distance by data scale
      float dist = (lines[i]->distance)*measurement1D_scale ;
      // highlight if in range
      if (dist>minmax[0] && dist<=minmax[1]) {
	n++ ;
	dtkMsg.add(DTKMSG_NOTICE,
		   "measurement1D::highlightOnMode(): "
		   "line with distance %f highlighted\n",
		   dist) ;
	highlightLine(lines[i],true) ;
      }
    }
    
    dtkMsg.add(DTKMSG_NOTICE,
	       "measurement1D::highlightOnMode(): "
	       "%d lines highlighted, with range %f to %f\n",
	       n, minmax[0], minmax[1]) ;
  }

  // only one highlightOn at a time
  mode = lastMode ;
  if (shm_measure) {
    shm_measure->write(&mode);
  }
}

////////////////////////////////////////////////////////////////////////
void measurement1D::highlightOffMode(void) {
  //printf("measurement1D::highlightOffMode\n") ;
  pointerVisibility(false) ;

  // turn off highlighting on all other nodes
  for (int i=0; i<numLines; i++) 
    highlightLine(lines[i],false) ;

  // only one highlightOff at a time
  mode = lastMode ;
  if (shm_measure) {
    shm_measure->write(&mode);
  }
  dtkMsg.add(DTKMSG_NOTICE,
	     "measurement1D::highlightOffMode(): "
	     "highlighting turned off for all lines\n") ;
}

////////////////////////////////////////////////////////////////////////
void measurement1D::allocLines(int n) {
  static int num = 0 ;
  if (n>num) {
    num += 128 ;
    dtkMsg.add(DTKMSG_NOTICE,
	       "measurement1D::allocLines:  realloc, for %d lines\n",num) ;
    lines = (line**) realloc(lines,sizeof(line*)*num) ;
  }
}

////////////////////////////////////////////////////////////////////////
// if the pointer is near enough to a point to select a point it sets the
// pointer to the line containing the point, and a pointer to the point in
// the line
 
void measurement1D::getClosestPoint(line **l, pfVec3 **p) {
  
  //printf("%s %d\n",__FILE__,__LINE__) ;
  // cut-off distance, squared
  float nearEnoughSq = (.03f)*(.03f) ;

  // keep track of closest distance squared
  float closestSq = 2.f*nearEnoughSq ;

  // and pointer to closet line and point
  line *closestLine = NULL ;
  pfVec3 *closestPoint = NULL ;
  pfCoord foo = pointerPosition() ;
  pfVec3 pointer = foo.xyz ;

  float dsq ;
  for (int i=0; i<numLines; i++) {
    for (int j=0; j<2; j++) {
      dsq = pointer.sqrDistance(lines[i]->p[j]) ;
      if (dsq<nearEnoughSq) {
	if (dsq<closestSq) {
	  closestSq = dsq ;
	  closestLine = lines[i] ;
	  closestPoint = &(lines[i]->p[j]) ;
	}
      }
    }
  }
  *l = closestLine ;
  *p = closestPoint ;
  //printf("%s %d: closestLine = %p, closestPoint = %p\n",
  //__FILE__,__LINE__,closestLine,closestPoint) ;
  return ;
}

////////////////////////////////////////////////////////////////////////
// update the lines' 3 DCS nodes so the line goes through the points and the
// markers are at the points.  update distance and distance text
// argument is index into lines array
void  measurement1D::updateLine(line *l) {

  // do the rubberband line
  pfCoord lineCoord ;
  lineCoord.xyz = l->p[0] ;
  
  // get the offset relative to the origin
  pfVec3 p = l->p[1] - l->p[0] ;
  lineCoord.hpr[1] = findPitch(p);
  lineCoord.hpr[0] = findHeading(p);
  lineCoord.hpr[2] = 0.0;
  
  l->lineDCS->setCoord(&lineCoord);

  // calculate distance
  float s = l->distance =  p.length() ;
  l->lineDCS->setScale(0.2, 0.2, s);
    
  // update pfString with new distance
  // the labeled distance is scaled by the overall scaling factor
  
  // try it with a local variable
  char *c = new char[10] ;
  //~char c[10] ;
  float9(c, s*measurement1D_scale) ;
  l->label->setString(c) ;

  // now the markers at the end of each line
  // reuse hpr from rubberband line
  pfMatrix mat ;

  lineCoord.xyz = l->p[0] ;
  mat.makeCoord(&lineCoord) ;
  l->markerDCS[0]->setMat(mat) ;
  
  lineCoord.xyz = l->p[1] ;
  mat.makeCoord(&lineCoord) ;
  l->markerDCS[1]->setMat(mat) ;

}

////////////////////////////////////////////////////////////////////////
// return pitch of rubberband line
float measurement1D::findPitch(pfVec3 v) {
  float theta ;
  float ratio ;
  if (abs(v[2]) < 1.0e-6) {
    theta = 90.0;
  }
  else {
    ratio = sqrt(v[0]*v[0] + v[1]*v[1])/v[2];
    theta = (180.0/M_PI)*atan(ratio);
    if (v[2] < 0.0)
      theta = 180.0 - abs(theta);
  }

  return theta;
}

////////////////////////////////////////////////////////////////////////
// return heading of rubberband line
float measurement1D::findHeading(pfVec3 v) {
  float theta ;
  float ratio ;
  // if the z value is zero: return 90
  if (abs(v[1]) < 1.0e-6) {
    // either +90 or -90
    theta = 90.0;
    if (v[0] < 0.0)
      theta = -90.0;
  }
  else {
    ratio = (v[0]/v[1]);
    theta = (180.0/M_PI)*atan(ratio);
    if (v[1] > 0.0)
      theta = 180.0 - theta;
    else
      theta = -theta;
  }

  return theta;
}

////////////////////////////////////////////////////////////////////////
// moves the distance label of a line to the indicated position
void measurement1D::setLabelPosition(line *l, pfVec3 v) {
  pfMatrix mat ;
  mat.makeTrans(v[0], v[1], v[2]) ;
  l->distDCS->setMat(mat) ;
}
 
////////////////////////////////////////////////////////////////////////
// highlights a line
void measurement1D::highlightLine(line *l, bool s) {

  // used when highlighting by range or for deletion
  static pfNode *hl_markerNode = NULL ;
  static pfNode *hl_lineNode = NULL ;

  if (!hl_lineNode &&
      !(hl_lineNode = loadFile("MEASUREMENT1D_HIGHLIGHTLINE", "highLightLine.savg"))) {
    dtkMsg.add(DTKMSG_ERROR, "measurement1D::highlightLine: failed- "
	       "unable to load highlight line\n") ;
    return ;
  }
  if (!hl_markerNode && 
      !(hl_markerNode = loadFile("arrowHighLightName", "highlightsphere.savg"))) {
    dtkMsg.add(DTKMSG_ERROR, "measurement1D::highlightLine failed- "
	       "unable to load highlight marker\n") ;
    return ;
  }
  
  if (!l)
    return ;
  if (l->hLight != s) {
    if (s) {
      l->lineDCS->addChild(hl_lineNode) ;
      l->markerDCS[0]->addChild(hl_markerNode) ;
      l->markerDCS[1]->addChild(hl_markerNode) ;
    }
    else {
      l->lineDCS->removeChild(hl_lineNode) ;
      l->markerDCS[0]->removeChild(hl_markerNode) ;
      l->markerDCS[1]->removeChild(hl_markerNode) ;
    }
    l->hLight = s ;
  }
  return ;
}

////////////////////////////////////////////////////////////////////////
// deletes a line from the arry and removes it from the scenegraph
void measurement1D::deleteLine(line *l) {

  int i ;
  for (i=0; i<numLines; i++) {
    if (lines[i]==l) 
      break ;
  }

  if (i<numLines) {
    dtkMsg.add(DTKMSG_NOTICE,
	       "measurement1D::deleteLine: deleting line %d\n",i) ;
    // remove from scenegraph
    world->removeChild(l->lineDCS) ;
    world->removeChild(l->markerDCS[0]) ;
    world->removeChild(l->markerDCS[1]) ;
    world->removeChild(l->distDCS) ;
    // slide the rest of the array down by one
    numLines-- ;
    for (int j=i; j<numLines; j++) {
      //printf("line %d replaced by line %d\n",j,j+1) ;
      lines[j] = lines[j+1] ;
    }
    //~ deleting the line causes a crash.  For now, just remove from scenegraph
    //~delete l ;
  }
  else {
    dtkMsg.add(DTKMSG_ERROR,
	       "measurement1D::deleteLine: can't find line!\n") ;
  }

}

////////////////////////////////////////////////////////////////////////
// indicates that the point next to the pointer can be selected. if
// visible, the DCS is updated every time, so this can be called to update
// the position of the selection marker as well as turn it on/off
void measurement1D::updateSelectionMarker(pfVec3 *p, bool s) {
  // display when wand intersects a point
  static pfNode *selection_markerNode = NULL ; 
  if (!selection_markerNode &&
      !(selection_markerNode = loadFile("redObjectName", "purplesphere.savg"))) {
    dtkMsg.add(DTKMSG_ERROR, "measurement1D::updateSelectionMarker failed- "
	       "unable to load selection marker\n") ;
    return ;
  }
  
  static pfDCS *offset = NULL ;
  static bool v = false ;
  if (!offset) {
    offset = new pfDCS ;
    world->addChild(offset) ;
  }

  if (s)
    offset->setTrans((*p)[0], (*p)[1], (*p)[2]) ;
  
  if (v != s) {
    if (s) {
      //printf("measurement1D::updateSelectionMarker: turning on selection marker, "
      //"offset = %f %f %f\n",(*p)[0], (*p)[1], (*p)[2]) ;
      offset->addChild(selection_markerNode) ;
    }
    else
      offset->removeChild(selection_markerNode) ;
  v = s ;
  }
  return ;
}

////////////////////////////////////////////////////////////////////////
/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new measurement1D(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
