#include <math.h>
#include <values.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

/* 

   same as nistRav3Display but just uses one X server- only change is wsName

   */

/*
   DSO to set default graphics state for the NIST RAVE.
   NIST's RAVE has three screens, front and left walls and a floor.
   
   The center of the dpf cube is 4 feet back, 4 feet up, and 4 feet in from
   the either wall.
   
   one DIVERSE unit is 4 feet, or 1.21 meters
   
   The NIST RAVE uses 3 pipes, one per screen.

   pipe 0: front wall, 1280x1024
     front (right) wall is 96" wide, 93 high"
     contains one pwin, filling the whole pipe
       contains one screen, xo = 111, yo = 0, xs = 1058, ys = 1024
   
     the vertical center of the screen is 93/2 = 46.5" above the
     floor.  The origin is at 48", so 46.5-48 = -1.5" = vertical
     center below origin.  -1.5/48 = -.03125 dpf units
   
     So, center of the screen is x=0, y=1, z=-.03125.
     hpr = 0,0,0
    
   pipe 1: left wall, 1280x1024
     left wall is 96" wide, 93 high"
     contains one pwin filling entire window
       contains one screen, xo = 111, yo = 0, xs = 1058, ys = 1024
   
     like front wall, center is x=-1, y=0, z=-.03125. hpr = 90,0,0
   
     floor is 90" wide (across front wall) and 96" deep (along left wall).
     projector is oriented so that width of projector is across left wall,
     height is across front wall
   
   pipe 2: floor, 1280x1024
     contains one pwin filling entire window
       contains one screen, xo = 111, yo = 0, xs = 1058, ys = 991
   
     the screen goes through two rotations to get it to the right
     position.  The center of the screen in dpf coordinates is at x=
     (90/2-48) = -3" = -3/48 = -.0625, y=0, z=-1.  hpr=90,0,-90
   
*/

/*************** DTK C++ DSO loader/unloader functions ***********/

/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  dpfDisplay *display = app->display();
  if(display == NULL) return DTKDSO_LOAD_ERROR;

  // one DIVERSE unit is 4 feet, or 1.21 meters
  display->meters(1.21f) ;

  dpfPipe *pipe ;
  dpfPwin *pwin ;
  dpfScreen *screen ;

  pfCoord offset ;

  // three pipes:
  // pipe 0 is front wall, pipe 1 is left wall, pipe 2 is floor
  if (display->pipes(3)) return DTKDSO_LOAD_ERROR;

  // assume all pixels & screens are the same size
  float pixels = 1058.f ;  //actual displayable area == 2 dpf units
  printf("nistRave3Display: %f pixels per inch\n",pixels/96.f) ;
  float width = 1280.f ;
  float height = 1024.f ;
  float leftBlank = 111.f ;

  // front wall
  {
    pipe = display->first() ;
    pipe->wsName(":0.0") ;
    
    // one window in the pipe
    pipe->pwins(1) ;
    
    // use all of the screen
    pwin = pipe->first() ;
    
    pwin->name("Front") ;
    float left =   leftBlank/width ;
    float right =  (leftBlank+pixels)/width ;
    float bottom = 0.f ;
    float top =    1.f ;
    pwin->setOriginSize(0, 0, (int)width, (int)height) ;
    pwin->noBorder(1) ;
    pwin->stereo(1) ;
    pwin->clear(1) ;
    pwin->cursor(0) ;
    pwin->screens(1) ;
    
    screen = pwin->first() ;
    screen->setViewport(left, right, bottom, top) ;
    screen->simpleFrustum(0) ;
    offset.xyz = pfVec3(0.f, 0.f, 0.f) ;
    offset.hpr = pfVec3(0.f, 0.f, 0.f) ;
    screen->offset(offset) ;
    screen->center(pfVec3(0.f, 1.f, -.03125f)) ;
    screen->width(2.f*(right-left)*(width/pixels)) ;
    screen->height(2.f*(top-bottom)*(height/pixels)) ;
    printf("Front wall:\n") ;
    screen->print() ;
  }

  // left wall- pipe 1
  {
    pipe = pipe->next() ;
    pipe->wsName(":0.1") ;
    
    // one window in the pipe
    pipe->pwins(1) ;
    
    // use all of the screen
    pwin = pipe->first() ;
    pwin->name("Left") ;
    float left =   leftBlank/width ;
    float right =  (leftBlank+pixels)/width ;
    float bottom = 0.f ;
    float top =    1.f ;
    pwin->setOriginSize(0, 0, (int)width, (int)height) ;
    pwin->noBorder(1) ;
    pwin->stereo(1) ;
    pwin->clear(1) ;
    pwin->screens(1) ;
    pwin->cursor(0) ;
    
    screen = pwin->first() ;
    screen->setViewport(left, right, bottom, top) ;
    screen->simpleFrustum(0) ;
    offset.xyz = pfVec3(0.f, 0.f, 0.f) ;
    offset.hpr = pfVec3(90.f, 0.f, 0.f) ;
    screen->offset(offset) ;
    screen->center(pfVec3(-1.f, 0.f, -.03125f)) ;
    screen->width(2.f*(right-left)*(width/pixels)) ;
    screen->height(2.f*(top-bottom)*(height/pixels)) ;
    printf("Left wall:\n") ;
    screen->print() ;
  }

  // floor- pipe 2
  {
    pipe = pipe->next() ;
    pipe->wsName(":0.2") ;
    
    // one window in the pipe
    pipe->pwins(1) ;
    
    // use all of the screen
    pwin = pipe->first() ;
    pwin->name("Floor") ;
    float left =   leftBlank/width ;
    float right =  (leftBlank+pixels)/width ;
    float bottom = (990.f-height)/height ;
    float top =    1.f ;
    pwin->setOriginSize(0, 0, (int)width, (int)height) ;
    pwin->noBorder(1) ;
    pwin->stereo(1) ;
    pwin->clear(1) ;
    pwin->screens(1) ;
    pwin->cursor(0) ;
    
    screen = pwin->first() ;
    screen->setViewport(left, right, bottom, top) ;
    screen->simpleFrustum(0) ;
    offset.xyz = pfVec3(0.f, 0.f, 0.f) ;
    offset.hpr = pfVec3(90.f, -90.f, 0.f) ;
    screen->offset(offset) ;
    screen->center(pfVec3(0.f, 0.f, -1.f)) ;
    screen->width(2.f*(right-left)*(width/pixels)) ;
    screen->height(2.f*(top-bottom)*(height/pixels)) ;
    printf("Floor:\n") ;
    screen->print() ;
  }

  return DTKDSO_LOAD_UNLOAD;
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  return DTKDSO_UNLOAD_CONTINUE;
}


