// objSwitchExec - 
// A DIVERSE DSO which loads sets of files and commands specified 
// by an environment variable, DPF_ABSWITCH_FILES. Originally, the first
// of each pair is displayed. When the wand selects an object using
// button 1 on the mouse,
// its partner object will be displayed instead. If there is a command
// associated with these objects, it will be implemented when the object
// is selected.
//
// Authors:
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//
//    Steve Satterfield, John Hagedorn and John Kelso (SAVG)
//    provided the idea for the program
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
//#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "testSuffix.C"
#include "testM4.C"

#define ONKEY  1
#define OFFKEY  2
#define EXECKEY  3

class objSwitchExec : public dtkAugment {
public:
  objSwitchExec(dpf *app) ;
  ~objSwitchExec(void) ;
  int preFrame(void);
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;
  //  Shared memory segment for the spin rate
  dtkInLocator *wand ;
  dtkInLocator *head ;
  //  Shared memory segment used to request the wand pointer from another program
  dtkSharedMem* shm2;
  //  Shared memory segment used to request a system call
  dtkSharedMem* shm4;
  //  Shared memory segment to put the call
  dtkSharedMem* shm_call;
  //  Shared memory segment used to request a system call
  dtkSharedMem* shmm4;
  //  Shared memory segment to put the call
  dtkSharedMem* shm_call4;
  // Shared memory for parameters for boundingBox
  //dtkSharedMem* shmwire;

  unsigned char selected;
  unsigned char newCallSet;
  dtkInButton *button;
  dtkDequeuer *dequeuer;
  int objcnt;
  int ccnt;
  int iccnt;

  pfNode** objectPtrs;
  char** commandList;
  char** initCommandList;
  char* groupList;
  int* objNumbers;
  int** onClick;
  char** filenames;
  bool* isObjOn;
  bool* selectable;

  pfNode* node;
  char group;
  char* cmmnd;
  char* fname;

  pfCoord wandCoord;
  pfCoord headCoord;
  pfCoord* hdcoord;
  pfCoord ptCoord;
  pfMatrix headWorld;
  pfMatrix wandWorld;
  pfGroup* rootGroup;
  pfVec3 check;
  pfVec3 upVector;
  pfVec3 checkhead;
  pfVec3 centerVector;
  void resetAorB();

  int pargc ;
  char **pargv ;
  void readInit();
  void readOnOff();
  void readExec();
  int isFileNew();  
  void checkdcs();

  void addTheNode();
  void removeTheNode();
  void executeTheCommand();

  int selectSelectable();
  void setPtCoord();
  float wandData[6];
  float wandPos[6];
  pfDCS *wanddcs;
  float headData[6];
  float headPos[6];
  float headPos2[6];
  pfDCS *headdcs;
  pfDCS *headdcs2;

  bool objectSelected;
  int last_one_hit;
  bool button1On; // indicates if button 1 is being pressed
  bool button1LastFrame; // indicates old position of button
  char* file_a;
  char* file_b; 
  char* commandA;

  // names of input arguments
  char *command;
  char *selection;
  char *response;
  char *parameter;

  pfBox box;
  float objRadius;
  float diff;
  pfVec3 centerPt;
  pfVec3 intPt;
  float distToWand;
  float d;
  int times;
  int last;
  int j, num, numA, numB;
  bool checkNodeName;
} ;

// create object and set description.
// be sure to validate it!
objSwitchExec::objSwitchExec(dpf *app_in) :
  dtkAugment("objSwitchExec")
{
  setDescription("loads files specified by DPF_SWITCH_FILES"
		 " environment variable and enables you to "
		 "establish links between pairs of objects."
		 " The objects that are displayed have one "
		 "of several functions: when selected with "
		 "the wand, they can turn on (make visible)"
		 " an object, turn off (make invisible) an "
		 "object, or issue a command. "
		 " This version was compiled on 6/18/04.") ;
  app = app_in;
  validate();
  
  // counters for the numbers of objects, commands
  objcnt = 0;
  ccnt = 0;
  iccnt = 0;

  // arrays for object nodes, commands, group names
  objectPtrs = NULL;
  commandList = NULL;
  initCommandList = NULL;
  groupList = NULL;
  objNumbers = NULL;
  onClick = NULL;
  filenames = NULL;
  isObjOn = NULL;
  selectable = NULL;
  pargv = NULL ;

  file_b = new char[100];

  shm2 = new dtkSharedMem(sizeof(unsigned char), "pointerOn");
  if(!shm2 || shm2->isInvalid()) {
      shm2 = NULL;
  }
  else
    shm2->flush();

  shm4 = new dtkSharedMem(sizeof(unsigned char), "runCommandFlag");
  if(!shm4 || shm4->isInvalid()) {
      shm4 = NULL;
  }
  else
    shm4->flush();

  shm_call = new dtkSharedMem(sizeof(char)*1024, "runCommandString");
  if(!shm_call || shm_call->isInvalid())
    shm_call = NULL ;
  else
    shm_call->flush();


  shmm4 = new dtkSharedMem(sizeof(unsigned char), "call-m4");
  if(!shmm4 || shmm4->isInvalid()) {
      shmm4 = NULL;
  }
  else
    shmm4->flush();

  shm_call4 = new dtkSharedMem(sizeof(char)*100, "m4-command");
  if(!shm_call4 || shm_call4->isInvalid())
    shm_call4 = NULL ;
  else
    shm_call4->flush();

  // turn buttons off to start for wandPointer
  unsigned char turnit = 2;
  if (shm2)
     shm2->write(&turnit);

  newCallSet = 1;
  if (shm4)
    shm4->write(&newCallSet);
  button = NULL;
  dequeuer = NULL;
  objectSelected = false;
  last_one_hit = -1;
  button1On = false;
  button1LastFrame = false;

  //bsph = new pfSphere();
  wandPos[0] = wandPos[1] = wandPos[2] = wandPos[3] = 
    wandPos[4] = wandPos[5] = 0.0f ;
  headPos[0] = headPos[1] = headPos[2] = headPos[3] = 
    headPos[4] = headPos[5] = 0.0f ;
  headPos2[0] = headPos2[1] = headPos2[2] = headPos2[3] = 
    headPos2[4] = headPos2[5] = 0.0f ;

  wanddcs = NULL;
  headdcs = NULL;
  headdcs2 = NULL;
  checkNodeName = false;
  command = new char[100];
  selection = new char[100];
  response = new char[100];
  parameter = new char[100];
} ;

// destructor
objSwitchExec::~objSwitchExec(void) {
  dtkSharedMem_destroy("pointerOn");
  dtkSharedMem_destroy("call-m4");
  dtkSharedMem_destroy("m4-command");
}

#define MIN_BUTTONS 3

int objSwitchExec::preFrame(void)
{
  dequeuer = new dtkDequeuer(app->record());
  if(!dequeuer || dequeuer->isInvalid()) return DTK_CALLBACK_ERROR;

  display = app->display();
  if(!display) return DTK_CALLBACK_ERROR;

  button = (dtkInButton *) app->get("buttons", DTKINBUTTON_TYPE);
  if(!button) return DTK_CALLBACK_ERROR;

   if(button->number() < MIN_BUTTONS)
     {
       dtkMsg.add(DTKMSG_ERROR,
		  "buttonNavControl::preConfig() failed:"
		  " the dtkInButton named \"%s\" has "
		  "just %d buttons"
		  " which is less than the %d needed.\n",
		  button->getName(), button->number(),
		  MIN_BUTTONS);
      return DTK_CALLBACK_ERROR;
    }
  button->queue();

  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  head = (dtkInLocator *) app->get("head", DTKINLOCATOR_TYPE);

  if(!wand || wand->isInvalid() || !head || head->isInvalid()) {
    dtkMsg.add(DTKMSG_ERROR,
	       "objSwitchExec::preFrame() failed:"
	       " cannot local wand or head\n");
    return DTK_CALLBACK_ERROR;
  }

  // just call buttonNavControl::preFrame() once at the start.
  return DTK_REMOVE_CALLBACK;
}

int objSwitchExec::postConfig(void) {
  
  display = app->display();
  rootGroup = new pfGroup;
  display->world()->addChild(rootGroup);

  // init file loaders
  char *env = getenv("DPF_SWITCH_FILES") ;
  char *original_list ;
  char *line = new char[2047];

  // possible input lines
  //-----------------------------------------------------
  //  INIT       Selection  Response ResponseParameters  node
  //  SELECT      fileA      ON       fileB  
  //  SELECT      fileA      ON       fileB             groupB
  //  INIT   ON   fileA            
  //  INIT   ON   fileA                                 groupB
  //  SELECT      fileA      OFF      fileB
  //  INIT   EXEC command   
  //-----------------------------------------------------

  bool testm4;
  string openThisOne;

  if (env) {
    original_list = strdup(env) ;
    // strtok() is broken in GNU libc, so we roll our own here.
    char *end = original_list ;
    for(;*end;) {
      for(; *end != ':' && *end; end++);
      if(*end) { *end = '\0'; end++; }
      // add to the list of files to read
      if (strlen(original_list)>0) {
	// read each file and fill the filea and fileb arrays with names
	testm4 = testForM4(original_list);
	openThisOne = original_list;
	if (testm4) {
	  // set the flag to tell the other program to read a call
	  newCallSet = 2;
	  if (shm4)
	    shm4->write(&newCallSet);
	  
	  // use m4 to get the expected files
	  openThisOne = getM4FilteredFile(original_list,shmm4,shm_call4);

	  // wait for hev-m4 to finish
	  if (shmm4) {
	    while (newCallSet == 2) {
	      shmm4->read(&newCallSet);
	    }
	  }
	}
	//FILE *fp = fopen(original_list, "r");  
	FILE *fp = fopen(openThisOne.c_str(), "r");              
        if (!fp) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "objSwitchExec::preConfig: error opening input file %s\n",
		     original_list) ;
	  return DTK_CALLBACK_ERROR;
        }
	
	// printf("next file off of the list \"%s\"\n",original_list) ;
	while ((fgets(line,250,fp)!=NULL)) {
	  //printf("got the next line: \"%s\"\n", line);
	  // parse the line into 2 objects and a command if there
	  dpf_charsToArgcArgv(line, &pargc, &pargv) ;

	  // check for the correct number of parameters
	  if (pargc >2 && pargc < 7) {
	    strcpy(command,pargv[0]);
	    strcpy(selection,pargv[1]);
	    strcpy(response,pargv[2]);
	    //printf("command: %s\n",command);
	    //printf("selection: %s\n",selection);
	    //printf("response: %s\n",response);
	    if (pargc > 3)
	      strcpy(parameter,pargv[3]);

	    if (strcasecmp(command,"INIT")==0) {
	      // parse initialization line
	      readInit();
	    }
	    else if (strcasecmp(response,"ON")==0 || 
		     strcasecmp(response,"OFF")==0) {
	      // parse OFF command line
	      readOnOff();
	    }
	    else if (strcasecmp(response,"EXEC")==0) {
	      // parse EXEC line
	      readExec();
	    }
	    else {
	      // not a valid command in the line
	      dtkMsg.add(DTKMSG_ERROR,
		       "objSwitchExec:postConfig: input file error :"
		       "This line does not have a valid command.\n");
	    }
	  }
	}
	original_list = end;
      }
    }
  }

  char *old_path = (char *) pfGetFilePath();
  char *path = NULL;
  if(old_path)
    {
      // copy old_path from static memory to the heap.
      old_path = dtk_strdup(old_path);
      path = (char *) dtk_malloc(strlen(old_path) +
				 strlen("%s:.:"DPF_DATA_DIR));
      sprintf(path,"%s:.:"DPF_DATA_DIR, old_path);
      pfFilePath(path);
    }
  else // old_path == NULL
    pfFilePath(".:"DPF_DATA_DIR);

  return DTK_REMOVE_CALLBACK ; 
}

void objSwitchExec::readInit() {
  // first parse the selection - either ON or EXEC
  //printf("start of readInit: select is: %s\n",selection);
  if (strcasecmp(selection,"ON")==0) {
    // first parse the line in its entities
    // third parameter is file name - see if this is new
    fname = pargv[2];
    group = 'w';
    checkNodeName = false;
    if (pargc == 4) {
      group = pargv[3][0];
      checkNodeName = true;
      // see if this is a special node
      if (strcasecmp(pargv[3],"wand")==0)
	group = 'a';
      // see if this is a special node
      if (strcasecmp(pargv[3],"head_no_hpr")==0)
	group = 'r';
    }
    num = isFileNew();
    checkdcs();
    
    // num is the object number in the list: make it visible
    node = objectPtrs[num];
    group = groupList[num];
    isObjOn[num] = true;
    addTheNode();
  }
  else if (strcasecmp(selection,"EXEC")==0) {
    // store the command stored in response
    initCommandList = (char **) realloc(initCommandList, sizeof(char*)*(iccnt+1));
    initCommandList[iccnt] = strdup(response);
    ++iccnt;
  }
}

void objSwitchExec::readOnOff() {
  // line is: fileA ON fileB (group)
  // see if fileA is on the list
  fname = pargv[1];
  group = 'w';
  if (pargc > 4) {
    group = pargv[4][0];
    // see if this is a special node
    if (strcasecmp(pargv[4],"wand")==0)
      group = 'a';
    if (strcasecmp(pargv[4],"head_no_hpr")==0)
      group = 'r';
  }
  checkNodeName = false;
  numA = isFileNew();
  selectable[numA] = true;

  fname = pargv[3];
  if (pargc > 4) 
    checkNodeName = true;
  numB = isFileNew();
  checkdcs();

  // put b on a's list to turn on
  num = objNumbers[numA] + 1;
  onClick[numA] = (int *) realloc(onClick[numA], sizeof(int)*(2*num));
  num = objNumbers[numA];
  // store ON, OFF, and objB number
  onClick[numA][2*num] = ONKEY;
  if (strcasecmp(pargv[2],"OFF")==0)
    onClick[numA][2*num] = OFFKEY;
  onClick[numA][2*num+1] = numB;
  ++objNumbers[numA];
}

void objSwitchExec::checkdcs() {
  if (group == 'a' && wanddcs == NULL) {
    // define the dcs
    wanddcs = new dpfDCS(wandPos, /* position */
			 1.0f, /* overall scale */
			 /* file  scale xyz position xyz rotation hpr */
			 NULL, DPF_IDENTITY,
			 NULL);

    wanddcs->setName("wanddcs"); 
    display->scene()->addChild(wanddcs);
  }

  if (group == 'h' && headdcs == NULL) {
    // define the head dcs
    headdcs = new dpfDCS(headPos, /* position */
			 1.0f, /* overall scale */
			 /* file  scale xyz position xyz rotation hpr */
			 NULL, DPF_IDENTITY,
			 NULL);
  
    headdcs->setName("headdcs");	
    display->scene()->addChild(headdcs);
  }

  if (group == 'r' && headdcs2 == NULL) {
    // define the head (no hpr) dcs
    headdcs2 = new dpfDCS(headPos2, /* position */
			 1.0f, /* overall scale */
			 /* file  scale xyz position xyz rotation hpr */
			 NULL, DPF_IDENTITY,
			 NULL);
	  
    headdcs2->setName("head_no_hprdcs"); 
    display->scene()->addChild(headdcs2);
  }
}

void objSwitchExec::readExec() {
  // line is fileA EXEC command
  // see if fileA is on the list
  fname = pargv[1];
  group = 'w';
  checkNodeName = false;
  if (pargc == 5) {
    group = pargv[4][0];
    checkNodeName = true;
  }
  numA = isFileNew();
  selectable[numA] = true;

  // store the command
  commandList = (char **) realloc(commandList, sizeof(char*)*(ccnt+1));
  commandA = strdup(pargv[3]);
  commandList[ccnt] = commandA;
  ++ccnt;

  // store the EXEC = 3
  num = objNumbers[numA];
  onClick[numA] = (int *) realloc(onClick[numA], sizeof(int)*(2*num+4));
  onClick[numA][2*num] = EXECKEY;
  onClick[numA][2*num+1] = ccnt - 1;
  ++objNumbers[numA];
}

int objSwitchExec::isFileNew() {
  num = -1;
  for ( int i = 0; i < objcnt; ++i ) {
    if (strcasecmp(fname, filenames[i])==0) {
      num = i;
      break;
    }
  }

  // if not found: add to list
  if (num < 0) {
    num = objcnt;
    // test suffixes and store files
    testSuffix(fname);
    objectPtrs = (pfNode **) realloc(objectPtrs, sizeof(pfNode*)*(objcnt+1));
    groupList = (char *) realloc(groupList, sizeof(char)*(objcnt+1));	
    objNumbers = (int *) realloc(objNumbers, sizeof(int)*(objcnt+1));
    onClick = (int **) realloc(onClick, sizeof(int*)*(objcnt+1));
    onClick[objcnt] = new int[2];
    filenames = (char **) realloc(filenames, sizeof(char*)*(objcnt+1));
    isObjOn = (bool *) realloc(isObjOn, sizeof(bool)*(objcnt+1));
    selectable = (bool *) realloc(selectable, sizeof(bool)*(objcnt+1));
    objectPtrs[objcnt] = pfdLoadFile(fname);
    file_a = strdup(fname) ;
    filenames[objcnt] = file_a;
    groupList[objcnt] = group;
    objNumbers[objcnt] = 0;
    isObjOn[objcnt] = false;
    selectable[objcnt] = false;

    num = objcnt;
    ++objcnt;
  }
  else if (checkNodeName) {
    // see if the group is updated
    group = pargv[4][0];
    if (group != groupList[num]) {
      // remove object from the old node
      group = groupList[num];
      node = objectPtrs[num];
      if (group != 'a' && group != 'h' && group != 'r') {
	// remove and add the node if it is visible
	if (isObjOn[num]) {	
	  removeTheNode();

	  // update the group
	  group = pargv[4][0];
	  groupList[num] = group;
	
	  // hang object on new node
	  addTheNode();
	}
	else {
	  // just update the group
	  group = pargv[4][0];
	  groupList[num] = group;
	}
      }
    }
  }

  return num;
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int objSwitchExec::postFrame(void) {

  // first read the button (whichbutton) on the mouse/wand
  int whichButton;
  dtkRecord_event *e = dequeuer->getNextEvent(button);
  int buttonState = button->read(&whichButton, e);

  //--------------------------------------------------
  // button1On; indicates if button 1 is being pressed
  // button1LastFrame; indicates old position of button 1
  //--------------------------------------------------

  if(whichButton == 1) {
    // first see if button 1 is currently pressed
    if (button->read(1,e)) {
      if (!button1LastFrame) {
	// switch button1ON: if off, turn it on; if on, turn it off
	if (button1On) 
	  button1On = false;
	else
	  button1On = true;
      }
      button1LastFrame = true;
    }
    else
      button1LastFrame = false;
  }
  //--------------------------------------------------

  // read the wand direction 
  if (wand)
      wand->read(wandData);
  //printf("objSwitchExec wand data: %f %f %f\n",wandData[3],wandData[4],wandData[5]);
  // read the head direction 
  if (head)
      head->read(headData);

  // create a vector in the +y direction and move according to the wand
  pfMatrix wandWorld;
  wandWorld.makeEuler(wandData[3],wandData[4],wandData[5]);
 
  upVector.set(0.f, 1.f, 0.f);
  check.xformVec(upVector, wandWorld);

  wandPos[0] = wandData[0];
  wandPos[1] = wandData[1];
  wandPos[2] = wandData[2];
  wandPos[3] = wandData[3];
  wandPos[4] = wandData[4];
  wandPos[5] = wandData[5];

  headPos[0] = headData[0];
  headPos[1] = headData[1];
  headPos[2] = headData[2];
  headPos[3] = headData[3];
  headPos[4] = headData[4];
  headPos[5] = headData[5];

  headPos2[0] = headData[0];
  headPos2[1] = headData[1];
  headPos2[2] = headData[2];
  headPos2[3] = 0.0;
  headPos2[4] = 0.0;
  headPos2[5] = 0.0;

  if (button1On) {
    // make the wandPointer appear
    if (shm2) {
      selected = 1;
      shm2->write(&selected);
    }

    // do the intersection and find the visible object that is closest
    // to the wand
    // the next routine reset the last_one_hit object number if found
    int last = selectSelectable();

    // get new value of newCallSet
    if (shm4)
      shm4->read(&newCallSet);

    if (last > -1) {
      last_one_hit = last;
      objectSelected = true;
      // reset the toggle
      resetAorB();

      // make a sound to indicate that the object was selected
      button1On = false;
      // make the wandPointer disappear
      if (shm2) {
	selected = 2;
	shm2->write(&selected);
      }
    }
  }
  else {
    // button1Off - restart
    objectSelected = false;
  } 

  // initialization - a one time thing - execute initial commands
  if (iccnt > 0) {
    for ( int i = 0; i < iccnt; ++i ) {
      commandA = initCommandList[i];
      if (commandA != NULL) {
	// set the flag to tell the other program to read a call
	newCallSet = 2;
	if (shm4)
	  shm4->write(&newCallSet);
	if (shm_call) {
	  //printf("writing to dtk-runCommand: %s\n",commandA);
	  shm_call->write(commandA);

	  // now wait until it is sent before sending the next one
	  while (newCallSet == 2)
	    shm4->read(&newCallSet);
	}
      }  
    }
    iccnt = 0;
  }
  return DTK_CONTINUE ;
}

void objSwitchExec::resetAorB() {
  // go through the list of things to do for the chosen node
  for (j = 0; j < objNumbers[last_one_hit]; ++j) {
    numA = onClick[last_one_hit][2*j];
    numB = onClick[last_one_hit][2*j+1];

    if (numA == ONKEY) {
      // turn on node numB
      node = objectPtrs[numB];
      group = groupList[numB];
      addTheNode();
      isObjOn[numB] = true;
    }
    else if (numA == OFFKEY) {
      // turn on node numB
      node = objectPtrs[numB];
      group = groupList[numB];
      removeTheNode();
      isObjOn[numB] = false;
    }
    else if (numA == EXECKEY) {
      // execute the command
      commandA = commandList[numB];

      // as long as the call is not NULL, 
      // read it out so dtk-runCommand can run it
      if (commandA != NULL) {
	// set the flag to tell the other program to read a call
	newCallSet = 2;
	if (shm4)
	  shm4->write(&newCallSet);
	if (shm_call) {
	  shm_call->write(commandA);
	}
      }  
    }
  }
}

void objSwitchExec::addTheNode() {
  if (group == 'w') 
    display->world()->addChild(node);
  
  if (group == 's') 
    display->scene()->addChild(node);
  
  if (group == 'n')
    display->nav()->addChild(node);
  
  if (group == 'e')
    display->ether()->addChild(node);
  
  if (group == 'a')
    wanddcs->addChild(node);
  
  if (group == 'h')
    headdcs->addChild(node);
  
  if (group == 'r')
    headdcs2->addChild(node);
}

void objSwitchExec::removeTheNode() {
  if (group == 'w') 
    display->world()->removeChild(node);
  
  if (group == 's') 
    display->scene()->removeChild(node);
  
  if (group == 'n')
    display->nav()->removeChild(node);
  
  if (group == 'e')
    display->ether()->removeChild(node);

  if (group == 'a')
    wanddcs->removeChild(node);
  
  if (group == 'h')
    headdcs->removeChild(node);
  
  if (group == 'r')
    headdcs2->removeChild(node);
}

int objSwitchExec::selectSelectable() {
  int last_one_hit = -1;

  // for each visible object, see if the check vector is within
  // its bounding sphere, and keep track of the one that is
  // the closest to the wand

  // loop through each pair and find the visible one
  distToWand = 1000000.0;

  for (j = 0; j < objcnt; j++ ) {
    if (last_one_hit > -1) break;
    if (isObjOn[j] && selectable[j]) {
      group = groupList[j];
      node = objectPtrs[j];
      if (node == NULL) continue;

      // find the center point of the object in world coordinates
      setPtCoord();
      
      // Is the distance from centerPt to the wand is less than the radius?
      
      // This code is from the objectMover: OMS.C
      // the line from the wand is defined by a point(wandCoord) 
      // and a direction (check) r = r0 + d(direction);
      // so, need to find the distance d up the line that is perpendicular
      // to the center point, that is at the point P=(a,b,c)
      
      // the dot product of the P-centr with P-wand = 0 
      // if the vectors are perpendicular
      // did some algebra and came up with an expression for d:
      float d = check[0]*(ptCoord.xyz[0]-wandData[0]) +
	check[1]*(ptCoord.xyz[1]-wandData[1]) +
	check[2]*(ptCoord.xyz[2]-wandData[2]);
      d = d/(check[0]*check[0] + check[1]*check[1] + check[2]*check[2]);
      
      // this gives the point of intersection with the wand, int: now find
      // the distance from the centerpoint to int
      intPt[0] = wandData[0] + d*check[0];
      intPt[1] = wandData[1] + d*check[1];
      intPt[2] = wandData[2] + d*check[2];
      //printf("current check: %f %f %f\n",check[0],check[1],check[2]);
      float dist = sqrt((ptCoord.xyz[0]-intPt[0])*(ptCoord.xyz[0]-intPt[0]) +
			(ptCoord.xyz[1]-intPt[1])*(ptCoord.xyz[1]-intPt[1]) +
			(ptCoord.xyz[2]-intPt[2])*(ptCoord.xyz[2]-intPt[2]));
      //printf("dist: %f \n",dist);
      
      // see if the sphere hits the wand
      if (dist < objRadius && d > 0) {
	// this object hits!
	// save distance = d
	if (dist < distToWand) {
	  distToWand = dist;
	  last_one_hit = j;
	  break;
	}
      }
    }
  }

  return last_one_hit;
}

void objSwitchExec::setPtCoord() {

  // find bounding sphere
  box = dpf_getBBox(node);
 
  // find the center point of the box
  centerPt[0] = (box.max[0]+box.min[0])/2.0;
  centerPt[1] = (box.max[1]+box.min[1])/2.0;
  centerPt[2] = (box.max[2]+box.min[2])/2.0;

  diff = abs(box.max[0]-box.min[0]);
  if (abs(box.max[1]-box.min[1]) > diff)
    diff = abs(box.max[1]-box.min[1]);
  if (abs(box.max[2]-box.min[2]) > diff)
    diff = abs(box.max[2]-box.min[2]);
  objRadius = diff/1.5;
  //printf("new radius: %f\n",objRadius);
  //printf("center point for object 1: %f %f %f\n",centerPt[0],centerPt[1],centerPt[2]);
  
  ptCoord.xyz.set(centerPt[0], centerPt[1], centerPt[2]);
  
  if (group == 'h' || group == 'r') {
    hdcoord = display->headPos();

    if (group == 'h') {
      // make a matrix for the head rotation
      headWorld.makeEuler(hdcoord->hpr[0],hdcoord->hpr[1],hdcoord->hpr[2]);
      // transform the center point by this matrix
      centerVector.set(centerPt[0], centerPt[1], centerPt[2]);
      checkhead.xformVec(centerVector, headWorld);
      ptCoord.xyz.set(checkhead[0], checkhead[1], checkhead[2]);
    }
    //printf("ptcoord in head: %f %f %f\n",ptCoord.xyz[0],ptCoord.xyz[1],ptCoord.xyz[2]);
    // move the object with the head xyz
    ptCoord.xyz[0] += hdcoord->xyz[0];
    ptCoord.xyz[1] += hdcoord->xyz[1];
    ptCoord.xyz[2] += hdcoord->xyz[2];
  }

  else if (group == 'e') {
    display->etherToDpf(&ptCoord);
  }

  //printf("center point for object is: %f %f %f\n",ptCoord.xyz[0],ptCoord.xyz[1],ptCoord.xyz[2]);
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new objSwitchExec(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
