//////////////////////////////////////////////////////////////////////////
//
// rubberbandBox DSO - provides user-stretchable box
//
// Overview:
//
//    A wireframe axis-aligned box is displayed.  A 3D cursor is displayed 
//    and hovers 0.2 units out from the wand.
//
//    The user moves the 3D cursor to the box and presses a button to
//    interact with it.  There are two types of interaction:
// 
//        Reposition Box:  The user moves the cursor to the interior
//                  of the box, presses the button and drags the box
//
//	  Resize Box:  The user moves the cursor to one of the faces
//		    of the box, presses the button and drags the face.
//
//    For both types of interactions, the box highlights in appropriate
//    ways when the cursor is in pickable proximity of something that
//    can be picked and dragged.
//
//    Wand position comes in through the dtkInLocator, the button comes
//    in through dtk shared memory (see slb DSO). 
//
//    The box goes out to other DSOs by a pointer provided by the
//    function getRubberbandBoxPtr.  (Should probably change this.)
//
//    The box is always aligned with the world coordinate axes and
//    the coordinates are always given in world coordinates.
//
//////////////////////////////////////////////////////////////////////////

/*

  To Do:

	- change name and mechanism for passing PublicBox ptr to other DSOs
	- frame accuracy (synchronization)
	- should currentBox and PublicBox be separate?
	- should file global vars be class variables?  
	   (interactionIsInProgress, PublicBox)
	- should there be a shm output option for box?
	- mechanism for specifying init pos of box?
	- should Box typedef be pfBox?  (Probably so.)
	- remove dependence on noClipWorld DSO? 
*/



#include <dpf.h>
#include <dtk/dtkDSO_loader.h>



//
// 
// The class rubberbandBox : public dtkAugment can only instantiated once.
// 
// Given that only one rubberbandBox exists at any one time, is there
// really any difference between class variables and variables that
// are declared global (but static) in this source file?
// 
// Is it better programming practice to use class variables?
// 
// I'm not really sure.
// 
// The use of the file global variables does seem to simplify some of the 
// programming.
// 
// 



#define PT_IN_BOX(p,b) \
  (\
  ((p)[0]>=(b)[0][0])&&\
  ((p)[0]<=(b)[1][0])&&\
  ((p)[1]>=(b)[0][1])&&\
  ((p)[1]<=(b)[1][1])&&\
  ((p)[2]>=(b)[0][2])&&\
  ((p)[2]<=(b)[1][2])\
  )



//
// Indicates if the user is in the process of interacting with the box.
//
static int interactionIsInProgress = 0;

//
// Data type for a box:  box[0][*] is min corner, box[1][*] is max corner
//
typedef double Box[2][3];

//
// Specifies which type of object the user has picked.
//
typedef enum
  {
  PICK_NONE = 0,
  PICK_ENTIRE = 1,
  PICK_FACE = 2
  } PickItem;


// This is the publically accessible version of the current box.
static Box *PublicBox;

// This routine provides the current box to other DSOs.
Box *getRubberbandBoxPtr ()
  {
  //printf ("getRubberbandBoxPtr %p first element %f\n", 
  //				PublicBox, (*PublicBox)[0][0]);
  return PublicBox;
  }



//
// Here's the class on which the DSO is based.
//
class rubberbandBox : public dtkAugment {

public:

  // Here are the methods needed by dtkDSO_loader
  rubberbandBox (dpf *app) ;
  int preConfig (void);
  int postConfig (void);
  int preFrame (void);
  int postFrame (void);
  ~rubberbandBox ();


private:


  dpf* app;
  dpfDisplay *display;

  // we get the wand position directly from dtk
  dtkInLocator *wand;

  // we get the wand button through shared memory
  dtkSharedMem *buttonShm;
  dtkSharedMem *boxCoordShm;

  // The current box as specified interactively by the user
  Box currentBox;

  double boxSize[3];  	// This is redundant with currentBox; it is here
			// for convenience.  It must be kept consistent
			// with currentBox.


  // These bounding boxes indicate the pickable 3D regions that correspond
  // to each of the pickable items.
  Box entirePickBox;    	// the pick box for picking the whole box
  Box facePickBox[2][3];	// the pick boxes for each face
  Box *currentPickBox;  	// This is the pick box that the wand
				// is currently inside

  // Highlights box features based on wand pos
  void doHighlighting (pfVec3 *xyz, 
		       PickItem *whichItem, int *faceI, int *faceJ);

  // Figure out whichItem has been picked
  void determineWhichPick (Box *currentPickBox, PickItem *whichItem, 
						   int *faceI, int *faceJ);

  // Various Performer data items that we need:
  pfDCS *wireframeDCS;
  pfGeode *nohl_entire;
  pfGeode *hl_entire;
  pfGeode *hl_face[2][3];
  pfGeode *currentChild;
  pfDCS *cursorDCS;
  pfGeode *cursorNode;

  void writeBoxCoordShm ();

} ;


//////////////////////////////////////////////////////////////////////

//
// setPickBoxes figures out the 3D pick regions for each of the 
// pickable items associated with the indicated Box.  The pickable
// items are each of the six faces and the entire box.
//
// This routine makes sure that each of these 3D pick regions is
// large enough and not so overlapping with its neighbors so that it can 
// be practically picked by the user.
//
// It also forces a minimal overlap between neighbors.  This is to provide
// a certain amount of "stickiness" to the old region as the user
// moves the wand position into a new region.
//
static void
setPickBoxes (float scale, Box box, Box entirePickBox, Box facePickBox[2][3])
  {
  Box localBox;
  int i, ii;
  double size, center;

  memcpy (localBox, box, sizeof(Box));
  for (ii = 0; ii < 3; ii++)
	{
	localBox[0][ii] *= scale;
	localBox[1][ii] *= scale;
	}


// In each dimension, each pick box must be have least 0.06 in dpf coords
//  that does not overlap with neighbors,
//  and faces must overlap entire by 0.02 in dpf coords
// This means that entirePickBox must be at leas 0.1 in each dimension


  for (i = 0; i < 3; i++)
    {
    size = localBox[1][i] - localBox[0][i];
    center = localBox[0][i] + (size/2);

    if (size < 0.1)
	{
	size = 0.1;
	}

    entirePickBox[0][i] = center - size/2;
    entirePickBox[1][i] = center + size/2;
    }


  for (i = 0; i < 3; i++)
    {
    // Set face pick box to be extent of original box, then
    // we adjust the extent only along the axis perpendicular
    // to the face.
    memcpy (facePickBox[0][i],entirePickBox, sizeof(Box));
    facePickBox[0][i][0][i] = entirePickBox[0][i] - 0.06;
    facePickBox[0][i][1][i] = entirePickBox[0][i] + 0.02;
    for (ii = 0; ii < 3; ii++)
	{
	facePickBox[0][i][0][ii] /= scale;
	facePickBox[0][i][1][ii] /= scale;
	}

    memcpy (facePickBox[1][i],entirePickBox, sizeof(Box));
    facePickBox[1][i][0][i] = entirePickBox[1][i] - 0.02;
    facePickBox[1][i][1][i] = entirePickBox[1][i] + 0.06;
    for (ii = 0; ii < 3; ii++)
	{
	facePickBox[1][i][0][ii] /= scale;
	facePickBox[1][i][1][ii] /= scale;
	}

    }

  for (ii = 0; ii < 3; ii++)
	{
	entirePickBox[0][ii] /= scale;
	entirePickBox[1][ii] /= scale;
	}


  }  // end of setPickBoxes

void
rubberbandBox::writeBoxCoordShm ()
	{
	float fbox[6];

	if (boxCoordShm != NULL)
		{
		fbox[0] = currentBox[0][0];
		fbox[1] = currentBox[0][1];
		fbox[2] = currentBox[0][2];
		fbox[3] = currentBox[1][0];
		fbox[4] = currentBox[1][1];
		fbox[5] = currentBox[1][2];
		boxCoordShm->write (fbox);
		}

	}  // end of rubberbandBox::writeBoxCoordShm


// Figures out whichItem has been picked (and associated face indices)
// based on Box *box.
void
rubberbandBox::determineWhichPick (Box *box, PickItem *whichItem, 
						   int *faceI, int *faceJ)
	{
	int i, j;

	if (box == &entirePickBox)
		{
		*whichItem = PICK_ENTIRE;
		return;
		}
	else
		{
		for (i = 0; i < 2; i++)
			{
			for (j = 0; j < 3; j++)
				{
				if (box == &facePickBox[i][j])
					{
					*whichItem = PICK_FACE;
					*faceI = i;
					*faceJ = j;
					return;
					}
				}
			}
		}

	*whichItem = PICK_NONE;
	}  // end of determineWhichPick

//
// Highlight the drawn box based on the given xyz position.
//
void 
rubberbandBox::doHighlighting 
  (
  pfVec3 *xyz, 
  PickItem *whichItem, 
  int *faceI, 
  int *faceJ  
  )
  {
  int i, j;
  static PickItem savedWhich = PICK_NONE;
  static int savedFaceI = 0;
  static int savedFaceJ = 0;


  // This first section figures out whether we are still in the same
  // pick region that we were in the last time.  If so, then we want
  // want to stay there, and we don't want to change the Performer stuff.
  //
  if (currentPickBox != NULL)
	{
	if (PT_IN_BOX (*xyz, *currentPickBox))
		{
		if (savedWhich == PICK_ENTIRE)
		  {
		  if (currentPickBox == &entirePickBox)
			{
			savedWhich = *whichItem = PICK_ENTIRE;
			return;
			}
		  }

		if (savedWhich == PICK_FACE)
		  {
		  if (currentPickBox == 
			    &facePickBox[savedFaceI][savedFaceJ])
			{
			savedWhich = *whichItem = PICK_FACE;
			savedFaceI = *faceI = savedFaceI;
			savedFaceJ = *faceJ = savedFaceJ;
			return;
			}
		  }

                determineWhichPick (currentPickBox, whichItem, 
							   faceI, faceJ);
		savedWhich = *whichItem;
		savedFaceI = *faceI;
		savedFaceJ = *faceJ;
		return;
		}
	}


  // If we have gotten here, we are not inside the same pick region
  // that we were the last time through here.  So we need to check
  // the point against each of the 7 pick regions.

  // If we find that we are in one of those regions, we must change
  // the Performer scene graph to do the highlighting.
  //

  if (PT_IN_BOX (*xyz, entirePickBox))
	{
	currentPickBox = &entirePickBox;
	savedWhich = *whichItem = PICK_ENTIRE;

  	int rtn = wireframeDCS->removeChild (currentChild);
  	wireframeDCS->addChild (hl_entire);
  	currentChild = hl_entire;

	return;
	}
  

  for (i = 0; i < 2; i++)
	{
	for (j = 0; j < 3; j++)
		{
  		if (PT_IN_BOX (*xyz, facePickBox[i][j]))
			{
			currentPickBox = &facePickBox[i][j];
			savedWhich = *whichItem = PICK_FACE;
			savedFaceI = *faceI = i;
			savedFaceJ = *faceJ = j;

  			int rtn = wireframeDCS->removeChild (currentChild);
  			wireframeDCS->addChild (hl_face[i][j]);
  			currentChild = hl_face[i][j];

			return;
			}
		}
	}

  if (currentPickBox != NULL)
    {
    currentPickBox = NULL;
    savedWhich = *whichItem = PICK_NONE;

    int rtn = wireframeDCS->removeChild (currentChild);
    wireframeDCS->addChild (nohl_entire);
    currentChild = nohl_entire;

    }


  }  // end of rubberbandBox::doHighlighting









//////////////////////////////////////////////////////////////////////

// Constructor
//
// Create object and set description.  
// Be sure to validate it!
//
rubberbandBox::rubberbandBox (dpf *app_in) :
  dtkAugment ("rubberbandBox")
{
  setDescription (
    "rubberbandBox DSO: Provides an axis-aligned box that can be\n"
  "  resized and repositioned interactively.\n");

  app = app_in;

  // All of the rest of this really gets set in preConfig and preFrame

  display = NULL;
  wand = NULL;
  buttonShm = NULL;
  boxCoordShm = NULL;

  int i;

  // We are going to initialize the box based on an environment variable.
  // It should contain two points opposite corners of the box, six coords, 
  // separated either by 
  // semicolons or by white space.
  char *boxInitStr = getenv ("RUBBERBAND_BOX_COORD_INIT");


  if (boxInitStr == NULL)
	{
printf ("\n\n didn't get RUBBERBAND_BOX_COORD_INIT\n\n\n");

	// If the env var isn't there, we init to a 0.1 box centered at origin
  	for (i = 0; i < 3; i++)
		{
		// We initialize a box centered at the origin of size 0.1
		boxSize[i] = 0.1;
  		currentBox[0][i] = - boxSize[i]/2;
  		currentBox[1][i] =   boxSize[i]/2;
		}
	}
  else
	{

printf ("\n\n got RUBBERBAND_BOX_COORD_INIT |%s|\n\n\n", boxInitStr);
	// First copy into a local string and change ; to blank

	char boxInitStrLocal[1024];
	strcpy (boxInitStrLocal, boxInitStr);
	char *c;
	for (c = boxInitStrLocal; *c; c++)
		{
		if ((*c == ';')||(*c == ':'))
			{
			*c = ' ';
			}
		}

	// now read ascii to doubles

printf ("\n\n boxInitStrLocal =  |%s|\n\n\n", boxInitStrLocal);

	if (sscanf (boxInitStrLocal, "%lf %lf %lf %lf %lf %lf",
		&(currentBox[0][0]),
		&(currentBox[0][1]),
		&(currentBox[0][2]),
		&(currentBox[1][0]),
		&(currentBox[1][1]),
		&(currentBox[1][2])) == 6)
		{
		// If the read works, then we make sure that
		// the two points are lower left & upper right
		// and set boxSize
		for (i = 0; i < 3; i++)
			{
			if (currentBox[1][i] < currentBox[0][i])
				{
				double tmp;
				tmp = currentBox[1][i];
				currentBox[1][i] = currentBox[0][i];
				currentBox[0][i] = tmp;
				}
			boxSize[i] = currentBox[1][i] - currentBox[0][i];
			}
		}
	else
		{

printf ("\n\n bad read\n\n\n");

		// If the read didn't work we put out an error message
		// and revert to the default box.
        	dtkMsg.add (DTKMSG_INFO,
                   "rubberbandBox: "
                   "Unable to decode env variable RUBBERBAND_BOX_COORD_INIT.\n"
		   "        Using default initial box.\n\n");
  		for (i = 0; i < 3; i++)
			{
			//We initialize a box centered at the origin of size 0.1
			boxSize[i] = 0.1;
  			currentBox[0][i] = - boxSize[i]/2;
  			currentBox[1][i] =   boxSize[i]/2;
			}
		}
	}

  

printf ("\n\n first corner = %f %f %f\n\n\n",
currentBox[0][0],
currentBox[0][1],
currentBox[0][2]);


  setPickBoxes (1.0, currentBox, entirePickBox, facePickBox);

  currentPickBox = NULL;


  wireframeDCS = NULL;
  nohl_entire = NULL;
  hl_entire = NULL;
  hl_face[0][0] = NULL;
  hl_face[0][1] = NULL;
  hl_face[0][2] = NULL;
  hl_face[1][0] = NULL;
  hl_face[1][1] = NULL;
  hl_face[1][2] = NULL;
  currentChild = NULL;
  cursorDCS = NULL;
  cursorNode = NULL;


  // Be sure to validate it!
  validate ();
} // end of constructor


rubberbandBox::~rubberbandBox ()
{

// We should probably free some stuff here.

}  // end of destructor
//////////////////////////////////////////////////////////////////////









//////////////////////////////////////////////////////////////////////

//
// preConfig is called only once by dpf, after pfInit() and 
//  dpfDisplay:preConfig() but before pfConfig()
//
int rubberbandBox::preConfig (void) {

  int i;


  // We allocate space for publically accessible box

  PublicBox = (Box *) pfMalloc(sizeof(Box));

#if 0

  // Set up initial box 
  for (i = 0; i < 3; i++)
	{
	// We initialize a box centered at the origin of size 0.1
	boxSize[i] = 0.1;
  	currentBox[0][i] = - boxSize[i]/2;
  	currentBox[1][i] =   boxSize[i]/2;
	}
#endif

  setPickBoxes (1.0, currentBox, entirePickBox, facePickBox);


  currentPickBox = NULL;

  // Update PublicBox from currentBox
  memcpy (PublicBox, currentBox, sizeof (Box));
  


  return dtkAugment::CONTINUE ; 
}   // end of preConfig


//
// postConfig is called after pfConfig
//
int rubberbandBox::postConfig (void) {

  //
  // Get access to wand data stream
  //
  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand || wand->isInvalid())
    {
    wand = NULL ;
    dtkMsg.add(DTKMSG_ERROR,
                 "rubberbandBox::postConfig() failed:"
                 "Unable to get valid wand pointer.\n");
    return dtkAugment::ERROR_;
    }


  return dtkAugment::REMOVE_CALLBACK ;
}  // end of postConfig

//
//////////////////////////////////////////////////////////////////////








//////////////////////////////////////////////////////////////////////


//
// This is a utility routine that creates the Performer nodes for
// displaying the visible box, both highlighted and unhighlighted.
//
// They are placed under a DCS that is then used for resizing and
// repositioning the drawn box.
//
// It also creates a 3D cursor geode and DCS.
//
static void
makePerformerNodes (
pfDCS **dcs, 
pfGeode **nohlEntireNode, 
pfGeode **hlEntireNode,
pfGeode *hlFaceNode[2][3],
pfDCS **cursorDCS, 
pfGeode **cursorNode
)
  {

  *dcs = new pfDCS;

  
  // hightlighted and unhighted entire boxes use same vertices and colors

  int *lengths = (int *) pfMalloc(sizeof(int)*2);
  // 8 coords, 7 segs
  lengths[0] = 8 ;
  lengths[1] = 8 ;

  // the vertices
  pfVec3 *lines = (pfVec3*) pfMalloc(sizeof(pfVec3)*(16)) ;

  lines[0]  = pfVec3 (0.0, 0.0, 0.0);
  lines[1]  = pfVec3 (1.0, 0.0, 0.0);
  lines[2]  = pfVec3 (1.0, 1.0, 0.0);
  lines[3]  = pfVec3 (0.0, 1.0, 0.0);
  lines[4]  = pfVec3 (0.0, 1.0, 1.0);
  lines[5]  = pfVec3 (0.0, 0.0, 1.0);
  lines[6]  = pfVec3 (0.0, 0.0, 0.0);
  lines[7]  = pfVec3 (0.0, 1.0, 0.0);

  lines[8]  = pfVec3 (1.0, 1.0, 1.0);
  lines[9]  = pfVec3 (0.0, 1.0, 1.0);
  lines[10] = pfVec3 (0.0, 0.0, 1.0);
  lines[11] = pfVec3 (1.0, 0.0, 1.0);
  lines[12] = pfVec3 (1.0, 0.0, 0.0);
  lines[13] = pfVec3 (1.0, 1.0, 0.0);
  lines[14] = pfVec3 (1.0, 1.0, 1.0);
  lines[15] = pfVec3 (1.0, 0.0, 1.0);

  // the colors- one color per vertex
  pfVec4 *colors = (pfVec4*) pfMalloc(sizeof(pfVec4)*(2)) ;
  colors[0] = pfVec4(1.0, 1.0, 1.0, 1.0) ;
  colors[1] = pfVec4(1.0, 1.0, 1.0, 1.0) ;

  pfVec4 *hl_colors = (pfVec4*) pfMalloc(sizeof(pfVec4)*(2)) ;
  hl_colors[0] = pfVec4(1.0, 1.0, 0.3, 1.0) ;
  hl_colors[1] = pfVec4(1.0, 1.0, 0.3, 1.0) ;





  // first the no highlight version
  pfGeoSet *nohl_gset = new pfGeoSet ;
  pfGeoState *nohl_gstate = new pfGeoState ;
  nohl_gstate->setMode (PFSTATE_ENLIGHTING, PF_OFF) ;
  nohl_gset->setGState (nohl_gstate) ;

  // a series of polylines
  nohl_gset->setPrimType(PFGS_LINESTRIPS) ;

  // two polylines
  nohl_gset->setNumPrims(2) ;

  // pass lengths array to geoset
  nohl_gset->setPrimLengths(lengths) ;

  // put vertex data into geoset
  nohl_gset->setAttr(PFGS_COORD3,  PFGS_PER_VERTEX, lines, NULL) ;

  // put color data into geoset
  nohl_gset->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, colors, NULL) ;


  // the node to hold the geoset(s)
  *nohlEntireNode = new pfGeode ;

  // add the geoset to the geode
  (*nohlEntireNode)->addGSet(nohl_gset) ;



 
  // Now the highlighted version.  Same except for line width
  pfGeoSet *hl_gset = new pfGeoSet ;
  pfGeoState *hl_gstate = new pfGeoState ;
  hl_gstate->setMode (PFSTATE_ENLIGHTING, PF_OFF) ;
  hl_gset->setGState (hl_gstate) ;


  hl_gset->setLineWidth (4.0) ;
  // hl_gset->setLineWidth (6.0) ;



  // a series of polylines
  hl_gset->setPrimType(PFGS_LINESTRIPS) ;

  // two polylines
  hl_gset->setNumPrims(2) ;

  // pass lengths array to geoset
  hl_gset->setPrimLengths(lengths) ;

  // put vertex data into geoset
  hl_gset->setAttr(PFGS_COORD3,  PFGS_PER_VERTEX, lines, NULL) ;

  // put color data into geoset
  hl_gset->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, hl_colors, NULL) ;

  // the node to hold the geoset(s)
  *hlEntireNode = new pfGeode ;

  // add the geoset to the geode
  (*hlEntireNode)->addGSet(hl_gset) ;



  //
  // Create each highlighted face
  //
  for (int i = 0; i < 2; i++)
    {
    for (int j = 0; j < 3; j++)
      {

      
      pfGeoSet *gsetBox = new pfGeoSet ;
      pfGeoState *gstateBox = new pfGeoState ;
      gstateBox->setMode (PFSTATE_ENLIGHTING, PF_OFF) ;
      gsetBox->setGState (gstateBox) ;
      gsetBox->setLineWidth (1.0) ;
      gsetBox->setPrimType(PFGS_LINESTRIPS) ;
      gsetBox->setNumPrims(2) ;
      gsetBox->setPrimLengths(lengths) ;
      gsetBox->setAttr(PFGS_COORD3,  PFGS_PER_VERTEX, lines, NULL) ;
      gsetBox->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, colors, NULL) ;


      // now the highlighted face:
      pfVec3 *faceLines = (pfVec3*) pfMalloc(sizeof(pfVec3)*(5)) ;


      float corner[3], *x, *y;

      corner[j] = i;

      switch (j)
	{
	case 0:
	  x = &corner[1];	
	  y = &corner[2];	
	  break;

	case 1:
	  x = &corner[0];	
	  y = &corner[2];	
	  break;

	case 2:
	  x = &corner[0];	
	  y = &corner[1];	
	  break;
	}

      *x = 0; *y = 0;
      faceLines[0]  = pfVec3 (corner[0], corner[1], corner[2]);
      *x = 1; *y = 0;
      faceLines[1]  = pfVec3 (corner[0], corner[1], corner[2]);
      *x = 1; *y = 1;
      faceLines[2]  = pfVec3 (corner[0], corner[1], corner[2]);
      *x = 0; *y = 1;
      faceLines[3]  = pfVec3 (corner[0], corner[1], corner[2]);
      *x = 0; *y = 0;
      faceLines[4]  = pfVec3 (corner[0], corner[1], corner[2]);

      int *faceLengths = (int *) pfMalloc(sizeof(int)*1);
      faceLengths[0] = 5 ;

      pfVec4 *faceColors = (pfVec4*) pfMalloc(sizeof(pfVec4)*(1)) ;
      // faceColors[0] = pfVec4(1.0, 1.0, 1.0, 1.0) ;
      faceColors[0] = pfVec4(1.0, 1.0, 0.3, 1.0) ;


      pfGeoSet *gsetFace = new pfGeoSet ;
      pfGeoState *gstateFace = new pfGeoState ;
      gstateFace->setMode (PFSTATE_ENLIGHTING, PF_OFF) ;
      gsetFace->setGState (gstateFace) ;
      gsetFace->setLineWidth (4.0) ;
      // gsetFace->setLineWidth (6.0) ;
      gsetFace->setPrimType(PFGS_LINESTRIPS) ;
      gsetFace->setNumPrims(1) ;
      gsetFace->setPrimLengths(faceLengths) ;
      gsetFace->setAttr(PFGS_COORD3,  PFGS_PER_VERTEX, faceLines, NULL) ;
      gsetFace->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, faceColors, NULL) ;

      hlFaceNode[i][j] = new pfGeode;
      hlFaceNode[i][j]->addGSet (gsetBox);
      hlFaceNode[i][j]->addGSet (gsetFace);

      }
    }


  // now the cursor
  *cursorDCS = new pfDCS;

  int *cursorLengths = (int *) pfMalloc(sizeof(int)*3);
  cursorLengths[0] = 2 ;
  cursorLengths[1] = 2 ;
  cursorLengths[2] = 2 ;

  // the vertices
  pfVec3 *cursorLines = (pfVec3*) pfMalloc(sizeof(pfVec3)*(6)) ;

  cursorLines[0]  = pfVec3 ( -0.03,  0.00,  0.00);
  cursorLines[1]  = pfVec3 (  0.03,  0.00,  0.00);
  cursorLines[2]  = pfVec3 (  0.00, -0.03,  0.00);
  cursorLines[3]  = pfVec3 (  0.00,  0.03,  0.00);
  cursorLines[4]  = pfVec3 (  0.00,  0.00, -0.03);
  cursorLines[5]  = pfVec3 (  0.00,  0.00,  0.03);



  // the colors- one color per vertex
  pfVec4 *cursorColors = (pfVec4*) pfMalloc(sizeof(pfVec4)*(3)) ;
  cursorColors[0] = pfVec4(1.0, 1.0, 0.3, 1.0) ;
  cursorColors[1] = pfVec4(1.0, 1.0, 0.3, 1.0) ;
  cursorColors[2] = pfVec4(1.0, 1.0, 0.3, 1.0) ;

  pfGeoSet *gsetCursor = new pfGeoSet ;
  pfGeoState *gstateCursor = new pfGeoState ;
  gstateCursor->setMode (PFSTATE_ENLIGHTING, PF_OFF) ;
  gsetCursor->setGState (gstateCursor) ;
  gsetCursor->setLineWidth (2.0) ;
  gsetCursor->setPrimType(PFGS_LINESTRIPS) ;
  gsetCursor->setNumPrims(3) ;
  gsetCursor->setPrimLengths(cursorLengths) ;
  gsetCursor->setAttr(PFGS_COORD3,  PFGS_PER_VERTEX, cursorLines, NULL) ;
  gsetCursor->setAttr(PFGS_COLOR4,  PFGS_PER_PRIM, cursorColors, NULL) ;

  *cursorNode = new pfGeode;
  (*cursorNode)->addGSet (gsetCursor);

  }  // end of makePerformerNodes




//
// Put the indicated pfNode under the named parentNode
//
static int
putInSceneGraph (dpfDisplay *display, pfNode *node, char *parentName)
  {
  pfGroup *parent = NULL;

  parent = (pfGroup*) display->world()->
                         lookup (parentName, pfGroup::getClassType()) ;
  if (!parent)
      {
      parent= (pfGroup*) display->world()->
			 lookup (parentName, pfDCS::getClassType()) ;
      }

  if (!parent)
      {
      parent= (pfGroup*) display->world()->
			 lookup (parentName, pfSCS::getClassType()) ;
      }

  if (!parent)
      {
      return -1;
      }


  parent->addChild (node);
  return 0;

  }  // end of putInSceneGraph


//
// Match the displayed box to the given size and position.
//
static void
matchBox (double corner[3], double size[3], pfDCS *dcs)
	{
	dcs->setScale (size[0], size[1], size[2]);
	dcs->setTrans (corner[0], corner[1], corner[2]);
	}  // end of matchBox


//////////////////////////////////////////////////////////////



//
// We're calling preFrame only once: to get display , setup shared mem,
// and create some Performer data items.
//
int rubberbandBox::preFrame (void)
{
  display = app->display ();
  if (!display) 
    {
    dtkMsg.add(DTKMSG_ERROR,
                 "rubberbandBox::preFrame() failed:"
                 "Unable to get display pointer.\n");
    return dtkAugment::ERROR_;
    }





  // We get the wand button from shared memory.  The name of the
  // shared memory is:
  //     the value of the env variable RUBBERBAND_BOX_BUTTON_SHM
  //
  // or if that doesn't exist:
  //     it is "rubberbandBoxButton"
  //
  char *shmName = getenv ("RUBBERBAND_BOX_BUTTON_SHM");
  if (shmName == NULL)
	{
	shmName = "rubberbandBoxButton";
	}

  buttonShm =  new dtkSharedMem (shmName);
  int buttonOK = 0;
  if (buttonShm)
	{
	buttonOK = ! buttonShm->isInvalid();
	}

  if ( ! buttonOK )
	{
        dtkMsg.add(DTKMSG_ERROR,
                 "rubberbandBox::preFrame() failed:"
                 " Couldn't open shared memory segment\"%s\"\n",
                 shmName) ;
	return dtkAugment::ERROR_;
	}




  // Now we get the shared memory for the rubber band box coords
  // if the environment variable RUBBERBAND_BOX_COORD_SHM is not
  // set, then this DSO does not write the coords of the box to
  // shared memory.

  shmName = getenv ("RUBBERBAND_BOX_COORD_SHM");
  if (shmName == NULL)
	{
        dtkMsg.add (DTKMSG_INFO,
                 "rubberbandBox: "
                 "No name found for box coord shared memory.  "
		 "External shared memory will not be used");
	boxCoordShm = NULL;
	}
  else
	{
	boxCoordShm = new dtkSharedMem (6*sizeof(float), shmName);
	int boxCoordOK = 0;
	if (boxCoordShm)
		{
		boxCoordOK = ! boxCoordShm->isInvalid ();
		}
	
	if ( ! boxCoordOK )
		{
        	dtkMsg.add(DTKMSG_ERROR,
                 	"rubberbandBox::preFrame() failed:"
                 	" Couldn't open shared memory segment\"%s\"\n",
                 	shmName) ;
		return dtkAugment::ERROR_;
		}

	writeBoxCoordShm ();
	}



  // Setup the Peformer stuff for the drawn box & cursor
  makePerformerNodes (&wireframeDCS, &nohl_entire, &hl_entire, hl_face, 
			&cursorDCS, &cursorNode);
  if (putInSceneGraph (display, wireframeDCS, "noClip"))
    {
    if (putInSceneGraph (display, wireframeDCS, "world"))
      {
      dtkMsg.add (DTKMSG_ERROR,
        "rubberbandBox::preFrame() failed: "
        "Unable to place box wireframe node under "
        "either world or noClip nodes.");
      return dtkAugment::ERROR_;
      }
    }

  matchBox (currentBox[0], boxSize, wireframeDCS);

  wireframeDCS->addChild (nohl_entire);
  currentChild = nohl_entire;

  display->ether()->addChild (cursorDCS);
  cursorDCS->addChild (cursorNode);


  // The following return means that this preFrame callback is called only once.
  return dtkAugment::REMOVE_CALLBACK;
}    // end of preFrame


//
// We call postFrame every frame to get wand position and adjust the box
//
int rubberbandBox::postFrame (void) {

  float wandData[6];
  // float offsetWand[3];
  static pfCoord wandCoord;
  static char buttonPressed = 0;
  static char buttonWasPressed = 0;
  static double boxWandOffset[3] = {0,0,0};
  static PickItem which;
  static int faceI, faceJ;
  static double faceWandOffset = 0;

  static dtkMatrix *matrix = NULL;

	
  if (buttonShm)
    {
    if (buttonShm->read (&buttonPressed))
      {
      dtkMsg.add (DTKMSG_ERROR, 
        "rubberbandBox::postFrame failed: "
        " Unable to read button data from shared memory.");
      return dtkAugment::ERROR_;
      }
    }
  else
    {
    dtkMsg.add (DTKMSG_ERROR, 
        "rubberbandBox::postFrame failed: "
	"Button shared memory pointer is NULL.");
    return dtkAugment::ERROR_;
    }



  // read the wand position
  if (wand)
      wand->read (wandData);

  if (matrix == NULL)
	{
	matrix = new dtkMatrix;
  	wandCoord.hpr.set (0, 0, 0);
	}
  else
	{
	matrix->identity ();
	}
  matrix->rotateHPR(wandData[3], wandData[4], wandData[5]);

  // we want the 3D cursor to float 0.2 units out from the front of the wand:

  wandCoord.xyz.set (
  			wandData[0] + matrix->element (0, 1)*0.2,
  			wandData[1] + matrix->element (1, 1)*0.2,
  			wandData[2] + matrix->element (2, 1)*0.2 
		    );
  display->dpfToWorld (&wandCoord);
  // wandCoord is now the position of the 3D cursor in world coordinates

  // We move the drawn 3D cursor to the appropriate DPF coords:
  cursorDCS->setTrans   (
  			wandData[0] + matrix->element (0, 1)*0.2,
  			wandData[1] + matrix->element (1, 1)*0.2,
  			wandData[2] + matrix->element (2, 1)*0.2 
			);


  //
  // We only do highlighting and derive a new "which" when a current
  // interaction is not currently in progress.
  //
  if ( ! interactionIsInProgress )
    {
    // highlight box based on wand, figure out which thing would be picked
    doHighlighting (&wandCoord.xyz, &which, &faceI, &faceJ);
    }



  // If the button is pressed, we need to do the interaction:

  if (buttonPressed)
    {

    if (which == PICK_NONE)
      {
      // This is a small trick to ensure that when a user presses the
      // button when he/she is not pointing at something that is 
      // pickable, then no pick will occur until he/she releases the
      // button, repositions the wand into a pickable region, then
      // presses the button again.
      //
      interactionIsInProgress = 1;
      }
    else if (which == PICK_ENTIRE)
      {

      // We need to drag relative to the initial position when the button 
      // was pressed.  So here is where we grab the pos when the button
      // is first pressed, then we move relative to this pos.
      if ( ! buttonWasPressed )
	{
	// set up relative motion
	boxWandOffset[0] = currentBox[0][0] - wandCoord.xyz[PF_X];
	boxWandOffset[1] = currentBox[0][1] - wandCoord.xyz[PF_Y];
	boxWandOffset[2] = currentBox[0][2] - wandCoord.xyz[PF_Z];
	interactionIsInProgress = 1;
	}



      if (interactionIsInProgress)
        {
	// We drag the box without changing size
        int i;
        for (i = 0; i < 3; i++)
	  {
          currentBox[0][i] = wandCoord.xyz[i] + boxWandOffset[i];
          currentBox[1][i] = currentBox[0][i] + boxSize[i];
	  }

  	matchBox (currentBox[0], boxSize, wireframeDCS);
        }

      // update publically available box from currentBox
      memcpy (PublicBox, currentBox, sizeof (Box));
      writeBoxCoordShm ();

      }  // end of PICK_ENTIRE
    else if (which == PICK_FACE)
      {
      // We are draggin just one face, thereby stretching the box.

      if ( ! buttonWasPressed )
	{
	// On first button press, we ensure that subsequent movements
	// are done relative to the first pos
	faceWandOffset = currentBox[faceI][faceJ] - wandCoord.xyz[faceJ];
	interactionIsInProgress = 1;
	}


      // faceI and faceJ indicate which face is being dragged.


      if (interactionIsInProgress)
	{
	currentBox[faceI][faceJ] = wandCoord.xyz[faceJ] + faceWandOffset;


	// Handling cross-over of the opposite sides - when the user
	// drags one face beyond its opposite (stationary) face.
	// 
	// The logic here is a little tricky, but it make sense.  
	//
        // When faceI is 0
	// when we are dragging the lower face.  This means that 
        // the difference below should be negative and the "<"
        // is true.  So the == returns false .  When the difference
        // becomes positive, then the faces have crossed over and
	// we need to swap the roles of the faces.
	//
	// Similar logic when faceI is 1.
	//
	if ( ( (currentBox[faceI][faceJ] - 
                currentBox[1-faceI][faceJ] ) < 0 ) == faceI ) 
	    {
	    double tmp;

	    // we reach here only if we have dragged one side past its
	    // opposite side.  In this case, we need to flip the roles of 
	    // the two sides
	    tmp = currentBox[faceI][faceJ];
	    currentBox[faceI][faceJ] = currentBox[1-faceI][faceJ];
	    currentBox[1-faceI][faceJ] = tmp;
	    faceI = 1-faceI;
	    }



	// recalc box size for J dimension and make the drawn box match
	boxSize[faceJ] = currentBox[1][faceJ] - currentBox[0][faceJ];
  	matchBox (currentBox[0], boxSize, wireframeDCS);
	}



      // update publically available box from currentBox
      memcpy (PublicBox, currentBox, sizeof (Box));
      writeBoxCoordShm ();
      }  // end of PICK_FACE


    }  // end of if (buttonPressed)
  else
    {

    // button is not pressed, so interaction (if any) is over.

    if (interactionIsInProgress)
	{
  	setPickBoxes (1.0, currentBox, entirePickBox, facePickBox);
        // update publically available box from currentBox
  	memcpy (PublicBox, currentBox, sizeof (Box));
        writeBoxCoordShm ();
        interactionIsInProgress = 0;
	}
    }

  buttonWasPressed = buttonPressed;


  return dtkAugment::CONTINUE ; 
}    // end of postFrame







//////////////////////////////////////////////////////////////////////




/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new rubberbandBox (app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
