
// screenContactAlert - This is a DIVERSE DSO which monitors the physical
//     position of the wand and head tracked glasses.  If either device comes
//     close to physically touching the screen, a large bright red Stop Sign is
//     displayed while the devices are within proximity of the screen.
//     Hopefully, this "screen contact alert" will prevent users from actually
//     touching the screen.
//
// Authors:
//
//    Julien Franiatte did the major implementation work by modifying the
//                example generic dtkAugment DSO provided with the DIVERSE
//                distribution.
//
//    Steve Satterfield provided minor assistance, created the stop sign,
//                and made small modifications for the distribution.
//
//    John Hagedorn provided some motivation by suggesting this one day while
//                  we were watching a demo.
//
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//      
//    
//    John Kelso  had the idea before any of us and provided assistance in
//               in getting it going.  He rewrote from scratch it for dpf 2.2
//
//
// For information about this DSO, contact Steve Satterfield, steve@nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."
// 
// 
//

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

// how close to get to the screen before displaying stop sign
#define METERS (.35)



// example screenContactAlert dtkAugment DSO

class screenContactAlert : public dtkAugment {
public:

  screenContactAlert(dpf *app) ;
  int postConfig(void);
  int postFrame(void);

private:
  dpf *app;
  dpfDisplay *display ;
  pfScene *scene ;
  dtkInLocator *wand;
  dtkInLocator *head;
  float headPos[6];
  float wandPos[6];
  dpfScreen **screen;
  int nScreens ;
  dtkColor color;
  pfSCS **scs;
  int *displayed;
  float distance;

} ;

// create object and set description.
// be sure to validate it!
screenContactAlert::screenContactAlert(dpf *app_in) :
  dtkAugment("generic"), color(stdout)
{
  app = app_in ;
  setDescription("Displays a Stop sign when too close from the wall" );

  screen = NULL ;
  nScreens = 0 ;
  scs = NULL ;
  setDescription("generic dtkAugment DSO") ;
  displayed = NULL;
  scene = NULL ;

  // dtkAugment starts out invalid.
  validate();

} ;

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig()
int screenContactAlert::postConfig(void) {

  distance = METERS/app->display()->meters() ;

  if (!(display = app->display()))
  return dtkAugment::ERROR_ ;
  if (!(scene = display->scene()))
  return dtkAugment::ERROR_ ;

  pfFilePath(".:"DPF_DATA_DIR);

  for (dpfPipe *p=display->first(); p; p=p->next())
    for (dpfPwin *w=p->first(); w; w=w->next())
      for (dpfScreen *s=w->first(); s; s=s->next()) {
	if (!s->simpleFrustum()) {
	  
	  screen = (dpfScreen **) realloc(screen, sizeof(dpfScreen *)*(nScreens+1)) ;
	  screen[nScreens] = s ;
	  
	  scs = (pfSCS **) realloc(scs, sizeof(pfSCS *)*(nScreens+1)) ;
#if 0
	  pfCoord pos ;
	  pos.xyz = s->center() ;
	  pos.hpr = (s->offset()).hpr ;
	  pfMatrix mat ;
	  mat.makeCoord(&pos) ;
#endif
	  pfMatrix mat ;
	  mat = s->rotMatrix() ;
	  float height = s->height() ;
	  float width = s->width() ;
	  pfVec3 center = s->center() ;
	  mat.postTrans(mat,
			   center[PF_X], center[PF_Y], center[PF_Z]) ;
	  mat.preScale(width, 1.0f, height, mat) ;

	  scs[nScreens] = new pfSCS(mat) ;

	  scs[nScreens]->addChild(pfdLoadFile("stopsign.pfb")) ;
	  displayed = (int *) realloc(displayed, sizeof(int)*(nScreens+1)) ;
	  displayed[nScreens] = 0 ;

	  nScreens++ ;
	}
      }

  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand) return dtkAugment::ERROR_ ;
  head = (dtkInLocator *) app->get("head", DTKINLOCATOR_TYPE);
  if(!head) return dtkAugment::ERROR_ ;

  return dtkAugment::REMOVE_CALLBACK ;
}

// called every display loop, as a pfFrame callback, after pfDraw()
int screenContactAlert::postFrame(void) {
  wand->read(wandPos);
  head->read(headPos);
  
  int i ;
  for (i=0; i<nScreens; i++) {
    float wdistance, hdistance ;
    pfVec3 point ;
    point.set(wandPos[0], wandPos[1], wandPos[2]) ;
    wdistance = (screen[i]->normal()).dot(screen[i]->center()-point) ;
    point.set(headPos[0], headPos[1], headPos[2]) ;
    hdistance = (screen[i]->normal()).dot(screen[i]->center()-point) ;
    if ((wdistance<distance) || (hdistance<distance)) {  // too close
      if (!displayed[i]) {
	scene->addChild(scs[i]) ;
	displayed[i] = 1 ;
      }
    } else { // too far
      if (displayed[i]) {
	scene->removeChild(scs[i]) ;
	displayed[i] = 0 ;
      }
    }
  }
  return dtkAugment::CONTINUE ;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new screenContactAlert(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
