// sgPicker - 
// A DIVERSE DSO which loads sets of files and commands specified 
// by an environment variable, DPF_ABSWITCH_FILES. Originally, the first
// of each pair is displayed. When the wand selects an object using
// button 1 on the mouse,
// its partner object will be displayed instead. If there is a command
// associated with these objects, it will be implemented when the object
// is selected.
//
// Authors:
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//
//    Steve Satterfield, John Hagedorn and John Kelso (SAVG)
//    provided the idea for the program
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "testSuffix.C"

#define ONKEY  1
#define OFFKEY  2
#define EXECKEY  3
#define GROUPMAX 100
#define LOSTMAX 100

class sgPicker : public dtkAugment {
public:
  sgPicker(dpf *app) ;
  ~sgPicker(void) ;
  int preFrame(void);
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;
  //  Shared memory segment for the spin rate
  dtkInLocator *wand ;
  dtkInLocator *head ;
  //  Shared memory segment used to request the wand pointer from another program
  dtkSharedMem* shm2;
  //  Shared memory segment used to request a system call
  dtkSharedMem* shm4;
  //  Shared memory segment to put the call
  dtkSharedMem* shm_call;

  unsigned char selected;
  unsigned char newCallSet;
  dtkInButton *button;
  dtkDequeuer *dequeuer;
  int objcnt;
  int ccnt;
  int iccnt;
  bool objectCheck;

  pfNode** objectPtrs;
  char** commandList;
  char** initCommandList;
  string groupList[GROUPMAX];
  string lostgroup[LOSTMAX];
  int lostgroupobjnum[LOSTMAX];
  int* groupnum;
  int* objNumbers;
  int** onClick;
  char** filenames;
  bool* isObjOn;
  bool* selectable;
  int findGroupInScenegraph();
  int findGnum();
  int gnum;
  int groupcnt;
  int grouptotal;
  int lostcnt;

  pfNode* node;
  pfNode* node2;
  string group;
  string lostname;

  pfCoord wandCoord;
  pfCoord headCoord;
  pfCoord* hdcoord;
  pfCoord ptCoord;
  pfMatrix headWorld;
  pfMatrix wandWorld;
  pfGroup* rootGroup;
  pfVec3 check;
  pfVec3 upVector;
  pfVec3 checkhead;
  pfVec3 centerVector;
  void resetAorB();
  int pargc ;
  char **pargv ;
  int ppargc ;
  char **ppargv ;
  void readInit();
  void readOnOff();
  void readExec();
  int isObjectNew();
  void addObjectToList();
  void addLostObject();
  void updateOnClick();
  bool isNodeOnGroup(const char* name, pfGroup* grp);
  void checkdcs();
  void addWanddcs();
  void addHeaddcs();
  void addHeadnohprdcs();
  bool checkFunction();
  bool checkSelection();
  void doObjectCheck();
  void addToDisplay();
  void removeTheNode();
  void executeTheCommand();
  pfGroup* findFirstNamedParent();
  int selectSelectable();
  void resetIsObjOn();
  void setPtCoord();
  float wandData[6];
  float wandPos[6];
  pfDCS *wanddcs;
  float headData[6];
  float headPos[6];
  float headPos2[6];
  pfDCS *headdcs;
  pfDCS *headnohprdcs;
  pfMatrix dcsMatrix;
  pfMatrix scsMatrix;
  pfVec3 row3;

  bool objectSelected;
  int last_one_hit;
  bool button1On; // indicates if button 1 is being pressed
  bool button1LastFrame; // indicates old position of button
  char* commandA;

  // names of input arguments
  string function;
  string selection;
  string selNodeName;
  string objNodeName;
  string command;
  string filename;
  pfGroup* parentgroup;
  pfGroup *newgroup;
  pfDCS* newdcs;
  pfSCS* newscs;
  pfMatrix coordMatrix;
  int pcnt;
  int framenum;
  int count, oldcount ;
  pfBox box;
  float objRadius;
  float diff;
  pfVec3 centerPt;
  pfVec3 intPt;
  float distToWand;
  float d;
  int last;
  int i, j, k, ii, q, num, numA, numB;
  bool foundit;
  bool isit;
  bool findParent();
} ;

// create object and set description.
// be sure to validate it!
sgPicker::sgPicker(dpf *app_in) :
  dtkAugment("sgPicker")
{
  setDescription("loads files specified by DPF_SGPICKER_FILES"
		 " environment variable and enables you to "
		 "establish links between pairs of objects."
		 " The objects that are displayed have one "
		 "of several functions: when selected with "
		 "the wand, they can turn on (make visible)"
		 " an object, turn off (make invisible) an "
		 "object, or issue a command. "
		 " This version was compiled on 12/02/04.") ;
  app = app_in;
  validate();
  
  // counters for the numbers of objects, commands
  objcnt = 0;
  ccnt = 0;
  iccnt = 0;
  objectCheck = false;

  // arrays for object nodes, commands, group names
  objectPtrs = NULL;
  initCommandList = NULL;
  objNumbers = NULL;
  onClick = NULL;
  filenames = NULL;
  groupnum = NULL;
  isObjOn = NULL;
  selectable = NULL;
  commandA = new char[1024];
  commandList = new char*[100];
  pargv = NULL ;
  ppargv = NULL ;

  shm2 = new dtkSharedMem(sizeof(unsigned char), "pointerOn");
  if(!shm2 || shm2->isInvalid()) {
      shm2 = NULL;
  }
  else
    shm2->flush();

  shm4 = new dtkSharedMem(sizeof(unsigned char), "runCommandFlag");
  if(!shm4 || shm4->isInvalid()) {
      shm4 = NULL;
  }
  else
    shm4->flush();

  shm_call = new dtkSharedMem(sizeof(char)*1024, "runCommandString");
  if(!shm_call || shm_call->isInvalid())
    shm_call = NULL ;
  else
    shm_call->flush();

  // turn buttons off to start for wandPointer
  unsigned char turnit = 2;
  if (shm2)
     shm2->write(&turnit);

  newCallSet = 1;
  if (shm4)
    shm4->write(&newCallSet);
  button = NULL;
  dequeuer = NULL;
  objectSelected = false;
  last_one_hit = -1;
  button1On = false;
  button1LastFrame = false;
  count = 0;
  oldcount = 0;

  wandPos[0] = wandPos[1] = wandPos[2] = wandPos[3] = 
    wandPos[4] = wandPos[5] = 0.0f ;
  headPos[0] = headPos[1] = headPos[2] = headPos[3] = 
    headPos[4] = headPos[5] = 0.0f ;
  headPos2[0] = headPos2[1] = headPos2[2] = headPos2[3] = 
    headPos2[4] = headPos2[5] = 0.0f ;

  wanddcs = NULL;
  headdcs = NULL;
  headnohprdcs = NULL;
  groupList[0] = "world";
  groupList[1] = "scene";
  groupList[2] = "ether";
  groupList[3] = "nav";
  groupList[4] = "wand";
  groupList[5] = "head";
  groupList[6] = "head_no_hpr";
  groupcnt = 7;
  grouptotal = GROUPMAX;
  lostcnt = 0;
  framenum = 0;
} ;

// destructor
sgPicker::~sgPicker(void) {
  dtkSharedMem_destroy("pointerOn");
}

#define MIN_BUTTONS 3

int sgPicker::preFrame(void)
{
  dequeuer = new dtkDequeuer(app->record());
  if(!dequeuer || dequeuer->isInvalid()) return DTK_CALLBACK_ERROR;

  display = app->display();
  if(!display) return DTK_CALLBACK_ERROR;

  button = (dtkInButton *) app->get("buttons", DTKINBUTTON_TYPE);
  if(!button) return DTK_CALLBACK_ERROR;

   if(button->number() < MIN_BUTTONS)
     {
       dtkMsg.add(DTKMSG_ERROR,
		  "buttonNavControl::preConfig() failed:"
		  " the dtkInButton named \"%s\" has "
		  "just %d buttons"
		  " which is less than the %d needed.\n",
		  button->getName(), button->number(),
		  MIN_BUTTONS);
      return DTK_CALLBACK_ERROR;
    }
  button->queue();

  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand || wand->isInvalid())

  dequeuer = new dtkDequeuer(app->record());
  if(!dequeuer || dequeuer->isInvalid()) return DTK_CALLBACK_ERROR;

  display = app->display();
  if(!display) return DTK_CALLBACK_ERROR;

  button = (dtkInButton *) app->get("buttons", DTKINBUTTON_TYPE);
  if(!button) return DTK_CALLBACK_ERROR;

   if(button->number() < MIN_BUTTONS)
     {
       dtkMsg.add(DTKMSG_ERROR,
		  "buttonNavControl::preConfig() failed:"
		  " the dtkInButton named \"%s\" has "
		  "just %d buttons"
		  " which is less than the %d needed.\n",
		  button->getName(), button->number(),
		  MIN_BUTTONS);
      return DTK_CALLBACK_ERROR;
    }
  button->queue();

  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand || wand->isInvalid())
    wand = NULL ;

  head = (dtkInLocator *) app->get("head", DTKINLOCATOR_TYPE);
  if(!head || head->isInvalid())
    head = NULL ;

  // init file loaders
  char *filelist = getenv("DPF_SGPICKER_FILES") ;
  char *newfilelist = new char[250];
  char* nextfile = new char[100];
  char *line = new char[2047];

  // possible input lines
  //-----------------------------------------------------
  //  INIT   ON   nodeA   fileA   groupA             
  //  INIT   OFF  nodeB   fileB 
  //  INIT   EXEC command                                  
  //  SELECT ON nodeA nodeB groupB 
  //  SELECT OFF nodeA nodeB
  //  SELECT EXEC nodeA command
  //-----------------------------------------------------

  if (filelist) {
    //parse this list and take out the filenames
    //printf("in sgPicker: filelist: %s\n",filelist);
    strcpy(newfilelist, filelist);
    
    // replace : with blank
    int len = strlen(filelist);
    for ( int i = 0; i < len; ++i ) {
      if (filelist[i] == ':') {
	newfilelist[i] = ' ';
      }
    }
    
    // parse
    dpf_charsToArgcArgv(newfilelist, &ppargc, &ppargv) ;
    // read through each file
    for ( int i = 0; i < ppargc; ++i ) {
      nextfile = ppargv[i];
      //printf("in sgPicker: nextfile: %s\n",nextfile);
      if (strlen(nextfile)>0) {
	
	// see if the file is really there 
	FILE *fp = pfdOpenFile(nextfile);              
        if (!fp) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "sgPicker::preConfig: error opening input file %s\n",
		     nextfile) ;
        }

	while ((fgets(line,250,fp)!=NULL)) {
	  //printf("got the next line from input: %s\n", line);
	  dpf_charsToArgcArgv(line, &pargc, &pargv) ;
	  // check for the correct parameters and then process
	  if (pargc >2 && pargc < 6) {
	    // check if function is defined
	    selection = pargv[0];
	    // check if selection if ON, OFF, EXEC
	    function = pargv[1];
	    if (!checkFunction() || !checkSelection()) {
	      dtkMsg.add(DTKMSG_ERROR,
		       "sgPicker:postConfig: input file error: "
		       "function or selection.\n");
	    }
	    else {
	      // continue to parse
	      if (selection == "INIT") {
		//printf("got an INIT\n");
		if (function == "EXEC")
		  command = pargv[2];
		else {
		  selNodeName = pargv[2];
		  if (pargc > 3)
		    filename = pargv[3];
		  if (pargc > 4)
		    group = pargv[4];
		  else
		    group = "world";
		}
		readInit();
	      }
	      else if (selection == "SELECT") {
		//printf("found a select\n");
		if (function == "ON" || 
		    function == "OFF") {
		  selNodeName = pargv[2];
		  //printf("selNodeName: %s\n",selNodeName.c_str());
		  if (pargc > 3) {
		    objNodeName = pargv[3];
		    //printf("got an object: %s\n",objNodeName.c_str());
		  }
		  if (pargc > 4)
		    group = pargv[4];
		  else
		    group = "world";
		  readOnOff();
		}
		else if (function == "EXEC") {
		  //printf("found a select exec: %s\n",pargv[2]);
		  selNodeName= pargv[2];
		  if (pargc > 3)
		    command= pargv[3];

		  readExec();
		}
	      }
	      else {
		// not a valid command in the line
		dtkMsg.add(DTKMSG_ERROR,
			   "sgPicker:postConfig: input file error :"
			   "This line does not have a valid command.\n");
	      }
	    }
	  }
	}
      }
    }
  }

  //-debugging----------------------------------------
  //for ( int jj = 0; jj< objcnt; ++jj) {
  //  printf("next object: %s\n",objectPtrs[jj]->getName());
  //  for ( int jjj = 0; jjj < objNumbers[jj]; ++jjj) {
  //    printf("%d  %d  %d    \n",onClick[jj][3*jjj],onClick[jj][3*jjj+1],onClick[jj][3*jjj+2]);
  //  }
  //}
  //-------------------------------------------------

  //printf("check commandList: ccnt %d\n",ccnt);
  //for (int rr = 0; rr < ccnt; ++rr) {
  //printf("%d  %s  \n",rr,commandList[rr]);
  //}
  // just call buttonNavControl::preFrame() once at the start.
  return DTK_REMOVE_CALLBACK;
}

int sgPicker::postConfig(void) {

  display = app->display();
  rootGroup = new pfGroup;
  display->world()->addChild(rootGroup);

  char *old_path = (char *) pfGetFilePath();
  char *path = NULL;
  if(old_path)
    {
      // copy old_path from static memory to the heap.
      old_path = dtk_strdup(old_path);
      path = (char *) dtk_malloc(strlen(old_path) +
				 strlen("%s:.:"DPF_DATA_DIR));
      sprintf(path,"%s:.:"DPF_DATA_DIR, old_path);
      pfFilePath(path);
    }
  else // old_path == NULL
    pfFilePath(".:"DPF_DATA_DIR);

  return DTK_REMOVE_CALLBACK ; 
}

bool sgPicker::checkFunction() {
   bool okay = false;
   if (selection == "INIT" ||
       selection == "SELECT")
     okay = true;
   return okay;
}

bool sgPicker::checkSelection() {
   bool okay = false;
   if (function == "ON" ||
       function == "OFF" ||
       function == "EXEC")
     okay = true;
   return okay;
}

void sgPicker::readInit() {
  // every readInit will either 
  // load in a file to a named node or execute a command
  if (function == "ON" ||
      function == "OFF") {
    // first load the object

    // see if this is a good file name
    FILE *fp = fopen(filename.c_str(), "r");
    if (!fp) {
      dtkMsg.add(DTKMSG_ERROR,
		 "sgPicker:: error opening input file %s\n",filename.c_str());
      return;
    }
    else {
      char* testit = new char[100];
      strcpy(testit,filename.c_str());
      testSuffix(testit);
      node = pfdLoadFile(filename.c_str());
      node->setName(selNodeName.c_str());
      numA = isObjectNew();

      // add object to the object list
      addObjectToList();
      checkdcs();
    }

    // if selection is on, add it onto the scenegraph
    if (function == "ON") {
      // make object selectable
      selectable[abs(numA)] = true;
      isObjOn[abs(numA)] = true;
      addToDisplay();
    }
  }
  else if (function == "EXEC") {
    // store the command stored in response
    initCommandList = (char **) realloc(initCommandList, sizeof(char*)*(iccnt+1));
    initCommandList[iccnt] = strdup(command.c_str());
    ++iccnt;
  }
}

void sgPicker::readOnOff() {
  // this is for line with SELECT ON or SELECT OFF
  // first find the object node
  node = display->world()->lookup(selNodeName.c_str(),pfNode::getClassType());
  if (node != NULL) {
    //printf("found node: %s in sgPicker\n",node->getName());
    numA = isObjectNew();

    // if it is found, and has a parent, set isObjOn to true
    pcnt = node->getNumParents();
    if (pcnt > 0)
      isObjOn[abs(numA)] = true;
  }
  else {
    // put on lostlist
    //printf("put name on lost list: %s\n",selNodeName.c_str());
    lostname = selNodeName;
    addLostObject();
    numA = num;
  }
  selectable[abs(numA)] = true;

  // find the objectNode
  //printf("look for the objectnode: %s\n",objNodeName.c_str());
  node = display->world()->lookup(objNodeName.c_str(),pfNode::getClassType());
  if (node != NULL) {
    numB = isObjectNew();
  }
  else {
    // put on lostlist
    lostname = objNodeName;
    addLostObject();
    numB = num;
  }
  checkdcs();

  // put b on a's list to turn on
  num = objNumbers[abs(numA)] + 1;
  onClick[numA] = (int *) realloc(onClick[numA], sizeof(int)*(3*num));
  num = objNumbers[abs(numA)];
  // store ON, OFF, and objB number
  if (function == "ON")
    onClick[abs(numA)][3*num] = ONKEY;
  if (function == "OFF")
    onClick[abs(numA)][3*num] = OFFKEY;
  onClick[abs(numA)][3*num+1] = numB;
  onClick[abs(numA)][3*num+2] = findGnum();
  ++objNumbers[abs(numA)];
}

void sgPicker::checkdcs() {
  if (group == "wand" && wanddcs == NULL)
       addWanddcs();
  else if (group == "head" && headdcs == NULL)
    addHeaddcs();
  else if (group == "head_no_hpr" && headnohprdcs == NULL)
    addHeadnohprdcs();
}

void sgPicker::addWanddcs() {
  // first see if there is a wand dcs from another DSO
  isit = isNodeOnGroup("savg_wand", display->scene());
  if (isit)
    wanddcs = (dpfDCS*)node2;
  else {
    // define the dcs
    wanddcs = new dpfDCS(wandPos, /* position */
			 1.0f, /* overall scale */
			 /* file  scale xyz position xyz rotation hpr */
			 NULL, DPF_IDENTITY,
			 NULL);
    
    wanddcs->setName("savg_wand");  
    display->scene()->addChild(wanddcs);
  }
}

void sgPicker::addHeaddcs() {
  // first see if there is a head dcs from another DSO
  isit = isNodeOnGroup("savg_head", display->scene());
  if (isit)
    headdcs = (dpfDCS*)node2;
  else {
    // define the head dcs
    headdcs = new dpfDCS(headPos, /* position */
			 1.0f, /* overall scale */
			 /* file  scale xyz position xyz rotation hpr */
			 NULL, DPF_IDENTITY,
			 NULL);
    
    headdcs->setName("savg_head");
    display->scene()->addChild(headdcs);
  }
}

void sgPicker::addHeadnohprdcs() {
  // first see if there is a headnohpr dcs from another DSO
  isit = isNodeOnGroup("savg_headnohpr", display->scene());
  if (isit)
    headnohprdcs = (dpfDCS*)node2;
  else {
    // define the head (no hpr) dcs
    headnohprdcs = new dpfDCS(headPos2, /* position */
			      1.0f, /* overall scale */
			      /* file  scale xyz position xyz rotation hpr */
			      NULL, DPF_IDENTITY,
			      NULL);
    
    headnohprdcs->setName("savg_headnohpr");  
    display->scene()->addChild(headnohprdcs);
  }
}

void sgPicker::readExec() {
  // find the node and make it selectable
  node = display->world()->lookup(selNodeName.c_str(),pfNode::getClassType());
  if (node != NULL) {
    //printf("in readExec: found node: %s\n",selNodeName.c_str());

    numA = isObjectNew();
    // see if the numA object is visible
    parentgroup = NULL;
    pcnt = node->getNumParents();
    //printf("parent number: %d\n",pcnt);
    if (pcnt > 0) {
      isObjOn[abs(numA)] = true;
      // if parent is not on group list, add it

      parentgroup = findFirstNamedParent();
      if (parentgroup != NULL) {
	//parentgroup = node->getParent(0);
	bool found = false;
	if (parentgroup->getName() != NULL) {
	  //printf("print the parent: %s\n",parentgroup->getName());
	  for ( int ii = 0; ii < groupcnt; ++ii ) {
	    if (strcasecmp(groupList[ii].c_str(),parentgroup->getName())==0) {
	      found = true;
	      break;
	    }
	  }
	}
	if (!found) {
	  //printf("adding group to list: %s\n",parentgroup->getName());
	  groupList[groupcnt] = parentgroup->getName();
	  ++groupcnt;
	}
      }
    }
  }
  else {
    // put on lostlist
    //printf("in sgPicker: lost node: %s\n",selNodeName.c_str());
    lostname = selNodeName;
    addLostObject();
    numA = num;
  }
  selectable[abs(numA)] = true;

  // store the command
  if (ccnt > 100)
    commandList = (char **) realloc(commandList, sizeof(char*)*(ccnt+1));
  commandList[ccnt] = strdup(pargv[3]);
  ++ccnt;

  // store the EXEC = 3
  num = objNumbers[abs(numA)];
  if (num == 0)
    onClick[abs(numA)] = new int[3];
  else
    onClick[abs(numA)] = (int *) realloc(onClick[abs(numA)], sizeof(int)*(3*num+4));
  onClick[abs(numA)][3*num] = EXECKEY;
  onClick[abs(numA)][3*num+1] = ccnt - 1;
  onClick[abs(numA)][3*num+2] = 0;
  ++objNumbers[abs(numA)];
}

int sgPicker::isObjectNew() {
  num = -1;
  for ( i = 0; i < objcnt; ++i ) {
    if (strcasecmp(node->getName(),objectPtrs[i]->getName())==0) {
      num = i;
      break;
    }
  }

  // if not found: add to list
  if (num < 0) {
    //printf("this is a new one\n");
    addObjectToList();
  }

  return num;
}

void sgPicker::addObjectToList() {
  num = objcnt;
  objectPtrs = (pfNode **) realloc(objectPtrs, sizeof(pfNode*)*(objcnt+1));
  groupnum = (int *) realloc(groupnum, sizeof(int)*(objcnt+1));
  objNumbers = (int *) realloc(objNumbers, sizeof(int)*(objcnt+1));
  onClick = (int **) realloc(onClick, sizeof(int*)*(objcnt+1));
  onClick[objcnt] = new int[3];
  isObjOn = (bool *) realloc(isObjOn, sizeof(bool)*(objcnt+1));
  selectable = (bool *) realloc(selectable, sizeof(bool)*(objcnt+1));
  objectPtrs[objcnt] = node;
  groupnum[objcnt] = findGnum();
  objNumbers[objcnt] = 0;
  isObjOn[objcnt] = false;
  selectable[objcnt] = false;
  ++objcnt;
}

void sgPicker::addLostObject() {
  num = 100;
  // see if this node is already on the lost list
  isit = false;
  for (i = 0; i < lostcnt; ++i) {
    //printf("compare: %s %s\n",lostname.c_str(),lostgroup[i].c_str());
    if (strcasecmp(lostname.c_str(),lostgroup[i].c_str())==0) {
      isit = true;
      num = lostgroupobjnum[i];
    }
  }

  if (!isit) {
    num = -objcnt;
    lostgroupobjnum[lostcnt] = num;
    objectPtrs = (pfNode **) realloc(objectPtrs, sizeof(pfNode*)*(objcnt+1));
    groupnum = (int *) realloc(groupnum, sizeof(int)*(objcnt+1));
    objNumbers = (int *) realloc(objNumbers, sizeof(int)*(objcnt+1));
    onClick = (int **) realloc(onClick, sizeof(int*)*(objcnt+1));
    onClick[objcnt] = new int[3];
    isObjOn = (bool *) realloc(isObjOn, sizeof(bool)*(objcnt+1));
    selectable = (bool *) realloc(selectable, sizeof(bool)*(objcnt+1));
    objectPtrs[objcnt] = node;
    groupnum[objcnt] = 0;
    objNumbers[objcnt] = 0;
    isObjOn[objcnt] = false;
    selectable[objcnt] = false;
    ++objcnt;

    // add to lost object list
    lostgroup[lostcnt] = lostname;
    ++lostcnt;
  }
}

pfGroup* sgPicker::findFirstNamedParent() {
   pfGroup* grp = NULL;
   grp = node->getParent(0);
   //printf("in findFirstNamedParent: pcnt: %d, name: %s\n",pcnt,grp->getName());
   while (pcnt > 0 && grp->getName() == NULL) {
     pcnt = grp->getNumParents();
     //printf("next pcnt: %d\n",pcnt);
     if (pcnt > 0)
       grp = grp->getParent(0);
     //printf("got next parent up: %s\n",grp->getName());
   }
   if (grp->getName() == NULL) grp = NULL;
   return grp;
}

int sgPicker::findGroupInScenegraph() {
  gnum = -1;
  // if there is no group defined, and the node has a parent,
  // look for one
  bool gotone = false;
  pcnt = node->getNumParents();
  if (pcnt > 0) {
    pfNode* savenode = node;
    while (pcnt > 0 && !gotone) {
      pfGroup* ggg = node->getParent(0);
      node = (pfNode*) ggg;
      const char* parname = node->getName();
      if (parname != NULL)
	gotone = findParent();
      if (gotone) {
	group = parname;
	break;
      }
      pcnt = node->getNumParents();
    }
    node = savenode;
  }

  if (gotone) {
    // if this is a new group, add it to the list
    if (gnum < 0 && strcasecmp(group.c_str(),"")!=0) {
      groupList[groupcnt] = group;
      gnum = groupcnt;
      ++groupcnt;
    }
  }

  return gnum;
}

int sgPicker::findGnum() {
  gnum = -1;
  
  // given a group, find the number in the groupList
  if (strcasecmp(group.c_str(),"")!=0) {
    for (q = 0; q < groupcnt; ++q) {
      if (strcasecmp(group.c_str(),groupList[q].c_str())==0)
	gnum = q;
    }
  }

  return gnum;
}

void sgPicker::updateOnClick() {
  // first see if there is a toggle node holding objects from 
  // another DSO
  dpfToggle* tog = 
    (dpfToggle*)display->world()->lookup("hidden_node",
					 dpfToggle::getClassType());

  int j = 0;
  // for each object, go back and find any that may have been loaded by
  // a different DSO (in the lookup table, but not loaded with this DSO)
  for ( i = 0; i < objcnt; ++i ) {
    // find objectPointers that are null and try to fill them in
    if (objectPtrs[i] == NULL) {
      node2 = display->world()->lookup(lostgroup[j].c_str(),
				       pfNode::getClassType());
      if (node2 == NULL) {
	dtkMsg.add(DTKMSG_ERROR,
		   "sgPicker:: Cannot locate the node: %s\n",lostgroup[j].c_str());
      }
      else {
	// node must be visible if attached here
	objectPtrs[i] = node2;
	isObjOn[i] = true;
	++j;
      }
    }
  } // loop over all objects
}

bool sgPicker::isNodeOnGroup(const char* name, pfGroup* grp) {
  bool isit = false;
  if (grp != NULL) {
    int scnt = grp->getNumChildren();
    node2 = NULL;
    for (k = 0; k < scnt; k++ ) {
      if (strcasecmp(grp->getChild(k)->getName(),name) == 0) {
	node2 = grp->getChild(k);
	isit = true;
      }
    }
  }

  return isit;
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int sgPicker::postFrame(void) {
  oldcount = count;
  count = pfGetFrameCount() ;
  // first read the button (whichbutton) on the mouse/wand
  int whichButton;
  dtkRecord_event *e = dequeuer->getNextEvent(button);

  // just skip out if no new data
  //if (!e)
  //  return DTK_CONTINUE ;
  
  dtkRecord_event *lastEvent ;
  
  // if we have data, toss all but the last one
  if ((oldcount+1) != count && e != NULL) {
    while (e) {
      lastEvent = e ;
      e = dequeuer->getNextEvent(button);
    }
    e = lastEvent ;
  }
  
  int buttonState = button->read(&whichButton, e);

  //--------------------------------------------------
  // button1On; indicates if button 1 is being pressed
  // button1LastFrame; indicates old position of button 1
  //--------------------------------------------------

  if(whichButton == 1) {
    //printf("whichButton is 1 \n");
    // first see if button 1 is currently pressed
    if (button->read(1,e)) {
      //printf("read1,e, buttonstate 1 = pressed:%d\n",buttonState);
      if (!button1LastFrame) {
	//printf("first frame with button 1 ");
	// switch button1ON: if off, turn it on; if on, turn it off
	if (button1On) {
	  //printf("button 1 on: turn it and the wandPointer off\n");
	  button1On = false;

	  // make sure the wandPointer is off
	  if (shm2) {
	    selected = 2;
	    shm2->write(&selected);
	  }  
	}
	else {
	  resetIsObjOn();
	  button1On = true;
	  //printf("button 1 not on\n");
	  // make sure the wandPointer is on
	  if (shm2) {
	    selected = 1;
	    shm2->write(&selected);
	  }  
	}
      }
      button1LastFrame = true;
    }
    else {
      //printf("just set lastfrmae false\n");
      button1LastFrame = false;
    }
  }
  //--------------------------------------------------

  // read the wand direction 
  if (wand)
      wand->read(wandData);

  // read the headd direction 
  if (head)
      head->read(headData);

  // create a vector in the +y direction and move according to the wand
  pfMatrix wandWorld;
  wandWorld.makeEuler(wandData[3],wandData[4],wandData[5]);
  //printf("wand data should change: %f %f %f\n",wandData[3],wandData[4],wandData[5]);
  upVector.set(0.f, 1.f, 0.f);
  check.xformVec(upVector, wandWorld);

  wandPos[0] = wandData[0];
  wandPos[1] = wandData[1];
  wandPos[2] = wandData[2];
  wandPos[3] = wandData[3];
  wandPos[4] = wandData[4];
  wandPos[5] = wandData[5];

  headPos[0] = headData[0];
  headPos[1] = headData[1];
  headPos[2] = headData[2];
  headPos[3] = headData[3];
  headPos[4] = headData[4];
  headPos[5] = headData[5];

  headPos2[0] = headData[0];
  headPos2[1] = headData[1];
  headPos2[2] = headData[2];
  headPos2[3] = 0.0;
  headPos2[4] = 0.0;
  headPos2[5] = 0.0;

  if (button1On) {
    //printf("button on\n");
    // make the wandPointer appear
    if (shm2) {
      selected = 1;
      shm2->write(&selected);
    }

    // do the intersection and find the visible object that is closest
    // to the wand
    // the next routine reset the last_one_hit object number if found
    int last = selectSelectable();
    // get new value of newCallSet
    if (shm4)
      shm4->read(&newCallSet);

    if (last > -1) {
      last_one_hit = last;
      objectSelected = true;

      // reset the toggle
      resetAorB();

      // make a sound to indicate that the object was selected
      button1On = false;
      // make the wandPointer disappear
      if (shm2) {
	selected = 2;
	shm2->write(&selected);
      }
    }
  }
  else {
    // button1Off - restart
    objectSelected = false;
  } 
  // initialization - a one time thing - execute initial commands
  // and then update the onClick list with any new objects
  if (iccnt > 0) {
    if (framenum == 0) {
      for ( i = 0; i < iccnt; ++i ) {
	commandA = initCommandList[i];
	if (commandA != NULL) {
	  // set the flag to tell the other program to read a call
	  newCallSet = 2;
	  if (shm4)
	    shm4->write(&newCallSet);
	  if (shm_call) {
	    shm_call->write(commandA);
	    // now wait until it is sent before sending the next one
	    while (newCallSet == 2)
	      shm4->read(&newCallSet);
	  }
	}  
      }
      framenum = 1;
    }
    else {
      if (framenum < 300)
	++framenum;
      else {
	// update the onClick list
	updateOnClick();
	iccnt = 0;
      }
    }
  }

  if (iccnt == 0 && !objectCheck) {
    if (framenum >= 1) {
      // one time - check all objects
      doObjectCheck();
      objectCheck = true;
    }
    else 
      ++framenum;
  }

  if (framenum < 300)
    ++framenum;

  return DTK_CONTINUE ;
}

void sgPicker::doObjectCheck() {
  gnum = -1;
  // for each object, check to see if it has a parent defined
  for ( i = 0; i < objcnt; ++i ) {
    node = objectPtrs[i];
    // find objectPointers that are null and try to fill them in
    if (groupnum[i] == -1 && node != NULL) {
      pcnt = node->getNumParents();
      if (pcnt > 0) {
	group = "";
	gnum = findGroupInScenegraph();
	if (gnum > -1) {
	  groupnum[i] = gnum;
	}
      }
    }
  } // loop over all objects
}

bool sgPicker::findParent() {
  // is this name node on the list?
  gnum = -1;  
  bool found = false;
  const char* pp = node->getName();
  for ( int ii = 0; ii < groupcnt; ++ii ) {
    if (strcasecmp(groupList[ii].c_str(),pp)==0) {
      gnum = ii;
      found = true;
      break;
    }
  }
  
  return found;
}

void sgPicker::resetIsObjOn() {
  // go through the selectable items and see which ones
  // are hooked into the scenegraph
  for (j = 0; j < objcnt; j++ ) {
    if (selectable[j]) {
      node = objectPtrs[j];
      if (node == NULL) continue;

      pcnt = node->getNumParents();
      if (pcnt == 0)
	isObjOn[j] = false;
      else
	isObjOn[j] = true;
    }
  }
}

void sgPicker::resetAorB() {

  // before you start, initialize the call button for executables
  newCallSet = 1;
  if (shm4)
    shm4->write(&newCallSet);
  
  // go through the list of things to do for the chosen node
  for (j = 0; j < objNumbers[last_one_hit]; ++j) {
    numA = abs(onClick[last_one_hit][3*j]);
    numB = abs(onClick[last_one_hit][3*j+1]);
    //printf("in reset: lastonehit = %d; numA: %d  numB: %d\n",last_one_hit,numA,numB);
    gnum = 0;
    if (onClick[last_one_hit][3*j+2] <= groupcnt)
      gnum = onClick[last_one_hit][3*j+2];
    if (gnum < 0) {
      // find group for last one hit
      node = objectPtrs[last_one_hit];
      gnum = findGroupInScenegraph();
      if (gnum < 0)
	gnum = 0;
    }
    group = groupList[gnum];
    if (numA == ONKEY) {
      // turn on node numB
      node = objectPtrs[numB];
      addToDisplay();
      isObjOn[numB] = true;
      // update where the node is on the groupnum list
      groupnum[numB] = gnum;
    }
    else if (numA == OFFKEY) {
      // turn on node numB
      node = objectPtrs[numB];
      removeTheNode();
      isObjOn[numB] = false;
    }
    else if (numA == EXECKEY) {
      // execute the command
      commandA = commandList[numB];

      if (shm4) {
	while (newCallSet == 2) {
	  shm4->read(&newCallSet);
	}
      }

      // as long as the call is not NULL, 
      // read it out so dtk-runCommand can run it
      if (commandA != NULL) {
	// set the flag to tell the other program to read a call
	newCallSet = 2;
	if (shm4)
	  shm4->write(&newCallSet);
	if (shm_call) {
	  shm_call->write(commandA);
	}
      }  
    }
  }
}

void sgPicker::addToDisplay() {
  if (strcasecmp(group.c_str(),"world")==0)
    display->world()->addChild(node);
  else if (group == "scene")
    display->scene()->addChild(node);
  else if (group == "ether")
    display->ether()->addChild(node);
  else if (group == "nav")
    display->nav()->addChild(node);
  else if (group == "wand")
    wanddcs->addChild(node);
  else if (group == "head")
    headdcs->addChild(node);
  else if (group == "head_no_hpr")
    headnohprdcs->addChild(node);
  else {
    // find the right dcs or scs or group
    // search for a dcs
    newdcs = NULL;
    newdcs = (pfDCS*)display->scene()->lookup(group.c_str(),pfDCS::getClassType());
    if (newdcs == NULL) {
      // search for an scs
      newscs = NULL;
      newscs = (pfSCS*)display->scene()->lookup(group.c_str(),pfSCS::getClassType());
      if (newscs == NULL) {
	// search for a group
	newgroup = NULL;
	newgroup = (pfGroup*)display->world()->lookup(group.c_str(),pfGroup::getClassType());
	if (newgroup == NULL) {
	  dtkMsg.add(DTKMSG_ERROR,
		   "sgPicker:: Cannot locate the group: %s\n",group.c_str());
	}
	else {
	  newgroup->addChild(node);
	}
      }
      else
	newscs->addChild(node);
    }
    else
      newdcs->addChild(node);
  }
}

void sgPicker::removeTheNode() {
  // find the parent
  parentgroup = NULL;
  pcnt = node->getNumParents();
  if (pcnt > 0)
    parentgroup = node->getParent(0);
  
  if (node != NULL && parentgroup != NULL) 
    parentgroup->removeChild(node);
  else {
    dtkMsg.add(DTKMSG_ERROR,
	       "sgPicker:: Cannot locate the parent for node: %s\n",node->getName());
  }
}

int sgPicker::selectSelectable() {
  int last_one_hit = -1;
  // for each visible object, see if the check vector is within
  // its bounding sphere, and keep track of the one that is
  // the closest to the wand
  // loop through each pair and find the visible one
  distToWand = 1000000.0;

  for (j = 0; j < objcnt; j++ ) {
    if (last_one_hit > -1) break;
    if (isObjOn[j] && selectable[j]) {
      node = objectPtrs[j];
      gnum = groupnum[j];
      //printf("next node: %s and number: %d\n",node->getName(),gnum);
      if (gnum > -1)
	group = groupList[gnum];
      else {
	gnum = findGroupInScenegraph();
	if (gnum > -1)
	  group = groupList[gnum];
	else
	  group = "world";
      }
      //printf("print the group: %s\n",group.c_str());
      if (node == NULL) continue;
      // find the center point of the object in world coordinates
      setPtCoord();  
      // Is the distance from centerPt to the wand is less than the radius?
      
      // This code is from the objectMover: OMS.C
      // the line from the wand is defined by a point(wandCoord) 
      // and a direction (check) r = r0 + d(direction);
      // so, need to find the distance d up the line that is perpendicular
      // to the center point, that is at the point P=(a,b,c)
      
      // the dot product of the P-centr with P-wand = 0 
      // if the vectors are perpendicular
      // did some algebra and came up with an expression for d:
      float d = check[0]*(ptCoord.xyz[0]-wandData[0]) +
	check[1]*(ptCoord.xyz[1]-wandData[1]) +
	check[2]*(ptCoord.xyz[2]-wandData[2]);
      d = d/(check[0]*check[0] + check[1]*check[1] + check[2]*check[2]);
      //printf("dist: %f\n",d);
      // this gives the point of intersection with the wand, int: now find
      // the distance from the centerpoint to int
      intPt[0] = wandData[0] + d*check[0];
      intPt[1] = wandData[1] + d*check[1];
      intPt[2] = wandData[2] + d*check[2];
      //printf("current check: %f %f %f\n",check[0],check[1],check[2]);
      float dist = sqrt((ptCoord.xyz[0]-intPt[0])*(ptCoord.xyz[0]-intPt[0]) +
			(ptCoord.xyz[1]-intPt[1])*(ptCoord.xyz[1]-intPt[1]) +
			(ptCoord.xyz[2]-intPt[2])*(ptCoord.xyz[2]-intPt[2]));
      //printf("dist: %f \n",dist);
      
      // see if the sphere hits the wand
      if (dist < objRadius && d > 0) {
	// this object hits!
	// save distance = d
	if (dist < distToWand) {
	  distToWand = dist;
	  last_one_hit = j;
	  break;
	}
      }
    }
    }
  
  return last_one_hit;
}

void sgPicker::setPtCoord() {

  // find bounding sphere
  box = dpf_getBBox(node);
 
  // find the center point of the box
  centerPt[0] = (box.max[0]+box.min[0])/2.0;
  centerPt[1] = (box.max[1]+box.min[1])/2.0;
  centerPt[2] = (box.max[2]+box.min[2])/2.0;

  diff = abs(box.max[0]-box.min[0]);
  if (abs(box.max[1]-box.min[1]) > diff)
    diff = abs(box.max[1]-box.min[1]);
  if (abs(box.max[2]-box.min[2]) > diff)
    diff = abs(box.max[2]-box.min[2]);
  objRadius = diff/1.5;
  //printf("new radius: %f\n",objRadius);
  //printf("center point for object 1: %f %f %f\n",centerPt[0],centerPt[1],centerPt[2]);
  
  ptCoord.xyz.set(centerPt[0], centerPt[1], centerPt[2]);
  
  if (group == "head" || group == "head_no_hpr") {
    hdcoord = display->headPos();

    if (group == "head") {
      // make a matrix for the head rotation
      headWorld.makeEuler(hdcoord->hpr[0],hdcoord->hpr[1],hdcoord->hpr[2]);
      // transform the center point by this matrix
      centerVector.set(centerPt[0], centerPt[1], centerPt[2]);
      checkhead.xformVec(centerVector, headWorld);
      ptCoord.xyz.set(checkhead[0], checkhead[1], checkhead[2]);
    }
    //printf("ptcoord in head: %f %f %f\n",ptCoord.xyz[0],ptCoord.xyz[1],ptCoord.xyz[2]);
    // move the object with the head xyz
    ptCoord.xyz[0] += hdcoord->xyz[0];
    ptCoord.xyz[1] += hdcoord->xyz[1];
    ptCoord.xyz[2] += hdcoord->xyz[2];
  }

  else if (group == "ether") {
    display->etherToDpf(&ptCoord);
  }

  else if (group == "world") {
    display->worldToDpf(&ptCoord);
  }

  else if (gnum > 6) {
    // check for object hanging on a dcs or scs
    newdcs = (pfDCS*)display->scene()->lookup(group.c_str(),pfDCS::getClassType());
    if (newdcs != NULL) {
      // put object on dcs
      // make a matrix for the dcs
      newdcs->getMat(dcsMatrix);
      // transform the center point by this matrix
      centerVector.set(centerPt[0], centerPt[1], centerPt[2]);
      checkhead.xformVec(centerVector, dcsMatrix);
	// translate
      dcsMatrix.getRow(3,row3);
      //printf("move %s by %f %f %f\n",node->getName(),row3[0],row3[1],row3[2]);
      ptCoord.xyz.set(checkhead[0]+row3[0], 
		      checkhead[1]+row3[1], checkhead[2]+row3[2]);
    }
    else {
      newscs = (pfSCS*)display->scene()->lookup(group.c_str(),pfSCS::getClassType());
      if (newscs != NULL) {
	// put object on scs
	// make a matrix for the scs
	newscs->getMat(scsMatrix);
	// transform the center point by this matrix;
	centerVector.set(centerPt[0], centerPt[1], centerPt[2]);
	checkhead.xformVec(centerVector, scsMatrix);
	// translate
	scsMatrix.getRow(3,row3);
	ptCoord.xyz.set(checkhead[0]+row3[0], 
			checkhead[1]+row3[1], checkhead[2]+row3[2]);
      }
    }
  }
  //printf("center point for object is: %f %f %f\n",ptCoord.xyz[0],ptCoord.xyz[1],ptCoord.xyz[2]);
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new sgPicker(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
