/* Switches navigations and resets location and navscale using button presses.
   using the dtkShmNavSwitch DSO
*/

#define SHM_NAV_NAME "shmNavControl"
#define SHM_TEXT_NAME "shmHeadsUpText" 

#define MIN_BUTTONS 3

// how long does the text remain visible?
#define TIMEOUT (2.f)

#include <dtk.h>
#include <dtk/dtkDSO_loader.h>

class buttonNavSwitch : public dtkAugment {
public:
  buttonNavSwitch(dtkManager *) ;
  ~buttonNavSwitch(void) ;

  int preFrame(void) ;
  int postFrame(void) ;

private:

  dtkManager *mgr ;
  dtkNavList *navList;
  dtkTime t;

  dtkSharedMem *shmNav ;
  dtkSharedMem *shmText ;
  bool visible ;
  dtkInButton *button ;
  dtkDequeuer *dequeuer ;
} ;


buttonNavSwitch::buttonNavSwitch(dtkManager *m) :
  dtkAugment("buttonNavSwitch") {
  setDescription("adds location, navscale reset and navigation control"
		 " to the current dtkNav navigation:"
		 " 'button-0' =  reset location and navscale"
		 " 'button-2' =  next navigation") ;
  mgr = m ;
  dequeuer = NULL ;
  shmNav = NULL ;
  shmText = NULL ;
  visible = false ;
  button = NULL ;

  // try to load shmNavControl if it is not already loaded
  if (!mgr->check("shmNavControl") && (mgr->load("shmNavControl"))) 
    return ;

  // try to load shmHeadsUpText if it is not already loaded
  if (!mgr->check("shmHeadsUpText") && (mgr->load("shmHeadsUpText"))) 
    return ;

  // dtkAugment::dtkAugment() will not validate the object
  validate() ;
}


buttonNavSwitch::~buttonNavSwitch(void) {
  if (dequeuer) {
    delete dequeuer ;
    dequeuer = NULL ;
  }
}

int buttonNavSwitch::preFrame(void) {

  /********************* Get a dtkNavList ***********************/
  navList = (dtkNavList *) mgr->check("dtkNavList", DTKNAVLIST_TYPE) ;
  if (!navList) {
    navList = new dtkNavList;
    if (mgr->add(navList)) {
      delete navList;
      navList = NULL;
      return dtkAugment::ERROR_;
    }
  }

  // needed??
  t.reset(0.0,1);

  // presumptuously unload buttonNavControl
  mgr->remove("buttonNavControl") ;
  
  // try to connect to shared memory created by shmNavControl
  shmNav =  new dtkSharedMem(SHM_NAV_NAME) ;
  if (!shmNav)
    dtkMsg.add(DTKMSG_ERROR,
	       "buttonNavSwitch::preFrame() failed:"
	       " couldn't open shared memory segment %s\n",SHM_NAV_NAME) ; 

  // try to connect to shared memory created by shmHeadsUpText
  shmText =  new dtkSharedMem(SHM_TEXT_NAME) ;
  if (!shmText)
    dtkMsg.add(DTKMSG_ERROR,
	       "buttonNavSwitch::preFrame() failed:"
	       " couldn't open shared memory segment %s\n",SHM_TEXT_NAME) ; 

  dequeuer = new dtkDequeuer(mgr->record()) ;
  if (!dequeuer || dequeuer->isInvalid()) 
    return ERROR_ ;

  button = (dtkInButton *) mgr->get("buttons", DTKINBUTTON_TYPE) ;
  if (!button) 
    return ERROR_ ;
  
  if (button->number() < MIN_BUTTONS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "buttonNavSwitch::preFrame() failed:"
	       " the dtkInButton named \"%s\" has "
	       "just %d buttons"
	       " which is less than the %d needed.\n",
	       button->getName(), button->number(),
	       MIN_BUTTONS) ;
    return ERROR_ ;
  }
  button->queue() ;
  
  // just call buttonNavSwitch::preFrame() once at the start.
  return REMOVE_CALLBACK ;
}


int buttonNavSwitch::postFrame(void) {

  if ((float)t.get()>TIMEOUT && visible) {
    shmText->write("") ;
    visible = false ;
  }

  dtkNav *nav = navList->current() ;
  
  // nothing to do if no navigation!
  if (!nav)
    return CONTINUE ;

  dtkRecord_event *e ;
  while ((e = dequeuer->getNextEvent(button))) {
    int whichButton ;
    int buttonState = button->read(&whichButton, e) ;
    if (buttonState && whichButton < 3) { // if pressed button 0 or 1 or 2
      if (whichButton == 0) {
	// location and scale reset for button 0
	dtkMsg.add(DTKMSG_INFO, "buttonNavSwitch::postFrame(): "
		   "set scale 1\n") ;
	if (shmNav->write("set scale 1")) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "buttonNavSwitch::postFrame() failed:"
		     " unable to write to shared memory") ;
	  return ERROR_ ;
	}
	dtkMsg.add(DTKMSG_INFO, "buttonNavSwitch::postFrame(): "
		   "set location 0 0 0 0 0 0\n") ;
	if (shmNav->write("set location 0 0 0 0 0 0")) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "buttonNavSwitch::postFrame() failed:"
		     " unable to write to shared memory") ;
	  return ERROR_ ;
	}
	shmText->write("reset location and navscale") ;
	visible = true ;
	t.reset(0.0,1);
      } else if (whichButton == 2) {
	// switch navigation for button 2
	dtkMsg.add(DTKMSG_INFO, "buttonNavSwitch::postFrame(): "
		   "navigation next\n") ;
	if (shmNav->write("navigation next")) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "buttonNavSwitch::postFrame() failed:"
		     " unable to write to shared memory") ;
	  return ERROR_ ;
	}
	dtkNav *next_nav = navList->next(nav);
	if (!next_nav)
	  next_nav = navList->first();
	shmText->write(next_nav->getName()) ;
	visible = true ;
	t.reset(0.0,1);
#if 0
      // for debugging
      } else if (whichButton == 1) {
	shmNav->write("FOO") ; 
	shmNav->write("BAR") ;
	shmNav->write("BAZ") ;
#endif	
      }
    }
  } 
  return CONTINUE ;
}


/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  return new buttonNavSwitch(manager) ;
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment ;
  return DTKDSO_UNLOAD_CONTINUE ;
}


