// set the navigation's pivot point based on the valeu of six floats in
// shared memory.  THE NUMBERS REPRESENT VALUES IN WORLD COORDINATES!
// The environment variable "SHM_NAV_WORLD_PIVOT" contains the name
// of the segment to read.  If the variable isn't set, the pivot point is
// set to 0,0,0 0,0,0. 

#include <sys/types.h>
#include <stdlib.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

#define SHM "SHM_NAV_WORLD_PIVOT"

class shmNavWorldPivot : public dtkAugment {
public:
  shmNavWorldPivot(dpf *);
  int postFrame(void);
  
private:
  
  dpf *app ;
  dpfDisplay *display ;
  dtkSharedMem *shm ;
  float shmPivotFloats[6] ;
  dtkCoord shmPivot ;
  dtkNavList *navList ;
};


shmNavWorldPivot::shmNavWorldPivot(dpf *app_in) :
  dtkAugment("shmNavWorldPivot") {
  app = app_in ;
  display = app->display() ;
  navList = NULL ;
  setDescription("sets the pivot point based on the value of shared "
		 "memory segment named $"SHM", which are in world coordinates") ;

  shmPivotFloats[0] = shmPivotFloats[1] = shmPivotFloats[2] = 
    shmPivotFloats[3] = shmPivotFloats[4] = shmPivotFloats[5] = 0.f ;
  shmPivot.set(shmPivotFloats) ;

  shm = NULL ;
  char *shmName = getenv(SHM) ;
  if (shmName) {
    shm = new dtkSharedMem(shmName) ;
    if(!shm || shm->isInvalid()) {
      dtkMsg.add(DTKMSG_ERROR,"shmNavWorldPivot: unable to get valid shared "
		 "memory \"%s\".\n", shmName) ;
      return ;
    }
    if (shm->getSize() == 3*sizeof(float)) {
      dtkMsg.add(DTKMSG_INFO,"shmNavWorldPivot: shared "
		 "memory is of size 3 floats, only pivot position will be updated-"
		 "HPR will remain (0,0,0)\n") ;
    } else if (shm->getSize() == 6*sizeof(float)) {
      dtkMsg.add(DTKMSG_NOTICE,"shmNavWorldPivot: shared "
		 "memory is of size 6 floats, pivot position and orientation will be updated\n") ;
    } else {
      dtkMsg.add(DTKMSG_ERROR,"shmNavWorldPivot: shared "
		 "memory \"%s\" is of incorrect size- should be 3 or 6 floats, is %d.\n",
		 shmName, shm->getSize()) ;
      return ;
    }
  }
  // dtkAugment::dtkAugment() will not validate the object
  validate() ;
}

int shmNavWorldPivot::postFrame(void) {
  // this should get done just once- first pass
  if (!navList) {
    navList = (dtkNavList *) app->check("dtkNavList", DTKNAVLIST_TYPE) ;
    if (!navList) {
      navList = new dtkNavList ;
      if (app->add(navList)) {
	delete navList ;
	navList = NULL ;
	return ERROR_ ;
      }
    }
  }
  
  if (shm && shm->read(shmPivotFloats)) 
    return ERROR_ ;
  shmPivot.set(shmPivotFloats) ;
  display->worldToDpf(&shmPivot) ;

  // just set the point, but not the hpr
  shmPivot.h =  shmPivot.p =  shmPivot.r = 0.f ;

  navList->current()->pivot(&shmPivot) ;
  
  
  //shmPivot.print() ;
  //navList->current()->pivot()->print() ;
  
  return CONTINUE ;
}



/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new shmNavWorldPivot(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
