// simple DSO to reroute button data into seperate shared memory segments,
// one per button.  A button device can have up to 32 buttons, one per bit
// of a u_int32_t (unsigned 32 bit int), number 0-31.

// one env var is set per button.  if the env var isn't set for a particular
// button, data isn't rerouted for that button.

// the env var for button N is SLB_N, without leading zeros, so SLB_0 to
// SLB_31 are evaluated.

// if you specify more env vars than buttons on the device we unload.

#include <stdlib.h>
#include <stdio.h>

#include <dtk.h>
#include <dtk/dtkDSO_loader.h>

class slb : public dtkAugment {
public:
  slb(dtkManager *) ;
  
  int preFrame(void) ;
  int postFrame(void) ;
  
private:
  dtkManager *mgr ;
  dtkRecord *record ;
  dtkInButton *button ;
  int buttons ;
  dtkDequeuer *dequeuer ;
  char *envvar[32] ;
  dtkSharedMem *shm[32] ;
  // button state the previous frame
  char last_b[32] ;
} ;

slb::slb(dtkManager *m) :
  dtkAugment("slb") {

  mgr = m ;
  button = NULL ;
  dequeuer = NULL ;

  char slbenvvar[32][10] ;
  for (int i = 0; i<32; i++) {
    shm[i] = NULL ;
    last_b[i] = 0 ;
    strcpy(slbenvvar[i], "SLB_") ;
    int n = strlen(slbenvvar[i]) ;
    if (i<10) {
      slbenvvar[i][n] = i+'0' ;
      slbenvvar[i][n+1] = '\0' ;
    } else {
      slbenvvar[i][n] = (i/10)+'0' ;
      slbenvvar[i][n+1] = (i%10)+'0' ;
      slbenvvar[i][n+2] = '\0' ;
    }
    envvar[i] = getenv(slbenvvar[i]) ;
  }

  setDescription("reroutes button data to seperate shared memory segments") ;
  
  // dtkAugment::dtkAugment() will not validate the object
  validate() ;
}

int slb::preFrame(void) {

  dequeuer = new dtkDequeuer(mgr->record()) ;
  if (!dequeuer || dequeuer->isInvalid()) 
    return ERROR_ ;
  
  button = (dtkInButton *) mgr->get("buttons", DTKINBUTTON_TYPE) ;
  if (!button) 
    return ERROR_ ;
  
  buttons = button->number() ;
  bool bomb = false ;
  for (int i = buttons; i<32; i++) {
    if (envvar[i]) {
      dtkMsg.add(DTKMSG_ERROR,
		 "slb::preFrame() failed:"
		 " device has %d buttons, but action is specified for"
		 " button %d\n", buttons, i) ;
      bomb = true ;
    }
  }   

  if (bomb) 
    return ERROR_ ;

  // WHY DO WE DO THIS????  BECAUSE IF WE DON'T IT DOESN'T WORK IN THE RAVE!!
  button->queue() ;

  char zero ;
  zero = 0 ;
  for (int i = 0; i < buttons; i++) {
    if (envvar[i]) {
      dtkMsg.add(DTKMSG_INFO,
		 "slb::preFrame(): "
		 "button %d writes to DTK shared memory \"%s\"\n",
		 i,envvar[i]) ;
      shm[i] =  new dtkSharedMem(1, envvar[i], &zero) ;
      if (!shm[i] || shm[i]->isInvalid()) {
	dtkMsg.add(DTKMSG_ERROR,
		   "slb::preFrame() failed:"
		   " couldn't open shared memory segment\"%s\"\n",
		   envvar[i]) ; 
	return ERROR_ ;
      }
    }
  }
  return REMOVE_CALLBACK ;
}


int slb::postFrame(void) {
  
  dtkRecord_event *e ;
  // if we have a button event
  while ((e = dequeuer->getNextEvent(button))) {
    //printf("slb: got a button press\n") ;
    // loop through all buttons we're watching 
    for (int i = 0; i < buttons; i++) {
      // if we're watching this button
      if (shm[i]) {
	// get the status for this button
	char b = (char) button->read(i,e) ;
	if (b!=last_b[i]) {
	  last_b[i] = b ;
	  printf("slb: writing button %d, value = %d\n",i,b) ;
	  // and write it to the slb shared memory
	  if (shm[i]->write(&b)) {
	    dtkMsg.add(DTKMSG_ERROR,
		       "slb::postFrame() failed:"
		       " unable to write to shared memory \"%s\"\n",
		       envvar[i]) ;
	    return ERROR_ ;
	  }
	}
      }
    }
  }  
  
  return CONTINUE ;
}

/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p) {
  return new slb(manager) ;
}

static int dtkDSO_unloader(dtkAugment *augment) {
  delete augment ;
  return DTKDSO_UNLOAD_CONTINUE ;
}

