/*

SOME COMMENTS ON THE USE OF THE Fl_Pack CLASS IN FLTK

When using Fl_Pack, you need to create the Fl_Pack with exactly
the right dimensions for the contained children or things don't
pack and resize properly.

Rules for packing Vertically:

	* make sure all children have same width (not sure this is
		necessary, but I think it is)

	* make sure that the height of the Fl_Pack is the sum
		of the heights of the contained children

	* I'm not sure what happens if you introduce "spacing"

	* create all children with x,y position of 0,0

Horizontal packing is done in the corresponding way.

You know, if you're going to do all of the above, you might
as well use an Fl_Group and position the widgets yourself.

Looking at the fltk code, the only difference between Fl_Group and
Fl_Pack is that Fl_Pack reimplements the redraw method.  The Fl_Pack
version of this method goes through a whole resizing procedure
for all of the children.  I don't see why this is necessary for
every redraw.  Why not do the positioning (the packing) of the
the children at the time of the "add", then let the Fl_Group
resize and redraw methods serve for Fl_Pack.

In other words, Fl_Pack should override the add method, not the
redraw method.  If the positions of the children relative to
the group are set at the time of "add", then after that, you
just have a regular Fl_Group.

(I guess you'd have to override insert and so on as well.)

Or maybe you override "resize".

Or maybe you just fix Fl_Pack::redraw.

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>


#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Light_Button.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Counter.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Text_Buffer.H>
#include <FL/Fl_Text_Display.H>
#include <FL/Fl_Image.H>
#include <FL/Fl_XBM_Image.H>
#include <FL/Fl_GIF_Image.H>
#include <FL/Fl_XPM_Image.H>
#include <FL/Fl_JPEG_Image.H>
#include <FL/Fl_PNG_Image.H>
#include <FL/Fl_PNM_Image.H>

#define MIN(x,y)	(((x)<(y))?(x):(y))
#define MAX(x,y)	(((x)>(y))?(x):(y))



#define MAX_POINTS 1000



typedef struct measurement
	{
	float distance;
	float point[2][3];
	} Measurement;

typedef struct img_hl_desc {
	Fl_Image *img;
	char * pixelArray;
	int width;
	int height;
	int depth;
	int count;
	int leftHL;
	int rightHL;
	int movingLeft;
	int leftEdgeX;
	int rightEdgeX;
	double leftEdgeData;
	double rightEdgeData;
	} ImgHLDesc;

ImgHLDesc ImageHLDescriptor;

static Fl_Image *NoDataImg = NULL;
static Fl_Image *CurrHistogramImg = NULL;
static Fl_Box *ImgBox = NULL;




static int NumPts = 0;

static Measurement MeasureList[MAX_POINTS];

#if 0
static int BufferPosition[MAX_POINTS+1];
static Fl_Text_Buffer *TxtBuffer;
static int MinPos, MaxPos;
static float MinVal, MaxVal;
static Fl_Value_Slider *MinSlider;
static Fl_Value_Slider *MaxSlider;
#endif

static Fl_Widget *MeanLabel;
static Fl_Widget *StdDevLabel;

static Fl_Image *getImage (char *fileName);

static void
setupImgHLDesc (ImgHLDesc *imgHLDescriptor, Fl_Image *img, double dataLeft, double dataRight)
	{
	imgHLDescriptor->img = img;
	imgHLDescriptor->pixelArray = (char *)((Fl_RGB_Image *)img)->array;
	imgHLDescriptor->width = img->w();
	imgHLDescriptor->height = img->h();
	imgHLDescriptor->depth = img->d();
	imgHLDescriptor->count = img->count();
	imgHLDescriptor->leftHL = -1;
	imgHLDescriptor->rightHL = -1;
	imgHLDescriptor->movingLeft = 0;
	imgHLDescriptor->leftEdgeX = 29;
	imgHLDescriptor->rightEdgeX = 344;
	imgHLDescriptor->leftEdgeData = dataLeft;
	imgHLDescriptor->rightEdgeData = dataRight;
	}  // end of setupImgHLDescForNewImg


static int 
convertDpEpsToPng (char *epsFN, char *pngFN)
	{

	char cmd[1000];

printf ("******************************\n\n\n\n\n");

	sprintf (cmd, 
	    "gs -dEPSCrop -sDEVICE=png256 -sOutputFile=%s.tmp1.png "
	    "-sNOPAUSE  -q %s -c showpage -c quit", 
	    pngFN, epsFN);


system ("which gs");
system ("printenv PATH");

printf ("%s\n", cmd);
	system (cmd);
printf ("done with gs\n\n");

#if 0
	sprintf (cmd, 
	  "/usr/local/nekoware/bin/convert -rotate 90 -crop 360x200+59+42 "
	  "%s.tmp1.png PNG:%s", pngFN, pngFN);
#else
	sprintf (cmd, 
	  "convert -rotate 90 -crop 360x200+59+42 "
	  "%s.tmp1.png PNG:%s", pngFN, pngFN);
#endif
printf ("%s\n", cmd);
	system (cmd);

#if 1
	sprintf (cmd, "rm %s.tmp1.png", pngFN);
#endif
printf ("%s\n", cmd);
	system (cmd);

#if 1
#if 0
	sprintf (cmd, "/usr/local/nekoware/bin/convert PNG:%s  GIF:%s.gif",
		pngFN, pngFN);
#else
	sprintf (cmd, "convert PNG:%s  GIF:%s.gif",
		pngFN, pngFN);
#endif
printf ("%s\n", cmd);
	system (cmd);
#endif

printf ("******************************\n\n\n\n\n");
	return 0;
	}


static int
makeNoDataImg ()
	{

	FILE *fp;

	if (NoDataImg != NULL)
		{
		return 0;
		}


	if ((fp = fopen ("dpScript", "w")) == NULL)
		{
		fprintf (stderr, 
		 "Unable to open output dataplot script file dpScript.\n");
		return -1;
		}

		

	fprintf (fp, "WINDOW COORDINATES 0 56 60 100 \n");
	fprintf (fp, "set IPL1NA tmp.eps\n");
	fprintf (fp, "device 2 POSTSCRIPT ENCAPSULATED\n");
	fprintf (fp, "legend size 8\n");
	fprintf (fp, "legend case asis\n");
	fprintf (fp, "legend 1 coordinates  30 50\n");
	fprintf (fp, "legend 1 No data present.\n");
	fprintf (fp, "x1frame off\n");
	fprintf (fp, "x2frame off\n");
	fprintf (fp, "y1frame off\n");
	fprintf (fp, "y2frame off\n");
	fprintf (fp, "let x = sequence 1 1 2\n");
	fprintf (fp, "char blank\n");
	fprintf (fp, "line blank\n");
	fprintf (fp, "plot x\n");
	fprintf (fp, "device 2 close\n");
	fprintf (fp, "quit\n");

	fclose (fp);

	system ("dataplot < dpScript > /dev/null");

	sleep (3);

	if (convertDpEpsToPng ("tmp.eps", "tmp.png"))
		{
		return -1;
		}

	NoDataImg = getImage ("tmp.png");

	return (NoDataImg == NULL);

	}  // end of makeNoDataImg





static int
writeFakeData  (int numPts, char *fileName)
	{
	FILE *fp;
	int i, j;

	if ( (fp = fopen (fileName, "w")) == NULL)
		{
		fprintf (stderr,"Unable to open fake file < %s >.\n", fileName);
		return -1;
		}

	for (i = 0; i < numPts; i++)
		{
		for (j = 0; j < 6; j++)
			{
			fprintf (fp, "%f ", drand48());
			}
		fprintf (fp, "\n");
		}

	fclose (fp);
	return 0;
	}  // end of writeFakeData




int getMeasurements ()
        {
        int i;
        char line[1000];
        char token[100];
        FILE *fp;
        static int pid = 0;
        static int seq = 0;
        char fileName[1000];

        //
        // Fill in
        //    NumPts
        //    MeasureList[i].point[j][k]
        //    MeasureList[i].distance
        //



        if (pid == 0)
                {
                pid = getpid ();
                }

        seq++;

        sprintf (fileName, "%d.%04d.M1D.dat", pid, seq);

#if 1
        char cmd[2000];
        sprintf (cmd, "hev-measurement1D save %s", fileName);
        system (cmd);
#else

	writeFakeData  (50, fileName);

#endif


        sleep (1);


        if ( (fp = fopen (fileName, "r")) == NULL)
                {
                fprintf (stderr,
                  "Error opening measurement file %s\n", fileName);
                return -1;
                }

        i = 0;
        while (fgets (line, sizeof(line), fp) != NULL)
          {
printf ("got line < %s >\n", line);
          if (sscanf (line, "%s", token) != 0)
            {
            if (token[0] != '#')
                {
                if (sscanf (line, "%f %f %f %f %f %f",
                        &(MeasureList[i].point[0][0]),
                        &(MeasureList[i].point[0][1]),
                        &(MeasureList[i].point[0][2]),
                        &(MeasureList[i].point[1][0]),
                        &(MeasureList[i].point[1][1]),
                        &(MeasureList[i].point[1][2]) ) != 6 )
                  {
                  fprintf (stderr,
                           "Bad line in measurement file; line skipped.\n");
                  }
                else
                  {
                  MeasureList[i].distance = sqrt (
                   (MeasureList[i].point[0][0]-MeasureList[i].point[1][0]) *
                   (MeasureList[i].point[0][0]-MeasureList[i].point[1][0])    +
                   (MeasureList[i].point[0][1]-MeasureList[i].point[1][1]) *
                   (MeasureList[i].point[0][1]-MeasureList[i].point[1][1])    +
                   (MeasureList[i].point[0][2]-MeasureList[i].point[1][2]) *
                   (MeasureList[i].point[0][2]-MeasureList[i].point[1][2])    );printf ("got Measure[%d] = (%f %f %f)   (%f %f %f)   %f\n",
i,
MeasureList[i].point[0][0],
MeasureList[i].point[0][1],
MeasureList[i].point[0][2],
MeasureList[i].point[1][0],
MeasureList[i].point[1][1],
MeasureList[i].point[1][2],
MeasureList[i].distance);


                  i++;
                  }
                }
            }
          }

	fclose (fp);

        if (i == 0)
                {
                fprintf (stderr,
                  "No measurements found in < %s >.\n", fileName);
                }

	NumPts = i;

        return 0;
        }  // end of getMeasurements

static void
getLimits (double min, double max, double *limMin, double *limMax)
        {

        int power;
        int l = log10 ((double)(max-min));
        double incr;
        int iLimMin, iLimMax;

        incr = pow (10.0, (double)l);
        iLimMin = min/incr;
        iLimMax = 1 + max/incr;
        *limMin = iLimMin*incr;
        *limMax = iLimMax*incr;

        for (power = l-2; power <= l+2; power++)
                {
                incr = pow (10.0, (double) power);
                iLimMin = min/incr;
                iLimMax = 1 + max/incr;

printf ("incr = %f    iLimMin = %d   iLimMax = %d\n", incr, iLimMin, iLimMax);

                if ( ((iLimMax - iLimMin) >= 2) &&
                            ((iLimMax - iLimMin) <= 10 )  )
                        {
                        *limMin = iLimMin*incr;
                        *limMax = iLimMax*incr;
                        break;
                        }
                }



        }       /* end of getLimits */



static int
setupHistogramImage ()
	{
	FILE *fp;
	int i;

	if (NumPts <= 0)
		{
		fprintf (stderr, "No points present.\n");
		if ( (CurrHistogramImg != NULL) && 
			(CurrHistogramImg != NoDataImg) )
			{
printf ("changing image to NoData\n");
			ImgBox->deimage ();
			delete CurrHistogramImg;
			CurrHistogramImg = NoDataImg;
			ImgBox->image (CurrHistogramImg);
			ImgBox->redraw();
			setupImgHLDesc (&ImageHLDescriptor, 
				CurrHistogramImg, -1.0, -1.0);
			}
		return 0;
		}

	if ( (fp = fopen ("m1d.dpInput.dat", "w")) == NULL)
		{
		fprintf (stderr, "Unable to write dp input file.\n");
		return -1;
		}

	
	double mn = MeasureList[0].distance;
	double mx = MeasureList[0].distance;
	double mnDP, mxDP;

        for (i = 0; i < NumPts; i++)
                {
		fprintf (fp, "%f\n", MeasureList[i].distance);
		mn = MIN (mn, MeasureList[i].distance);
		mx = MAX (mx, MeasureList[i].distance);
		}
	fclose (fp);

	getLimits (mn, mx, &mnDP, &mxDP);

	if ((fp = fopen ("dpScript", "w")) == NULL)
		{
		fprintf (stderr, 
		 "Unable to open output dataplot script file dpScript.\n");
		return -1;
		}

	fprintf (fp, "WINDOW COORDINATES 0 56 60 100 \n");
	fprintf (fp, "set IPL1NA tmp.eps\n");
	fprintf (fp, "device 2 POSTSCRIPT ENCAPSULATED\n");
	fprintf (fp, "tic mark label size 5\n");
	fprintf (fp, "read m1d.dpInput.dat x\n");
	fprintf (fp, "xlimit %f %f\n", mnDP, mxDP);
	fprintf (fp, "histogram x\n");
	fprintf (fp, "device 2 close\n");
	fprintf (fp, "quit\n");

	fclose (fp);

	system ("dataplot < dpScript > /dev/null");

	sleep (3);

	if (convertDpEpsToPng ("tmp.eps", "tmp.png"))
		{
		return -1;
		}

	Fl_Image *newImg = getImage ("tmp.png");

	if (newImg == NULL)
		{
		fprintf (stderr, "Error generating image.\n");
		return -1;
		}

printf ("changing image to Histogram \n");

	ImgBox->deimage ();

	if ( (CurrHistogramImg != NULL) && (CurrHistogramImg != NoDataImg) )
		{
		delete CurrHistogramImg;
		}


	CurrHistogramImg = newImg;
	ImgBox->image (CurrHistogramImg);
	ImgBox->redraw();
	setupImgHLDesc (&ImageHLDescriptor, CurrHistogramImg, mnDP, mxDP);

	return 0;
	} // end of setupHistogramImage




#if 0



static int compareMeasurements(const void *a, const void *b)
	{
	if ( ((Measurement *)a)->distance < ((Measurement*)b)->distance )
		{
		return -1;
		}
	else
		{
		return 1;
		}
	}

static char *getMeasurementsString ()
        {
        static char str[20000];
        char line[1000];

        int i;

	if (NumPts == 0)
		{
		strcpy (str, "No Measurements");
		return str;
		}



        qsort (MeasureList, NumPts, sizeof (Measurement), compareMeasurements);



        str[0] = 0;
        for (i = 0; i < NumPts; i++)
                {
                BufferPosition[i] = strlen (str);
		sprintf (line, "Dist: %f    Pts: (%f, %f, %f)   (%f, %f, %f)\n",
			MeasureList[i].distance,
			MeasureList[i].point[0][0],
			MeasureList[i].point[0][1],
			MeasureList[i].point[0][2],
			MeasureList[i].point[1][0],
			MeasureList[i].point[2][1],
			MeasureList[i].point[3][2]);
                strcat (str, line);
                }
        BufferPosition[NumPts] = strlen (str);

        return str;
        }

void findMinMaxBufferPos ()
	{
	int i;

	MaxPos = BufferPosition[NumPts];
	for (i = 0; i < NumPts; i++)
		{
		if (MeasureList[i].distance > MaxVal)
			{
			MaxPos = BufferPosition[i];
			break;
			}
		}



	MinPos = BufferPosition[0];
	for (i = 0; i < NumPts; i++)
		{
		if (MeasureList[i].distance >= MinVal)
			{
			MinPos = BufferPosition[i];
			break;
			}
		}

	return;
	}  // end of findMinMaxBufferPos


void setSliders ()
	{
	float max, min;

	if (NumPts == 0)
		{
		max = 1;
		min = 0;
		}
	else
		{
		min = MeasureList[0].distance;
		max = MeasureList[NumPts-1].distance;
		}


	MinSlider->range (min, max);
	MaxSlider->range (min, max);

	MinVal = MAX (min, MinVal);
	MaxVal = MIN (max, MaxVal);

	findMinMaxBufferPos ();
	TxtBuffer->highlight (MinPos, MaxPos);
	}  // end of setSliders
#endif


void setMeanSD ()
	{
	int i;
	double mean, sd;
	double sum, d;
	static char meanString[100];
	static char stdDevString[100];



	if (NumPts > 0)
		{
		sum = 0;
		for (i = 0; i < NumPts; i++)
			{
			sum += MeasureList[i].distance;
			}
		mean = sum / NumPts;
		sprintf (meanString, "Mean: %f", mean);
		}
	else
		{
		strcpy (meanString, "Mean: ");
		}

	if (NumPts > 1)
		{
		sum = 0;
		for (i = 0; i < NumPts; i++)
			{
			d = MeasureList[i].distance - mean;
			sum += d*d;
			}
		sd = sqrt (sum/(NumPts - 1));
		sprintf (stdDevString, "Std Dev: %f", sd);
		}
	else
		{
		strcpy (stdDevString, "Std Dev: ");
		}
		

	MeanLabel->label (meanString);
	StdDevLabel->label (stdDevString);

	}  // end of setMeanSD


void getDataCB (Fl_Widget *w, void *data)
	{
	printf ("getDataCB called\n");

	getMeasurements ();

#if 0
	char *str = getMeasurementsString ();

	TxtBuffer->text (str);
#else
	setupHistogramImage ();

#endif

	// setSliders ();
	setMeanSD ();

	}  // end of getDataCB


void hiLightCB (Fl_Widget *w, void *data)
	{

	printf ("hiLightCB called\n");

	double minVal, maxVal;
	double d;
	char cmd[200];

	// if the current image is not a histogram, don't do anything
	if (CurrHistogramImg == NULL) 
		{
		return;
		}
	else if (CurrHistogramImg == NoDataImg)
		{
		return;
		}

	d = 
	((double)(ImageHLDescriptor.leftHL - ImageHLDescriptor.leftEdgeX)) /
	((double)(ImageHLDescriptor.rightEdgeX - ImageHLDescriptor.leftEdgeX));

	minVal = ImageHLDescriptor.leftEdgeData + 
	  d *(ImageHLDescriptor.rightEdgeData - ImageHLDescriptor.leftEdgeData);

	d = 
	((double)(ImageHLDescriptor.rightHL - ImageHLDescriptor.leftEdgeX)) /
	((double)(ImageHLDescriptor.rightEdgeX - ImageHLDescriptor.leftEdgeX));

	maxVal = ImageHLDescriptor.leftEdgeData + 
	  d *(ImageHLDescriptor.rightEdgeData - ImageHLDescriptor.leftEdgeData);



        system ("hev-measurement1D highlight off");
        usleep (200000);

	sprintf (cmd, "hev-measurement1D highlight %f %f", minVal, maxVal);
	printf ("%s\n", cmd);
	system (cmd);

	}  // end of getDataCB

#if 0


void valueSliderMaxCB (Fl_Widget *w, void *data)
	{


	MaxVal = ((Fl_Value_Slider *)w)->value();
	if (MaxVal < MinVal)
		{
		MaxVal = MinVal;
		((Fl_Value_Slider *)w)->value(MaxVal);
		}
	findMinMaxBufferPos ();
	TxtBuffer->highlight (MinPos, MaxPos);
	}

void valueSliderMinCB (Fl_Widget *w, void *data)
	{
	MinVal = ((Fl_Value_Slider *)w)->value();
	if (MaxVal < MinVal)
		{
		MinVal = MaxVal;
		((Fl_Value_Slider *)w)->value(MinVal);
		}
	findMinMaxBufferPos ();
	TxtBuffer->highlight (MinPos, MaxPos);
	}




static char *getTrialString ()
	{
	static char str[20000];
	char line[1000];

	int i;


	str[0] = 0;
	for (i = 0; i < 120; i++)
		{
		sprintf (line, "Line %3d abcd...xyz\n", i+1);
		strcat (str, line);
		}

	return str;
	}




static char *getTrialString2 ()
	{
	static char str[20000];
	char line[1000];

	int i;


	for (i = 0; i < 120; i++)
		{
		MeasureList[i].distance = drand48();
		}

	qsort (MeasureList, 120, sizeof (Measurement), compareMeasurements);



	str[0] = 0;
	for (i = 0; i < 120; i++)
		{
		BufferPosition[i] = strlen (str);
		sprintf (line, "%3d  Measurement %9.6f\n", i+1, MeasureList[i].distance);
		strcat (str, line);
		}
	BufferPosition[120] = strlen (str);


	return str;
	}


static char *getTrialString3 ()
	{
	static char str[20000];
	char line[1000];

	int i;



	qsort (MeasureList, 120, sizeof (Measurement), compareMeasurements);



	str[0] = 0;
	for (i = 0; i < 120; i++)
		{
		BufferPosition[i] = strlen (str);
		sprintf (line, "%3d  Measurement %9.6f\n", i+1, MeasureList[i].distance);
		strcat (str, line);
		}
	BufferPosition[120] = strlen (str);

	return str;
	}

#endif


static
Fl_Box *makeLabel (int x, int y, int w, int h, char *str)
        {
        Fl_Box *label = new Fl_Box (FL_FLAT_BOX, x, y, w, h, "");
	label->label(str);
        label->align (FL_ALIGN_INSIDE | FL_ALIGN_CENTER);

        return label;
        }       // end of makeLabel



#if 0
void makeVertSlider (int xPos, int yPos, int xSize, int ySize, 
	  char *label, Fl_Pack **groupWidget, Fl_Value_Slider **sliderWidget)
	{

	//Fl_Group *sliderGroup = new Fl_Pack (0, 0, xSize, ySize);
	Fl_Pack *sliderGroup = new Fl_Pack (xPos, yPos, xSize, ySize);
        sliderGroup->end ();

        sliderGroup->type (FL_VERTICAL);
	*groupWidget = sliderGroup;

        Fl_Value_Slider *slider =
                new Fl_Value_Slider (0, 0, 30, ySize-30);
	slider->type (FL_VERTICAL);

        Fl_Box *sliderLabel = new Fl_Box (
                                 0, 0, 50, 30);
        sliderLabel->type (FL_FLAT_BOX);
	sliderLabel->label (label);
        sliderLabel->align (FL_ALIGN_INSIDE | FL_ALIGN_CENTER);
	
        sliderGroup->add (sliderLabel);

        sliderGroup->add (slider);
        sliderGroup->resizable (slider);

	*sliderWidget = slider;
	}	// end of makeVertSlider



static Fl_Pack * makeTxtDisplayGroup ()
	{

	// Text_Display


	TxtBuffer = new Fl_Text_Buffer ();



	char *trialString = getTrialString2 ();





	TxtBuffer->text (trialString);
	TxtBuffer->highlight (10, 25);

	Fl_Widget *label = makeLabel (0, 0, 300, 30, "Distance between Points");

	Fl_Text_Display *txtDisplay = new Fl_Text_Display (0, 0, 300, 170);
	txtDisplay->buffer (TxtBuffer);

	Fl_Pack *txtDisplayGroup = new Fl_Pack (0, 0, 300, 200);
	txtDisplayGroup->type (FL_VERTICAL);
	txtDisplayGroup->end ();

	txtDisplayGroup->add (label);
	txtDisplayGroup->add (txtDisplay);
	txtDisplayGroup->resizable (txtDisplay);

	return txtDisplayGroup;

	}  // end of makeTxtDisplayGroup

#endif



//////////////////////////////////////////////////////////////////





class HEV_Fl_Event_Box : public Fl_Box
   {

    public:

        HEV_Fl_Event_Box (
           Fl_Boxtype b, int x, int y, int w, int h, const char *c) : 
           Fl_Box (b, x, y, w, h, c) {};

        int handle (int);

   };  // end of class HEV_Fl_Event_Box


int HEV_Fl_Event_Box::handle (int event)
        {
        // char *evName;
        int rtn;

        event = Fl::event();

        switch (event)
                {
                case FL_PUSH:
                        // evName = "PUSH";
			do_callback ();
                        rtn = 1;
                        break;

                case FL_RELEASE:
                        // evName = "RELEASE";
			do_callback ();
                        rtn = 1;
                        break;

                case FL_DRAG:
                        // evName = "DRAG";
			do_callback ();
                        rtn = 1;
                        break;

                default:
                        // printf ("Event %d\n", event);
                        return 0;
                        break;

                }

#if 0
        int x = Fl::event_x();
        int y = Fl::event_y();

        printf ("Event: %s  at %d %d\n", evName, x, y);
#endif

        return rtn;
        }  // end of HEV_Fl_Event_Box::handle


#if 1

//
// Create an Fl_Image object from the file.  Figure out file type from
// the file name.
//
static Fl_Image *getImage (char *fileName)
        {
        char *extension;
        int len;
        Fl_Image *img;

        len = strlen (fileName);
        if (len == 0)
                {
                return NULL;
                }

        extension = fileName + (len-1);

        for ( extension = fileName+(len-1); extension >= fileName; extension--)
                {
                if (*extension == '.')
                        {
                        extension++;
                        break;
                        }
                }

        if (extension <= fileName)
                {
                return NULL;
                }


        img = NULL;

        if (strcasecmp (extension, "XBM") == 0)
                {
                img = new Fl_XBM_Image (fileName);
                }
        else if (strcasecmp (extension, "GIF") == 0)
                {
                img = new Fl_GIF_Image (fileName);
                }
        else if (strcasecmp (extension, "XPM") == 0)
                {
                img = new Fl_XPM_Image (fileName);
                }
        else if (strcasecmp (extension, "JPEG") == 0)
                {
                img = new Fl_JPEG_Image (fileName);
                }
        else if (strcasecmp (extension, "JPG") == 0)
                {
                img = new Fl_JPEG_Image (fileName);
                }
        else if (strcasecmp (extension, "PNG") == 0)
                {
                img = new Fl_PNG_Image (fileName);
                }
        else if (strcasecmp (extension, "PNM") == 0)
                {
                img = new Fl_PNM_Image (fileName);
                }


        return img;

        }  // end of getImage

#else

unsigned char ImageRGBData[300][300][3];

static Fl_Image *getImage (char *fileName)
        {

	int i, j;
	unsigned char r, g, b;
	unsigned char *rgbArray = &(ImageRGBData[0][0][0]);

	b = 0;
	for (j = 0; j < 300; j++)
		{
		r = 255 * ((j/30) % 2);
		for (i = 0; i < 300; i++)
			{
			g = 255 * ((i/30) % 2);		
			ImageRGBData [i][j][0] = r;
			ImageRGBData [i][j][1] = g;
			ImageRGBData [i][j][2] = b;
			}
		}

	Fl_RGB_Image *img = new Fl_RGB_Image (rgbArray, 300, 300, 3, 0);
printf ("rgbArray = %p\n", rgbArray);
printf ("start of array %p    end of array %p\n",
&(ImageRGBData[0][0][0]), &(ImageRGBData[299][299][2]));
	return img;
	}





#endif







static void 
highlightImage (ImgHLDesc *imageDesc, int mn, int mx)
	{
	char *lineStart, *c;
	int lineCount;
	int y, i;


	if (mx < mn)
		{
		return;
		}


	lineStart = imageDesc->pixelArray + imageDesc->depth * mn;


	lineCount = (1+mx-mn) * imageDesc->depth;

	for (y = 0; y < imageDesc->height; y++)
		{

		c = lineStart;

		for (i = 0; i < lineCount; i++)
			{
			*c = 255 - *c;
			c++;
			}
		lineStart += imageDesc->depth * imageDesc->width;
		}


	imageDesc->img->uncache ();
	

	} // end of highlightImage




void imgBoxCB (Fl_Widget *w, void *data)
        {
// printf ("BoxCB called\n");


	// data is stored pixel interleaved  in the data array.
	// step through, invert all bytes except (last when img has depth 4)

	// if the current image is not a histogram, don't do anything
	if (CurrHistogramImg == NULL) 
		{
		return;
		}
	else if (CurrHistogramImg == NoDataImg)
		{
		return;
		}


        int event = Fl::event();

        int x = Fl::event_x() - (w->x());
        int y = Fl::event_y() - (w->y());

#if 0
	x = MIN (x, ImageHLDescriptor.width);
	x = MAX (x, 0);
#else
	x = MIN (x, ImageHLDescriptor.rightEdgeX);
	x = MAX (x, ImageHLDescriptor.leftEdgeX);
#endif

        switch (event)
            {
            case FL_PUSH:
                        // printf ("PUSH (%d %d)\n", x, y);
		if (ImageHLDescriptor.leftHL == -1)
			{
			// begin hl at x
			// highlight at x
			ImageHLDescriptor.leftHL=ImageHLDescriptor.rightHL = x;
			highlightImage (&ImageHLDescriptor, x, x);
			w->redraw();
			}
		else
			{
			if (x <= ImageHLDescriptor.leftHL)
				{
				// highlight from x to leftHL-1 inclusive
				highlightImage (&ImageHLDescriptor, 
					x, ImageHLDescriptor.leftHL-1);
				w->redraw();
				ImageHLDescriptor.leftHL = x;
				ImageHLDescriptor.movingLeft = 1;
				}
			else if (x >= ImageHLDescriptor.rightHL)
				{
				// hl from rightHL+1 to x inclusive
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.rightHL+1, x);
				w->redraw();
				ImageHLDescriptor.rightHL = x;
				ImageHLDescriptor.movingLeft = 0;
				}
			else if ((x-ImageHLDescriptor.leftHL) < 
					      (ImageHLDescriptor.rightHL-x))
				{
				// change leftHL
				// unhl from leftHL to x-1 inclusive
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.leftHL, x-1);
				w->redraw();
				ImageHLDescriptor.leftHL = x;
				ImageHLDescriptor.movingLeft = 1;
				}
			else
				{
				// change rightHL
				// unhl from x+1 to rightHL inclusive
				highlightImage (&ImageHLDescriptor, 
					x+1, ImageHLDescriptor.rightHL);
				w->redraw();
				ImageHLDescriptor.rightHL = x;
				ImageHLDescriptor.movingLeft = 0;
				}
			}
				
                break;

            case FL_RELEASE:
                // printf ("RELEASE %d %d\n", x, y);
		// nothing needed
                break;

            case FL_DRAG:
                // printf ("DRAG %d %d\n", x, y);
		// printf ("movingLeft = %d   leftHL %d  righHL %d\n", 
		//             movingLeft, leftHL, rightHL);


		if (ImageHLDescriptor.movingLeft)
			{

			// first check for crossover
			if (x > ImageHLDescriptor.rightHL)
				{
				// unhighlight everything
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.leftHL,
					ImageHLDescriptor.rightHL);

				// switch left & right boundaries
				ImageHLDescriptor.leftHL = 
					ImageHLDescriptor.rightHL;
				ImageHLDescriptor.rightHL = x;

				// rehighlight
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.leftHL,
					ImageHLDescriptor.rightHL);


				// switch movingLeft
				ImageHLDescriptor.movingLeft = 
					! ImageHLDescriptor.movingLeft;
				
				w->redraw();
				}
			else if (x < ImageHLDescriptor.leftHL)
				{
				// hl from x to leftHL-1 inclusive
				highlightImage (&ImageHLDescriptor, 
					x, ImageHLDescriptor.leftHL-1);
				w->redraw();
				ImageHLDescriptor.leftHL = x;
				}
			else if (x > ImageHLDescriptor.leftHL)
				{
				// unhl from leftHL to x-1 inclusive
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.leftHL, x-1);
				w->redraw();
				ImageHLDescriptor.leftHL = x;
				}
				
			}
		else
			{
			// first check for crossover
			if (x < ImageHLDescriptor.leftHL)
				{
				// unhighlight everything
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.leftHL,
					ImageHLDescriptor.rightHL);

				// switch left & right boundaries
				ImageHLDescriptor.rightHL = 
					ImageHLDescriptor.leftHL;
				ImageHLDescriptor.leftHL = x;

				// rehighlight
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.leftHL,
					ImageHLDescriptor.rightHL);

				// switch movingLeft
				ImageHLDescriptor.movingLeft = 
					! ImageHLDescriptor.movingLeft;
				
				w->redraw();
				}
			else if (x > ImageHLDescriptor.rightHL)
				{
				// hl from rightHL+1 to x inclusive
				highlightImage (&ImageHLDescriptor, 
					ImageHLDescriptor.rightHL+1, x);
				w->redraw();
				ImageHLDescriptor.rightHL = x;
				}
			else if (x < ImageHLDescriptor.rightHL)
				{
				// unhl from x+1 to rightHL inclusive
				highlightImage (&ImageHLDescriptor, 
					x+1, ImageHLDescriptor.rightHL);
				w->redraw();
				ImageHLDescriptor.rightHL = x;
				}
			}

	
		if (ImageHLDescriptor.leftHL > ImageHLDescriptor.rightHL)
			{
			int t;
			t = ImageHLDescriptor.leftHL;
			ImageHLDescriptor.leftHL = ImageHLDescriptor.leftHL;
			ImageHLDescriptor.rightHL = t;
			ImageHLDescriptor.movingLeft = 
				! ImageHLDescriptor.movingLeft;
			}


                break;

            default:
                        printf ("Event %d\n", event);
                        break;

            }



#if 0
	printf ("edges: %d %d     hl: %d %d\n",
		ImageHLDescriptor.leftEdgeX,
		ImageHLDescriptor.rightEdgeX,
		ImageHLDescriptor.leftHL,
		ImageHLDescriptor.rightHL);
#endif

        }  // end of imgBoxCB






static Fl_Pack * 
makeHistogramDisplayGroup ()
	{
	char *imageFileName = "histogram.png";


#if 0
        Fl_Image *img = getImage (imageFileName);
#else
	makeNoDataImg ();
        Fl_Image *img = NoDataImg;
#endif

        if (img == NULL)
                {
                fprintf (stderr, "Unable to open image file < %s >\n", 
                        imageFileName);
                return NULL;
                }

        int w = img->w();
        int h = img->h();


        HEV_Fl_Event_Box *imgBox = 
		new HEV_Fl_Event_Box (FL_NO_BOX, 0, 0, w, h, NULL);

	ImgBox = imgBox;

        imgBox->image (img);
	CurrHistogramImg = NoDataImg;


	setupImgHLDesc (&ImageHLDescriptor, CurrHistogramImg, -1.0, -1.0);

        // imgBox->align (FL_ALIGN_LEFT | FL_ALIGN_BOTTOM);

        // The fltk code is a little confused about positioning
        // when combining text and image in a label.
        // The following combination of alignment flags seems to be
        // a reasonable solution.
        imgBox->align   (
                        FL_ALIGN_IMAGE_OVER_TEXT | 
                        FL_ALIGN_TOP | 
                        FL_ALIGN_LEFT | 
                        FL_ALIGN_INSIDE
                        );

        imgBox->callback (imgBoxCB, NULL);




	Fl_Pack *txtDisplayGroup = new Fl_Pack (0, 0, 360, 200);
	txtDisplayGroup->type (FL_VERTICAL);
	txtDisplayGroup->end ();

	txtDisplayGroup->add (imgBox);


	return txtDisplayGroup;
	}  // makeHistogramDisplayGroup


//////////////////////////////////////////////////////////////////

static Fl_Pack * makeListGroup ()
	{


	// Create the Pack group to contain the three widgets

	Fl_Pack *listGroup = new Fl_Pack (0, 0, 360, 200);
	listGroup->type (FL_HORIZONTAL);
 listGroup->end ();



	// Create the child widgets


#if 0
	// Slider: Min

	// Fl_Value_Slider *valueSlider1;
	Fl_Pack *groupWidget1;
	makeVertSlider (0, 0, 30, 200, "MIN", &groupWidget1, &MinSlider);

	MinSlider->callback (valueSliderMinCB, NULL);
	MinSlider->type (FL_VERTICAL);
	MinSlider->range (0.0, 1.0);
#endif




#if 0
	Fl_Pack * txtDisplay = makeTxtDisplayGroup ();
#else
	Fl_Pack * txtDisplay = makeHistogramDisplayGroup ();

#endif

 // return txtDisplay;



#if 0

	// Slider: Max

        // Fl_Value_Slider *valueSlider2;
        Fl_Pack *groupWidget2;
        makeVertSlider (0, 0, 30, 200, "MAX", &groupWidget2, &MaxSlider);

	MaxSlider->callback (valueSliderMaxCB, NULL);
	MaxSlider->type (FL_VERTICAL);
	MaxSlider->range (0.0, 1.0);

#endif


	// Add the children to the group
	// listGroup->add (groupWidget1);
 listGroup->add (txtDisplay);
	// listGroup->add (groupWidget2);

	// Make the Fl_Text_Display widget govern the resizing
	listGroup->resizable (txtDisplay);

// listGroup->end ();

	return listGroup;
	} // end of makeListGroup









static Fl_Pack * makeStatGroup ()
	{
#if 0
	Fl_Widget *mean = makeLabel (0, 0, 180, 50, "Mean: 0.500");
	Fl_Widget *sd = makeLabel (0, 0, 180, 50, "Std Dev: 0.200");
#else
	MeanLabel = makeLabel (0, 0, 180, 50, "Mean: ");
	StdDevLabel = makeLabel (0, 0, 180, 50, "Std Dev: ");
	Fl_Widget *mean = MeanLabel;
	Fl_Widget *sd = StdDevLabel;
#endif

	Fl_Pack *statGroup = new Fl_Pack (0, 0, 360, 50);
	statGroup->type (FL_HORIZONTAL);
	statGroup->end ();

	statGroup->add(mean);
	statGroup->add(sd);

	statGroup->resizable (statGroup);

	return statGroup;
	
	}


static Fl_Group * makeButtonGroup ()
	{

	Fl_Button *getDataButton = new Fl_Button (0, 0, 120, 50, 
		"Get Data");

	Fl_Button *hiLightButton = new Fl_Button (0, 0, 120, 50, 
		"Highlight Range");

	Fl_Light_Button *autoButton = new Fl_Light_Button (0, 0, 120, 50, 
		"Auto Update");


	Fl_Pack *buttonGroup = new Fl_Pack (0, 0, 360, 50);
	buttonGroup->type (FL_HORIZONTAL);
	buttonGroup->end ();


	buttonGroup->add (getDataButton);
	buttonGroup->add (hiLightButton);
	buttonGroup->add (autoButton);

	buttonGroup->resizable (buttonGroup);

	return buttonGroup;
	}


static Fl_Group * makeButtonGroup2 ()
	{

	Fl_Button *getDataButton = new Fl_Button (5, 10, 110, 30, 
		"Get Data");
	getDataButton->callback (getDataCB);

	Fl_Button *hiLightButton = new Fl_Button (125, 10, 110, 30, 
		"Highlight Range");
	hiLightButton->callback (hiLightCB);

	Fl_Light_Button *autoButton = new Fl_Light_Button (245, 10, 110, 30, 
		"Auto Update");

	Fl_Group *buttonGroup = new Fl_Group (0, 0, 360, 50);
	buttonGroup->end ();

	buttonGroup->add (getDataButton);
	buttonGroup->add (hiLightButton);
	buttonGroup->add (autoButton);

	buttonGroup->resizable (buttonGroup);

	return buttonGroup;
	}











static Fl_Pack *makeUIGroup ()
	{
	Fl_Pack * listGroup = makeListGroup ();
	Fl_Pack * statGroup = makeStatGroup ();
	Fl_Group * buttonGroup = makeButtonGroup2 ();

	Fl_Pack *uiGroup = new Fl_Pack (0, 0, 360, 300);
	uiGroup->type (FL_VERTICAL);
	uiGroup->end ();

	uiGroup->add (listGroup);
	uiGroup->add (statGroup);
	uiGroup->add (buttonGroup);
	uiGroup->resizable (listGroup);

	return uiGroup;
	}	// end of makeUIGroup



static Fl_Window *makeUI ()
	{
	Fl_Window *window = new Fl_Window (360, 300);

	window->resizable (window);
	window->size_range (50, 50);
	window->end();

	Fl_Pack * uiGroup = makeUIGroup ();
	window->add (uiGroup);

	return window;
	}  // end of makeUI



int main(int argc, char **argv)
	{



	Fl_Window *window = makeUI ();



	window->show (argc, argv);

	// we'll use Fl::run for now, but need to look at idle & timer
	// pseudo events

	return Fl::run();


	}


