// scales the world around the pivot point based on Y joystick input
// the wand position is the pivot point if it isn't defined

#define DEAD_ZONE (0.1f)
#define SHM_TEXT_NAME "shmHeadsUpText" 

// how long does the text remain visible?
#define TIMEOUT (2.f)

#define SCALE_SCALE (.2f)
#define MIN_SCALE (.01f)
#define MAX_SCALE (1.f/MIN_SCALE)

#include <sys/types.h>
#include <stdlib.h>
#include <math.h>
#include <dtk.h>
#include <dtk/dtkDSO_loader.h>
#include <Performer/pr/pfLinMath.h>

class wandJoystickScaleNav : public dtkNav {
public:
  wandJoystickScaleNav(dtkManager *);
  
  int postConfig(void);
  int postFrame(void);
  int activate(void);
  
private:
  
  dtkManager *mgr ;
  // dtkTime is a gettimeofday() wrapper
  dtkTime t1, t2;
  dtkInLocator *wand ;
  dtkInValuator *joystick ;
  dtkCoord wand_xyzhpr ;
  dtkCoord pivot_xyzhpr ;
  dtkSharedMem *shmText ;
  bool visible ;
  char *line ;
};


wandJoystickScaleNav::wandJoystickScaleNav(dtkManager *m) :
  dtkNav(m, "wandJoystickScaleNav") {
  mgr = m ;
  shmText = NULL ;
  visible = false ;
  // try to load shmHeadsUpText if it is not already loaded
  if (!mgr->check("shmHeadsUpText") && (mgr->load("shmHeadsUpText"))) 
    return ;

  setDescription("scales the world around the pivot point based on Y joystick input" );
}

int wandJoystickScaleNav::postConfig(void) {
  
  wand = (dtkInLocator *) mgr->get("wand", DTKINLOCATOR_TYPE);
  joystick = (dtkInValuator *) mgr->get("joystick", DTKINVALUATOR_TYPE);
  
  if(!wand || !joystick)
    return ERROR_;

  if(joystick->number() != 2) {
    dtkMsg.add(DTKMSG_ERROR,
	       "The dtkInValuator named \"joystick\" has %d values"
	       " which is not 2 values.",joystick->number());
    return ERROR_;
  }
  
  // init pivot to be wand position
  float pos[6];
  wand->read(pos);
  defPivot->set(pos) ;
  t1.reset(0.0,1);
  return dtkNav::postConfig();
}


int wandJoystickScaleNav::activate(void) {
  t1.reset(0.0,1);
  // Install the call-backs.
  return dtkNav::activate() ;
}

int wandJoystickScaleNav::postFrame(void) {
  
  // only do this the first time
  if (!shmText) {
    // try to connect to shared memory created by shmHeadsUpText
    shmText =  new dtkSharedMem(SHM_TEXT_NAME) ;
    if (!shmText)
      dtkMsg.add(DTKMSG_ERROR,
		 "wandJoystickScaleNav::postFrame() failed:"
		 " couldn't open shared memory segment %s\n",SHM_TEXT_NAME) ; 
    
    line = (char *) malloc(shmText->getSize()) ;
  }

  float delta_t = (float) t1.delta();
  float joy_xy[2];
  
  if ((float)t2.get()>TIMEOUT && visible) {
    shmText->write("") ;
    visible = false ;
  }

  // poll current joystick value
  joystick->read(joy_xy);
  
  if (fabs(joy_xy[1]) > DEAD_ZONE) {
    
    // poll wand positon
    float pos[6];
    wand->read(pos);
    wand_xyzhpr.set(pos) ;
    
    // if using the default pivot point, use the wand dtkCoord, otherwise
    // use the pivot's dtkCoord.  if using the default pivot, set it too-
    // see below
    if (_pivot == defPivot) {
      // using the wand as pivot point
      pivot_xyzhpr = wand_xyzhpr ;
    } else {
      // using non-default pivot as the pivot point - don't change it!
      pivot_xyzhpr = *_pivot ;
      // but use the wand hpr for the rotation- we get to decide how to
      // use it.
      pivot_xyzhpr.h = wand_xyzhpr.h ;
      pivot_xyzhpr.p = wand_xyzhpr.p ;
      pivot_xyzhpr.r = wand_xyzhpr.r ;
    }
    // UPDATE THE DEFAULT PIVOT POINT WHETHER IT'S BEING USED OR NOT
    *defPivot = wand_xyzhpr ;
    
    //////////////////////////////////////////
    // Change scale at pivot point
    //////////////////////////////////////////
    float dx = (joy_xy[1] > 0.0f) ?
      ((joy_xy[1] - DEAD_ZONE) / (1.0f - DEAD_ZONE)) :
      ((joy_xy[1] + DEAD_ZONE) / (1.0f - DEAD_ZONE)) ;
    
    dx *= delta_t*SCALE_SCALE ;
    
    // get the old scale so we can compare change
    float oldScale;
    oldScale = scale ;
    
    // this sets nav scale
    scale += dx ;
    if (scale<MIN_SCALE) scale = MIN_SCALE ;
    if (scale>MAX_SCALE) scale = MAX_SCALE ;
    
    // using old scale, what would location be without a scale?
    dtkCoord realLoc ;
    realLoc.x = location.x + pivot_xyzhpr.x*(1.f - 1.f/oldScale) ;
    realLoc.y = location.y + pivot_xyzhpr.y*(1.f - 1.f/oldScale) ;
    realLoc.z = location.z + pivot_xyzhpr.z*(1.f - 1.f/oldScale) ;
    
    // using the unmodified location, what would the location be before the
    // new scale is applied?
    location.x = realLoc.x - pivot_xyzhpr.x*(1.f - 1.f/scale) ;
    location.y = realLoc.y - pivot_xyzhpr.y*(1.f - 1.f/scale) ;
    location.z = realLoc.z - pivot_xyzhpr.z*(1.f - 1.f/scale) ;
    
    sprintf(line,"%f",scale) ;
    shmText->write(line) ;
    visible = true ;
    t2.reset(0.0,1);
    
          } // if(fabs(joy_xy[1]) > DEAD_ZONE)
  return dtkNav::postFrame() ;
}



/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  return new wandJoystickScaleNav(manager);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
