// modifies the pivot point by mirroring changes to the wand's position and
// orientation.  Changes are additive- not moving the wand will cause no
// change in pivot point.  Changes are only made is the shared memory
// segment named by the WAND_NAV_PIVOT environment variable is non-zero.

#include <sys/types.h>
#include <stdlib.h>
#include <math.h>
#include <dtk.h>
#include <dtk/dtkDSO_loader.h>
#include <Performer/pr/pfLinMath.h>

class wandNavPivot : public dtkAugment {
public:
  wandNavPivot(dtkManager *app) ;
  
  int preFrame(void) ;
  int postFrame(void) ;
  
private:
  dtkManager *mgr ;
  dtkInLocator *wandLoc ;
  dtkCoord wand ;
  dtkNavList *navList;
  dtkCoord pivot ;
  char *shmName ;
  dtkSharedMem *shmData ;
  char on ;
  dtkMatrix initWandMatXYZInv ;
  dtkMatrix initWandMatHPRInv ;
  dtkMatrix initPivotXYZMat ;
  dtkMatrix initPivotHPRMat ;
};


wandNavPivot::wandNavPivot(dtkManager *m) :
  dtkAugment("wandNavPivot") {
  mgr = m;
  on = false ;

  if (!(shmName = getenv("WAND_NAV_PIVOT"))) {
    dtkMsg.add(DTKMSG_ERROR,
	       "WAND_NAV_PIVOT envvar not set") ;
    return ;
  }
  
  // try to connect to shared memory 
  char zero ;
  zero = 0 ;
  shmData =  new dtkSharedMem(sizeof(char), shmName, &zero) ;
  if (!shmData) {
    dtkMsg.add(DTKMSG_ERROR,
	       "Can't load shared memory \"%s\" as specified by "
	       "WAND_NAV_PIVOT", shmName) ;
    return ;
  }
  setDescription("modifies the pivot point by mirroring changes to the wand's position and orientation.") ;
  
  validate() ;
}

int wandNavPivot::preFrame(void) {
  
  wandLoc = (dtkInLocator *) mgr->get("wand", DTKINLOCATOR_TYPE);
  if(!wandLoc)
    return ERROR_;
  
  /********************* Get a dtkNavList ***********************/
  navList = (dtkNavList *) mgr->check("dtkNavList", DTKNAVLIST_TYPE) ;
  if (!navList) {
    navList = new dtkNavList;
    if (mgr->add(navList)) {
      delete navList;
      navList = NULL;
      return dtkAugment::ERROR_;
    }
  }
  pivot = *(navList->current()->pivot()) ;
  navList->current()->pivot(&pivot) ;

  return REMOVE_CALLBACK ;
  
}


int wandNavPivot::postFrame(void) {
  
  dtkCoord wandHPR ;
  dtkCoord wandXYZ ;
  dtkMatrix wandXYZMat ;
  dtkMatrix wandHPRMat ;
  dtkMatrix pivotMat ;

  // poll wand positon
  float pos[6];
  wandLoc->read(pos);
  wand.set(pos) ;

  char data ;
  shmData->read(&data) ;

  if (data) {
    // get the wand XYZ
    wandXYZ.x = wand.x ;
    wandXYZ.y = wand.y ;
    wandXYZ.z = wand.z ;
    wandXYZ.h = wandXYZ.p = wandXYZ.r = 0.f ;

    // get wand HPR
    wandHPR.x = wandHPR.y = wandHPR.z = 0.f ;
    wandHPR.h = wand.h ;
    wandHPR.p = wand.p ;
    wandHPR.r = wand.r ;

    if (!on) {

      // when first turned on, get initial values
      on = true ;

      // get the initial inverted wand position
      initWandMatXYZInv.identity() ;
      initWandMatXYZInv.coord(wandXYZ) ;
      initWandMatXYZInv.invert() ;

      // get the initial inverted wand orientation
      initWandMatHPRInv.identity() ;
      initWandMatHPRInv.coord(wandHPR) ;
      initWandMatHPRInv.invert() ;

      // get the initial pivot, which will be updated by
      // changes in the wand's position
      dtkCoord pivotHPR ;
      pivotHPR = *(navList->current()->pivot()) ;
      pivotHPR.x = pivotHPR.y = pivotHPR.z = 0.f ;
      
      dtkCoord pivotXYZ ;
      pivotXYZ = *(navList->current()->pivot()) ;
      pivotXYZ.h = pivotXYZ.p = pivotXYZ.r = 0.f ;
      
      initPivotXYZMat.identity() ;
      initPivotXYZMat.coord(pivotXYZ) ;

      initPivotHPRMat.identity() ;
      initPivotHPRMat.coord(pivotHPR) ;


    }

    // do this every frame while "on"

    // get the current wand position
    wandXYZMat.identity() ;
    wandXYZMat.coord(wandXYZ) ;
    
    // and orientation
    wandHPRMat.identity() ;
    wandHPRMat.coord(wandHPR) ;
    
    // calculate a new pivot point based on changes to wand and initial
    // pivot, starting with identity
    pivotMat.identity() ;

    // do all the rotations
    pivotMat.mult(&initPivotHPRMat);
    pivotMat.mult(&initWandMatHPRInv) ;
    pivotMat.mult(&wandHPRMat) ;

    // do all the translations
    pivotMat.mult(&initPivotXYZMat);
    pivotMat.mult(&initWandMatXYZInv) ;
    pivotMat.mult(&wandXYZMat) ;

    pivotMat.coord(&pivot) ;

  } else {
    on = false ;
  }
  
  return CONTINUE ;
}



/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *mgr, void *p)
{
  return new wandNavPivot(mgr);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
